-------------------------- MODULE Ring ---------------------------
\* This model contains the definitions for the ring structure  

(* The hierarchy of the modules are
 Ring (The basic parameters and operations for ring calculation)
   \subset LS (The leaf set utilities)
      \subset RT (The routing table utilities)
         \subset Msg (The messages definitions)
            \subset InitialStates (The initial state and TypeInvariant)
               \subset Actions (All the actions)
                  \subset MSPastry (The specification of MSPastry including propertie
                  \subset MCMSPastry (The model checking file) 
*)
EXTENDS 
   TLC,       \* To enable the pretty print
   FiniteSets,\* To understand IsFiniteSet()
   Integers   \* To resolve the operators, e.g. <, \div on natural numbers,etc.
CONSTANTS  
   I,    \* Identifiers (ID) for a node or a key 
   A,    \* The initial active nodes
   B,    \* For defining the base of ID in routing table
   M,    \* The maximal allowed exponents of 2, 
         \*     which won't cause an overflow of natural number
   L     \* The parameter l in paper, max. length of each leafset
--------------------------------------------------------------

(*Assumptions*)
ASSUME IsFiniteSet(I)
ASSUME I \in SUBSET Nat
ASSUME A \in SUBSET I 
ASSUME M \in Nat
ASSUME L \in Nat
ASSUME B \in Nat
ASSUME B < M /\ M % B = 0
ASSUME L >0  
ASSUME Cardinality(A)>0
--------------------------------------------------------------
\*Definitions
Base == 2^B     \* Base of the number: if B=3, its base is then 8.
RingCap == 2^M  \* The capacity of the Ring

Dist(x, y) 
(*Result gives two information: positive sign represents 
  clockwise relation, value represents absolute distance*)
== LET 
       diff == (x%RingCap) - (y%RingCap)
       \*diff == x - y \* this is the case if x and y \in I
       half == 2^(M-1)
   IN  IF diff < 0-half 
       THEN diff + RingCap
       ELSE IF diff > half THEN diff - RingCap ELSE diff

\*ToTheRightOf(x,y) ==  Dist(x, y)  < 0 
(* (nodeA, nodeB) returns if the shortest way from A to B is clockwise on the ring 
   The opposite nodes in the ring is always to the left*)

AbsDist(x, i) == 
(* Absolute distance between two nodes in x ring, it takes the shortest bow to calculate *)
   LET hd == Dist(x, i) IN IF hd<0 THEN 0-hd ELSE hd

CwDist(x, i) == 
(*The clockwise distance from x to i*)
   IF Dist(x, i)  < 0 
   THEN RingCap - AbsDist(x, i) 
   ELSE AbsDist(x, i) 

==============================================================
