---------------------------- MODULE ProofLSetInv ----------------------------
(* The hierarchy of the Proofs are
 ProofBasic  
  \subset ProofRingProp (The basic parameters and operations for ring calculation)
   \subset ProofLSetProp (The leaf set local properties of data structure)
    \subset ProofProp (reductions)
     \subset ProofType (The Proof of TypeInvariant)
      \subset ProofLSetInv (The proof of Inv of Leaf Set properties)
*)
EXTENDS ProofType 


LEMMA InitInv == Init/\TypeInvariant  => Inv
<1>1. SUFFICES 
      ASSUME  
      receivedMsgs = {},
      status = [i \in I |-> IF i \in A THEN "ready" ELSE "dead"], 
      lease = [i \in I |-> IF i \in A
                          THEN A
                          ELSE {}],
      grant = [i \in I |-> IF i \in A
                          THEN A
                          ELSE {}],
     rtable = [i \in I |-> IF i \in A 
                           THEN AddToTable(A, InitRTable, i) 
                           ELSE AddToTable({i}, InitRTable, i)],
     lset = [i \in I |->  IF i \in A 
                          THEN AddToLSet(A, EmptyLS(i)) 
                          ELSE EmptyLS(i)],
     probing = [i \in I |-> {}],
     failed = [i \in I |-> {}],
     toj = [i \in I |-> i],
     TypeInvariant
     PROVE Inv
   BY DEF Init
<1>. \A i \in I: lset[i] \in LSet /\ lset[i].left \in SUBSET I /\ lset[i].right \in SUBSET I
    /\ lset[i].node = i
   BY TypeInvariant DEF LSet, TypeInvariant
<1>2. A = NodesReady 
   BY AType, <1>1 DEF NodesReady
<1>3. LSetInv
   \* BY  <1>1, NotAddSelf, EmptyLSNoSelf, AType, EmptyLST
   <2>1. \A i \in I: ~i \in (EmptyLS(i).left) /\ ~i \in (EmptyLS(i).right) 
      BY EmptyLSNoSelf
   <2>2. \A i \in I: 
         /\ ~ EmptyLS(i).node \in (AddToLSet(A, EmptyLS(i)).left) 
         /\ ~ EmptyLS(i).node  \in (AddToLSet(A, EmptyLS(i)).right)
      BY NotAddSelf, AType, EmptyLST
   <2>3. \A i \in I: ~i \in (AddToLSet(A, EmptyLS(i)).left) 
                  /\ ~i \in (AddToLSet(A, EmptyLS(i)).right)
      BY <2>2, EmptyLST
   <2>9. QED BY <2>1,<2>3, <1>1 DEF LSetInv
<1>5. NodesOK = {} /\ NodesWait = {} /\ 
      /\ NodesReady = A /\ NonDead = A /\ ReadyOK = A
   BY <1>1, AType, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
<1>6.  \A i \in I \ A: lset[i] = EmptyLS(i)
   BY <1>1, AType DEF TypeInvariant
<1>7. \A i \in A: GetLSetContent(lset[i]) = A
   BY <1>1, AType, EmptyAddAIsA DEF TypeInvariant
<1>80. SemLRply 
   BY <1>1  DEF SemLRply
<1>81. LeaseGrant 
   BY <1>1 DEF LeaseGrant

<1>83. SemJRply 
   BY <1>1 DEF SemJRply 
<1>84. SemJoinNeighbor 
   BY <1>1 DEF SemJoinNeighbor
<1>85. SemSingleReadyJoin
   BY <1>1 DEF SemSingleReadyJoin  
<1>86. SemEmptyLSetSingleNode
   BY <1>1   DEF SemEmptyLSetSingleNode
<1>87. SemJoinLeafSet
   BY <1>1 DEF SemJoinLeafSet
<1>82. GrantReady
   BY <1>1 DEF GrantReady, NodesReady, TypeInvariant
<1>90. GrantNeighbor
   <2>1. SUFFICES ASSUME NEW k \in NodesReady, 
         NEW i \in I,
         ~ i \in NodesReady,
         i \in grant[k]
         PROVE  \/ i = LeftNeighbor(lset[k])
                \/ i = RightNeighbor(lset[k])
      BY DEF GrantNeighbor
   <2>2. CASE i \in A
      BY <2>1, <2>2, <1>2
   <2>3. CASE ~i \in A
      BY <1>1, <2>1, <1>2, <2>3, AType DEF TypeInvariant
   <2>4. QED BY <2>3, <2>2
<1>91. SingleReadyNode
   <2>1. CASE Cardinality(A) =< 1
      <3>0. Cardinality(A) = 1
         BY <2>1, EnoughReadyNodes, Unequality6, CardinalityType, AType
      <3>01. PICK k \in A: A = {k}
         BY <1>5, ReadyInI, <3>0, CardinalityOneNodeCo
      <3>2. lset[k] = AddToLSet(A, EmptyLS(k))
         BY <1>1, AType
      <3>3. lset[k] = EmptyLS(k)
         BY <3>01, <3>2, EmptyAddEmtpy, AType
      <3>4. LeftNeighbor(lset[k]) = k /\ RightNeighbor(lset[k]) =k
         BY <3>3, EmptyLSNoNeighbor, EmptyLST, AType
      <3>9. QED BY <3>01, <3>4, <1>6, EmptyLSNoNeighbor, EmptyLST, AType DEF SingleReadyNode
   <2>2. CASE ~(Cardinality(A) =< 1)
      BY CardinalityOneNodeExt, <1>2, <2>2 DEF SingleReadyNode
   <2>3. QED BY <2>1,<2>2
<1>95. ReadyNeighborRight
   <2>8. \A r \in NodesReady: RightNeighbor(lset[r]) \in NodesReady
      <3>1. SUFFICES ASSUME NEW r \in NodesReady,
            NEW rn \in  I,
            rn = RightNeighbor(lset[r])
            PROVE rn \in NodesReady
         BY RightNeighborType, ReadyInI
      <3>2. rn \in GetLSetContent(lset[r])
         BY <3>1, NeigborsInLset, ReadyInI
      <3>3. rn \in A
         BY  <1>7, <3>2, <1>2, AType
      <3>9. QED BY <3>3, <1>2
   <2>9. QED BY <2>8 DEF ReadyNeighborRight
<1>96. ReadyNeighborLeft
   <2>8. \A r \in NodesReady: LeftNeighbor(lset[r]) \in NodesReady
      <3>1. SUFFICES ASSUME NEW r \in NodesReady,
            NEW ln \in  I,
            ln = LeftNeighbor(lset[r])
            PROVE ln \in NodesReady
         BY LeftNeighborType, ReadyInI
      <3>2. ln \in GetLSetContent(lset[r])
         BY <3>1, NeigborsInLset, ReadyInI
      <3>3. ln \in A
         BY  <1>7, <3>2, <1>2, AType
      <3>9. QED BY <3>3, <1>2
   <2>9. QED BY <2>8 DEF ReadyNeighborLeft
<1>97. HalfNeighbor
   <2>1. CASE Cardinality(A) =< 1
      <3>0. Cardinality(A) = 1
         BY <2>1, EnoughReadyNodes, Unequality6, CardinalityType, AType
      <3>01. PICK k \in A: A = {k}
         BY <1>5, ReadyInI, <3>0, CardinalityOneNodeCo
      <3>1. SUFFICES 
             /\ LeftNeighbor(lset[k]) =k
             /\ RightNeighbor(lset[k]) = k
             /\ ~\E m \in receivedMsgs: m.mreq.type = "LSProbeReply" 
             /\ ~\E m \in receivedMsgs: m.mreq.type = "LSProbe" /\ m.destination # k
             /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
             /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
             /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
         BY <3>01, <1>5 DEF HalfNeighbor
      <3>2. lset[k] = AddToLSet(A, EmptyLS(k))
         BY <1>1, AType
      <3>3. lset[k] = EmptyLS(k)
         BY <3>01, <3>2, EmptyAddEmtpy, AType
      <3>4. LeftNeighbor(lset[k]) = k /\ RightNeighbor(lset[k]) =k
         BY <3>3, EmptyLSNoNeighbor, EmptyLST, AType
      <3>5. receivedMsgs = {}
         BY <1>1
      <3>9. QED BY <3>5,<3>4, <1>5 DEF HalfNeighbor
   <2>2. CASE ~(Cardinality(A) =< 1)
      <3>1.  Cardinality(A) > 1
         BY Unequality3, <2>2, AType, CardinalityType
      <3>2. \A k \in A: 
            /\ lset[k] \in LSet 
            /\ EmptyLS(k) \in LSet 
            /\ A \in SUBSET I
            /\ lset[k] = AddToLSet(A, EmptyLS(k))
         BY EmptyLST, AType, <1>1
      <3>4. \A i \in A: RightNeighbor(lset[i]) # i /\ LeftNeighbor(lset[i]) # i
         BY AddMoreHasNeighbor, <3>2, <3>1
      <3>8. \A i \in ReadyOK : RightNeighbor(lset[i]) # i /\ LeftNeighbor(lset[i]) # i
         BY <1>5, <3>4, ReadyInI
      <3>9. QED BY <3>8 DEF HalfNeighbor
   <2>3. QED BY <2>1,<2>2
<1>98. NeighborClosest
   <2>5. \A i \in A: lset[i] = AddToLSet(A, EmptyLS(i)) 
      BY <1>1, AType
   <2>6. \A i \in A: 
        /\ \A m \in (EmptyLS(i).left) \cup A:
                  CwDist(LeftNeighbor(AddToLSet(A, EmptyLS(i))), AddToLSet(A, EmptyLS(i)).node)
                   =< CwDist(m, AddToLSet(A, EmptyLS(i)).node)
        /\ \A n \in (EmptyLS(i).right) \cup A:
                  CwDist(AddToLSet(A, EmptyLS(i)).node, RightNeighbor(AddToLSet(A, EmptyLS(i))))
                   =< CwDist(AddToLSet(A, EmptyLS(i)).node, n)   
      BY  AType, AddToLSetInv, EmptyLST, AddToLSetType
   <2>8. \A x, y \in A: /\ CwDist(LeftNeighbor(lset[x]), x)=< CwDist(y, x)
                      /\ CwDist(x, RightNeighbor(lset[x])) =< CwDist(x, y)
      BY <2>6, EmptyLSIsEmpty, <2>5, AType
   <2>9. QED BY <2>8, <1>2 DEF NeighborClosest
   
<1>92. CoverToJ
   <2>1. SUFFICES ASSUME NEW r \in NodesReady, 
         NEW k \in I,
         lset[k] # EmptyLS(k),
         k # r,
         Covers(lset[r], k)
         PROVE k = toj[r]
      BY DEF CoverToJ
   <2>2. /\ lset[r] \in LSet
         /\ lset[r].node = r
      BY <2>1, ReadyInI
   <2>6. CASE k \in A
      <3>1. \/ CwDist(r, k) < CwDist(r, RightNeighbor(lset[r])) 
            \/ CwDist(k, r) < CwDist(LeftNeighbor(lset[r]), r)
         BY NeighborCoverageCo, <2>2, <2>1
      <3>2. k \in NodesReady
         BY <2>6, <1>2
      <3>3. ~(Cardinality(NodesReady) =< 1)
         BY <2>1, <3>2, CardinalityTwoNode, ReadyInI
      <3>4. /\ CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, k)  
            /\ CwDist(LeftNeighbor(lset[r]), r) =< CwDist(k, r)
         BY <3>3, <3>2, <2>1, <1>98 
         DEF  NeighborClosest
      <3>9. QED BY <3>1, <3>4, <2>2, <2>1,<3>2,
             LeftNeighborType,RightNeighborType,
            ReadyInI, IisNat, CwDistType, Unequality2
   <2>7. CASE ~k \in A
      BY <1>1, <2>1, <1>2, <2>7, AType DEF TypeInvariant
   <2>4. QED BY <2>6, <2>7
<1>93. CompleteLeafSet
   <2>1. SUFFICES ASSUME NEW i \in I,
         NEW r \in I,
         NEW j \in NodesReady,
         i # j,
         lset[i] # EmptyLS(i),
         RightNeighbor(lset[i]) = r,
         CwDist(i, j) < CwDist(i, r)
         PROVE FALSE
       BY DEF CompleteLeafSet
   <2>2. /\ lset[i] \in LSet
         /\ lset[i].node = i
      BY <2>1, ReadyInI
   <2>4. CASE i \in A
     <3>8.  CwDist(i, r) =< CwDist(i, j)  
        BY <1>2, <2>1,<2>4, ReadyInI, CardinalityTwoNode, <1>98 
        DEF NeighborClosest
     <3>9. QED BY <3>8, <2>2, <2>1, LeftNeighborType,RightNeighborType,
            ReadyInI, IisNat, CwDistType, Unequality2
   <2>5. CASE ~i  \in A
      <3>7. i # r
         BY <2>1, IisNat, ReadyInI, CwDistZeroCompare
      <3>8. RightNeighbor(lset[i]) = i
         BY <2>5, <1>1, <2>2, EmptyLSNoNeighbor DEF TypeInvariant
      <3>9. QED BY <3>8, <3>7, <2>1
   <2>9. QED BY <2>4,<2>5
<1>94. CompleteLeafSetL
   <2>1. SUFFICES ASSUME NEW i \in I,
         NEW l \in I,
         NEW j \in NodesReady,
         i # j,
         lset[i] # EmptyLS(i),
         LeftNeighbor(lset[i]) = l,
         CwDist(j, i) < CwDist(l, i)
         PROVE FALSE
       BY DEF CompleteLeafSetL
   <2>2. /\ lset[i] \in LSet
         /\ lset[i].node = i
      BY <2>1, ReadyInI
   <2>4. CASE i \in A
     <3>8.  CwDist(l, i) =< CwDist(j, i) 
        BY <1>2, <2>1,<2>4, ReadyInI, CardinalityTwoNode, <1>98 
        DEF NeighborClosest
     <3>9. QED BY <3>8, <2>2, <2>1, LeftNeighborType,RightNeighborType,
            ReadyInI, IisNat, CwDistType, Unequality2
   <2>5. CASE ~i  \in A
      <3>7.l # i
         BY <2>1, IisNat, ReadyInI, CwDistZeroCompareCo
      <3>8. LeftNeighbor(lset[i]) = i
         BY <2>5, <1>1, <2>2, EmptyLSNoNeighbor DEF TypeInvariant
      <3>9. QED BY <3>8, <3>7, <2>1
   <2>9. QED BY <2>4,<2>5
<1>88. NoCCJoin
   <2>1. SUFFICES ASSUME NEW i \in A,
         NEW r \in A,
         NEW k \in I,
         lset[k] # EmptyLS(k),
         r = RightNeighbor(lset[i]),
         CwDist(i, k) < CwDist(i, r)
         PROVE FALSE
      BY <1>2, <1>6, ReadyInI DEF NoCCJoin
   <2>5. k \in A
      BY <1>1, <2>1, TypeInvariant DEF TypeInvariant
   <2>7. /\ lset[i] = AddToLSet(A, EmptyLS(i))
         /\ lset[i] \in LSet
         /\ EmptyLS(i) \in LSet
      BY <1>1, <2>1, EmptyLST, TypeInvariant DEF TypeInvariant
   <2>8. CwDist(i, r) =< CwDist(i, k) 
      BY AddToLSetInv, <2>1, AType, <2>5, <2>7
   <2>9. QED BY CwDistType, AType, <2>1, <2>8, IisNat, Unequality2
<1>89. SemPRplyNonEmpty
   BY <1>1 DEF SemPRplyNonEmpty    
<1>78. SemJRplyLSnonEmpty
   BY <1>1 DEF SemJRplyLSnonEmpty
<1>79. LSnonEmpty
   <2>1. SUFFICES ASSUME NEW i \in I, 
         NEW j \in I,
         i \in GetLSetContent(lset[j]),
         i # j
         PROVE lset[i] # EmptyLS(i)
      BY DEF LSnonEmpty
   <2>2. /\ lset[i] \in LSet
         /\ A \in SUBSET I
         /\ lset[i].node = i
         /\ lset[j] \in LSet
         /\ lset[j].node = j
         BY AType, <1>1,<1>2 DEF TypeInvariant
   <2>3. lset[j] # EmptyLS(j)
      BY <2>1, <2>2, NonEmptyLSet
   <2>4.  lset[j] = AddToLSet(A, EmptyLS(j))
         BY <2>3, <1>1 DEF TypeInvariant
   <2>6. j \in A
         BY <2>4, <1>1, <2>3 DEF TypeInvariant
   <2>7. i \in A
         BY <2>1, <1>1, <2>4, AType, AddEffect DEF TypeInvariant
   <2>8. lset[i] = AddToLSet(A, EmptyLS(i))
         BY <2>7, <1>1 DEF TypeInvariant
   <2>9. QED BY <2>8,<2>6, <2>7,<2>1,<2>2, AddSingleNodeNoEmpty
<1>61. SingleReadyNodeToj 
   <2>1. SUFFICES ASSUME NEW k \in I, 
         NEW r \in I,
         k # r,
         NodesReady = {r},
         \*lset[r] = EmptyLS(r),
         lset[k] # EmptyLS(k)
         PROVE FALSE
      BY DEF SingleReadyNodeToj
   <2>2. k # r
      BY <2>1
   <2>3. ~k \in A
      BY <2>1, <1>2, <2>2
   <2>4. lset[k] = EmptyLS(k)
      BY <2>3, <1>1 DEF TypeInvariant
   <2>9. QED BY <2>1,<2>4
<1>62. NoCCJoinL
   <2>1. SUFFICES ASSUME NEW i \in A,
         NEW l \in A,
         NEW k \in I,
         lset[k] # EmptyLS(k),
         l = LeftNeighbor(lset[i]),
         CwDist(k, i) < CwDist(l, i)
         PROVE FALSE
      BY <1>2, <1>6, ReadyInI DEF NoCCJoinL
   <2>5. k \in A
      BY <1>1, <2>1, TypeInvariant DEF TypeInvariant
   <2>7. /\ lset[i] = AddToLSet(A, EmptyLS(i))
         /\ lset[i] \in LSet
         /\ EmptyLS(i) \in LSet
      BY <1>1, <2>1, EmptyLST, TypeInvariant DEF TypeInvariant
   <2>8. CwDist(l, i) =< CwDist(k, i)
      BY AddToLSetInv, <2>1, AType, <2>5, <2>7
   <2>9. QED BY CwDistType, AType, <2>1, <2>8, IisNat, Unequality2
<1>63. SemJReq
   BY <1>1 DEF SemJReq
<1>64. OneJRply
   BY <1>1 DEF OneJRply
<1>65. OneJReq
   BY <1>1 DEF OneJReq
<1>66. JoinMsgDual
   BY <1>1 DEF JoinMsgDual
<1>67. AtleastOneReady
   BY <1>2, AisNotEmpty DEF AtleastOneReady
<1>99. QED BY <1>3,
      <1>97,\*<1>98,
      <1>96,<1>95,<1>94,<1>93,<1>92,<1>91,<1>90,
      <1>80,<1>81,<1>82,<1>83,<1>84,
      <1>85,<1>86,<1>87, <1>88, <1>89,
      <1>79, <1>78, <1>61,<1>62, <1>63, <1>64, 
      <1>65, <1>66, <1>67
      DEF Inv


 
LEMMA InvInv ==
 ASSUME TypeInvariant, 
        [Next]_vars,
        Inv
 PROVE  Inv'
   <1>1. ASSUME NEW i \in I, NEW j \in I,
                 RouteLookup(i,j), 
                 TypeInvariant,
                 Inv
         PROVE  Inv'
         <3>. DEFINE  nh ==  FindNext(j, i)
         <3>. nh \in I 
           BY FindNextType
         <3>1. ASSUME  status[i] = "ready", 
            NEW m \in receivedMsgs,
            m.destination = i,
            m.mreq.type ="Lookup", 
            m.mreq.node = j, 
            ~Covers(lset[i], j),
            receivedMsgs'= (receivedMsgs \{m}) \cup 
              IF nh # i
              THEN {[destination |-> nh, mreq |-> m.mreq]}
              ELSE {[destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> j]]},
            UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>,
                    TypeInvariant, Inv
         PROVE  Inv'
            <4>1. \A k \in I: /\ lset'[k] = lset[k]
                              /\ grant'[k] = grant[k]
                              /\ toj'[k] = toj[k]
                              /\ lease'[k] = lease[k]
               BY <3>1, UNCHANGED <<status, rtable, lset, 
                        probing, failed, lease,grant, toj>>
            <4>2. /\NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK 
                  /\ NonDead'= NonDead
                  /\ NodesWait'= NodesWait
               BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>5. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                         \/ ms.mreq.type = "LSProbeReply" 
                         \/ ms.mreq.type = "LSProbe" 
                         \/ ms.mreq.type = "JoinReply"
                         \/ ms.mreq.type = "BroadcastLSet"
                         \/ ms.mreq.type = "JoinRequest"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                        (IF nh # i
                        THEN {[destination |-> nh, mreq |-> m.mreq]}
                        ELSE {[destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> j]]})
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> nh, mreq |-> m.mreq].mreq.type = "Lookup"
                     BY <3>1
                  <6>3. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> j]].mreq.type = "NoLegalRoute"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2,<6>3
            \*<4>6. NeighborClosest'
            \*   BY <3>1, <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            <4>7. HalfNeighbor'  
               <5>2.\A k \in I:
                    /\ status'[i] = status[i]
                    /\ probing'[i] = probing[i]
                    /\ ReadyOK'= ReadyOK
                    /\ lset[k]' = lset[k]
                    /\ NodesWait'= NodesWait
                  BY <3>1 DEF ReadyOK, NodesReady, NodesOK, NodesWait
               <5>4. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                  BY <5>2, <5>4, ReadyInI DEF HalfNeighbor
               <5>5. CASE  \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k
                         /\ RightNeighbor(lset[k]) = k
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                  <6>1. PICK k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k
                         /\ RightNeighbor(lset[k]) = k
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                     BY <5>5
                  <6>2. /\ReadyOK'= {k}
                        /\ LeftNeighbor(lset'[k]) = k
                        /\ RightNeighbor(lset'[k]) = k
                     BY <6>1, <5>2, ReadyInI
                  <6>3. ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                     BY <6>1, <4>5
                  <6>4. ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # k
                     BY <6>1, <4>5, <5>5
                  <6>5. \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     BY <6>1, <4>5, <5>5
                  <6>6. ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1, <4>5, <5>5
                  <6>7. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                      BY <6>1, <5>2, ReadyInI
                  <6>9. QED BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, <6>7 DEF HalfNeighbor
               <5>9. QED BY <5>4, <5>5, HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <3>1, <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <3>1, <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <3>1, <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <3>1, <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2 DEF SemLRply, Inv
               <5>9. QED BY <4>5,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LeaseGrant'
               BY <3>1, <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>60. GrantReady'
               BY <3>1, <4>1, <4>2, ReadyInI DEF GrantReady, Inv
            <4>61. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node => 
                            toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node => 
                            toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg 
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>5,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>5, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>5, <5>3, <5>6, <5>8, 
                    <4>1, <4>2, SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>5, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>5, <5>3, <5>7, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. GrantNeighbor'
                BY <3>1, <4>1, <4>2, ReadyInI, GrantNeighbor DEF GrantNeighbor, Inv
            <4>67. SingleReadyNode'
               BY <3>1, <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
            <4>68. CoverToJ'
                BY <3>1, <4>1, <4>2, ReadyInI, CoverToJ DEF CoverToJ, Inv
            <4>69. NoCCJoin'
               BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve DEF NoCCJoin, Inv
            <4>53. NoCCJoinL'
               BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoinL, SimplifyAndSolve DEF NoCCJoinL, Inv
            <4>70. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>71. LSnonEmpty'
               BY <3>1, <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>72. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A k \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[k] # EmptyLS(k)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>5, <5>3, <5>8, <4>1, SlowestZenon
                    DEF SemJRplyLSnonEmpty, Inv
            <4>73. SingleReadyNodeToj'
               BY <3>1, <4>1, <4>2, ReadyInI DEF SingleReadyNodeToj, Inv
            <4>74. OneJRply'
               BY <4>5,<4>1,<4>2 DEF OneJRply, Inv
            <4>75. OneJReq'
               BY <4>5,<4>1,<4>2 DEF OneJReq, Inv
            <4>76. SemJReq'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type  = "JoinRequest"
                     PROVE
                       /\ mm.mreq.node \in NodesWait'
                       /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY <4>1, <4>2 DEF SemJReq, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JReq
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.node \in I
                  BY <5>5 DEF JReq
               <5>9. QED BY <4>5,<5>3,<5>6,<4>1,<4>2 DEF SemJReq, Inv
            <4>77. JoinMsgDual'
               BY <4>5, <4>1, <4>2 DEF JoinMsgDual, Inv
            <4>78. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv
            <4>9. QED BY <4>3,<4>7,\*<4>6, 
                  \*<4>51,<4>52,
                  <4>53,<4>54,<4>55,<4>56,<4>57,<4>58,<4>59,<4>60,<4>61,<4>62,<4>63,<4>64, 
                  <4>65,<4>66,<4>67, <4>68, <4>69,<4>70, 
                  <4>71, <4>72, <4>73, <4>74, <4>75, 
                  <4>76, <4>77, <4>78 DEF Inv
         <3>2.QED BY <3>1, <1>1 DEF  RouteLookup
   <1>2.ASSUME NEW i \in I, NEW k \in I,
                RouteJoinRequest(i,k),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>. DEFINE nh ==  FindNext(k, i)
      <3>0. nh \in I 
        BY FindNextType
      <3>1. ASSUME NEW m \in receivedMsgs,
                       m.destination = i,
                       m.mreq.type = "JoinRequest",
                       m.mreq.node = k, 
                       receivedMsgs'= (receivedMsgs \{m}) \cup 
                          IF nh # i
                          THEN {[destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]}
                          ELSE {[destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]},
                       UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>,
                 TypeInvariant,
                 Inv
         PROVE  Inv'
            <4>1. \A j \in I: /\ lset'[j] = lset[j]
                              /\ lease'[j] = lease[j]
                              /\ grant'[j] = grant[j]
                              /\ toj'[j] = toj[j]
               BY <3>1, UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>
            <4>2. /\ NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK 
                  /\ NonDead'= NonDead 
                  /\ NodesWait'= NodesWait
               BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>5. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                        (IF nh # i
                         THEN {[destination |-> nh, 
                                mreq |-> [type |-> "JoinRequest",
                                          rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                          node |-> k]]}
                         ELSE {[destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]})
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> nh, 
                                mreq |-> [type |-> "JoinRequest",
                                          rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                          node |-> k]].mreq.type = "JoinRequest"
                     BY <3>1
                  <6>3. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]].mreq.type = "NoLegalRoute"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2,<6>3
            \*<4>6. NeighborClosest'
            \*   BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            
            <4>7. HalfNeighbor'        
               <5>2./\ ReadyOK' = ReadyOK 
                    /\ NodesWait'= NodesWait
                    /\ \A j \in I: lset'[j] = lset[j]
                  BY <4>2, <4>1
               <5>4. CASE \A j \in ReadyOK : RightNeighbor(lset[j]) # j /\ LeftNeighbor(lset[j]) # j
                  BY <5>2, <5>4, ReadyInI DEF HalfNeighbor
               <5>5. CASE  \E j \in ReadyOK:
                             /\ ReadyOK = {j} 
                             /\ LeftNeighbor(lset[j]) = j
                             /\ RightNeighbor(lset[j]) = j
                             /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                             /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # j
                             /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                   => GetLSetContent(mj.mreq.lset) = {j}
                             /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                             /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {j, w}
                  <6>1. PICK j \in ReadyOK:
                             /\ ReadyOK = {j} 
                             /\ LeftNeighbor(lset[j]) = j
                             /\ RightNeighbor(lset[j]) = j
                             /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                             /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination #j
                             /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                   => GetLSetContent(mj.mreq.lset) = {j}
                             /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                             /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {j, w}
                     BY <5>5
                  <6>2. /\ReadyOK'= {j}
                        /\ LeftNeighbor(lset'[j]) = j
                        /\ RightNeighbor(lset'[j]) = j
                     BY <6>1, <5>2, ReadyInI
                  <6>3. ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                     BY <6>1, <4>5
                  <6>4. ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # j
                     BY <6>1, <4>5, <5>5
                  <6>5. /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                               => GetLSetContent(mj.mreq.lset) = {j}
                     BY <6>1, <4>5, <5>5
                  <6>6. ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1, <4>5, <5>5
                  <6>7. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {j, w}
                     BY <5>2, ReadyInI, <6>1
                  <6>9. QED BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, <6>7 DEF HalfNeighbor
               <5>9. QED BY <5>4, <5>5, HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \* <4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
                BY <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
                BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>5,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>61. GrantReady'
                BY <4>1, <4>2, <3>1, ReadyInI DEF GrantReady, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>5, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>5, <5>3, <5>6, <5>8, <4>1, <4>2, SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>5, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>5, <5>3, <5>7, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>5,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <4>1, <4>2, <3>1, ReadyInI DEF GrantNeighbor, Inv
            <4>68. SingleReadyNode'
                BY <4>1, <4>2, <3>1, ReadyInI DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               BY <4>1, <4>2, <3>1, ReadyInI, CoverToJ DEF CoverToJ, Inv
            <4>70. NoCCJoin'
               BY <4>1, <4>2, <3>1, ReadyInI, NoCCJoin DEF NoCCJoin, Inv
            <4>53. NoCCJoinL'
               BY <4>1, <4>2, <3>1, ReadyInI DEF NoCCJoinL, Inv
            <4>71. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>72. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>5, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>73. SingleReadyNodeToj'
               BY <4>1, <4>2, <3>1, ReadyInI DEF SingleReadyNodeToj, Inv 
            <4>74. OneJRply'
               BY <4>5,<4>1,<4>2 DEF OneJRply, Inv
            <4>75. OneJReq'
               <5>0. m #  [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]
                  <6>1. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]].mreq.type
                        = "NoLegalRoute"
                     OBVIOUS
                  <6>2. m.mreq.type = "JoinRequest"
                     BY <3>1
                  <6>3. QED BY <6>1, <6>2
               <5>1. SUFFICES ASSUME 
                     NEW m1 \in receivedMsgs', 
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type = "JoinRequest",
                     m2.mreq.type = "JoinRequest",
                     m1.mreq.node = m2.mreq.node
                  PROVE m1 = m2
                  BY DEF OneJReq, Inv
               <5>2. m1 # [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]
                  <6>8. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", 
                           key |-> k]].mreq.type = "NoLegalRoute" 
                     OBVIOUS
                  <6>9. QED BY <6>8, <5>1
               <5>3. m2 # [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]
                  <6>8. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", 
                           key |-> k]].mreq.type = "NoLegalRoute" 
                     OBVIOUS
                  <6>9. QED BY <6>8, <5>1
               <5>6. CASE m1 = [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. SUFFICES ASSUME m1 # m2 PROVE FALSE
                     OBVIOUS
                  <6>2. m2 \in receivedMsgs
                     BY <6>1, <5>1, <5>3, <5>6, <3>1
                  <6>3. m1.mreq.node = k
                     BY <5>6
                  <6>4. CASE m2 # m
                      <7>5. m.mreq.node = k
                         BY <3>1
                      <7>6. m2.mreq.node # k
                         BY <6>4, <7>5, <3>1, <5>1, <6>2 DEF OneJReq, Inv
                      <7>9. QED BY <5>1,<6>3,<7>6
                  <6>5. CASE m2 = m
                     <7>4. m =[destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                        BY <3>1, <5>0, <6>5, <5>1
                     <7>9. QED BY <7>4, <6>5, <5>6, <6>1
                  <6>9. QED BY <6>4,<6>5
               <5>7. CASE m2 = [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. SUFFICES ASSUME m1 # m2 PROVE FALSE
                     OBVIOUS
                  <6>2. m1 \in receivedMsgs
                     BY <6>1, <5>1, <5>2, <5>7, <3>1
                  <6>3. m2.mreq.node = k
                     BY <5>7
                  <6>4. CASE m1 # m
                      <7>5. m.mreq.node = k
                         BY <3>1
                      <7>6. m1.mreq.node # k
                         BY <6>4, <7>5, <3>1, <5>1, <6>2 DEF OneJReq, Inv
                      <7>9. QED BY <5>1,<6>3,<7>6
                  <6>5. CASE m1 = m
                     <7>4. m =[destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                        BY <3>1, <5>0, <6>5, <5>1
                     <7>9. QED BY <7>4, <6>5, <5>7, <6>1
                  <6>9. QED BY <6>4,<6>5
               <5>8. CASE /\ m1 # [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                          /\ m2 # [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. m1 \in receivedMsgs /\ m2 \in receivedMsgs
                     BY <5>8, <5>2,<5>3, <3>1, <5>1
                  <6>2. QED BY <5>1, <6>1 DEF OneJReq, Inv
               <5>9. QED BY <5>6, <5>7, <5>8
            <4>76. SemJReq'
               <5>1. /\ m.mreq.node = k
                     /\ k \in NodesWait
                     /\ lset[k] = EmptyLS(k)
                  BY <3>1 DEF SemJReq, Inv
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type  = "JoinRequest"
                     PROVE 
                      /\ mm.mreq.node \in NodesWait
                      /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node) 
                  BY <4>2 DEF SemJReq, Inv 
               <5>4. CASE mm = [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. mm.mreq.node = k
                     BY <5>4
                  <6>2. QED BY <6>1, <5>1, <4>1
               <5>5. CASE mm # [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. mm # [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]
                     BY <5>3, 
                        [destination |-> i, mreq |-> [type |-> "NoLegalRoute", 
                           key |-> k]].mreq.type = "NoLegalRoute"
                  <6>2. mm \in receivedMsgs
                     BY <6>1, <3>1, <5>5, <5>3
                  <6>5. mm.mreq \in JReq
                      BY SlowestZenon,<5>3, <6>2 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>6. mm.mreq.node \in I
                      BY <6>5 DEF JReq
                  <6>9. QED BY <6>6, <5>3, <4>1, <6>2
                        DEF SemJReq, Inv
               <5>9. QED BY <5>4,<5>5
            <4>77. SingleReadyNodeToj'
               BY <4>1, <4>2, <3>1, ReadyInI DEF SingleReadyNodeToj, Inv
            <4>78. JoinMsgDual'
               <5>1. SUFFICES ASSUME NEW m1 \in receivedMsgs',
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type  = "JoinReply",
                     m2.mreq.type = "JoinRequest"
                     PROVE m2.mreq.node # m1.destination
                  BY DEF JoinMsgDual
               <5>2. m1 \in receivedMsgs
                  BY <4>5, <5>1
               <5>3. m2 # [destination |-> i, mreq |-> [type |-> "NoLegalRoute", key |-> k]]
                  <6>8. [destination |-> i, mreq |-> [type |-> "NoLegalRoute", 
                           key |-> k]].mreq.type = "NoLegalRoute" 
                     OBVIOUS
                  <6>9. QED BY <6>8, <5>1
               <5>7. CASE m2 = [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. m2.mreq.node = k
                     BY <5>7
                  <6>2. m.mreq.node = k
                     BY <3>1
                  <6>3. m.mreq.node # m1.destination
                     BY <3>1, <5>1, <5>2 DEF JoinMsgDual, Inv
                  <6>9. QED BY <6>1,<6>2,<6>3
               <5>8. CASE m2 # [destination |-> nh, mreq |-> 
                                   [type |-> "JoinRequest",
                                    rtable |-> AddToTable(GetRTableContent(rtable[i]), m.mreq.rtable, i),
                                    node |-> k]]
                  <6>1. m2 \in receivedMsgs
                     BY <5>8, <5>3, <3>1, <5>1
                  <6>2. QED BY <5>1, <5>2, <6>1 DEF JoinMsgDual, Inv
               <5>9. QED BY <5>7, <5>8
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv            
            <4>9. QED BY <4>3,<4>7,\*<4>6, 
                  \*<4>51,<4>52,
                  <4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,<4>59,
                  <4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, 
                  <4>70,<4>71, <4>72, <4>73,
                  <4>74,<4>75, 
                  <4>76, <4>77, <4>78, <4>80
                  DEF Inv
      <3>2.QED BY <3>1, <1>2 DEF RouteJoinRequest, Inv
   <1>3. ASSUME NEW i \in I, NEW j \in I,
                Deliver(i,j),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <2>1. ASSUME NEW m \in receivedMsgs,
                   receivedMsgs'= receivedMsgs \{m},
                 UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>,
                 TypeInvariant,
                 Inv
         PROVE  Inv' 
            <4>1. \A k \in I: /\ lset'[k] = lset[k]
                              /\ lease'[k] = lease[k]
                              /\ grant'[k] = grant[k]
                              /\ toj'[k] = toj[k]
               BY <2>1, UNCHANGED <<status, rtable, lset, probing, failed, lease, toj>>
            <4>2./\ NodesReady' = NodesReady 
                 /\ ReadyOK' = ReadyOK 
                 /\ NonDead'= NonDead
                 /\ NodesWait' = NodesWait
               BY <2>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <4>1, <4>2, LSetInv DEF LSetInv, Inv

            <4>5. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinRequest"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
               BY <2>1 
            <4>4. HalfNeighbor' 
               <5>2./\ ReadyOK' = ReadyOK 
                    /\ NodesWait'= NodesWait
                    /\ \A jj \in I: lset'[jj] = lset[jj]
                  BY <4>2, <4>1
               <5>4. CASE \A jj \in ReadyOK : RightNeighbor(lset[jj]) # jj 
                          /\ LeftNeighbor(lset[jj]) # jj
                  BY <5>2, <5>4, ReadyInI DEF HalfNeighbor
               <5>5. CASE  \E jj \in ReadyOK:
                             /\ ReadyOK = {jj} 
                             /\ LeftNeighbor(lset[jj]) = jj
                             /\ RightNeighbor(lset[jj]) = jj
                             /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                             /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # jj
                             /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                   => GetLSetContent(mj.mreq.lset) = {jj}
                             /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                             /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {jj, w}
                  <6>1. PICK jj \in ReadyOK:
                             /\ ReadyOK = {jj} 
                             /\ LeftNeighbor(lset[jj]) = jj
                             /\ RightNeighbor(lset[jj]) = jj
                             /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                             /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination #jj
                             /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                   => GetLSetContent(mj.mreq.lset) = {jj}
                             /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                             /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {jj, w}
                     BY <5>5
                  <6>2. /\ReadyOK'= {jj}
                        /\ LeftNeighbor(lset'[jj]) = jj
                        /\ RightNeighbor(lset'[jj]) = jj
                     BY <6>1, <5>2, ReadyInI
                  <6>3. ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                     BY <6>1, <4>5
                  <6>4. ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # jj
                     BY <6>1, <4>5, <5>5
                  <6>5. /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                               => GetLSetContent(mj.mreq.lset) = {jj}
                     BY <6>1, <4>5, <5>5
                  <6>6. ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1, <4>5, <5>5
                  <6>7. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {jj, w}
                     BY <5>2, ReadyInI, <6>1
                  <6>9. QED BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, <6>7 DEF HalfNeighbor
               <5>9. QED BY <5>4, <5>5, HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>6. NeighborClosest'
            \*   BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <2>1, <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <2>1, <4>1, <4>2, ReadyInI, CompleteLeafSetL  DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <2>1, <4>1, <4>2, ReadyInI, ReadyNeighborRight  DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <2>1, <4>1, <4>2, ReadyInI, ReadyNeighborLeft  DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>5,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF LSnonEmpty, Inv, TypeInvariant
            <4>60. LeaseGrant'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF LeaseGrant, Inv, TypeInvariant
            <4>61. GrantReady'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF GrantReady, Inv, TypeInvariant
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>5, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>5, <5>3, <5>6, <5>8, 
                  <4>1, <4>2, SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>5, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>5, <5>3, <5>7, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>5,<5>3,<5>7, <5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF GrantNeighbor, Inv, TypeInvariant
            <4>68. SingleReadyNode'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF SingleReadyNode, Inv, TypeInvariant
            <4>69. CoverToJ'
               BY <2>1, <4>1, <4>2, ReadyInI, CoverToJ  
                  DEF CoverToJ, Inv, TypeInvariant
            <4>70. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>5, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>5, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>71. NoCCJoin'
               BY <2>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve
                  DEF NoCCJoin, Inv, TypeInvariant
            <4>53. NoCCJoinL'
               BY <2>1, <4>1, <4>2, ReadyInI, SimplifyAndSolve
                  DEF NoCCJoinL, Inv, TypeInvariant
            <4>72. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>5, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>73. SingleReadyNodeToj'
               BY <2>1, <4>1, <4>2, ReadyInI  
                  DEF SingleReadyNodeToj, Inv, TypeInvariant
            <4>74. OneJRply'
               BY <4>5,<4>1,<4>2 DEF OneJRply, Inv
            <4>75. OneJReq'
               BY <4>5,<4>1,<4>2 DEF OneJReq, Inv
            
            <4>76. SemJReq'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type  = "JoinRequest"
                     PROVE
                       /\ mm.mreq.node \in NodesWait'
                       /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY <4>1, <4>2 DEF SemJReq, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>5, <5>3
               <5>5. mm.mreq \in JReq
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.node \in I
                  BY <5>5 DEF JReq
               <5>9. QED BY <4>5,<5>3,<5>6,<4>1,<4>2 DEF SemJReq, Inv
            <4>77. JoinMsgDual'
               BY <4>5,<4>1,<4>2 DEF JoinMsgDual, Inv
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv                        
            <4>9. QED BY <4>3,<4>4,\*<4>6,
                  \*<4>51,<4>52,
                  <4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,<4>59,
                  <4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, <4>70, <4>71, 
                  <4>72, <4>73, <4>74,<4>75,
                  <4>76, <4>77, <4>80 DEF Inv
      <2>9. QED BY <1>3, <2>1 DEF Deliver, Inv
   <1>4. ASSUME NEW i \in I,
                ReceiveLSProbe(i),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
                   m.destination = i,
                   m.mreq.type = "LSProbe",
                   ReceiveLSProbeInner(i, m),
                   status[i] # "dead",
                   ~i \in NodesReady => lset[i] # EmptyLS(i),
                   m.mreq.node # i,
                   m.mreq.node = m.mreq.lset.node,
                    TypeInvariant,
                 Inv
         PROVE  Inv'
         <4>. DEFINE j == m.mreq.node
         <4>. DEFINE fi == failed[i] \ {j}
         <4>. DEFINE ls1 == IF j \in GetLSetContent(lset[i]) 
                              THEN lset[i]
                              ELSE AddToLSet({j}, lset[i])
         <4>. DEFINE prb1 == GetLSetContent(ls1) \cap m.mreq.failed
         <4>. DEFINE prb2 == GetLSetContent(AddToLSet((GetLSetContent(m.mreq.lset) \ fi), ls1))
                               \ GetLSetContent(ls1)
         <4>. DEFINE prb == (prb1 \cup prb2) \ (probing[i] \cup fi) 
         <4>. DEFINE newm == [type |-> "LSProbeReply",  
                  node |-> i,
                  lset |-> lset[i],
                  failed |-> fi]
         <4>01. failed' = [failed EXCEPT ![i] = fi]
            BY <3>1 DEF ReceiveLSProbeInner
         <4>02. rtable' = [rtable EXCEPT ![i] = AddToTable({j}, @, i)]
            BY <3>1 DEF ReceiveLSProbeInner
         <4>03. lset'   = [lset EXCEPT ![i] = ls1]
            BY <3>1 DEF ReceiveLSProbeInner
         <4>04. probing'= [probing EXCEPT ![i] = @ \cup prb] 
            BY <3>1 DEF ReceiveLSProbeInner
         <4>06. receivedMsgs' = (receivedMsgs \ {m}) 
                           \cup { [destination |-> j, mreq |-> newm]} 
                           \cup Probe(i, ls1, fi, prb)
            BY <3>1 DEF ReceiveLSProbeInner
         <4>07. UNCHANGED <<lease, status, grant, toj>>
            BY <3>1 DEF ReceiveLSProbeInner
         <4>08. /\ NodesReady' = NodesReady 
                /\ ReadyOK' = ReadyOK 
                /\ NodesWait' = NodesWait
                /\ NonDead'= NonDead
               BY <4>07, ReadyInI 
               DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
         <4>1.  m.mreq \in Prb
            BY SlowestZenon, <3>1 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
         <4>2. /\ j \in I
               /\ m.mreq.failed \in SUBSET I
               /\ m.mreq.lset \in LSet
            BY <3>1, <4>1 DEF Prb
         <4>3. /\ fi \in SUBSET I
               /\ ls1 \in LSet
               /\ lset[i] \in LSet
            BY <3>1, <4>2, AddToLSetType, RemoveFromLSetType,
            SlowestZenon DEF TypeInvariant    
         <4>4. prb1 \in SUBSET I
            BY <3>1, <4>2
         <4>5. prb2 \in SUBSET I
           <5>1. GetLSetContent(m.mreq.lset) \in SUBSET I
             BY <4>2, GetLSetContentType
           <5>2. AddToLSet((GetLSetContent(m.mreq.lset) \ fi), ls1) \in LSet
             BY <5>1, <4>3, AddToLSetType
           <5>3. QED
             BY <3>1, <5>2, GetLSetContentType
         <4>6. newm \in PRpl
             BY <3>1, <4>3 DEF PRpl
         <4>7. prb \in SUBSET I
            BY <4>3,<4>4,<4>5 DEF TypeInvariant
       <4>10.  ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "JoinRequest"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
          <6>. HIDE DEF ls1, fi, prb, newm
          <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                { [destination |-> j, mreq |-> newm]} 
                   \cup Probe(i, ls1, fi, prb)
             BY <4>06, <4>3 DEF TypeInvariant
          <6>2. [destination |-> j, mreq |-> newm].mreq.type = "LSProbeReply"
             BY <3>1, <4>6 DEF PRpl
          <6>3. \A ms \in Probe(i, ls1, fi, prb): ms.mreq.type = "LSProbe"
            <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb)
                    BY ONLY ProbeMsgType, <4>3, <4>7
            <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
          <6>9. QED BY <6>1,<6>2,<6>3
       <4>11. \A k \in I: ~k \in (lset[k].left) /\ ~k \in (lset[k].right) 
          BY LSetInv DEF LSetInv, Inv
        <4>12. /\ ~ lset[i].node \in (ls1.left) 
             /\ ~ lset[i].node  \in (ls1.right)
             /\ ls1.node = lset[i].node
           <5>. HIDE DEF ls1
           <5>1. CASE ls1 = AddToLSet({j}, lset[i])
              BY <5>1, AddNodeNotChange, NotAddSelf, <4>2,<4>3 DEF TypeInvariant
           <5>2. CASE ls1 = lset[i]
              BY <5>2,<4>2,<4>3, LSetInv DEF LSetInv, TypeInvariant
           <5>9. QED BY <5>1,<5>2 DEF ls1
        <4>13. lset'[i] = AddToLSet({j}, lset[i]) \/ lset'[i] = lset[i]
           BY <4>03 DEF TypeInvariant
       
       <4>15. lset[i]' = IF j \in GetLSetContent(lset[i]) 
                         THEN lset[i]
                         ELSE  AddToLSet({j}, lset[i])
          BY <4>03 DEF TypeInvariant
        <4>. HIDE DEF ls1
       <4>16. lset[i]' = ls1 /\ ls1.node = i /\ lset[i].node = i
          BY <4>03, <4>12,  TypeInvariant DEF TypeInvariant
       <4>17. /\ ~ i \in (lset[i]'.left) 
             /\ ~ i \in (lset[i]'.right)
          BY <4>12, <4>16
       <4>18. \A k \in I \{i}: 
             /\ ~k \in (lset'[k].left) 
             /\ ~k \in (lset'[k].right) 
          BY <4>11, <4>03, TypeInvariant DEF TypeInvariant
        <4>20. LSetInv'
        BY <4>18,<4>17 DEF LSetInv
        \* \A i \in I: ~i \in (lset[i].left) /\ ~i \in (lset[i].right)
        <4>21.\A k \in I: 
               /\ lset'[k] \in LSet 
               /\ lset[k] \in LSet
               /\ lset[k].node = k
           BY <4>03, <4>3 DEF TypeInvariant 
        <4>22. \A k \in I\ {i}: lset'[k] = lset[k]
           BY <4>03 DEF TypeInvariant
        <4>95. HalfNeighbor' 
           <5>2. CASE /\\E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                                   => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
              <6>1. CASE ReadyOK = {i}
                 <7>1. ReadyOK'= {i}
                    BY <4>08, <6>1
                 <7>2. j # i
                    BY <3>1
                 <7>3. lset[i]'= ls1 
                    BY <4>16
                 <7>33. ls1 = IF j \in GetLSetContent(lset[i]) 
                              THEN lset[i]
                              ELSE AddToLSet({j}, lset[i])
                    BY DEF ls1
                 <7>4. /\ LeftNeighbor(lset[i]) =i
                       /\ RightNeighbor(lset[i]) = i
                    BY <5>2, <6>1
                 <7>5. lset[i] = EmptyLS(i)
                    BY  <4>21, <7>4, <4>16, EmptyLSNoNeighbor
                 <7>6. lset[i]'= AddToLSet({j}, EmptyLS(i))
                    <8>7. GetLSetContent(lset[i]) = {i}
                       BY <7>5, EmptyLSIsEmpty
                    <8>8. ~(j \in GetLSetContent(lset[i]))
                       BY <8>7, <7>2
                    <8>9. ls1 = AddToLSet({j}, lset[i])
                       BY <8>8, <7>33
                    <8>. HIDE DEF ls1
                    <8>11. QED BY <8>9, <7>3, <7>5
                 <7>7. RightNeighbor(AddToLSet({j}, EmptyLS(i))) =j  /\ LeftNeighbor(AddToLSet({j}, EmptyLS(i))) = j
                    BY <4>2,<1>4, AddSingleNodeToEmpty, SlowestZenon
                 <7>8. RightNeighbor(lset'[i]) =j  /\ LeftNeighbor(lset'[i]) = j
                    BY <7>6, <7>7, AddSingleNodeToEmpty, SlowestZenon
                 <7>9. QED BY <7>1, <7>2, <7>8 DEF HalfNeighbor
              <6>2. CASE ~ReadyOK = {i}
                 <7>1. PICK k \in ReadyOK: 
                     /\ k # i
                     /\ ReadyOK = {k} 
                     BY <5>2, <6>2
                 <7>2. m \in receivedMsgs /\ m.mreq.type = "LSProbe" /\ m.destination  # k
                    BY <7>1, <3>1
                 <7>8. ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                    BY <7>1, <5>2
                 <7>9. QED BY <7>1, <7>2, <7>8
              <6>9. QED BY <6>1,<6>2
           <5>3. CASE  \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k 
              <6>1. CASE i \in ReadyOK
                 <7>1. RightNeighbor(lset'[i]) # i /\ LeftNeighbor(lset'[i]) # i
                    BY <5>3, <6>1, <4>16, <4>3, <4>2, AddToLSetNotEmpty DEF ls1
                 <7>2. \A k \in ReadyOK'\ {i}: lset[k]' = lset[k]
                    BY  <4>22, ReadyInI, <4>08
                 <7>3. \A k \in ReadyOK'\ {i}: RightNeighbor(lset'[k]) # k
                                      /\LeftNeighbor(lset'[k]) # k
                    BY <7>2, <5>3, <4>08
                 <7>9. QED BY <7>3,<7>1 DEF HalfNeighbor
              <6>2. CASE ~i \in ReadyOK
                 BY <6>2, <4>22, ReadyInI, <4>08, <5>3 DEF HalfNeighbor, Inv
              <6>9. QED BY <6>1,<6>2
           <5>9. QED BY <5>2,<5>3, HalfNeighbor DEF HalfNeighbor, Inv

        <4>54. CompleteLeafSet'
            <5>1.  /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ m.mreq.failed \in SUBSET I
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ NodesReady' = NodesReady
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  RightNeighbor(lset'[i]) \in I
                   /\ RightNeighbor(lset[i]) \in I
                BY <4>16, <4>3, <4>2,<4>08, <4>22, RightNeighborType DEF TypeInvariant
             <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. SUFFICES ASSUME NEW r \in NodesReady,
                   i # r,
                   lset'[i] # EmptyLS(i),
                   CwDist(i, r) < CwDist(i, RightNeighbor(lset'[i]))
                   PROVE FALSE
                BY <5>1 DEF CompleteLeafSet, Inv
             <5>6. lset[i] # EmptyLS(i)
                <6>1. CASE i \in NodesReady
                    <7>1. /\ r \in ReadyOK
                          /\ i \in ReadyOK
                          /\ r \in I
                          /\ i \in I
                          /\ ReadyOK \in SUBSET I
                       BY <6>1, <5>3, ReadyInI
                    <7>3. ~(Cardinality(ReadyOK)=<1)
                       BY CardinalityTwoNode,<6>1, <5>3, ReadyInI
                    <7>4. RightNeighbor(lset[i]) # i /\ LeftNeighbor(lset[i]) # i
                       BY <7>3, <7>1, CardinalityOneNodeExt,<5>1, <5>3 DEF HalfNeighbor, Inv
                    <7>9. QED BY <7>4, <5>1, <5>3, EmptyNoNeighbor
                <6>2. CASE ~i \in NodesReady
                   BY <3>1, <5>3, <5>1,<6>2
                <6>9. QED BY <6>1,<6>2
             <5>7. ~(CwDist(i, r) < CwDist(i, RightNeighbor(lset[i])))
                BY <5>3, <5>6, <5>1, ReadyInI DEF CompleteLeafSet, Inv

             <5>8. ~(CwDist(i, r) < CwDist(i, RightNeighbor(lset'[i])))
                BY <5>7, <5>2, <5>1, TransLEQ, Unequality2, CwDistType, IisNat,  
                   ReadyInI  DEF TypeInvariant
             <5>9. QED BY <5>1, <5>8, <5>3 
        <4>55. CompleteLeafSetL'
        \* similar to <4>54
            <5>1.  /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ m.mreq.failed \in SUBSET I
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ NodesReady' = NodesReady
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  LeftNeighbor(lset'[i]) \in I
                   /\ LeftNeighbor(lset[i]) \in I
                BY <4>16, <4>3, <4>2,<4>08, <4>22, LeftNeighborType DEF TypeInvariant
             <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node ) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. SUFFICES ASSUME NEW l \in NodesReady,
                   i # l,
                   lset'[i] # EmptyLS(i),
                   CwDist(l, i) < CwDist(LeftNeighbor(lset'[i]), i)
                   PROVE FALSE
                BY <5>1 DEF CompleteLeafSetL, Inv
             <5>6. lset[i] # EmptyLS(i)
                <6>1. CASE i \in NodesReady
                    <7>1. /\ l \in ReadyOK
                          /\ i \in ReadyOK
                          /\ l \in I
                          /\ i \in I
                          /\ ReadyOK \in SUBSET I
                       BY <6>1, <5>3, ReadyInI
                    <7>3. ~(Cardinality(ReadyOK)=<1)
                       BY CardinalityTwoNode,<6>1, <5>3, ReadyInI
                    <7>4. LeftNeighbor(lset[i]) # i /\ RightNeighbor(lset[i]) # i
                       BY <7>3, <7>1, CardinalityOneNodeExt,<5>1, <5>3 DEF HalfNeighbor, Inv
                    <7>9. QED BY <7>4, <5>1, <5>3, EmptyNoNeighbor
                <6>2. CASE ~i \in NodesReady
                   BY <3>1, <5>3, <5>1,<6>2 
                <6>9. QED BY <6>1,<6>2
             <5>7. ~(CwDist(l, i) < CwDist(LeftNeighbor(lset[i]), i))
                BY <5>3, <5>6, <5>1, ReadyInI DEF CompleteLeafSetL, Inv

             <5>8. ~(CwDist(l, i) < CwDist(LeftNeighbor(lset'[i]), i))
                BY <5>7, <5>2, <5>1, TransLEQ, Unequality2, CwDistType, IisNat,  
                   ReadyInI  DEF TypeInvariant
             <5>9. QED BY <5>1, <5>8, <5>3 
        <4>56. ReadyNeighborRight'
            \* the same as ReceiveProbeReply
             <5>0. NodesReady' = NodesReady
                BY <4>07 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  RightNeighbor(lset'[i]) \in I
                   /\ RightNeighbor(lset[i]) \in I
                BY <4>7, <4>16, <4>3, <4>2,<4>08, <4>22,  RightNeighborType
             <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. CASE i \in NodesReady
                <6>3. \A k \in I:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~RightNeighbor(lset'[i]) \in NodesReady 
                      => CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, k) 
                    <7>1. SUFFICES ASSUME NEW k \in I,
                           NEW r \in I,
                           r = RightNeighbor(lset[i]),
                           k # r,
                           i # k,
                           lset[k] # EmptyLS(k),
                           ~RightNeighbor(lset'[i]) \in NodesReady
                       PROVE CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, k) 
                       BY <7>1, ReadyInI, <5>1, <5>2 DEF ReadyNeighborRight, Inv
                    <7>2. CASE  RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                       BY <7>1,<7>2,<5>3, ReadyInI DEF ReadyNeighborRight, Inv
                    <7>3. CASE  RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       <8>3. CASE r \in NodesReady
                          <9>1. SUFFICES ASSUME  CwDist(i, k) < CwDist(i, RightNeighbor(lset'[i]))
                                PROVE FALSE
                                BY <7>1, <5>1, IisNat, Unequality2,CwDistType         
                          <9>2. RightNeighbor(lset'[i]) # k
                             BY <9>1, IisNat, CwDistType, <5>1, <7>1, Unequality23
                          <9>3. CwDist(i, k) < CwDist(i, r)
                             BY <9>1, <5>2, TransLESS, <5>1, <7>1, IisNat, CwDistType
                          <9>4. lset[k] # EmptyLS(k)
                             BY <7>1
                          <9>41. i # r
                             BY CwDistZeroCompare, <9>3, IisNat
                          <9>5. \/ toj[i] = k /\ toj[r] = r
                                \/ toj[r] = k /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>4, <9>41, <9>3(*, NoCCJoin*) DEF NoCCJoin, Inv
                          <9>6. CwDist(i, RightNeighbor(lset'[i])) < CwDist(i, r)
                            \* BY <5>2, <7>3, <5>1, <7>1, CwDistType, IisNat, Unequality6, RightDistInjectivity
                             <10>2. /\ CwDist(i, RightNeighbor(lset'[i])) \in Nat
                                    /\ CwDist(i, r) \in Nat
                                 BY <5>1, <7>1, CwDistType, IisNat
                             <10>8. CwDist(i, RightNeighbor(lset'[i])) # CwDist(i, r)
                                 BY <5>1, <7>1, <7>3, IisNat, RightDistInjectivity
                             <10>9. QED BY <10>2, <10>8, <5>2, <7>1, Unequality6                              
                          <9>7. lset[RightNeighbor(lset'[i])] # EmptyLS(RightNeighbor(lset'[i]))
                             <10>6. /\ lset'[i] = AddToLSet({j}, lset[i])
                                    /\ j \in I
                                 BY <7>3, <4>16, RNInjectivity, <5>1, <7>3, <3>1 DEF ls1
                             <10>7. RightNeighbor(lset'[i]) = j
                                 BY <10>6, <5>1, <7>3, AddSingleNodeChange
                             <10>8. lset[j] # EmptyLS(j)
                                 BY <3>1, <4>3,SemPRplyNonEmpty DEF SemPRplyNonEmpty, Inv
                             <10>9. QED BY  <10>7,<10>8
                          <9>8. \/ /\toj[i] = RightNeighbor(lset'[i]) 
                                   /\ toj[r] = r
                                \/ /\ toj[r] = RightNeighbor(lset'[i]) 
                                   /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>7,<9>41, <9>6, <5>1(*, NoCCJoin*) DEF NoCCJoin, Inv
                          <9>9. QED BY <9>5,<9>8, <9>2, <9>3, <7>1
                       <8>4. CASE ~ r \in NodesReady
                          <9>2. CwDist(i, r) =< CwDist(i, k)
                            BY <7>1, <8>4, <5>1, <5>0,<5>3, ReadyInI
                            DEF ReadyNeighborRight, Inv
                          <9>9. QED BY <9>2, <5>2, <5>1, <7>1,TransLEQ, CwDistType, IisNat,
                             ReadyInI  DEF TypeInvariant
                       <8>9. QED BY <8>3,<8>4
                    <7>9. QED BY <7>2,<7>3
                 <6>4. \A r \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, i) 
                    <7>1. SUFFICES ASSUME NEW r \in NodesReady,
                       i # r,
                       lset'[i] # EmptyLS(i),
                       ~RightNeighbor(lset'[r]) \in NodesReady
                       PROVE CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, i)
                       OBVIOUS
                    <7>2. /\ r \in I
                          /\ lset[r] = lset'[r]
                          /\ lset[r] \in LSet
                          /\ RightNeighbor(lset[r]) \in I
                       BY <7>1, <5>1, ReadyInI, <4>21, RightNeighborType
                    <7>3. ~(Cardinality(NodesReady) =< 1)
                       BY <5>3, <7>1, ReadyInI, CardinalityTwoNode
                    <7>4. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                       BY <7>3, <7>1, <5>3, LemmaCLtoNC DEF NeighborClosest, Inv
                    <7>9. QED BY <7>4, <7>1, <7>2
                 <6>5. \A r \in NodesReady\{i}, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, k)
                    BY <5>1, <5>0, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
                 <6>9. QED BY <6>3, <6>4, <6>5, <5>1, <5>0, ReadyNeighborRight DEF ReadyNeighborRight, Inv
             <5>4. CASE ~i \in NodesReady
                <6>2. lset[i] # EmptyLS(i)
                   BY <3>1, <5>1, <5>4
                <6>4. \A r \in NodesReady: lset'[r] = lset[r]
                   BY <5>4, <5>1, ReadyInI
                <6>9. QED BY <3>1,<5>0,<6>2,<5>4, <5>1, ReadyInI DEF ReadyNeighborRight, Inv
             <5>9. QED BY <5>3,<5>4

        <4>57. ReadyNeighborLeft'
         \*   OMITTED similar to <4>56
             <5>0. NodesReady' = NodesReady
                BY <4>07 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  LeftNeighbor(lset'[i]) \in I
                   /\ LeftNeighbor(lset[i]) \in I
                BY <4>7, <4>16, <4>3, <4>2,<4>08, <4>22,  LeftNeighborType
             <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. CASE i \in NodesReady
                <6>3. \A k \in I:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~LeftNeighbor(lset'[i]) \in NodesReady 
                      => CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(k, i) 
                    <7>1. SUFFICES ASSUME NEW k \in I,
                           NEW l \in I,
                           l = LeftNeighbor(lset[i]),
                           k # l,
                           i # k,
                           lset[k] # EmptyLS(k),
                           ~LeftNeighbor(lset'[i]) \in NodesReady
                       PROVE CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(k, i) 
                       BY <7>1, ReadyInI, <5>1, <5>2 DEF ReadyNeighborRight, Inv
                    <7>2. CASE  LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                       BY <7>1,<7>2,<5>3, ReadyInI DEF ReadyNeighborLeft, Inv
                    <7>3. CASE  LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       <8>3. CASE l \in NodesReady
                          <9>1. SUFFICES ASSUME  CwDist(k, i) < CwDist(LeftNeighbor(lset'[i]), i)
                                PROVE FALSE
                                BY <7>1, <5>1, IisNat, Unequality2,CwDistType         
                          <9>2. LeftNeighbor(lset'[i]) # k
                             BY <9>1, IisNat, CwDistType, <5>1, <7>1, Unequality23
                          <9>3. CwDist(k, i) < CwDist(l, i)
                             BY <9>1, <5>2, TransLESS, <5>1, <7>1, IisNat, CwDistType
                          <9>4. lset[k] # EmptyLS(k)
                             BY <7>1
                          <9>41. i # l
                             BY CwDistZeroCompareCo, <9>3, IisNat
                          <9>5. \/ toj[i] = k /\ toj[l] = l
                                \/ toj[l] = k /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>4, <9>41, <9>3, NoCCJoinL DEF NoCCJoinL, Inv
                          <9>6. CwDist(LeftNeighbor(lset'[i]), i) < CwDist(l, i)
                             \*BY <5>2, <7>3, <5>1, <7>1, CwDistType, IisNat, Unequality6, CwDistInjectivity,
                             <10>2. /\ CwDist(LeftNeighbor(lset'[i]), i) \in Nat
                                    /\ CwDist(l, i) \in Nat
                                 BY <5>1, <7>1, CwDistType, IisNat
                             <10>8. CwDist(LeftNeighbor(lset'[i]), i) # CwDist(l, i)
                                 BY <5>1, <7>1, <7>3, IisNat, CwDistInjectivity
                             <10>9. QED BY <10>2, <10>8, <5>2, <7>1, Unequality6  
                          <9>7. lset[LeftNeighbor(lset'[i])] # EmptyLS(LeftNeighbor(lset'[i]))
                             <10>6. /\ lset'[i] = AddToLSet({j}, lset[i])
                                    /\ j \in I
                                 BY <7>3, <4>16, LNInjectivity, <5>1, <7>3, <3>1 DEF ls1
                             <10>7. LeftNeighbor(lset'[i]) = j
                                 BY <10>6, <5>1, <7>3, AddSingleNodeChangeCo
                             <10>8. lset[j] # EmptyLS(j)
                                 BY <3>1, <4>3,SemPRplyNonEmpty DEF SemPRplyNonEmpty, Inv
                             <10>9. QED BY  <10>7,<10>8
                          <9>8. \/ /\toj[i] = LeftNeighbor(lset'[i]) 
                                   /\ toj[l] = l
                                \/ /\ toj[l] = LeftNeighbor(lset'[i]) 
                                   /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>7,<9>41, <9>6, <5>1, NoCCJoinL DEF NoCCJoinL, Inv
                          <9>9. QED BY <9>5,<9>8, <9>2, <9>3, <7>1
                       <8>4. CASE ~ l \in NodesReady
                          <9>2. CwDist(l, i) =< CwDist(k, i)
                            BY <7>1, <8>4, <5>1, <5>0,<5>3, ReadyInI
                            DEF ReadyNeighborLeft, Inv
                          <9>9. QED BY <9>2, <5>2, <5>1, <7>1,TransLEQ, CwDistType, IisNat,
                             ReadyInI  DEF TypeInvariant
                       <8>9. QED BY <8>3,<8>4
                    <7>9. QED BY <7>2,<7>3
                 <6>4. \A l \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # l
                       /\~LeftNeighbor(lset'[l]) \in NodesReady 
                      => CwDist(LeftNeighbor(lset'[l]), l) =< CwDist(i, l) 
                    <7>1. SUFFICES ASSUME NEW l \in NodesReady,
                       i # l,
                       lset'[i] # EmptyLS(i),
                       ~LeftNeighbor(lset'[l]) \in NodesReady
                       PROVE  CwDist(LeftNeighbor(lset'[l]), l) =< CwDist(i, l)
                       OBVIOUS
                    <7>2. /\ l \in I
                          /\ lset[l] = lset'[l]
                          /\ lset[l] \in LSet
                          /\ LeftNeighbor(lset[l]) \in I
                       BY <7>1, <5>1, ReadyInI, <4>21, LeftNeighborType
                    <7>3. ~(Cardinality(NodesReady) =< 1)
                       BY <5>3, <7>1, ReadyInI, CardinalityTwoNode
                    <7>4. CwDist(LeftNeighbor(lset[l]), l) =< CwDist(i, l)
                       BY <7>3, <7>1, <5>3, LemmaCLtoNC DEF NeighborClosest, Inv
                    <7>9. QED BY <7>4, <7>1, <7>2
                 <6>5. \A l \in NodesReady\{i}, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # l
                       /\~LeftNeighbor(lset'[l]) \in NodesReady 
                      =>CwDist(LeftNeighbor(lset'[l]), l) =< CwDist(k, l)
                    BY <5>1, <5>0, ReadyInI DEF ReadyNeighborLeft, Inv
                 <6>9. QED BY <6>3, <6>4, <6>5, <5>1, <5>0 DEF ReadyNeighborLeft, Inv
             <5>4. CASE ~i \in NodesReady
                <6>2. lset[i] # EmptyLS(i)
                   BY <3>1, <5>1, <5>4
                <6>4. \A l \in NodesReady: lset'[l] = lset[l]
                   BY <5>4, <5>1, ReadyInI
                <6>9. QED BY <3>1,<5>0,<6>2,<5>4, <5>1, ReadyInI DEF ReadyNeighborLeft, Inv
             <5>9. QED BY <5>3,<5>4

        <4>58. SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                 NEW nn \in I, 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "BroadcastLSet",
                 mm.mreq.lset.node = nn,
                 mm.destination = ii
                PROVE
                    /\ nn \in NodesReady'
                    /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
              BY <4>07,<4>08, SimplifyAndSolve DEF SemLRply, Inv, TypeInvariant
           <5>9. QED BY <4>10,<5>2,<4>07,<4>08 DEF SemLRply, Inv, TypeInvariant
        <4>59. LSnonEmpty'
           <5>1. SUFFICES ASSUME NEW ii \in I, NEW jj \in I,
                 ii \in GetLSetContent(lset'[jj]),
                 ~ii \in NodesReady,
                 ii # jj
                 PROVE lset'[ii] # EmptyLS(ii)
              BY <4>08 DEF LSnonEmpty, Inv
           <5>2. /\ lset[i] \in LSet
                 /\ lset[i].node = i
                 /\ lset'[i] \in LSet
                 /\ lset'[i].node = i
                 /\ j \in I
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
              BY <4>22, <4>21, <4>16, <4>2
           <5>5. CASE ii = i
              <6>0. ii \in GetLSetContent(lset[jj])
                 BY <5>1, <5>2, <5>5 DEF TypeInvariant
              <6>1. lset[i] # EmptyLS(i)
                 BY <5>5, <5>1, <6>0 DEF LSnonEmpty, Inv
              <6>2. AddToLSet({j}, lset[i]) # EmptyLS(i)
                 BY <6>1, <5>2, AddToLSetNoEmpty
              <6>3. lset'[i] # EmptyLS(i)
                  BY <6>1,<6>2,<4>16, <3>1 DEF ls1 
              <6>9. QED BY <6>3, <5>5
           <5>6. CASE jj = i
              <6>5. CASE ii \in GetLSetContent(lset[i])
                 <7>1. ii # i
                    BY <5>6, <5>1
                 <7>2. lset[ii] # EmptyLS(ii)
                    BY <5>1, <6>5, <5>6 DEF LSnonEmpty, Inv
                 <7>3. QED BY <7>2, <7>1, <5>2
              <6>6. CASE ~ii \in GetLSetContent(lset[i])     
                 <7>2. lset'[i] # lset[i]
                    BY <6>6, <5>1, NotEqualLS, <5>2, <5>6
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2,<4>13
                 <7>5. ii \in GetLSetContent(AddToLSet({j}, lset[i]))
                    BY <5>1, <5>6,<7>3
                 <7>6. ii = j
                    BY <5>2, <7>3,<7>5, <6>6, TheAddedNode 
                 <7>8. lset[j] # EmptyLS(j)
                    BY <3>1, <5>2 DEF SemPRplyNonEmpty, Inv
                 <7>9. QED BY <7>6, <7>8, <5>1, <5>2, <5>6 
              <6>9. QED BY <6>5,<6>6
           <5>8. CASE ii # i /\ jj # i
                 BY <5>8, <5>1, <5>2 DEF TypeInvariant, LSnonEmpty, Inv
           <5>9. QED BY <5>5,<5>6,<5>8
        <4>60. LeaseGrant'
           BY <4>07,<4>08 DEF LeaseGrant, Inv, TypeInvariant
        <4>61. GrantReady'
           BY <4>07,<4>08 DEF GrantReady, Inv, TypeInvariant
        <4>62. SemJoinNeighbor'
           <5>3. SUFFICES ASSUME NEW ls \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <4>10, <5>3 DEF SemJoinNeighbor, Inv
        <4>63. SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady' = {mm.mreq.lset.node}
              BY <4>1 DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>8. /\ lset[mm.mreq.lset.node] = mm.mreq.lset
                 /\ NodesReady = {mm.mreq.lset.node}
              BY <5>3, <5>4 DEF SemSingleReadyJoin, Inv
           <5>9. mm.mreq.lset.node # i
              <6>1. SUFFICES ASSUME mm.mreq.lset.node = i PROVE FALSE
                 OBVIOUS
              <6>2. lset[i] = EmptyLS(i)
                 BY <5>3, <5>8, <6>1
              <6>3. toj[i] = mm.destination
                 BY <5>3,<6>1, <5>4 DEF SemJRply, Inv
              <6>4. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                 BY <3>1 DEF SemPRplyNonEmpty, Inv
              <6>6. toj[i] = m.mreq.node
                 BY <6>2, <6>4, <6>1, <5>8, <4>2 DEF SingleReadyNodeToj, Inv
              <6>7. lset[mm.destination] = EmptyLS(mm.destination)
                 BY <5>3, <5>4 DEF SemJRply, Inv
              <6>9. QED BY <6>3,<6>4, <6>6,<6>7
           <5>10. QED BY <5>7, <5>8, <5>9, <4>22, <4>08
        <4>64. SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <4>10, <5>3
                DEF SemEmptyLSetSingleNode, Inv
        <4>65. SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY <4>1 DEF SemJoinLeafSet, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>71. mm.mreq.lset.node \in NodesReady
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>8. CASE mm.mreq.lset.node # i
              <6>1. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                 BY <4>22, <5>7, <5>8
              <6>3. QED BY <6>1, <5>3, <5>4 DEF SemJoinLeafSet, Inv
           <5>9. CASE mm.mreq.lset.node = i
              <6>0. /\ LeftNeighbor(lset[i]) = LeftNeighbor(mm.mreq.lset)
                    /\ RightNeighbor(lset[i]) = RightNeighbor(mm.mreq.lset)
                 BY <5>3, <5>4, <5>9 DEF SemJoinLeafSet, Inv
              <6>1. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(LeftNeighbor(AddToLSet({j}, lset[i])), i)
                        =< CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(LeftNeighbor(lset'[i]), lset'[i].node)
                        =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChangeCo
                 <7>51. /\ LeftNeighbor(lset'[i])  = j
                        /\ LeftNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(j, i)=< CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(j, i) # CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. SUFFICES ASSUME  CwDist(j, i) = CwDist(LeftNeighbor(lset[i]), i)
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ LeftNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. LeftNeighbor(lset[i]) = j
                       BY <8>1, <8>7, CwDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE LeftNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF SemPRplyNonEmpty, Inv
                    <8>6. LeftNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompareCo, IisNat
                    <8>7. \/ toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          \/ toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoinL, Inv
                    <8>8. CASE toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                          <10>1. CASE LeftNeighbor(mm.mreq.lset) = mm.mreq.lset.node 
                              BY <6>0, <5>9, <10>1, <8>9
                          <10>2. CASE  LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                              BY <10>2, <5>3, <5>4 DEF SemJRply, Inv
                          <10>9. QED BY <10>1,<10>2
                       <9>2.  toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. LeftNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~LeftNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv
                    <8>8. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborLeft, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                 \* OMITTED  similar to <6>1
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(i, RightNeighbor(AddToLSet({j}, lset[i])))
                        =< CwDist(i, RightNeighbor(lset[i]))
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(lset'[i].node, RightNeighbor(lset'[i]))
                        =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. RightNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChange
                 <7>51. /\ RightNeighbor(lset'[i])  = j
                        /\ RightNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(i, j)=< CwDist(i, RightNeighbor(lset[i]))
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist( i, j) # CwDist(i, RightNeighbor(lset[i]))
                    <8>1. SUFFICES ASSUME  CwDist(i, j) = CwDist(i, RightNeighbor(lset[i]))
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ RightNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. RightNeighbor(lset[i]) = j
                       BY <8>1, <8>7, RightDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE RightNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF SemPRplyNonEmpty, Inv
                    <8>6. RightNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompare, IisNat
                    <8>7. \/ toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          \/ toj[RightNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoin, Inv
                    <8>8. CASE toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[RightNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                          <10>1. CASE RightNeighbor(mm.mreq.lset) = mm.mreq.lset.node 
                              BY <6>0, <5>9, <10>1, <8>9
                          <10>2. CASE  RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                              BY <10>2, <5>3, <5>4 DEF SemJRply, Inv
                          <10>9. QED BY <10>1,<10>2
                       <9>2.  toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. RightNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~RightNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv
                    <8>8. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborRight, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>9. QED BY <6>1, <6>2, <6>0, <5>9
           <5>10. QED BY <5>7, <5>8, <5>9, <4>22
        <4>66. SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE
                     /\ lset'[mm.destination] = EmptyLS(mm.destination)
                     /\ toj'[mm.mreq.lset.node] = mm.destination
                     /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                     /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                     /\ mm.mreq.lset.node \in NodesReady
                     /\ mm.destination \in NodesWait
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
              BY <4>08 DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. /\ mm.mreq.lset.node \in I
                 /\ mm.destination \in I
              BY <5>5, <5>4 DEF JRpl, LSet, DMsg, TypeInvariant
           <5>71. /\mm.mreq.lset.node \in NodesReady
                  /\ mm.destination \in NodesWait
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>72.   /\ LeftNeighbor(mm.mreq.lset) \in I
                    /\ RightNeighbor(mm.mreq.lset) \in I
                    /\ mm.mreq.lset.node \in I
                    /\ \A k \in I: toj'[k] = toj[k]
                    /\ \A k \in I: lease'[k] = lease[k]
              BY <5>6, <5>7, LeftNeighborType, RightNeighborType, <4>07 DEF TypeInvariant
           <5>8. CASE mm.mreq.lset.node # i
              <6>1. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                 BY <4>22, <5>7, <5>8
              <6>2. mm.destination # i
                 <7>1. SUFFICES ASSUME mm.destination = i PROVE FALSE
                    OBVIOUS
                 <7>2. lset[i] = EmptyLS(i)
                    BY <7>1, <5>4, <5>3 DEF SemJRply, Inv
                 <7>3. mm.destination \in NodesReady
                    BY <3>1, <7>2, <7>1
                 <7>4. ~ mm.destination \in NodesReady
                    BY <5>3, <5>4, <5>7, StatusDisjoint DEF SemJRply, Inv
                 <7>9. QED BY <7>3, <7>4
              <6>3.  lset'[mm.destination] =  lset[mm.destination]
                 BY  <4>22, <5>7, <6>2
              <6>9. QED BY <6>1, <6>3, <5>72, <5>3, <5>4, SlowestZenon DEF SemJRply, Inv
           <5>9. CASE mm.mreq.lset.node = i
              <6>0. /\ LeftNeighbor(lset[i]) = LeftNeighbor(mm.mreq.lset)
                    /\ RightNeighbor(lset[i]) = RightNeighbor(mm.mreq.lset)
                 BY <5>3, <5>4, <5>9 DEF SemJoinLeafSet, Inv
              <6>1. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(LeftNeighbor(AddToLSet({j}, lset[i])), i)
                        =< CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(LeftNeighbor(lset'[i]), lset'[i].node)
                        =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChangeCo
                 <7>51. /\ LeftNeighbor(lset'[i])  = j
                        /\ LeftNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(j, i)=< CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(j, i) # CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. SUFFICES ASSUME  CwDist(j, i) = CwDist(LeftNeighbor(lset[i]), i)
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ LeftNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. LeftNeighbor(lset[i]) = j
                       BY <8>1, <8>7, CwDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE LeftNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF SemPRplyNonEmpty, Inv
                    <8>6. LeftNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompareCo, IisNat
                    <8>7. \/ toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          \/ toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoinL, Inv
                    <8>8. CASE toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                          BY  <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. LeftNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~LeftNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv
                    <8>8. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborLeft, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                 \* OMITTED \* similar to <6>1
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(i, RightNeighbor(AddToLSet({j}, lset[i])))
                        =< CwDist(i, RightNeighbor(lset[i]))
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(lset'[i].node, RightNeighbor(lset'[i]))
                        =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. RightNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChange
                 <7>51. /\ RightNeighbor(lset'[i])  = j
                        /\ RightNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(i, j)=< CwDist(i, RightNeighbor(lset[i]))
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(i, j) # CwDist(i, RightNeighbor(lset[i]))
                    <8>1. SUFFICES ASSUME  CwDist(i, j) = CwDist(i, RightNeighbor(lset[i]))
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ RightNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. RightNeighbor(lset[i]) = j
                       BY <8>1, <8>7, RightDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE RightNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF SemPRplyNonEmpty, Inv
                    <8>6. RightNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompare, IisNat
                    <8>7. \/ toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          \/ toj[RightNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoin, Inv
                    <8>8. CASE toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[RightNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                          BY  <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. RightNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~RightNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv
                    <8>8. CwDist(i, RightNeighbor(lset[i])) =< CwDist( i, j)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborRight, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8                 
              <6>3. Covers(lset'[i], mm.destination)
                 <7>1. Covers(lset[i], mm.destination)
                    BY <5>3, <5>4, <5>9 DEF SemJRply, Inv
                 <7>2. QED BY <7>1, <6>1,<6>2, <5>7,<4>16,<4>21, CoverNeighborRel
              <6>4. toj'[i] = mm.destination
                 BY <5>3,<5>4, <5>72, <5>9, <6>0 DEF SemJRply, Inv
              <6>5. /\ RightNeighbor(lset'[i]) \in lease'[i]
                    /\ LeftNeighbor(lset'[i]) \in lease'[i]
                 <7>1. /\ RightNeighbor(lset[i]) \in lease[i]
                       /\ LeftNeighbor(lset[i]) \in lease[i]
                    BY <5>3,<5>4, <5>9 DEF SemJRply, Inv
                 <7>2. QED BY <6>1, <6>2,  <5>72, <7>1
              <6>6. mm.destination # i
                 BY <5>9,<5>71, StatusDisjoint
              <6>7.  lset'[mm.destination] =  lset[mm.destination]
                 BY  <4>22, <5>7, <6>6 
              <6>8. lset'[mm.destination]  = EmptyLS(mm.destination)
                 BY <6>7, <5>3,<5>4 DEF SemJRply, Inv
              <6>9. ASSUME LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                    PROVE toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                 BY <5>3,<5>4, <5>72, <5>9, <6>0, <6>9 DEF SemJRply, Inv
              <6>19. ASSUME RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                    PROVE  toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                 BY <5>3,<5>4, <5>72, <5>9, <6>0, <6>19 DEF SemJRply, Inv
              <6>99. QED BY <6>4, <6>5, <6>3, <6>8, <5>71, <5>9, <6>9, <6>19
           <5>10. QED BY <5>8, <5>9
        <4>67. GrantNeighbor'
           <5>1. SUFFICES ASSUME 
                 NEW kk \in NodesReady, 
                 NEW ii \in I, 
                 ~ii \in NodesReady,
                 ii \in grant[kk],
                 ii # LeftNeighbor(lset'[kk]),
                 ii # RightNeighbor(lset'[kk])
                 PROVE FALSE
              BY <4>07, <4>08, ReadyInI 
              DEF TypeInvariant, GrantNeighbor, Inv
           <5>2. /\ lset[kk] \in LSet
                 /\ lset[kk].node = kk
                 /\ j \in I
                 /\ {j} \in SUBSET I
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
                 /\ lset'[i].node = i
                 /\ lset'[i] \in LSet
                 /\ lset[i].node = i
                 /\ lset[i] \in LSet
              BY <4>16, <4>21,<4>2, <4>22, ReadyInI DEF TypeInvariant
           <5>3. \/ ii = LeftNeighbor(lset[kk])
                 \/ ii = RightNeighbor(lset[kk])
              BY <5>1, <5>2 DEF GrantNeighbor, Inv
           <5>5. CASE kk = i
              <6>5. CASE ii = LeftNeighbor(lset[i])
                 <7>1. SUFFICES ASSUME NEW k \in I,
                       k = LeftNeighbor(lset'[i]),
                       k # ii
                       PROVE FALSE
                    BY <5>1, <5>2,<5>5, LeftNeighborType
                 <7>2. lset'[i] = AddToLSet({j}, lset[i])
                    BY <4>03, <7>1, <6>5, LeftNeighborType, <5>2 DEF TypeInvariant, ls1   
                 <7>3. CwDist(LeftNeighbor(lset'[i]), lset'[i].node) =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                    BY AddToLSetInvCo,<5>2,<7>2, <5>5
                 <7>4. CwDist(k,i) =< CwDist(ii, i)
                    BY  <7>3, <7>1, <6>5,<5>5, <5>2
                 <7>5. CwDist(k,i) < CwDist(ii, i)
                       BY <7>4, <7>1, CwDistInjectivity, Unequality6,
                         CwDistType, IisNat
                 <7>6. k = j
                    BY <7>2, <7>1, <6>5, <5>2, <5>5, AddSingleNodeChangeCo, SimplifyAndSolve
                 <7>7. /\ ~ii \in NodesReady
                       /\ k # i
                       /\ lset[k] # EmptyLS(k)
                    <8>1. /\ lset[k] \in LSet
                          /\ lset[k].node = k
                       BY DEF TypeInvariant
                    <8>2. lset[k]# EmptyLS(k)
                       BY <7>6, <3>1 DEF SemPRplyNonEmpty,Inv
                    <8>9. QED BY <8>1, <8>2, <7>6, <3>1, <5>1
                 <7>8. CwDist(ii, i) =< CwDist(k, i)
                    BY <7>7, <5>1, <5>5, <6>5 DEF ReadyNeighborLeft, Inv
                 <7>9. QED BY <7>5,<7>8, Unequality2, CwDistType, ReadyInI, IisNat
              <6>6. CASE ii = RightNeighbor(lset[i])
                 <7>1. SUFFICES ASSUME NEW k \in I,
                       k = RightNeighbor(lset'[i]),
                       k # ii
                       PROVE FALSE
                    BY <5>1, <5>2,<5>5, RightNeighborType
                 <7>2. lset'[i] = AddToLSet({j}, lset[i])
                    BY <4>03, <7>1, <6>6, RightNeighborType, <5>2 DEF TypeInvariant, ls1   
                 <7>3. CwDist(lset'[i].node, RightNeighbor(lset'[i])) =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                    BY AddToLSetInvCo,<5>2,<5>5,<7>2
                 <7>4. CwDist(i, k) =< CwDist(i, ii)
                    BY  <7>3, <7>1, <6>6,<5>5, <5>2
                 <7>5. CwDist(i, k) < CwDist(i, ii)
                       BY <7>4, <7>1, RightDistInjectivity, Unequality6,
                         CwDistType, IisNat
                 <7>6. k = j
                    BY <7>2, <7>1, <6>6, <5>2, <5>5, AddSingleNodeChange, SimplifyAndSolve
                 <7>7. /\ ~ii \in NodesReady
                       /\ k # i
                       /\lset[k] # EmptyLS(k)
                    <8>1. /\ lset[k] \in LSet
                          /\ lset[k].node = k
                       BY DEF TypeInvariant
                    <8>2. lset[k]# EmptyLS(k)
                       BY <7>6, <3>1 DEF SemPRplyNonEmpty,Inv
                    <8>9. QED BY <8>1, <8>2, <7>6, <3>1, <5>1
                 <7>8. CwDist(i, ii) =< CwDist(i, k)
                    BY <7>7, <5>1, <5>5, <6>6 DEF ReadyNeighborRight, Inv
                 <7>9. QED BY <7>5,<7>8, Unequality2, CwDistType, ReadyInI, IisNat
              <6>9. QED BY <5>5, <6>6,<6>5,<5>3
           <5>7. CASE kk # i
              BY <5>1, <5>2, <5>3, <5>7, ReadyInI
           <5>9. QED BY <5>5, <5>7
        <4>68. SingleReadyNode'
           <5>3. CASE i \in NodesReady /\ NodesReady = {i}
              <6>1. /\\A k \in I: /\ lset'[k] \in LSet 
                                  /\ lset[k] \in LSet
                                  /\ lset[k].node = k
                    /\ j \in I
                    /\ LeftNeighbor(lset[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                    /\ i # j
                 BY <3>1, <4>21, <4>2, LeftNeighborType, RightNeighborType DEF TypeInvariant
              <6>2. CASE lset'[i] = AddToLSet({j}, lset[i]) /\ LeftNeighbor(lset[i]) # i
                 <7>1. LeftNeighbor(lset[i]) = RightNeighbor(lset[i])
                    BY <5>3 DEF SingleReadyNode, Inv
                 <7>2. lset[j] # EmptyLS(j)
                       BY <3>1 DEF SemPRplyNonEmpty, Inv
                 <7>7. LeftNeighbor(lset'[i]) = j
                    <8>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # j 
                                   PROVE FALSE
                       OBVIOUS
                    <8>2. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                       <9>1. LeftNeighbor(AddToLSet({j}, lset[i])) # j
                          BY <8>1, <6>2
                       <9>2. LeftNeighbor(AddToLSet({j}, lset[i]))
                             = LeftNeighbor(lset[i])
                          BY <9>1, AddSingleNodeChangeCo, <6>1 
                       <9>3. QED BY <9>2, <6>2
                    <8>3. RightNeighbor(lset[i]) # j
                       BY <7>1, <8>1, <8>2
                    <8>4. RightNeighbor(lset[i]) # i
                       BY <6>2, <7>1
                    <8>5. CASE CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~RightNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4
                       <9>4. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborRight, Inv
                       <9>9. QED BY <9>4, <8>5, IisNat, Unequality2, <6>1, CwDistType
                    <8>6. CASE CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~LeftNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4, <7>1
                       <9>4. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborLeft, Inv
                       <9>9. QED BY <9>4, <8>6, IisNat, Unequality2, <6>1, CwDistType
                    <8>9. QED BY <6>1, <7>1, <8>4, <8>5,<8>6,<8>3,Ring3NodesPos7
                 <7>8. RightNeighbor(lset'[i]) = j
                    \* OMITTED \* Similart to <7>7
                    <8>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # j 
                                   PROVE FALSE
                       OBVIOUS
                    <8>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                       <9>1. RightNeighbor(AddToLSet({j}, lset[i])) # j
                          BY <8>1, <6>2
                       <9>2. RightNeighbor(AddToLSet({j}, lset[i]))
                             = RightNeighbor(lset[i])
                          BY <9>1, AddSingleNodeChange, <6>1 
                       <9>3. QED BY <9>2, <6>2
                    <8>3. RightNeighbor(lset[i]) # j
                       BY <7>1, <8>1, <8>2
                    <8>4. RightNeighbor(lset[i]) # i
                       BY <6>2, <7>1
                    <8>5. CASE CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~RightNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4
                       <9>4. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborRight, Inv
                       <9>9. QED BY <9>4, <8>5, IisNat, Unequality2, <6>1, CwDistType
                    <8>6. CASE CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~LeftNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4, <7>1
                       <9>4. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborLeft, Inv
                       <9>9. QED BY <9>4, <8>6, IisNat, Unequality2, <6>1, CwDistType
                    <8>9. QED BY <6>1, <7>1, <8>4, <8>5,<8>6,<8>3,Ring3NodesPos7
                 <7>9. QED BY <7>7,<7>8, ReadyInI, <4>07,<4>08, <4>22 
                  DEF SingleReadyNode, Inv, TypeInvariant
              <6>3. CASE LeftNeighbor(lset[i]) = i /\ lset'[i] = AddToLSet({j}, lset[i])
                 <7>1. RightNeighbor(lset[i]) = LeftNeighbor(lset[i])
                    BY <5>3 DEF SingleReadyNode, Inv
                 <7>2. RightNeighbor(lset[i]) = i
                    BY <6>3, <7>1
                 <7>. HIDE DEF j
                 <7>5. lset[i] = EmptyLS(i)
                    BY <6>3, <7>2, EmptyLSNoNeighbor, <6>1 
                 <7>6. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY AddSingleNodeToEmpty, <7>5, <6>1
                 <7>7. RightNeighbor(AddToLSet({j}, lset[i])) = j 
                    BY AddSingleNodeToEmpty, <7>5, <6>1
                 <7>8. /\ LeftNeighbor(lset'[i]) = j
                       /\ RightNeighbor(lset'[i]) = j
                    BY <7>6, <7>7, <6>3 
                 <7>9. QED BY <7>8, <4>08,<5>3 DEF SingleReadyNode, Inv
              <6>5. CASE lset'[i] = lset[i]
                 BY <6>5, <5>3, ReadyInI, <4>07,<4>08, <4>22 
                  DEF SingleReadyNode, Inv, TypeInvariant
              <6>9. QED BY <6>2, <6>3, <6>5, <4>13
           <5>4. CASE ~ i \in NodesReady \/ NodesReady # {i}
              BY <5>4, ReadyInI, <4>07,<4>08, <4>22 DEF SingleReadyNode, Inv, TypeInvariant
           <5>9. QED BY <5>3,<5>4
        <4>69. CoverToJ'
           <5>0. SUFFICES ASSUME NEW rr \in NodesReady',NEW kk \in I
                 PROVE CoverToJ!(rr, kk)'
              BY DEF CoverToJ
           <5>1. /\ NodesReady' = NodesReady
                 /\ \A k \in I: toj'[k] = toj[k]
                 /\ \A k \in I\ {i}: lset'[k] = lset[k]
                 /\ \A k \in I: lset[k] \in LSet /\ lset'[k] \in LSet
                 /\ j \in I
                 /\ rr \in I
              BY <4>22, <4>21, <4>08, <4>07, <4>2, ReadyInI DEF TypeInvariant
           <5>5. CASE rr = i
              <6>1. SUFFICES ASSUME 
                    lset'[kk] # EmptyLS(kk),
                    Covers(lset'[i], kk),
                    kk # i
                    PROVE kk = toj[i]
                 BY <5>1, <5>0, <5>5 DEF CoverToJ, Inv
              <6>3. Covers(lset[i], kk)
                 <7>3. CASE lset'[i] = lset[i]
                    BY <6>1, <7>3
                 <7>4. CASE lset'[i] = AddToLSet({j}, lset[i])
                    <8>1. Covers(AddToLSet({j}, lset[i]), kk)
                       BY <6>1, <7>4, <5>1
                    <8>2. /\ lset[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ kk \in I
                       BY <5>1, <5>0
                    <8>9. QED BY <8>1, AddToLSetCover, <8>2
                 <7>5. QED BY <4>13, <7>3,<7>4
              <6>9. QED BY <6>3, <5>1, <5>5, <6>1 DEF CoverToJ, Inv 
           <5>6. CASE kk = i 
              <6>1. SUFFICES ASSUME 
                    lset'[i] # EmptyLS(i),
                    Covers(lset'[rr], i),
                    rr # i
                    PROVE i = toj[rr]
                 BY <5>1, <5>0, <5>5 DEF CoverToJ, Inv
              <6>3. ~i \in NodesReady
                 <7>1. SUFFICES ASSUME i \in NodesReady
                       PROVE FALSE
                    OBVIOUS
                 <7>2. /\ rr \in NodesReady
                       /\ i # rr
                       /\ lset[rr] \in LSet
                       /\ lset[rr].node = rr
                     BY <5>0, <5>1, <6>1 DEF TypeInvariant
                 <7>3. Covers(lset[rr], i)
                    BY <6>1, <5>1
                 <7>5. ~ (Cardinality(NodesReady) =< 1)
                    BY <7>1,<7>2, CardinalityTwoNode, ReadyInI
                 <7>7. CASE  CwDist(rr, i) < CwDist(rr, RightNeighbor(lset[rr]))
                    <8>6. CwDist(rr, RightNeighbor(lset[rr])) =< CwDist(rr, i)
                       BY <7>5, <7>1,<7>2, LemmaCLtoNC DEF NeighborClosest, Inv  
                    <8>9. QED BY <7>7,<8>6, <5>1, Unequality2, 
                     RightNeighborType, IisNat, CwDistType, ReadyInI
                 <7>8. CASE  CwDist(i, rr) < CwDist(LeftNeighbor(lset[rr]), rr)
                    <8>6. CwDist(LeftNeighbor(lset[rr]), rr) =< CwDist(i, rr)
                       BY <7>5, <7>1,<7>2, LemmaCLtoNC DEF NeighborClosest, Inv  
                    <8>9. QED BY <7>8,<8>6, <5>1, Unequality2, 
                     LeftNeighborType, IisNat, CwDistType, ReadyInI
                 <7>9. QED BY <7>7,<7>8,<7>3,NeighborCoverageCo, <6>1, <7>2
              <6>9. QED BY <3>1, <6>1, <6>3, <5>1 DEF CoverToJ, Inv
           <5>7. CASE rr # i /\  kk # i
              BY <5>7, <5>0, <5>1 DEF CoverToJ, Inv
           <5>9. QED BY <5>5,<5>6,<5>7
        <4>70. SemPRplyNonEmpty'
           <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 \/ mm.mreq.type = "LSProbeReply"
                 \/ mm.mreq.type = "LSProbe"
                 PROVE 
                    lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
              BY DEF SemPRplyNonEmpty, Inv
           <5>4. CASE mm \in receivedMsgs 
              <6>1.  mm.mreq \in Prb \/ mm.mreq \in PRpl
                 <7>1. CASE mm.mreq.type = "LSProbeReply"
                    BY SlowestZenon,<7>1, <5>4 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                 <7>2. CASE mm.mreq.type = "LSProbe"
                      BY SlowestZenon,<7>2, <5>4 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                 <7>9. QED BY <7>1,<7>2,<5>1
              <6>4. m.mreq \in Prb
                 BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
              <6>5. /\ mm.mreq.node \in I
                    /\ lset[i].node = i   
                    /\ m.mreq.node # i
                    /\ m.mreq.node \in I
                    /\ lset[i] \in LSet
                 BY <3>1, <4>21,<6>1, <6>4 DEF Prb, PRpl
              <6>6. lset[mm.mreq.node] # EmptyLS(mm.mreq.node)
                 BY <5>4, <5>1 DEF SemPRplyNonEmpty, Inv
              <6>7. CASE mm.mreq.node = i
                 <7>7. CASE lset'[i] = lset[i]
                    BY <6>6, <6>7, <7>7
                 <7>8. CASE lset'[i] = AddToLSet({m.mreq.node}, lset[i])
                    BY <6>6, <6>7, <7>8, <6>5, AddSingleNodeNoEmpty
                 <7>9. QED BY <7>7,<7>8, <4>13
              <6>8. CASE mm.mreq.node # i
                 BY <4>22, <6>6,<6>8, <6>5
              <6>9. QED BY <6>7,<6>8
           <5>5. CASE ~mm \in receivedMsgs
              <6>. HIDE DEF ls1, fi, prb, newm
              <6>01. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                   { [destination |-> j, mreq |-> newm]} 
                   \cup Probe(i, ls1, fi, prb)
                 BY <4>06, <4>3 DEF TypeInvariant
              <6>02. [destination |-> j, mreq |-> newm].mreq.type = "LSProbeReply"
                 BY <3>1, <4>6 DEF PRpl
              <6>03. \A ms \in Probe(i, ls1, fi, prb): ms.mreq.type = "LSProbe"
                 <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb)
                    BY ONLY ProbeMsgType, <4>3, <4>7
                 <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
              <6>04. \/ mm \in Probe(i, ls1, fi, prb)
                     \/ mm = [destination |-> j, mreq |-> newm]
                 BY <5>5, <5>1, <6>01
              <6>1. mm.mreq \in Prb \/ mm.mreq \in PRpl 
                 <7>1. CASE mm \in Probe(i, ls1, fi, prb)
                    <8>0. /\ ls1 \in LSet
                          /\ fi \in SUBSET I
                          /\ prb \in SUBSET I
                       BY <4>3,<4>7
                    <8>1. mm.mreq.type = "LSProbe"
                       BY <6>03, <7>1
                    <8>2. mm \in DMsg
                       BY <7>1, <8>0, ProbeType
                    <8>3. mm.mreq \in Prb
                        BY SlowestZenon,<8>1, <8>2
                           DEF TypeInvariant, DMsg, 
                          MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                    <8>9. QED BY <8>3
                 <7>2. CASE mm = [destination |-> j, mreq |-> newm]
                    <8>1. newm = [type |-> "LSProbeReply",  
                          node |-> i,
                          lset |-> lset[i],
                          failed |-> fi]
                       BY DEF newm
                    <8>3. QED BY <7>2, <5>4, <8>1 
                       DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR, fi
                 <7>9. QED BY <7>1,<7>2, <6>04
              <6>2. mm.mreq.node = i 
                 <7>1. CASE mm \in Probe(i, ls1, fi, prb)
                    <8>1. ProbeMsgTypeConclusion(i, ls1, fi, prb)
                    BY ONLY ProbeMsgType, <4>3, <4>7
                    <8>2. QED BY <8>1, <7>1 DEF ProbeMsgTypeConclusion
                 <7>2. CASE mm = [destination |-> j, mreq |-> newm]
                    <8>1. newm = [type |-> "LSProbeReply",  
                          node |-> i,
                          lset |-> lset[i],
                          failed |-> fi]
                       BY DEF newm
                    <8>2. newm.node = i
                       BY <8>1
                    <8>3. mm.mreq = newm
                       BY <7>2
                    <8>9. QED BY <8>2, <7>2, <8>3
                 <7>9. QED BY <7>1,<7>2, <6>04
              <6>4. m.mreq \in Prb
                 BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
              <6>5. /\ mm.mreq.node \in I
                    /\ lset[i].node = i   
                    /\ m.mreq.node # i
                    /\ m.mreq.node \in I
                    /\ lset[i] \in LSet
                 BY <3>1, <4>21,<6>1, <6>4 DEF Prb, PRpl
              <6>7. CASE j \in GetLSetContent(lset[i])
                 <7>2. j # i
                    BY <3>1
                 <7>8. lset[i] # EmptyLS(i)
                    BY <6>7, <7>2,<6>5, NonEmptyLSet
                 <7>9. QED BY <7>8, <6>2,<6>7, <4>15 
              <6>8. CASE ~j \in GetLSetContent(lset[i])
                    BY <6>2, <6>8, <6>5,<4>15, AddSingleNodeNoEmpty
              <6>9. QED BY <6>7,<6>8, <4>13
           <5>9. QED BY <5>4,<5>5
        <4>71. NoCCJoin'
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW rr \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoin!(ii,rr,kk)'
             BY DEF NoCCJoin
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ j \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I \{i}: lset'[k] = lset[k]
                /\  RightNeighbor(lset'[i]) \in I
                /\ RightNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ rr \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk         
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ j \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I \{i}: lset'[k] = lset[k]
                    /\  RightNeighbor(lset'[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                 BY <4>7, <4>16, <4>3, <4>2,<4>08, <4>22,  RightNeighborType, ReadyInI DEF  TypeInvariant
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>07 DEF TypeInvariant, NodesReady
              <6>3. /\ ii \in I
                    /\ rr \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3   
           <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1, <6>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
           <5>5. CASE ii = i
              <6>1. SUFFICES ASSUME i # kk,
                    lset[kk] # EmptyLS(kk),
                    rr = RightNeighbor(lset'[i]),
                    i # rr,
                    CwDist(i, kk) < CwDist(i, rr)
                    PROVE \/ toj[i] = kk  /\ toj[rr] = rr
                          \/ toj[rr] = kk /\ toj[i] = i
                 BY <5>5, <5>0, <5>1  DEF NoCCJoin
              <6>2. CwDist(i, kk) < CwDist(i, RightNeighbor(lset[i]))
                 BY <5>1, <5>2, <6>1, TransLESS, CwDistType, IisNat
              <6>3. rr = RightNeighbor(lset[i]) 
                 <7>1. SUFFICES ASSUME rr # RightNeighbor(lset[i]) 
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CASE RightNeighbor(lset[i])  \in NodesReady
                    <8>2. ~(Cardinality(NodesReady) =< 1)
                       BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <7>2, <7>1, <5>5
                    <8>3. CwDist(i, RightNeighbor(lset[i])) =<  CwDist(i, rr)
                       BY <8>2, <5>5, <5>1, <7>1, <5>0, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>4. CwDist(i, RightNeighbor(lset[i])) <  CwDist(i, rr)
                       BY <8>3, RightDistInjectivity, Unequality6, <7>1, <5>1, IisNat, CwDistType
                    <8>5. QED BY <8>4, <5>2, <6>1, CwDistType, <5>1, IisNat, Unequality2
                 <7>3. CASE ~RightNeighbor(lset[i])  \in NodesReady
                    <8>1. /\ i \in NodesReady
                          /\ kk \in I
                          /\ kk # i
                       BY <5>0, <5>1, <5>5, <6>1
                    <8>2. lset[kk] # EmptyLS(kk)
                       BY <6>1
                    <8>4. CwDist(i, RightNeighbor(lset[i]) ) =< CwDist(i, kk)
                       BY ReadyNeighborRight, <8>1, <8>2, <7>3 DEF ReadyNeighborRight, Inv
                    <8>5. QED BY <6>2, <8>4, Unequality2, CwDistType, IisNat, <5>1
                 <7>9. QED BY <7>3,<7>2
              <6>9. QED BY <5>1, <5>5, <6>1, <6>2, <6>3, NoCCJoin DEF NoCCJoin, Inv
           <5>6. CASE rr = i
              <6>1. SUFFICES ASSUME ii # kk,
                     lset'[kk] # EmptyLS(kk),
                     i = RightNeighbor(lset[ii]),
                     ii # i,
                     CwDist(ii, kk) < CwDist(ii, i)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[i] = i
                     \/ toj[i] = kk /\ toj[ii] = ii
                  BY <5>6, <5>0, <5>1 DEF NoCCJoin, Inv
              <6>2. kk # i
                 BY <6>1, <5>1, IisNat, CwDistZeroCompare
              <6>9.QED BY <6>1, <6>2, <5>0, <5>6, <5>1, NoCCJoin DEF  NoCCJoin, Inv
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     rr = RightNeighbor(lset[ii]),
                     ii # rr,
                     CwDist(ii, i) < CwDist(ii, rr)
                     PROVE 
                      \/ toj[ii] = i /\ toj[rr] = rr
                      \/ toj[rr] = i /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1  DEF NoCCJoin, Inv
              <6>2. ~(Cardinality(NodesReady) =< 1)
                 BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <6>1
              <6>7. ~i \in NodesReady
                 <7>1. SUFFICES ASSUME i \in NodesReady
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CwDist(ii, rr) =<  CwDist(ii, i)
                 BY <6>1, <5>1, <6>2,<7>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>2, <6>1, <5>1, IisNat, Unequality2, CwDistType
              <6>8. lset[i] # EmptyLS(i)
                 BY <6>7, <3>1
              <6>9. QED BY <6>1, <6>8, <5>1,  NoCCJoin DEF NoCCJoin, Inv
           <5>8. CASE ii # i /\ rr # i /\ kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    rr = RightNeighbor(lset[ii]),
                    ii # rr,
                    CwDist(ii, kk) < CwDist(ii, rr)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[rr] = rr
                     \/ toj[rr] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>8 DEF NoCCJoin
              <6>. QED BY <5>1,<6>1, NoCCJoin DEF NoCCJoin, Inv
           <5>. QED
             BY <5>5, <5>6, <5>7, <5>8
        <4>72. NoCCJoinL'
           \* OMITTED \* similar to <4>71
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW ll \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoinL!(ii,ll,kk)'
             BY DEF NoCCJoinL
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ j \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I \{i}: lset'[k] = lset[k]
                /\  LeftNeighbor(lset'[i]) \in I
                /\ LeftNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ ll \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk         
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ j \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I \{i}: lset'[k] = lset[k]
                    /\  LeftNeighbor(lset'[i]) \in I
                    /\ LeftNeighbor(lset[i]) \in I
                 BY <4>7, <4>16, <4>3, <4>2,<4>08, <4>22,  LeftNeighborType, ReadyInI DEF  TypeInvariant
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>07 DEF TypeInvariant, NodesReady
              <6>3. /\ ii \in I
                    /\ ll \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3   
           <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1, <6>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
           <5>5. CASE ii = i
              <6>1. SUFFICES ASSUME i # kk,
                    lset[kk] # EmptyLS(kk),
                    ll = LeftNeighbor(lset'[i]),
                    i # ll,
                    CwDist(kk, i) < CwDist(ll, i)
                    PROVE \/ toj[i] = kk  /\ toj[ll] = ll
                          \/ toj[ll] = kk /\ toj[i] = i
                 BY <5>5, <5>0, <5>1  DEF NoCCJoinL
              <6>2. CwDist(kk, i) < CwDist(LeftNeighbor(lset[i]), i)
                 BY <5>1, <5>2, <6>1, TransLESS, CwDistType, IisNat
              <6>3. ll = LeftNeighbor(lset[i]) 
                 <7>1. SUFFICES ASSUME ll # LeftNeighbor(lset[i]) 
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CASE LeftNeighbor(lset[i])  \in NodesReady
                    <8>2. ~(Cardinality(NodesReady) =< 1)
                       BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <7>2, <7>1, <5>5
                    <8>3. CwDist(LeftNeighbor(lset[i]), i) =<  CwDist(ll, i)
                       BY <8>2, <5>5, <5>1, <7>1, <5>0, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>4. CwDist(LeftNeighbor(lset[i]), i) <  CwDist(ll, i)
                       BY <8>3, CwDistInjectivity, Unequality6, <7>1, <5>1, IisNat, CwDistType
                    <8>5. QED BY <8>4, <5>2, <6>1, CwDistType, <5>1, IisNat, Unequality2
                 <7>3. CASE ~LeftNeighbor(lset[i])  \in NodesReady
                    <8>1. /\ i \in NodesReady
                          /\ kk \in I
                          /\ kk # i
                       BY <5>0, <5>1, <5>5, <6>1
                    <8>2. lset[kk] # EmptyLS(kk)
                       BY <6>1
                    <8>4. CwDist(LeftNeighbor(lset[i]), i ) =< CwDist(kk, i)
                       BY <8>1, <8>2, <7>3 DEF ReadyNeighborLeft, Inv
                    <8>5. QED BY <6>2, <8>4, Unequality2, CwDistType, IisNat, <5>1
                 <7>9. QED BY <7>3,<7>2
              <6>9. QED BY <5>1, <5>5, <6>1, <6>2, <6>3, NoCCJoinL DEF NoCCJoinL, Inv
           <5>6. CASE ll = i
              <6>1. SUFFICES ASSUME ii # kk,
                     lset'[kk] # EmptyLS(kk),
                     i = LeftNeighbor(lset[ii]),
                     ii # i,
                     CwDist(kk, ii) < CwDist(i, ii)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[i] = i
                     \/ toj[i] = kk /\ toj[ii] = ii
                  BY <5>6, <5>0, <5>1 DEF NoCCJoinL, Inv
              <6>2. kk # i
                 BY <6>1, <5>1, IisNat, CwDistZeroCompare
              <6>9.QED BY <6>1, <6>2, <5>0, <5>6, <5>1 DEF  NoCCJoinL, Inv
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     ll = LeftNeighbor(lset[ii]),
                     ii # ll,
                     CwDist(i, ii) < CwDist(ll, ii)
                     PROVE 
                      \/ toj[ii] = i /\ toj[ll] = ll
                      \/ toj[ll] = i /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1  DEF NoCCJoinL, Inv
              <6>2. ~(Cardinality(NodesReady) =< 1)
                 BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <6>1
              <6>7. ~i \in NodesReady
                 <7>1. SUFFICES ASSUME i \in NodesReady
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CwDist(ll, ii) =<  CwDist(i, ii)
                 BY <6>1, <5>1, <6>2,<7>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>2, <6>1, <5>1, IisNat, Unequality2, CwDistType
              <6>8. lset[i] # EmptyLS(i)
                 BY <6>7, <3>1
              <6>9. QED BY <6>1, <6>8, <5>1 DEF NoCCJoinL, Inv
           <5>8. CASE ii # i /\ ll # i /\ kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    ll = LeftNeighbor(lset[ii]),
                    ii # ll,
                    CwDist(kk, ii) < CwDist(ll, ii)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[ll] = ll
                     \/ toj[ll] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>8 DEF NoCCJoinL
              <6>. QED BY <5>1,<6>1 DEF NoCCJoinL, Inv
           <5>. QED
             BY <5>5, <5>6, <5>7, <5>8
        <4>73. SingleReadyNodeToj'
           <5>1. SUFFICES ASSUME NEW kk \in I, 
                 NEW rr \in I,
                 kk # rr,
                 NodesReady = {rr},
                 lset'[kk] # EmptyLS(kk)
                 PROVE toj[rr] = kk
              BY <4>07,<4>08 DEF SingleReadyNodeToj, Inv
           <5>2. /\ lset[i] \in LSet
                 /\ lset[i].node = i
                 /\ j \in I
                 /\ j # i
              BY <3>1, <4>2, <4>3,<4>16
           <5>4. lset'[i] # EmptyLS(i)
              <6>5. CASE j \in GetLSetContent(lset[i])
                 BY <4>15, <6>5, <5>2, NonEmptyLSet
              <6>6. CASE  ~ j \in GetLSetContent(lset[i])
                 BY <4>15, <6>6, <5>2, AddSingleNodeNoEmpty
              <6>9. QED BY <6>5,<6>6
           <5>5. CASE kk = i
              <6>5. CASE lset[i] = EmptyLS(i)
                 <7>1. i \in NodesReady
                    BY <3>1, <6>5
                 <7>2. i = rr
                    BY <7>1, <5>1
                 <7>3. QED BY <7>1,<7>2,<5>1, <5>5
              <6>6. CASE lset[i] # EmptyLS(i)
                 BY <5>1, <4>22, <6>6, <5>5 DEF SingleReadyNodeToj, Inv 
              <6>9. QED BY <6>5,<6>6
           <5>6. CASE rr = i
              <6>2. lset[kk] # EmptyLS(kk)
                 BY <5>1, <4>22, <5>6
              <6>3. QED BY <6>2, <5>1, SingleReadyNodeToj DEF SingleReadyNodeToj, Inv
           <5>7. CASE kk # i /\ rr # i
              BY <5>7, <4>22, <5>1 DEF SingleReadyNodeToj, Inv
           <5>9. QED BY <5>5,<5>6,<5>7
        <4>74. SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY  DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. ASSUME i \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}
                 PROVE lset'[i] # EmptyLS(i)
              <6>6. lset[i] # EmptyLS(i)
                 BY <5>3,<5>4, <5>9, <4>22
                    DEF SemJRplyLSnonEmpty, Inv
              <6>7. CASE lset'[i] = lset[i]
                 BY <6>7, <6>6
              <6>8. CASE lset'[i] = AddToLSet({j}, lset[i])
                 BY AddToLSetNoEmpty, <6>6, <6>8, <4>21, <4>2
              <6>9. QED BY <6>7,<6>8, <4>15
           <5>99. QED BY <5>3, <5>4, <5>8, <5>9, <4>22
                DEF SemJRplyLSnonEmpty, Inv
        
        <4>75. OneJRply'
           BY <4>10,<4>07,<4>08 DEF OneJRply, Inv, TypeInvariant
        <4>76. OneJReq'
           BY <4>10,<4>07,<4>08 DEF OneJReq, Inv, TypeInvariant
        <4>77. SemJReq'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type  = "JoinRequest"
                 PROVE
                   /\ mm.mreq.node \in NodesWait
                   /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY <4>08 DEF SemJReq, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JReq
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.node \in I
              BY <5>5 DEF JReq
           <5>7. /\ mm.mreq.node \in NodesWait
                 /\ ~ mm.mreq.node \in NodesReady
                 /\ lset[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY <5>3, <5>4, <5>6, StatusDisjoint DEF SemJReq, Inv
           <5>8. mm.mreq.node # i
              BY <5>7, <3>1, <5>6
           <5>9. QED BY <5>6,<5>7,<5>8, <4>22
        <4>78. JoinMsgDual'
           BY <4>10,<4>07,<4>08 DEF JoinMsgDual, Inv, TypeInvariant
        <4>80. AtleastOneReady'
           BY <4>08 DEF AtleastOneReady, Inv                    
        <4>99. QED BY <4>20,<4>95,\*<4>96,
                 \*<4>51,<4>52,<4>53,
                 <4>54,<4>55,<4>56,<4>57,<4>58,<4>59,
                  <4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, 
                  <4>70, <4>71, <4>72, <4>73, <4>74,
                  <4>75,<4>76,<4>77, <4>78, <4>80 DEF Inv         
      <3>2.QED BY <3>1, <1>4, SlowestZenon DEF ReceiveLSProbe, Inv, TypeInvariant
   <1>5. ASSUME NEW i \in I,
                ReceiveLSPrRpl(i),
                TypeInvariant,
                Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
                   lset[i] # EmptyLS(i),
                   status[i] # "dead",
                   m.destination = i,
                   m.mreq.type = "LSProbeReply",
                   m.mreq.node # i,
                   m.mreq.node \in probing[i],
                   ReceiveLSPrRplInner(i,m),
                    TypeInvariant,
                    Inv
         PROVE  Inv'
        (* The following auxiliary definitions are copied from the LET-bindings in the action definition. *)
        <4>. DEFINE j == m.mreq.node
        <4>. DEFINE Ls == m.mreq.lset
        <4>. DEFINE fi == failed[i] \ {j}
        <4>. DEFINE ls1 == IF j \in GetLSetContent(lset[i]) 
                              THEN lset[i]
                              ELSE AddToLSet({j}, lset[i])
        \*<4>. DEFINE ls2 == RemoveFromLSet(m.mreq.failed, ls1)
        <4>. DEFINE lprim == GetLSetContent(AddToLSet((GetLSetContent(Ls) \ fi), ls1))
        <4>. DEFINE prb1 == (GetLSetContent(ls1) \cap m.mreq.failed )\ (probing[i] \cup fi)
        <4>. DEFINE prb2 == lprim \ (GetLSetContent(ls1) \cup probing[i] \cup fi \cup prb1)
        <4>. DEFINE prb3 == ((probing[i] \cup prb1 \cup prb2) \ failed[i]) \ {j} 
        <4>. DEFINE shouldBeOKRest ==  /\ prb3={}
                                       /\ (Overlaps(ls1)\/IsComplete(ls1))
        <4>. DEFINE shouldBeOK == 
              /\ status[i] = "wait" 
              /\ shouldBeOKRest

        <4>01. rtable' = [rtable EXCEPT ![i] = AddToTable({j}, @, i)]
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>02. failed' = [failed EXCEPT ![i] = IF prb3={} /\ IsComplete(ls1) THEN {} ELSE fi]
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>03. lset'   = [lset EXCEPT ![i] = ls1]
          BY <3>1 DEF ReceiveLSPrRplInner 
        <4>04. probing'= [probing EXCEPT ![i] = prb3]
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>05. status'= [status EXCEPT ![i] = IF shouldBeOK THEN "ok" ELSE @]
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>06. receivedMsgs' = (receivedMsgs \ {m}) 
                              \cup Probe(i, ls1, fi, prb1)
                              \cup Probe(i, ls1, fi, prb2)
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>07. UNCHANGED <<lease,grant, toj>>
          BY <3>1 DEF ReceiveLSPrRplInner
        <4>08. /\ NodesReady' = NodesReady
               /\ \A k \in I\{i}: k \in NodesWait => k \in NodesWait'
           BY <4>05 DEF TypeInvariant, NodesReady, NodesWait, NodesOK
        <4>0. m.mreq.type = "LSProbeReply" /\ m \in DMsg
          BY <3>1 DEF TypeInvariant, DMsg
        <4>1. m.mreq \in PRpl
          BY <4>0 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
        (* We now prove typing information for each of the auxiliary operators defined above, then hide their definitions.
           This helps to keep the terms small enough for the prover to handle them. *)
        <4>2. /\ j \in I /\  j # i
              /\ m.mreq.failed \in SUBSET I
              /\ m.mreq.lset \in LSet
           BY <3>1, <4>1 DEF PRpl
        <4>3. j = m.mreq.node
           OBVIOUS
        <4>. HIDE DEF j
        <4>4. fi \in SUBSET I
          BY <4>2 DEF TypeInvariant
        <4>. HIDE DEF fi
        <4>5. ls1 \in LSet 
          BY <4>2, AddToLSetType DEF TypeInvariant
        <4>6. ls1.node = i
           BY <4>2,  
           AddToLSetNoChangeNode DEF TypeInvariant
        <4>7. lset'[i] = IF j \in GetLSetContent(lset[i]) 
                          THEN lset[i]
                          ELSE AddToLSet({j}, lset[i])
           BY <4>03 DEF TypeInvariant
        <4>. HIDE DEF ls1
        <4>49. m.mreq.type = "LSProbeReply" /\ m \in DMsg
          BY <3>1 DEF TypeInvariant, DMsg
        <4>50. m.mreq \in PRpl
          BY <4>49 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
        <4>. HIDE DEF j
        <4>45. lprim \in SUBSET I
          <5>1. (GetLSetContent(m.mreq.lset) \ fi) \in SUBSET I
            BY GetLSetContentType, <4>50, <4>4 DEF PRpl
          <5>2. AddToLSet((GetLSetContent(m.mreq.lset)\ fi), ls1) \in LSet
            BY AddToLSetType, <5>1, <4>5
          <5>3. QED BY <5>2, GetLSetContentType
        <4>. HIDE DEF lprim
        <4>46. prb1 \in SUBSET I
          BY GetLSetContentType, <4>5, <4>50, <4>4 DEF PRpl, TypeInvariant
        <4>. HIDE DEF prb1
        <4>47. prb2 \in SUBSET I
          BY <4>45, <4>5, <4>4, <4>46, GetLSetContentType DEF TypeInvariant
        <4>. HIDE DEF prb2
        <4>8.  /\ fi \in SUBSET I
               /\ ls1 \in LSet
               /\ lset[i] \in LSet
               /\ lset[i]' \in LSet
           BY <4>4, <4>03, <4>5 DEF TypeInvariant
        <4>10.  ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
                     \/ ms.mreq.type = "JoinRequest"
          <6>. HIDE DEF ls1, fi, prb1, prb2
          <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                     Probe(i, ls1, fi, prb1)
                \cup Probe(i, ls1, fi, prb2)
             BY <4>06, <4>3 DEF TypeInvariant
          <6>2. \A ms \in Probe(i, ls1, fi, prb1): ms.mreq.type = "LSProbe"
             <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb1)
                    BY ONLY ProbeMsgType, <4>4,<4>5,<4>46, <4>47
            <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
          <6>3. \A ms \in Probe(i, ls1, fi, prb2): ms.mreq.type = "LSProbe"
            <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb2)
                    BY ONLY ProbeMsgType, <4>4,<4>5,<4>46, <4>47
            <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
          <6>9. QED BY <6>1,<6>2,<6>3
        <4>11. \A k \in I: ~k \in (lset[k].left) /\ ~k \in (lset[k].right) 
          BY LSetInv DEF LSetInv, Inv
        <4>12. /\ ~ lset[i].node \in (ls1.left) 
               /\ ~ lset[i].node  \in (ls1.right)
               /\ ls1.node = lset[i].node
           <5>. HIDE DEF ls1
           <5>1. CASE ls1 = AddToLSet({j}, lset[i])
              BY <5>1, AddNodeNotChange, NotAddSelf, <4>2,<4>8 DEF TypeInvariant
           <5>2. CASE ls1 = lset[i]
              BY <5>2,<4>2,<4>8, LSetInv DEF LSetInv, TypeInvariant
           <5>9. QED BY <5>1,<5>2 DEF ls1
        <4>13. lset'[i] = AddToLSet({j}, lset[i]) \/ lset'[i] = lset[i]
           BY <4>7
        <4>14. /\ ~ ls1.node \in (ls1.left) 
             /\ ~ ls1.node  \in (ls1.right)
          BY <4>12


        <4>16. /\ lset[i]' = ls1 
               /\ ls1.node = i 
               /\ lset[i].node = i 
               /\ lset[i]'.node = i
          BY <4>03, <4>12, TypeInvariant DEF TypeInvariant
        <4>17. /\ ~ i \in (lset[i]'.left) 
               /\ ~ i \in (lset[i]'.right)
          BY <4>14, <4>16
        <4>18. \A k \in I \{i}: 
             /\ ~k \in (lset'[k].left) 
             /\ ~k \in (lset'[k].right) 
          BY <4>11, <4>03, TypeInvariant DEF TypeInvariant
        <4>20. LSetInv'
        BY <4>18,<4>17 DEF LSetInv
        \* \A i \in I: ~i \in (lset[i].left) /\ ~i \in (lset[i].right)
        <4>21.\A k \in I: /\lset'[k] \in LSet 
                          /\ lset[k] \in LSet
                          /\ lset[k].node = k
           BY <4>03, <4>8 DEF TypeInvariant 
        <4>22. \A k \in I\{i}: lset[k]' = lset[k]
           BY <4>03 DEF TypeInvariant
        <4>. HIDE DEF shouldBeOKRest

        <4>95. HalfNeighbor' 
          <5>0. \A k \in ReadyOK: RightNeighbor(lset[k]) # k 
                       /\ LeftNeighbor(lset[k]) # k
             BY <3>1, HalfNeighbor DEF HalfNeighbor, Inv
          <5>1. CASE /\ status[i] = "wait" 
                     /\ prb3={}
                     /\ (Overlaps(ls1)\/IsComplete(ls1))              
              <6>0. Overlaps(lset[i]') \/ IsComplete(lset[i]')
                 BY <4>16, <5>1  DEF shouldBeOKRest
              <6>1. RightNeighbor(lset[i]') # i /\ LeftNeighbor(lset[i]') # i
                 BY <6>0, CompleteHasNeighbor, OverlapNotEmpty, <4>8, <4>16
              <6>2. ReadyOK = ReadyOK' \{i} /\  i \in ReadyOK'
                 BY <5>1, <4>05 DEF shouldBeOKRest, NodesReady, ReadyOK, NodesOK, TypeInvariant
              <6>3. \A k \in ReadyOK'\{i}: lset'[k] = lset[k]
                    BY <5>0, <6>2, ReadyInI, <4>03 DEF TypeInvariant
              <6>4. \A k \in ReadyOK'\{i}: 
                     RightNeighbor(lset[k]') # k /\ LeftNeighbor(lset[k]') # k
                   BY <5>0, <6>2, <6>3
              <6>5. QED BY <6>1,<6>2, <6>4, <6>3, ReadyInI,  HalfNeighbor 
                      DEF HalfNeighbor
           <5>2. CASE ~(status[i] = "wait" /\ shouldBeOKRest)
              <6>1. status' = status
                 BY <4>05, <5>2 DEF TypeInvariant
              <6>2. ReadyOK' = ReadyOK /\  NonDead' = NonDead
                 BY <6>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NonDead, NodesWait
              <6>7. CASE i \in ReadyOK                
                 <7>1. /\lset[i] \in LSet
                       /\ j \in I
                       /\ lset[i].node = i
                       /\ lset[i]'= ls1
                    BY <4>16, <4>8, <4>2
                 <7>6. \A k \in ReadyOK \{i}: RightNeighbor(lset'[k])=RightNeighbor(lset[k])
                                      /\LeftNeighbor(lset'[k]) = LeftNeighbor(lset[k]) 
                       BY <4>03, ReadyInI DEF TypeInvariant
                 <7>7. RightNeighbor(AddToLSet({j},lset[i])) # lset[i].node /\ LeftNeighbor(AddToLSet({j}, lset[i])) # lset[i].node 
                       BY <6>7, <7>1, <5>0, AddToLSetNotEmpty 
                 <7>8. RightNeighbor(lset[i]) # lset[i].node /\ LeftNeighbor(lset[i]) # lset[i].node
                       BY <6>7, <7>1, <5>0, LSetInv DEF LSetInv
                 <7>9. RightNeighbor(lset'[i]) # i /\ LeftNeighbor(lset'[i]) # i
                       BY <7>7, <7>8, <4>16 DEF ls1
                 <7>10. QED BY <5>0, <6>1, <6>2, <7>6, <7>9, <7>1 DEF HalfNeighbor
              <6>8. CASE ~i \in ReadyOK
                 <7>1. \A k \in ReadyOK: lset'[k] = lset[k]
                    BY <6>8, ReadyInI, <4>03 DEF TypeInvariant
                 <7>2. QED BY <6>2, <6>8, <7>1, ReadyInI, <5>0 DEF HalfNeighbor, TypeInvariant
              <6>9. QED BY <6>7, <6>8 DEF HalfNeighbor
           <5>9. QED BY <5>1, <5>2 DEF shouldBeOKRest

        <4>54. CompleteLeafSet'
             <5>0. NodesReady' = NodesReady
                BY <4>05 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  RightNeighbor(lset'[i]) \in I
                   /\ RightNeighbor(lset[i]) \in I
                BY <4>7,<4>05,  <4>2,<4>8, <4>21, <4>22,<4>16, RightNeighborType
             <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. SUFFICES ASSUME NEW r \in NodesReady,
                   i # r,
                   lset'[i] # EmptyLS(i),
                   CwDist(i, r) < CwDist(i, RightNeighbor(lset'[i]))
                   PROVE FALSE
                BY <5>1, <5>0 DEF CompleteLeafSet, Inv
             <5>6.  lset[i] # EmptyLS(i)
                <6>1. CASE i \in NodesReady
                    <7>1. /\ r \in ReadyOK
                          /\ i \in ReadyOK
                          /\ r \in I
                          /\ i \in I
                          /\ ReadyOK \in SUBSET I
                       BY <6>1, <5>3, ReadyInI
                    <7>3. ~(Cardinality(ReadyOK)=<1)
                       BY CardinalityTwoNode,<6>1, <5>3, ReadyInI
                    <7>4. RightNeighbor(lset[i]) # i /\ LeftNeighbor(lset[i]) # i
                       BY <7>3, <7>1, CardinalityOneNodeExt,<5>1, <5>3 DEF HalfNeighbor, Inv
                    <7>9. QED BY <7>4, <5>1, <5>3, EmptyNoNeighbor
                <6>2. CASE ~i \in NodesReady
                   BY <3>1, <5>3, <5>1,<6>2
                <6>9. QED BY <6>1,<6>2
             <5>7. ~(CwDist(i, r) < CwDist(i, RightNeighbor(lset[i])))
                BY <5>3, <5>6, <5>1, ReadyInI DEF CompleteLeafSet, Inv
    
             <5>8. ~(CwDist(i, r) < CwDist(i, RightNeighbor(lset'[i])))
                BY <5>7, <5>2, <5>1, TransLEQ, Unequality2, CwDistType, IisNat,  
                   ReadyInI  DEF TypeInvariant
             <5>9. QED BY <5>1, <5>8, <5>3 
        <4>55. CompleteLeafSetL'
           \* OMITTED \* similar to <4>54
             <5>0. NodesReady' = NodesReady
                BY <4>05 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  LeftNeighbor(lset'[i]) \in I
                   /\ LeftNeighbor(lset[i]) \in I
                BY <4>7,<4>05,  <4>2,<4>8, <4>21, <4>22,<4>16, LeftNeighborType
             <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node) 
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. SUFFICES ASSUME NEW r \in NodesReady,
                   i # r,
                   lset'[i] # EmptyLS(i),
                   CwDist(r, i) < CwDist(LeftNeighbor(lset'[i]), i)
                   PROVE FALSE
                BY <5>1, <5>0 DEF CompleteLeafSetL, Inv
             <5>6.  lset[i] # EmptyLS(i)
                <6>1. CASE i \in NodesReady
                    <7>1. /\ r \in ReadyOK
                          /\ i \in ReadyOK
                          /\ r \in I
                          /\ i \in I
                          /\ ReadyOK \in SUBSET I
                       BY <6>1, <5>3, ReadyInI
                    <7>3. ~(Cardinality(ReadyOK)=<1)
                       BY CardinalityTwoNode,<6>1, <5>3, ReadyInI
                    <7>4. RightNeighbor(lset[i]) # i /\ LeftNeighbor(lset[i]) # i
                       BY <7>3, <7>1, CardinalityOneNodeExt,<5>1, <5>3 DEF HalfNeighbor, Inv
                    <7>9. QED BY <7>4, <5>1, <5>3, EmptyNoNeighbor
                <6>2. CASE ~i \in NodesReady
                   BY <3>1, <5>3, <5>1,<6>2
                <6>9. QED BY <6>1,<6>2
             <5>7. ~(CwDist(r, i) < CwDist(LeftNeighbor(lset[i]), i))
                BY <5>3, <5>6, <5>1, ReadyInI DEF CompleteLeafSetL, Inv
    
             <5>8. ~(CwDist(r, i) < CwDist(LeftNeighbor(lset'[i]), i))
                BY <5>7, <5>2, <5>1, TransLEQ, Unequality2, CwDistType, IisNat,  
                   ReadyInI  DEF TypeInvariant
             <5>9. QED BY <5>1, <5>8, <5>3            
        <4>56. ReadyNeighborRight' 
             <5>0. NodesReady' = NodesReady
                BY <4>05 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  RightNeighbor(lset'[i]) \in I
                   /\ RightNeighbor(lset[i]) \in I
                BY <4>7,<4>05,  <4>2,<4>8, <4>21,<4>22, <4>16, RightNeighborType
             <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. CASE i \in NodesReady
                <6>3. \A k \in I:
                       /\  lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~RightNeighbor(lset'[i]) \in NodesReady 
                      => CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, k) 
                    <7>1. SUFFICES ASSUME NEW k \in I,
                           NEW r \in I,
                           r = RightNeighbor(lset[i]),
                           k # r,
                           i # k,
                           lset[k] # EmptyLS(k),
                           ~RightNeighbor(lset'[i]) \in NodesReady
                       PROVE CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, k) 
                       BY <7>1, ReadyInI, <5>1, <5>2 DEF ReadyNeighborRight, Inv
                    <7>2. CASE  RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                       BY <7>1,<7>2,<5>3, ReadyInI DEF ReadyNeighborRight, Inv
                    <7>3. CASE  RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       <8>3. CASE r \in NodesReady
                          <9>1. SUFFICES ASSUME  CwDist(i, k) < CwDist(i, RightNeighbor(lset'[i]))
                                PROVE FALSE
                                BY <7>1, <5>1, IisNat, Unequality2,CwDistType         
                          <9>2. RightNeighbor(lset'[i]) # k /\  i # RightNeighbor(lset'[i])
                             BY <9>1, IisNat, <5>1, <7>1, CwDistZeroCompare
                          <9>3. CwDist(i, k) < CwDist(i, r)
                             BY <9>1, <5>2, TransLESS, <5>1, <7>1, IisNat, CwDistType
                          <9>4. lset[k] # EmptyLS(k)
                             BY <7>1
                          <9>41. i # r
                             BY CwDistZeroCompare, <9>3, IisNat
                          <9>5. \/ toj[i] = k /\ toj[r] = r
                                \/ toj[r] = k /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>4, <9>41, <9>3, NoCCJoin DEF NoCCJoin, Inv
                          <9>6. CwDist(i, RightNeighbor(lset'[i])) < CwDist(i, r)
                             <10>2. /\ CwDist(i, RightNeighbor(lset'[i])) \in Nat
                                    /\ CwDist(i, r) \in Nat
                                 BY CwDistType, IisNat, <5>1
                             <10>8.CwDist(i, RightNeighbor(lset'[i])) # CwDist(i, r)
                                 BY RightDistInjectivity,  <5>1, <7>3, <7>1, <10>2, IisNat
                             <10>9. QED BY <10>2, <10>8, <5>2,<7>1, Unequality6
                          <9>7. lset[RightNeighbor(lset'[i])] # EmptyLS(RightNeighbor(lset'[i]))
                             <10>6. /\ lset'[i] = AddToLSet({j}, lset[i])
                                    /\ j \in I
                                 BY <7>3, <4>16, RNInjectivity, <5>1, <7>3, <3>1 DEF ls1
                             <10>7. RightNeighbor(lset'[i]) = j
                                 BY <10>6, <5>1, <7>3, AddSingleNodeChange
                             <10>8. lset[j] # EmptyLS(j)
                                 BY <3>1, <4>3,SemPRplyNonEmpty DEF SemPRplyNonEmpty, Inv
                             <10>9. QED BY  <10>7,<10>8
                          <9>8. \/ /\ toj[i] = RightNeighbor(lset'[i]) 
                                   /\ toj[r] = r
                                \/ /\ toj[r] = RightNeighbor(lset'[i]) 
                                   /\ toj[i] = i
                             BY <9>1, <9>2, <7>1, <5>3, <8>3, <9>7,
                                <9>41, <9>6, <5>1, NoCCJoin, SlowestZenon
                                DEF NoCCJoin, Inv
                          <9>9. QED BY <9>5,<9>8, <9>2, <9>3, <7>1
                       <8>4. CASE ~ r \in NodesReady
                          <9>2. CwDist(i, r) =< CwDist(i, k)
                            BY <7>1, <8>4, <5>1, <5>0,<5>3, ReadyInI
                            DEF ReadyNeighborRight, Inv
                          <9>9. QED BY <9>2, <5>2, <5>1, <7>1,TransLEQ, CwDistType, IisNat,
                             ReadyInI  DEF TypeInvariant
                       <8>9. QED BY <8>3,<8>4
                    <7>9. QED BY <7>2,<7>3
                 <6>4. \A r \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, i) 
                    <7>1. SUFFICES ASSUME NEW r \in NodesReady,
                       i # r,
                       lset'[i] # EmptyLS(i),
                       ~RightNeighbor(lset'[r]) \in NodesReady
                       PROVE CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, i)
                       OBVIOUS
                    <7>2. /\ r \in I
                          /\ lset[r] = lset'[r]
                          /\ lset[r] \in LSet
                          /\ RightNeighbor(lset[r]) \in I
                       BY <7>1, <5>1, ReadyInI, <4>21, RightNeighborType
                    <7>3. ~(Cardinality(NodesReady) =< 1)
                       BY <5>3, <7>1, ReadyInI, CardinalityTwoNode
                    <7>4. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                       BY <7>3, <7>1, <5>3, LemmaCLtoNC DEF NeighborClosest, Inv
                    <7>9. QED BY <7>4, <7>1, <7>2
                 <6>5. \A r \in NodesReady\{i}, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, k)
                    BY <5>1, <5>0, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
                 <6>9. QED BY <6>3, <6>4, <6>5, <5>1, <5>0, ReadyNeighborRight DEF ReadyNeighborRight, Inv
             <5>4. CASE ~i \in NodesReady
                <6>2. lset[i] # EmptyLS(i)
                   BY <3>1,  <5>1
                <6>4. \A r \in NodesReady: lset'[r] = lset[r]
                   BY <5>4, <5>1, ReadyInI
                <6>9. QED BY <3>1,<5>0,<6>2,<5>4, <5>1, ReadyInI DEF ReadyNeighborRight, Inv
             <5>9. QED BY <5>3,<5>4

        <4>57. ReadyNeighborLeft'
           \* OMITTED \* Similar to <4>56
             <5>0. NodesReady' = NodesReady
                BY <4>05 DEF TypeInvariant, NodesReady
             <5>1. /\ lset'[i] \in LSet 
                   /\ lset[i] \in LSet
                   /\ j \in I
                   /\ lset'[i].node = i
                   /\ lset[i].node = i
                   /\ \A k \in I \{i}: lset'[k] = lset[k]
                   /\  LeftNeighbor(lset'[i]) \in I
                   /\ LeftNeighbor(lset[i]) \in I
                BY <4>7,<4>05,  <4>2,<4>8, <4>21,<4>22, <4>16, LeftNeighborType
             <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node) 
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
             <5>3. CASE i \in NodesReady
                <6>3. \A k \in I:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~LeftNeighbor(lset'[i]) \in NodesReady 
                      => CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(k, i) 
                    <7>1. SUFFICES ASSUME NEW k \in I,
                           NEW l \in I,
                           l = LeftNeighbor(lset[i]),
                           k # l,
                           i # k,
                            lset[k] # EmptyLS(k),
                           ~LeftNeighbor(lset'[i]) \in NodesReady
                       PROVE CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(k, i) 
                       BY <7>1, ReadyInI, <5>1, <5>2 DEF ReadyNeighborLeft, Inv
                    <7>2. CASE  LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                       BY <7>1,<7>2,<5>3, ReadyInI DEF ReadyNeighborLeft, Inv
                    <7>3. CASE  LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       <8>3. CASE l \in NodesReady
                          <9>1. SUFFICES ASSUME  CwDist(k, i) < CwDist(LeftNeighbor(lset'[i]), i)
                                PROVE FALSE
                                BY <7>1, <5>1, IisNat, Unequality2,CwDistType         
                          <9>2. LeftNeighbor(lset'[i]) # k /\  i # LeftNeighbor(lset'[i])
                             BY <9>1, IisNat, <5>1, <7>1, CwDistZeroCompareCo
                          <9>3. CwDist(k, i) < CwDist(l, i)
                             BY <9>1, <5>2, TransLESS, <5>1, <7>1, IisNat, CwDistType
                          <9>4. lset[k] # EmptyLS(k)
                             BY <7>1
                          <9>41. i # l
                             BY CwDistZeroCompareCo, <9>3, IisNat
                          <9>5. \/ toj[i] = k /\ toj[l] = l
                                \/ toj[l] = k /\ toj[i] = i
                             BY <9>1, <7>1, <5>3, <8>3, <9>4, <9>41, <9>3 DEF NoCCJoinL, Inv
                          <9>6. CwDist(LeftNeighbor(lset'[i]), i) < CwDist(l, i)
                             <10>2. /\ CwDist(LeftNeighbor(lset'[i]), i) \in Nat
                                    /\ CwDist(l, i) \in Nat
                                 BY CwDistType, IisNat, <5>1
                             <10>8. CwDist(LeftNeighbor(lset'[i]), i) # CwDist(l, i)
                                 BY CwDistInjectivity,  <5>1, <7>3, <7>1, <10>2, IisNat
                             <10>9. QED BY <10>2, <10>8, <5>2,<7>1, Unequality6
                          <9>7. lset[LeftNeighbor(lset'[i])] # EmptyLS(LeftNeighbor(lset'[i]))
                             <10>6. /\ lset'[i] = AddToLSet({j}, lset[i])
                                    /\ j \in I
                                 BY <7>3, <4>16, LNInjectivity, <5>1, <7>3, <3>1 DEF ls1
                             <10>7. LeftNeighbor(lset'[i]) = j
                                 BY <10>6, <5>1, <7>3, AddSingleNodeChangeCo
                             <10>8. lset[j] # EmptyLS(j)
                                 BY <3>1, <4>3,SemPRplyNonEmpty DEF SemPRplyNonEmpty, Inv
                             <10>9. QED BY  <10>7,<10>8
                          <9>8. \/ /\ toj[i] = LeftNeighbor(lset'[i]) 
                                   /\ toj[l] = l
                                \/ /\ toj[l] = LeftNeighbor(lset'[i]) 
                                   /\ toj[i] = i
                             BY <9>1, <9>2, <7>1, <5>3, <8>3, <9>7,
                                <9>41, <9>6, <5>1,  SlowestZenon
                                DEF NoCCJoinL, Inv
                          <9>9. QED BY <9>5,<9>8, <9>2, <9>3, <7>1
                       <8>4. CASE ~ l \in NodesReady
                          <9>2. CwDist(l, i) =< CwDist(k, i)
                            BY <7>1, <8>4, <5>1, <5>0,<5>3, ReadyInI
                            DEF ReadyNeighborLeft, Inv
                          <9>9. QED BY <9>2, <5>2, <5>1, <7>1,TransLEQ, CwDistType, IisNat,
                             ReadyInI  DEF TypeInvariant
                       <8>9. QED BY <8>3,<8>4
                    <7>9. QED BY <7>2,<7>3
                 <6>4. \A r \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # r
                       /\~LeftNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(i, r) 
                    <7>1. SUFFICES ASSUME NEW r \in NodesReady,
                       i # r,
                       lset'[i] # EmptyLS(i),
                       ~LeftNeighbor(lset'[r]) \in NodesReady
                       PROVE CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(i, r)
                       OBVIOUS
                    <7>2. /\ r \in I
                          /\ lset[r] = lset'[r]
                          /\ lset[r] \in LSet
                          /\ LeftNeighbor(lset[r]) \in I
                       BY <7>1, <5>1, ReadyInI, <4>21, LeftNeighborType
                    <7>3. ~(Cardinality(NodesReady) =< 1)
                       BY <5>3, <7>1, ReadyInI, CardinalityTwoNode
                    <7>4. CwDist(LeftNeighbor(lset[r]), r) =< CwDist(i, r)
                       BY <7>3, <7>1, <5>3, LemmaCLtoNC DEF NeighborClosest, Inv
                    <7>9. QED BY <7>4, <7>1, <7>2
                 <6>5. \A r \in NodesReady\{i}, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~LeftNeighbor(lset'[r]) \in NodesReady 
                      => CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(k, r)
                    BY <5>1, <5>0, ReadyInI DEF ReadyNeighborLeft, Inv
                 <6>9. QED BY <6>3, <6>4, <6>5, <5>1, <5>0 DEF ReadyNeighborLeft, Inv
             <5>4. CASE ~i \in NodesReady
                <6>2. lset[i] # EmptyLS(i)
                   BY <3>1, <5>1
                <6>4. \A r \in NodesReady: lset'[r] = lset[r]
                   BY <5>4, <5>1, ReadyInI
                <6>9. QED BY <3>1,<5>0,<6>2,<5>4, <5>1, ReadyInI DEF ReadyNeighborLeft, Inv
             <5>9. QED BY <5>3,<5>4
        <4>58. SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                 NEW nn \in I, 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "BroadcastLSet",
                 mm.mreq.lset.node = nn,
                 mm.destination = ii
                PROVE
                    /\ nn \in NodesReady'
                    /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
              BY <4>07,<4>08, SimplifyAndSolve DEF SemLRply, Inv, TypeInvariant
           <5>9. QED BY <4>10,<5>2,<4>07,<4>08 DEF SemLRply, Inv, TypeInvariant
        <4>59. LSnonEmpty' \* similar to ReceiveProbe
           <5>1. SUFFICES ASSUME NEW ii \in I, NEW jj \in I,
                 ii \in GetLSetContent(lset'[jj]),
                 ii # jj,
                 ~ii \in NodesReady 
                 PROVE lset'[ii] # EmptyLS(ii)
              BY <4>08 DEF LSnonEmpty, Inv
           <5>2. /\ lset[i] \in LSet
                 /\ lset[i].node = i
                 /\ lset'[i] \in LSet
                 /\ lset'[i].node = i
                 /\ j \in I
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
              BY <4>03, <4>21, <4>16, <4>2 DEF TypeInvariant
           <5>5. CASE ii = i
              <6>1. lset[i] # EmptyLS(i)
                 BY <5>5, <5>1, <5>2 DEF TypeInvariant, LSnonEmpty, Inv
              <6>2. AddToLSet({j}, lset[i]) # EmptyLS(i)
                 BY <6>1, <5>2, AddToLSetNoEmpty
              <6>3. lset'[i] # EmptyLS(i)
                  BY <6>1,<6>2,<4>16, <3>1 DEF ls1 
              <6>9. QED BY <6>3, <5>5
           <5>6. CASE jj = i
              <6>5. CASE ii \in GetLSetContent(lset[i])
                 <7>1. ii # i
                    BY <5>6, <5>1
                 <7>2. lset[ii] # EmptyLS(ii)
                    BY <5>1, <6>5, <5>6 DEF LSnonEmpty, Inv
                 <7>3. QED BY <7>2, <7>1, <5>2
              <6>6. CASE ~ii \in GetLSetContent(lset[i])
                
                 <7>2. lset'[i] # lset[i]
                    BY <6>6, <5>1, NotEqualLS, <5>2, <5>6
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2,<4>7
                 <7>5. ii \in GetLSetContent(AddToLSet({j}, lset[i]))
                    BY <5>1, <5>6,<7>3
                 <7>6. ii = j
                    BY <5>2, <7>3,<7>5, <6>6, TheAddedNode 
                 <7>8. lset[j] # EmptyLS(j)
                    BY <3>1, <5>2 DEF j,SemPRplyNonEmpty, Inv
                 <7>9. QED BY <7>6, <7>8, <5>1, <5>2, <5>6 
              <6>9. QED BY <6>5,<6>6
           <5>8. CASE ii # i /\ jj # i
                 BY <5>8, <5>1, <5>2 DEF TypeInvariant, LSnonEmpty, Inv
           <5>9. QED BY <5>5,<5>6,<5>8
        <4>60. LeaseGrant'
           BY <4>07, <4>05 DEF TypeInvariant, NodesReady, LeaseGrant, Inv
        <4>61. GrantReady'
           BY <4>07, <4>05 DEF TypeInvariant, NodesReady, GrantReady, Inv
        <4>62. SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE SemJRply!(mm)'
              BY <4>08 DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. /\ mm.mreq.lset.node \in I
                 /\ mm.destination \in I
              BY <5>5, <5>4 DEF JRpl, LSet, DMsg, TypeInvariant
           <5>71. /\mm.mreq.lset.node \in NodesReady
                  /\mm.destination \in NodesWait
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>72.   /\ LeftNeighbor(mm.mreq.lset) \in I
                    /\ RightNeighbor(mm.mreq.lset) \in I
                    /\ mm.mreq.lset.node \in I
                    /\ \A k \in I: toj'[k] = toj[k]
                    /\ \A k \in I: lease'[k] = lease[k]
              BY <5>6, <5>7, LeftNeighborType, RightNeighborType, <4>07 DEF TypeInvariant
           <5>8. CASE mm.mreq.lset.node # i
              <6>1. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                 BY <4>22, <5>7, <5>8
              <6>2. mm.destination # i
                 <7>1. SUFFICES ASSUME mm.destination = i PROVE FALSE
                    OBVIOUS
                 <7>2. lset[i] = EmptyLS(i)
                    BY <7>1, <5>4, <5>3 DEF SemJRply, Inv
                 <7>3. mm.destination \in NodesReady
                    BY <3>1, <7>2, <7>1
                 <7>4. mm.destination \in NodesWait
                    BY <5>3, <5>4 DEF SemJRply, Inv
                 <7>9. QED BY <7>3, <7>4,<5>7, StatusDisjoint
              <6>3.  lset'[mm.destination] =  lset[mm.destination]
                 BY  <4>22, <5>7, <6>2
              <6>9. QED BY <6>1, <5>7, <6>2, <6>3, <5>72, <5>3, <5>4, <4>08 DEF SemJRply, Inv
           <5>9. CASE mm.mreq.lset.node = i
              <6>0. /\ LeftNeighbor(lset[i]) = LeftNeighbor(mm.mreq.lset)
                    /\ RightNeighbor(lset[i]) = RightNeighbor(mm.mreq.lset)
                 BY <5>3, <5>4, <5>9 DEF SemJoinLeafSet, Inv
              <6>1. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(LeftNeighbor(AddToLSet({j}, lset[i])), i)
                        =< CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(LeftNeighbor(lset'[i]), lset'[i].node)
                        =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChangeCo
                 <7>51. /\ LeftNeighbor(lset'[i])  = j
                        /\ LeftNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(j, i)=< CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(j, i) # CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. SUFFICES ASSUME  CwDist(j, i) = CwDist(LeftNeighbor(lset[i]), i)
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ LeftNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. LeftNeighbor(lset[i]) = j
                       BY <8>1, <8>7, CwDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE LeftNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF j, SemPRplyNonEmpty, Inv
                    <8>6. LeftNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompareCo, IisNat
                    <8>7. \/ toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          \/ toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoinL, Inv
                    <8>8. CASE toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                          BY  <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. LeftNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~LeftNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1 DEF j
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv, j
                    <8>8. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborLeft, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                 \* OMITTED \* similar to <6>1
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(i, RightNeighbor(AddToLSet({j}, lset[i])))
                        =< CwDist(i, RightNeighbor(lset[i]))
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(lset'[i].node, RightNeighbor(lset'[i]))
                        =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. RightNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChange
                 <7>51. /\ RightNeighbor(lset'[i])  = j
                        /\ RightNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(i, j)=< CwDist(i, RightNeighbor(lset[i]))
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(i, j) # CwDist(i, RightNeighbor(lset[i]))
                    <8>1. SUFFICES ASSUME  CwDist(i, j) = CwDist(i, RightNeighbor(lset[i]))
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ RightNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. RightNeighbor(lset[i]) = j
                       BY <8>1, <8>7, RightDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE RightNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 DEF j, SemPRplyNonEmpty, Inv
                    <8>6. RightNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompare, IisNat
                    <8>7. \/ toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          \/ toj[RightNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoin, Inv
                    <8>8. CASE toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[RightNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                          BY  <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. RightNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~RightNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1 DEF j
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv, j
                    <8>8. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborRight, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>3. Covers(lset'[i], mm.destination)
                 <7>1. Covers(lset[i], mm.destination)
                    BY <5>3, <5>4, <5>9 DEF SemJRply, Inv
                 <7>2. QED BY <7>1, <6>1,<6>2, <5>7,<4>16,<4>21, CoverNeighborRel
              <6>4.  /\ toj'[i] = mm.destination
                     /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node => 
                        toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))                         
                     /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node => 
                         toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                 BY <5>3,<5>4, <5>72, <5>9 DEF SemJRply, Inv
              <6>5. /\ RightNeighbor(lset'[i]) \in lease'[i]
                    /\ LeftNeighbor(lset'[i]) \in lease'[i]
                 <7>1. /\ RightNeighbor(lset[i]) \in lease[i]
                       /\ LeftNeighbor(lset[i]) \in lease[i]
                    BY <5>3,<5>4, <5>9 DEF SemJRply, Inv
                 <7>2. QED BY <6>1, <6>2,  <5>72, <7>1
              <6>6. mm.destination # i
                 BY <6>6,<5>9,<5>71, StatusDisjoint
              <6>7.  lset'[mm.destination] =  lset[mm.destination]
                 BY  <4>22, <5>7, <6>6 
              <6>8. lset'[mm.destination]  = EmptyLS(mm.destination)
                 BY <6>7, <5>3,<5>4 DEF SemJRply, Inv
              <6>99. QED BY <5>7, <6>6, <6>4, <6>5, <6>3, <6>8, <5>71, <5>9, <4>08
           <5>10. QED BY <5>8, <5>9
        <4>63. SemJoinNeighbor'
           <5>3. SUFFICES ASSUME NEW ls \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <4>10, <5>3 DEF SemJoinNeighbor, Inv
        <4>64. SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady' = {mm.mreq.lset.node}
              BY <4>1 DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>8. /\ lset[mm.mreq.lset.node] = mm.mreq.lset
                 /\ NodesReady = {mm.mreq.lset.node}
              BY <5>3, <5>4 DEF SemSingleReadyJoin, Inv
           <5>9. mm.mreq.lset.node # i
              <6>1. SUFFICES ASSUME mm.mreq.lset.node = i PROVE FALSE
                 OBVIOUS
              <6>2. lset[i] = EmptyLS(i)
                 BY <5>3, <5>8, <6>1
              <6>3. toj[i] = mm.destination
                 BY <5>3,<6>1, <5>4 DEF SemJRply, Inv
              <6>4. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                 BY <3>1 DEF SemPRplyNonEmpty, Inv
              <6>6. toj[i] = m.mreq.node
                 BY <6>2, <6>4, <6>1, <5>8, <4>2 DEF SingleReadyNodeToj, Inv, j
              <6>7. lset[mm.destination] = EmptyLS(mm.destination)
                 BY <5>3, <5>4 DEF SemJRply, Inv
              <6>9. QED BY <6>3,<6>4, <6>6,<6>7
           <5>10. QED BY <5>7, <5>8, <5>9, <4>22, <4>08
        <4>65. SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <4>10, <5>3
                DEF SemEmptyLSetSingleNode, Inv
        <4>66. SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY <4>1 DEF SemJoinLeafSet, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>71. mm.mreq.lset.node \in NodesReady
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>8. CASE mm.mreq.lset.node # i
              <6>1. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                 BY <4>22, <5>7, <5>8
              <6>3. QED BY <6>1, <5>3, <5>4 DEF SemJoinLeafSet, Inv
           <5>9. CASE mm.mreq.lset.node = i
              <6>0. /\ LeftNeighbor(lset[i]) = LeftNeighbor(mm.mreq.lset)
                    /\ RightNeighbor(lset[i]) = RightNeighbor(mm.mreq.lset)
                 BY <5>3, <5>4, <5>9 DEF SemJoinLeafSet, Inv
              <6>1. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # LeftNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(LeftNeighbor(AddToLSet({j}, lset[i])), i)
                        =< CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(LeftNeighbor(lset'[i]), lset'[i].node)
                        =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChangeCo
                 <7>51. /\ LeftNeighbor(lset'[i])  = j
                        /\ LeftNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(j, i)=< CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(j, i) # CwDist(LeftNeighbor(lset[i]), i)
                    <8>1. SUFFICES ASSUME  CwDist(j, i) = CwDist(LeftNeighbor(lset[i]), i)
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ LeftNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. LeftNeighbor(lset[i]) = j
                       BY <8>1, <8>7, CwDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE LeftNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 
                       DEF SemPRplyNonEmpty, Inv, j
                    <8>6. LeftNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompareCo, IisNat
                    <8>7. \/ toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          \/ toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoinL, Inv
                    <8>8. CASE toj[i] = j /\ toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[LeftNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                          BY <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. LeftNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~LeftNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1 DEF j
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv, j
                    <8>8. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborLeft, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                 \* OMITTED \* similar to <6>1
                 <7>0.   /\ lset[i] \in LSet
                         /\ lset[i].node = i
                         /\ lset'[i] \in LSet
                         /\ lset'[i].node = i
                         /\ j \in I
                         /\ lset[j] \in LSet
                         /\ lset[j].node = j
                         /\ \A k \in I\{i}: lset'[k] = lset[k]
                         /\ LeftNeighbor(lset[i]) \in I
                         /\ RightNeighbor(lset[i]) \in I
                      BY <4>22, <4>21,<3>1, <4>16, <4>2, LeftNeighborType, RightNeighborType
                 <7>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # RightNeighbor(lset[i])
                       PROVE FALSE
                    OBVIOUS
                 <7>2. lset'[i] # lset[i]
                    BY <7>1
                 <7>3. lset'[i] = AddToLSet({j}, lset[i])
                    BY <7>2, <4>13
                 <7>4. CwDist(i, RightNeighbor(AddToLSet({j}, lset[i])))
                        =< CwDist(i, RightNeighbor(lset[i]))
                    <8>1. /\ lset'[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ lset[i] \in LSet
                       BY <7>3, <7>0
                    <8>2. CwDist(lset'[i].node, RightNeighbor(lset'[i]))
                        =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                       BY  AddToLSetInvCo, <7>3, <8>1
                    <8>9. QED BY <8>2, <7>0, <7>3
                 <7>5. RightNeighbor(AddToLSet({j}, lset[i])) = j
                    BY <7>1, <7>3, <7>0,AddSingleNodeChange
                 <7>51. /\ RightNeighbor(lset'[i])  = j
                        /\ RightNeighbor(lset[i]) # j
                    BY <7>5, <7>3, <7>1
                 <7>52. CwDist(i, j)=< CwDist(i, RightNeighbor(lset[i]))
                    BY <7>4, <7>5, <7>51, <7>0
                 <7>53. CwDist(i, j) # CwDist(i, RightNeighbor(lset[i]))
                    <8>1. SUFFICES ASSUME  CwDist(i, j) = CwDist(i, RightNeighbor(lset[i]))
                          PROVE FALSE
                       OBVIOUS
                    <8>7. /\  j \in Nat
                          /\ i \in Nat
                          /\ RightNeighbor(lset[i]) \in Nat
                       BY <7>0, IisNat 
                    <8>8. RightNeighbor(lset[i]) = j
                       BY <8>1, <8>7, RightDistInjectivity
                    <8>9. QED BY <8>8,<7>51
                 <7>6. CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                    BY <7>52, <7>53, <7>0, Unequality6, CwDistType, IisNat
                 <7>7. CASE RightNeighbor(lset[i]) \in NodesReady
                    <8>5. /\ i # j
                          /\ lset[j] # EmptyLS(j)
                          /\ i \in NodesReady
                          /\ j \in I
                       BY <7>0, <5>9, <5>71, <3>1 
                       DEF SemPRplyNonEmpty, Inv, j
                    <8>6. RightNeighbor(lset[i]) #  i
                       BY <7>6, <7>0, CwDistZeroCompare, IisNat
                    <8>7. \/ toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          \/ toj[RightNeighbor(lset[i])] = j /\ toj[i] = i 
                       <9>9. QED BY <7>6, <7>7, <8>5, <8>6 DEF NoCCJoin, Inv
                    <8>8. CASE toj[i] = j /\ toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                        <9>1. toj[i] = m.mreq.node
                           BY <3>1, <8>8 DEF j
                        <9>2. toj[i] = mm.destination
                           BY <5>3,<5>9, <5>4 DEF SemJRply, Inv
                        <9>6. lset[m.mreq.node] # EmptyLS(m.mreq.node)
                           BY <3>1 DEF SemPRplyNonEmpty, Inv
                        <9>7. lset[mm.destination] = EmptyLS(mm.destination)
                           BY <5>3, <5>4 DEF SemJRply, Inv
                        <9>9. QED BY <9>1,<9>2,<9>6,<9>7
                    <8>9. CASE toj[RightNeighbor(lset[i])] = j /\ toj[i] = i
                       <9>1.   toj[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                          BY <6>0, <5>9, <8>9, <5>3, <5>4 DEF SemJRply, Inv
                       <9>2.  toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                          BY <9>1, <6>0
                       <9>3. RightNeighbor(lset[i]) = j
                          BY <9>2, <8>9
                       <9>9. QED BY <9>3, <7>51
                    <8>99. QED BY <8>7, <8>8, <8>9
                 <7>8. CASE ~RightNeighbor(lset[i]) \in NodesReady
                    <8>1. i \in NodesReady
                       BY <5>9, <5>71
                    <8>2. i # j
                       BY <3>1 DEF j
                    <8>6. lset[j] # EmptyLS(j)
                       BY <3>1, <7>0 DEF SemPRplyNonEmpty, Inv, j
                    <8>8. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                       BY <8>1,<8>2,<8>6, <7>8, <7>0 DEF ReadyNeighborRight, Inv
                    <8>9. QED BY <8>8, <7>6, <7>0, IisNat, CwDistType, Unequality2
                 <7>9.  QED BY <7>7, <7>8
              <6>9. QED BY <6>1, <6>2, <6>0, <5>9
           <5>10. QED BY <5>7, <5>8, <5>9, <4>22
        <4>67. GrantNeighbor'\*  similar to ReceiveProbe
           <5>0. NodesReady' = NodesReady
                BY <4>05 DEF TypeInvariant, NodesReady
           <5>1. SUFFICES ASSUME 
                 NEW kk \in NodesReady, 
                 NEW ii \in I, 
                 ~ii \in NodesReady,
                 ii \in grant[kk],
                 ii # LeftNeighbor(lset'[kk]),
                 ii # RightNeighbor(lset'[kk])
                 PROVE FALSE
              BY <4>05, <4>07, ReadyInI, <5>0
              DEF TypeInvariant, GrantNeighbor, Inv
           <5>2. /\ lset[kk] \in LSet
                 /\ lset[kk].node = kk
                 /\ j \in I
                 /\ {j} \in SUBSET I
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
                 /\ lset'[i].node = i
                 /\ lset'[i] \in LSet
                 /\ lset[i].node = i
                 /\ lset[i] \in LSet
              BY <4>7,<4>05,  <4>2,<4>8, <4>21, <4>16,<4>22, ReadyInI DEF TypeInvariant
           <5>3. \/ ii = LeftNeighbor(lset[kk])
                 \/ ii = RightNeighbor(lset[kk])
              BY <5>1, <5>2 DEF GrantNeighbor, Inv
           <5>5. CASE kk = i
              <6>5. CASE ii = LeftNeighbor(lset[i])
                 <7>1. SUFFICES ASSUME NEW k \in I,
                       k = LeftNeighbor(lset'[i]),
                       k # ii
                       PROVE FALSE
                    BY <5>1, <5>2,<5>5, LeftNeighborType
                 <7>2. lset'[i] = AddToLSet({j}, lset[i])
                    BY <4>03, <7>1, <6>5, LeftNeighborType, <5>2 DEF TypeInvariant, ls1   
                 <7>3. CwDist(LeftNeighbor(lset'[i]), lset'[i].node) =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                    BY AddToLSetInvCo,<5>2,<7>2, <5>5
                 <7>4. CwDist(k,i) =< CwDist(ii, i)
                    BY  <7>3, <7>1, <6>5,<5>5, <5>2
                 <7>5. CwDist(k,i) < CwDist(ii, i)
                       BY <7>4, <7>1, CwDistInjectivity, Unequality6,
                         CwDistType, IisNat
                 <7>6. k = j
                    BY <7>2, <7>1, <6>5, <5>2, <5>5, AddSingleNodeChangeCo, SimplifyAndSolve
                 <7>7. /\ ~ii \in NodesReady
                       /\ k # i
                       /\ lset[k] # EmptyLS(k)
                    <8>1. /\ lset[k] \in LSet
                          /\ lset[k].node = k
                       BY DEF TypeInvariant
                    <8>2. lset[k]# EmptyLS(k)
                       BY <7>6, <3>1 DEF SemPRplyNonEmpty,Inv,j
                    <8>9. QED BY <8>1, <8>2, <7>6, <4>2, <5>1
                 <7>8. CwDist(ii, i) =< CwDist(k, i)
                    BY <7>7, <5>1, <5>5, <6>5 DEF ReadyNeighborLeft, Inv
                 <7>9. QED BY <7>5,<7>8, Unequality2, CwDistType, ReadyInI, IisNat
              <6>6. CASE ii = RightNeighbor(lset[i])
                 <7>1. SUFFICES ASSUME NEW k \in I,
                       k = RightNeighbor(lset'[i]),
                       k # ii
                       PROVE FALSE
                    BY <5>1, <5>2,<5>5, RightNeighborType
                 <7>2. lset'[i] = AddToLSet({j}, lset[i])
                    BY <4>03, <7>1, <6>6, RightNeighborType, <5>2 DEF TypeInvariant, ls1   
                 <7>3. CwDist(lset'[i].node, RightNeighbor(lset'[i])) =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                    BY AddToLSetInvCo,<5>2,<5>5,<7>2
                 <7>4. CwDist(i, k) =< CwDist(i, ii)
                    BY  <7>3, <7>1, <6>6,<5>5, <5>2
                 <7>5. CwDist(i, k) < CwDist(i, ii)
                       BY <7>4, <7>1, RightDistInjectivity, Unequality6,
                         CwDistType, IisNat
                 <7>6. k = j
                    BY <7>2, <7>1, <6>6, <5>2, <5>5, AddSingleNodeChange, SimplifyAndSolve
                 <7>7. /\ ~ii \in NodesReady
                       /\ k # i
                       /\  lset[k] # EmptyLS(k)
                    <8>1. /\ lset[k] \in LSet
                          /\ lset[k].node = k
                       BY DEF TypeInvariant
                    <8>2. lset[k]# EmptyLS(k)
                       BY <7>6, <3>1 DEF SemPRplyNonEmpty,Inv, j
                    <8>9. QED BY <8>1, <8>2, <7>6, <4>2, <5>1
                 <7>8. CwDist(i, ii) =< CwDist(i, k)
                    BY <7>7, <5>1, <5>5, <6>6 DEF ReadyNeighborRight, Inv
                 <7>9. QED BY <7>5,<7>8, Unequality2, CwDistType, ReadyInI, IisNat
              <6>9. QED BY <5>5, <6>6,<6>5,<5>3
           <5>7. CASE kk # i
              BY <5>1, <5>2, <5>3, <5>7, ReadyInI
           <5>9. QED BY <5>5, <5>7
        <4>68. SingleReadyNode' \* similar to ReceiveProbe
           <5>3. CASE i \in NodesReady /\ NodesReady = {i}
              <6>1. /\\A k \in I: /\ lset'[k] \in LSet 
                                  /\ lset[k] \in LSet
                                  /\ lset[k].node = k
                    /\ j \in I
                    /\ LeftNeighbor(lset[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                    /\ i # j
                 BY <3>1, <4>21, <4>2, LeftNeighborType, RightNeighborType DEF TypeInvariant
              <6>2. CASE lset'[i] = AddToLSet({j}, lset[i]) /\ LeftNeighbor(lset[i]) # i
                 <7>1. LeftNeighbor(lset[i]) = RightNeighbor(lset[i])
                    BY <5>3 DEF SingleReadyNode, Inv
                 <7>2. lset[j] # EmptyLS(j)
                       BY <3>1 DEF j,SemPRplyNonEmpty, Inv
                 <7>7. LeftNeighbor(lset'[i]) = j
                    <8>1. SUFFICES ASSUME LeftNeighbor(lset'[i]) # j 
                                   PROVE FALSE
                       OBVIOUS
                    <8>2. LeftNeighbor(lset'[i]) = LeftNeighbor(lset[i])
                       <9>1. LeftNeighbor(AddToLSet({j}, lset[i])) # j
                          BY <8>1, <6>2
                       <9>2. LeftNeighbor(AddToLSet({j}, lset[i]))
                             = LeftNeighbor(lset[i])
                          BY <9>1, AddSingleNodeChangeCo, <6>1 
                       <9>3. QED BY <9>2, <6>2
                    <8>3. RightNeighbor(lset[i]) # j
                       BY <7>1, <8>1, <8>2
                    <8>4. RightNeighbor(lset[i]) # i
                       BY <6>2, <7>1
                    <8>5. CASE CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~RightNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4
                       <9>4. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborRight, Inv
                       <9>9. QED BY <9>4, <8>5, IisNat, Unequality2, <6>1, CwDistType
                    <8>6. CASE CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1 
                       <9>3. ~LeftNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4, <7>1
                       <9>4. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborLeft, Inv
                       <9>9. QED BY <9>4, <8>6, IisNat, Unequality2, <6>1, CwDistType
                    <8>9. QED BY <6>1, <7>1, <8>4, <8>5,<8>6,<8>3,Ring3NodesPos7
                 <7>8. RightNeighbor(lset'[i]) = j
                    \* OMITTED \* Same as <7>7
                    <8>1. SUFFICES ASSUME RightNeighbor(lset'[i]) # j 
                                   PROVE FALSE
                       OBVIOUS
                    <8>2. RightNeighbor(lset'[i]) = RightNeighbor(lset[i])
                       <9>1. RightNeighbor(AddToLSet({j}, lset[i])) # j
                          BY <8>1, <6>2
                       <9>2. RightNeighbor(AddToLSet({j}, lset[i]))
                             = RightNeighbor(lset[i])
                          BY <9>1, AddSingleNodeChange, <6>1 
                       <9>3. QED BY <9>2, <6>2
                    <8>3. RightNeighbor(lset[i]) # j
                       BY <7>1, <8>1, <8>2
                    <8>4. RightNeighbor(lset[i]) # i
                       BY <6>2, <7>1
                    <8>5. CASE CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~RightNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4
                       <9>4. CwDist(i, RightNeighbor(lset[i])) =< CwDist(i, j)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborRight, Inv
                       <9>9. QED BY <9>4, <8>5, IisNat, Unequality2, <6>1, CwDistType
                    <8>6. CASE CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                       <9>1. /\ i \in NodesReady
                             /\ j \in I
                             /\ j # i
                          BY <5>3, <6>1
                       <9>3. ~LeftNeighbor(lset[i]) \in NodesReady
                          BY <5>3, <8>4, <7>1
                       <9>4. CwDist(LeftNeighbor(lset[i]), i) =< CwDist(j, i)
                          BY <9>1,<7>2,<9>3 DEF  ReadyNeighborLeft, Inv
                       <9>9. QED BY <9>4, <8>6, IisNat, Unequality2, <6>1, CwDistType
                    <8>9. QED BY <6>1, <7>1, <8>4, <8>5,<8>6,<8>3,Ring3NodesPos7
                 <7>9. QED BY <7>7,<7>8, ReadyInI, <4>07,<4>08, <4>22 
                  DEF SingleReadyNode, Inv, TypeInvariant
              <6>3. CASE LeftNeighbor(lset[i]) = i /\ lset'[i] = AddToLSet({j}, lset[i])
                 <7>1. RightNeighbor(lset[i]) = LeftNeighbor(lset[i])
                    BY <5>3 DEF SingleReadyNode, Inv
                 <7>2. RightNeighbor(lset[i]) = i
                    BY <6>3, <7>1
                 <7>5. lset[i] = EmptyLS(i)
                    BY <6>3, <7>2, EmptyLSNoNeighbor, <6>1
                 <7>6. LeftNeighbor(AddToLSet({j}, lset[i])) = j
                    BY AddSingleNodeToEmpty, <7>5, <6>1
                 <7>7. RightNeighbor(AddToLSet({j}, lset[i])) = j 
                    BY AddSingleNodeToEmpty, <7>5, <6>1
                 <7>8. /\ LeftNeighbor(lset'[i]) = j
                       /\ RightNeighbor(lset'[i]) = j
                    BY <7>6, <7>7, <6>3 
                 <7>9. QED BY <7>8, <4>08,<5>3 DEF SingleReadyNode, Inv
              <6>5. CASE lset'[i] = lset[i]
                 BY <6>5, <5>3, ReadyInI, <4>07,<4>08, <4>22 
                  DEF SingleReadyNode, Inv, TypeInvariant
              <6>9. QED BY <6>2, <6>3, <6>5, <4>13
           <5>4. CASE ~ i \in NodesReady \/ NodesReady # {i}
              BY <5>4, ReadyInI, <4>07,<4>08, <4>22 DEF SingleReadyNode, Inv, TypeInvariant
           <5>9. QED BY <5>3,<5>4
        <4>69. CoverToJ'
           <5>0. SUFFICES ASSUME NEW kk \in I, NEW rr \in NodesReady'
                 PROVE CoverToJ!(rr, kk)'
              BY DEF CoverToJ

           <5>2. /\ NodesReady' = NodesReady
                 /\ \A k \in I \{i}: lset'[k] = lset[k]
                 /\ lset[i] # EmptyLS(i)
                 /\ \A k \in I: toj'[k] = toj[k]
                 /\ j \in I
                 /\ lset[i] \in LSet
              <6>1. /\ \A k \in I \{i}: lset'[k] = lset[k]
                    /\ \A k \in I: toj'[k] = toj[k]
                    /\ lset[i] \in LSet
                 BY <4>03, <4>07, <4>8 DEF TypeInvariant
              <6>2. NodesReady' = NodesReady
                 BY <4>05, ReadyInI DEF shouldBeOK, NodesReady, NodesWait, NodesOK, TypeInvariant
              <6>3. /\ lset[i] # EmptyLS(i)
                    /\ j \in I
                 BY <3>1, <4>2
              <6>9. QED BY <6>1,<6>2,<6>3
           <5>3. CASE kk = i
              <6>1. SUFFICES ASSUME
                     lset'[i] # EmptyLS(i), 
                     Covers(lset'[rr], i),
                     i # rr
                     PROVE i = toj'[rr] 
                  BY <5>2,<5>0, <5>3 DEF CoverToJ, Inv
              <6>9. QED BY <5>2, <6>1,<5>0, ReadyInI DEF CoverToJ, Inv
           <5>4. CASE rr = i
              <6>1. SUFFICES ASSUME
                     lset'[kk] # EmptyLS(kk), 
                     Covers(lset'[i], kk),
                     i # kk
                     PROVE kk = toj'[i] 
                  BY <5>2,<5>0, <5>4 DEF CoverToJ, Inv
              <6>3. Covers(lset[i], kk)
                 <7>3. CASE lset'[i] = lset[i]
                    BY <6>1, <7>3
                 <7>4. CASE lset'[i] = AddToLSet({j}, lset[i])
                    <8>1. Covers(AddToLSet({j}, lset[i]), kk)
                       BY <6>1, <7>4, <5>2
                    <8>2. /\ lset[i] \in LSet
                          /\ {j} \in SUBSET I
                          /\ kk \in I
                       BY  <5>0, <5>2
                    <8>9. QED BY <8>1, AddToLSetCover, <8>2
                 <7>5. QED BY <4>7, <7>3,<7>4
              <6>9. QED BY <5>2, <6>1,<5>0,<6>3,<5>4, ReadyInI DEF CoverToJ, Inv
           <5>5. CASE kk # i /\ rr # i
              BY <5>5,<5>2, ReadyInI DEF CoverToJ, Inv
           <5>9. QED BY <5>3,<5>4, <5>5, <5>2, <5>0
        <4>70. SemPRplyNonEmpty'
          <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 \/ mm.mreq.type = "LSProbeReply"
                 \/ mm.mreq.type = "LSProbe"
                 PROVE 
                    lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
              BY DEF SemPRplyNonEmpty, Inv
           <5>4. CASE mm \in receivedMsgs 
              <6>1.  mm.mreq \in Prb \/ mm.mreq \in PRpl
                 <7>1. CASE mm.mreq.type = "LSProbeReply"
                    BY SlowestZenon,<7>1, <5>4 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                 <7>2. CASE mm.mreq.type = "LSProbe"
                      BY SlowestZenon,<7>2, <5>4 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                 <7>9. QED BY <7>1,<7>2,<5>1
              <6>4. m.mreq \in PRpl
                 BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
              <6>5. /\ mm.mreq.node \in I
                    /\ lset[i].node = i   
                    /\ m.mreq.node # i
                    /\ m.mreq.node \in I
                    /\ lset[i] \in LSet
                 BY <3>1, <4>21,<6>1, <6>4 DEF PRpl,Prb
              <6>6. lset[mm.mreq.node] # EmptyLS(mm.mreq.node)
                 BY <5>4, <5>1 DEF SemPRplyNonEmpty, Inv
              <6>7. CASE mm.mreq.node = i
                 <7>7. CASE lset'[i] = lset[i]
                    BY <6>6, <6>7, <7>7
                 <7>8. CASE lset'[i] = AddToLSet({m.mreq.node}, lset[i])
                    BY <6>6, <6>7, <7>8, <6>5, AddSingleNodeNoEmpty
                 <7>9. QED BY <7>7,<7>8, <4>13 DEF j
              <6>8. CASE mm.mreq.node # i
                 BY <4>22, <6>6,<6>8, <6>5
              <6>9. QED BY <6>7,<6>8
           <5>5. CASE ~mm \in receivedMsgs
              <6>. HIDE DEF ls1, fi, prb1, prb2
              <6>01. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                   Probe(i, ls1, fi, prb1)
                   \cup Probe(i, ls1, fi, prb2)
                 BY <4>06, <4>3 DEF TypeInvariant
              <6>02. \A ms \in Probe(i, ls1, fi, prb1): ms.mreq.type = "LSProbe"
                 <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb1)
                    BY ONLY ProbeMsgType, <4>3, <4>4, <4>5, <4>46
                 <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
              <6>03. \A ms \in Probe(i, ls1, fi, prb2): ms.mreq.type = "LSProbe"
                 <7>1. ProbeMsgTypeConclusion(i, ls1, fi, prb2)
                    BY ONLY ProbeMsgType, <4>3,<4>4,<4>5, <4>47
                 <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
              <6>04. \/ mm \in Probe(i, ls1, fi, prb1)
                     \/ mm \in Probe(i, ls1, fi, prb2)
                 BY <5>5, <5>1, <6>01
              <6>1. mm.mreq \in Prb
                 <7>1. CASE mm \in Probe(i, ls1, fi, prb1)
                    <8>0. /\ ls1 \in LSet
                          /\ fi \in SUBSET I
                          /\ prb1 \in SUBSET I
                       BY <4>3,<4>4, <4>5, <4>46
                    <8>1. mm.mreq.type = "LSProbe"
                       BY <6>02, <7>1
                    <8>2. mm \in DMsg
                       BY <7>1, <8>0, ProbeType
                    <8>3. mm.mreq \in Prb
                        BY SlowestZenon,<8>1, <8>2
                           DEF TypeInvariant, DMsg, 
                          MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                    <8>9. QED BY <8>3
                 <7>2. CASE mm \in Probe(i, ls1, fi, prb2)
                    <8>0. /\ ls1 \in LSet
                          /\ fi \in SUBSET I
                          /\ prb2 \in SUBSET I
                       BY <4>3,<4>4,<4>5,<4>47
                    <8>1. mm.mreq.type = "LSProbe"
                       BY <6>03, <7>2
                    <8>2. mm \in DMsg
                       BY <7>2, <8>0, ProbeType
                    <8>3. mm.mreq \in Prb
                        BY SlowestZenon,<8>1, <8>2
                           DEF TypeInvariant, DMsg, 
                          MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                    <8>9. QED BY <8>3
                 <7>9. QED BY <7>1,<7>2, <6>04
              <6>2. mm.mreq.node = i 
                 <7>1. CASE mm \in Probe(i, ls1, fi, prb1)
                    <8>1. ProbeMsgTypeConclusion(i, ls1, fi, prb1)
                    BY ONLY ProbeMsgType, <4>3, <4>4,<4>5,<4>46
                    <8>2. QED BY <8>1, <7>1 DEF ProbeMsgTypeConclusion
                 <7>2. CASE mm \in Probe(i, ls1, fi, prb2)
                    <8>1. ProbeMsgTypeConclusion(i, ls1, fi, prb2)
                    BY ONLY ProbeMsgType, <4>3, <4>4, <4>5,<4>47
                    <8>2. QED BY <8>1, <7>2 DEF ProbeMsgTypeConclusion
                 <7>9. QED BY <7>1,<7>2, <6>04
              <6>4. m.mreq \in PRpl
                 BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                    MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
              <6>5. /\ mm.mreq.node \in I
                    /\ lset[i].node = i   
                    /\ m.mreq.node # i
                    /\ m.mreq.node \in I
                    /\ lset[i] \in LSet
                 BY <3>1, <4>21,<6>1, <6>4 DEF Prb, PRpl
              <6>7. CASE j \in GetLSetContent(lset[i])
                 <7>2. j # i
                    BY <3>1 DEF j
                 <7>8. lset[i] # EmptyLS(i)
                    BY <6>7, <7>2,<6>5, NonEmptyLSet
                 <7>9. QED BY <7>8, <6>2,<6>7, <4>21, <4>7
              <6>8. CASE ~j \in GetLSetContent(lset[i])
                    BY <6>2, <6>8, <6>5, AddSingleNodeNoEmpty, <4>7, <4>21 DEF j
              <6>9. QED BY <6>7,<6>8, <4>13
           <5>9. QED BY <5>4,<5>5
        <4>71. NoCCJoin'
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW rr \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoin!(ii,rr,kk)'
             BY DEF NoCCJoin
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ j \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I \{i}: lset'[k] = lset[k]
                /\  RightNeighbor(lset'[i]) \in I
                /\ RightNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ rr \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk         
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ j \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I \{i}: lset'[k] = lset[k]
                    /\  RightNeighbor(lset'[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                 BY <4>7,<4>05,  <4>2,<4>8, <4>21, <4>22,<4>16, RightNeighborType, ReadyInI DEF  TypeInvariant
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>07, <4>05 DEF TypeInvariant, NodesReady
              <6>3. /\ ii \in I
                    /\ rr \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3   
           <5>2.  CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, RightNeighbor(lset[i]))
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(lset'[i].node, RightNeighbor(lset'[i])) 
                       =< CwDist(lset[i].node, RightNeighbor(lset[i]))
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1, <6>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
           <5>5. CASE ii = i
              <6>1. SUFFICES ASSUME i # kk,
                    lset[kk] # EmptyLS(kk),
                    rr = RightNeighbor(lset'[i]),
                    i # rr,
                    CwDist(i, kk) < CwDist(i, rr)
                    PROVE \/ toj[i] = kk  /\ toj[rr] = rr
                          \/ toj[rr] = kk /\ toj[i] = i
                 BY <5>5, <5>0, <5>1  DEF NoCCJoin
              <6>2. CwDist(i, kk) < CwDist(i, RightNeighbor(lset[i]))
                 BY <5>1, <5>2, <6>1, TransLESS, CwDistType, IisNat
              <6>3. rr = RightNeighbor(lset[i]) 
                 <7>1. SUFFICES ASSUME rr # RightNeighbor(lset[i]) 
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CASE RightNeighbor(lset[i])  \in NodesReady
                    <8>2. ~(Cardinality(NodesReady) =< 1)
                       BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <7>2, <7>1, <5>5
                    <8>3. CwDist(i, RightNeighbor(lset[i])) =<  CwDist(i, rr)
                       BY <8>2, <5>5, <5>1, <7>1, <5>0, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>4. CwDist(i, RightNeighbor(lset[i])) <  CwDist(i, rr)
                       BY <8>3, RightDistInjectivity, Unequality6, <7>1, <5>1, IisNat, CwDistType
                    <8>5. QED BY <8>4, <5>2, <6>1, CwDistType, <5>1, IisNat, Unequality2
                 <7>3. CASE ~RightNeighbor(lset[i])  \in NodesReady
                    <8>1. /\ i \in NodesReady
                          /\ kk \in I
                          /\ kk # i
                       BY <5>0, <5>1, <5>5, <6>1
                    <8>2. lset[kk] # EmptyLS(kk)
                       BY <6>1
                    <8>4. CwDist(i, RightNeighbor(lset[i]) ) =< CwDist(i, kk)
                       BY ReadyNeighborRight, <8>1, <8>2, <7>3 DEF ReadyNeighborRight, Inv
                    <8>5. QED BY <6>2, <8>4, Unequality2, CwDistType, IisNat, <5>1
                 <7>9. QED BY <7>3,<7>2
              <6>9. QED BY <5>1, <5>5, <6>1, <6>2, <6>3, NoCCJoin DEF NoCCJoin, Inv
           <5>6. CASE rr = i
              <6>1. SUFFICES ASSUME ii # kk,
                     lset'[kk] # EmptyLS(kk),
                     i = RightNeighbor(lset[ii]),
                     ii # i,
                     CwDist(ii, kk) < CwDist(ii, i)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[i] = i
                     \/ toj[i] = kk /\ toj[ii] = ii
                  BY <5>6, <5>0, <5>1 DEF NoCCJoin, Inv
              <6>2. kk # i
                 BY <6>1, <5>1, IisNat, CwDistZeroCompare
              <6>9.QED BY <6>1, <6>2, <5>0, <5>6, <5>1, NoCCJoin DEF  NoCCJoin, Inv
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     rr = RightNeighbor(lset[ii]),
                     ii # rr,
                     CwDist(ii, i) < CwDist(ii, rr)
                     PROVE 
                      \/ toj[ii] = i /\ toj[rr] = rr
                      \/ toj[rr] = i /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1  DEF NoCCJoin, Inv
              <6>2. ~(Cardinality(NodesReady) =< 1)
                 BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <6>1
              <6>7. ~i \in NodesReady
                 <7>1. SUFFICES ASSUME i \in NodesReady
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CwDist(ii, rr) =<  CwDist(ii, i)
                 BY <6>1, <5>1, <6>2,<7>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>2, <6>1, <5>1, IisNat, Unequality2, CwDistType
              <6>8. lset[i] # EmptyLS(i)
                 BY <6>7, <3>1
              <6>9. QED BY <6>1, <6>8, <5>1,  NoCCJoin DEF NoCCJoin, Inv              
           <5>8. CASE ii # i /\ rr # i /\ kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    rr = RightNeighbor(lset[ii]),
                    ii # rr,
                    CwDist(ii, kk) < CwDist(ii, rr)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[rr] = rr
                     \/ toj[rr] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>8 DEF NoCCJoin
              <6>. QED BY <5>1,<6>1, NoCCJoin DEF NoCCJoin, Inv
           <5>. QED
             BY <5>5, <5>6, <5>7, <5>8
        <4>72. NoCCJoinL'
          \*  OMITTED \* similar to <4>71
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW ll \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoinL!(ii,ll,kk)'
             BY DEF NoCCJoinL
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ j \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I \{i}: lset'[k] = lset[k]
                /\ LeftNeighbor(lset'[i]) \in I
                /\ LeftNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ ll \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk         
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ j \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I \{i}: lset'[k] = lset[k]
                    /\  LeftNeighbor(lset'[i]) \in I
                    /\ LeftNeighbor(lset[i]) \in I
                 BY <4>7,<4>05,  <4>2,<4>8, <4>21, <4>22,<4>16, LeftNeighborType, ReadyInI DEF  TypeInvariant
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>07, <4>05 DEF TypeInvariant, NodesReady
              <6>3. /\ ii \in I
                    /\ ll \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3   
           <5>2.  CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(LeftNeighbor(lset[i]), i)
                <6>1. /\ lset'[i] \in LSet 
                      /\ lset[i] \in LSet
                      /\ {j} \in SUBSET I
                   BY <5>1
                <6>7. CASE lset'[i] = AddToLSet({j}, lset[i])
                   <7>1.  CwDist(LeftNeighbor(lset'[i]), lset'[i].node ) 
                       =< CwDist(LeftNeighbor(lset[i]), lset[i].node)
                     BY <6>1, <6>7, AddToLSetInvCo, SlowestZenon
                   <7>2.QED BY <7>1, <5>1, <6>1
                <6>8. CASE lset'[i] = lset[i]
                      BY <6>8, Unequality6, 
                      CwDistType, <5>1, IisNat
                <6>9. QED BY <6>7, <6>8, <4>16 DEF ls1
           <5>5. CASE ii = i
              <6>1. SUFFICES ASSUME i # kk,
                    lset[kk] # EmptyLS(kk),
                    ll = LeftNeighbor(lset'[i]),
                    i # ll,
                    CwDist(kk, i) < CwDist(ll, i)
                    PROVE \/ toj[i] = kk  /\ toj[ll] = ll
                          \/ toj[ll] = kk /\ toj[i] = i
                 BY <5>5, <5>0, <5>1  DEF NoCCJoinL
              <6>2. CwDist(kk, i) < CwDist(LeftNeighbor(lset[i]), i)
                 BY <5>1, <5>2, <6>1, TransLESS, CwDistType, IisNat
              <6>3. ll = LeftNeighbor(lset[i]) 
                 <7>1. SUFFICES ASSUME ll # LeftNeighbor(lset[i]) 
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CASE LeftNeighbor(lset[i])  \in NodesReady
                    <8>2. ~(Cardinality(NodesReady) =< 1)
                       BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <7>2, <7>1, <5>5
                    <8>3. CwDist(LeftNeighbor(lset[i]), i) =<  CwDist(ll, i)
                       BY <8>2, <5>5, <5>1, <7>1, <5>0, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>4. CwDist(LeftNeighbor(lset[i]), i) <  CwDist(ll, i)
                       BY <8>3, CwDistInjectivity, Unequality6, <7>1, <5>1, IisNat, CwDistType
                    <8>5. QED BY <8>4, <5>2, <6>1, CwDistType, <5>1, IisNat, Unequality2
                 <7>3. CASE ~LeftNeighbor(lset[i])  \in NodesReady
                    <8>1. /\ i \in NodesReady
                          /\ kk \in I
                          /\ kk # i
                       BY <5>0, <5>1, <5>5, <6>1
                    <8>2. lset[kk] # EmptyLS(kk)
                       BY <6>1
                    <8>4. CwDist(LeftNeighbor(lset[i]) , i) =< CwDist(kk, i)
                       BY <8>1, <8>2, <7>3 DEF ReadyNeighborLeft, Inv
                    <8>5. QED BY <6>2, <8>4, Unequality2, CwDistType, IisNat, <5>1
                 <7>9. QED BY <7>3,<7>2
              <6>9. QED BY <5>1, <5>5, <6>1, <6>2, <6>3 DEF NoCCJoinL, Inv
           <5>6. CASE ll = i
              <6>1. SUFFICES ASSUME ii # kk,
                     lset'[kk] # EmptyLS(kk),
                     i = LeftNeighbor(lset[ii]),
                     ii # i,
                     CwDist(kk, ii) < CwDist(i, ii)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[i] = i
                     \/ toj[i] = kk /\ toj[ii] = ii
                  BY <5>6, <5>0, <5>1 DEF NoCCJoinL, Inv
              <6>2. kk # i
                 BY <6>1, <5>1, IisNat, CwDistZeroCompareCo
              <6>9.QED BY <6>1, <6>2, <5>0, <5>6, <5>1 DEF  NoCCJoinL, Inv
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     ll = LeftNeighbor(lset[ii]),
                     ii # ll,
                     CwDist(i, ii) < CwDist(ll, ii)
                     PROVE 
                      \/ toj[ii] = i /\ toj[ll] = ll
                      \/ toj[ll] = i /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1  DEF NoCCJoinL, Inv
              <6>2. ~(Cardinality(NodesReady) =< 1)
                 BY CardinalityTwoNode, ReadyInI, <5>0, <5>1, <6>1
              <6>7. ~i \in NodesReady
                 <7>1. SUFFICES ASSUME i \in NodesReady
                       PROVE FALSE
                    OBVIOUS
                 <7>2. CwDist(ll, ii) =<  CwDist(i, ii)
                 BY <6>1, <5>1, <6>2,<7>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>2, <6>1, <5>1, IisNat, Unequality2, CwDistType
              <6>8. lset[i] # EmptyLS(i)
                 BY <6>7, <3>1
              <6>9. QED BY <6>1, <6>8, <5>1 DEF NoCCJoinL, Inv              
           <5>8. CASE ii # i /\ ll # i /\ kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    ll = LeftNeighbor(lset[ii]),
                    ii # ll,
                    CwDist(kk, ii) < CwDist(ll, ii)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[ll] = ll
                     \/ toj[ll] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>8 DEF NoCCJoinL
              <6>. QED BY <5>1,<6>1 DEF NoCCJoinL, Inv
           <5>. QED
             BY <5>5, <5>6, <5>7, <5>8
        <4>73. SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY  DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. ASSUME i \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}
                 PROVE lset'[i] # EmptyLS(i)
              <6>6. lset[i] # EmptyLS(i)
                 BY <5>3,<5>4, <5>9, <4>22
                    DEF SemJRplyLSnonEmpty, Inv
              <6>7. CASE lset'[i] = lset[i]
                 BY <6>7, <6>6
              <6>8. CASE lset'[i] = AddToLSet({j}, lset[i])
                 BY AddToLSetNoEmpty, <6>6, <6>8, <4>21, <4>2
              <6>9. QED BY <6>7,<6>8, <4>7 
           <5>99. QED BY <5>3, <5>4, <5>8, <5>9, <4>22
                DEF SemJRplyLSnonEmpty, Inv
        <4>74. SingleReadyNodeToj'
           <5>1. SUFFICES ASSUME NEW kk \in I, 
                 NEW rr \in I,
                 NodesReady = {rr},
                 lset'[kk] # EmptyLS(kk),
                 kk # rr
                 PROVE toj[rr] = kk
              BY <4>07,<4>08 DEF SingleReadyNodeToj, Inv
           <5>2. /\ lset[i] \in LSet
                 /\ lset[i].node = i
                 /\ j \in I
                 /\ j # i
              BY <3>1, <4>2, <4>3,<4>8,<4>16 DEF j
           <5>4. lset'[i] # EmptyLS(i)
              <6>5. CASE j \in GetLSetContent(lset[i])
                 BY <4>7, <6>5, <5>2, NonEmptyLSet DEF j
              <6>6. CASE  ~ j \in GetLSetContent(lset[i])
                 BY <4>7, <6>6, <5>2, AddSingleNodeNoEmpty DEF j
              <6>9. QED BY <6>5,<6>6
           <5>5. CASE kk = i
              <6>5. CASE lset[i] = EmptyLS(i)
                 <7>1. i \in NodesReady
                    BY <3>1, <6>5
                 <7>2. i = rr
                    BY <7>1, <5>1
                 <7>3. QED BY <7>1,<7>2,<5>1, <5>5
              <6>6. CASE lset[i] # EmptyLS(i)
                 BY <5>1, <4>22, <6>6, <5>5 DEF SingleReadyNodeToj, Inv 
              <6>9. QED BY <6>5,<6>6
           <5>6. CASE rr = i
              <6>2. lset[kk] # EmptyLS(kk)
                 BY <5>1, <4>22, <5>6
              <6>3. QED BY <6>2, <5>1, SingleReadyNodeToj DEF SingleReadyNodeToj, Inv
           <5>7. CASE kk # i /\ rr # i
              BY <5>7, <4>22, <5>1 DEF SingleReadyNodeToj, Inv
           <5>9. QED BY <5>5,<5>6,<5>7
        <4>75.OneJRply'
           BY <4>10,<4>07,<4>08 DEF OneJRply, Inv, TypeInvariant 
        <4>76.OneJReq'
           BY <4>10,<4>07,<4>08 DEF OneJReq, Inv, TypeInvariant 
        <4>77. SemJReq'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type  = "JoinRequest"
                 PROVE
                   /\ mm.mreq.node \in NodesWait'
                   /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY DEF SemJReq, Inv
           <5>4. mm \in receivedMsgs
              BY <4>10, <5>3
           <5>5. mm.mreq \in JReq
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.node \in I
              BY <5>5 DEF JReq
           <5>7. /\ mm.mreq.node \in NodesWait
                 /\ ~ mm.mreq.node \in NodesReady
                 /\ lset[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY <5>3, <5>4, <5>6, StatusDisjoint DEF SemJReq, Inv
           <5>8. mm.mreq.node # i
              BY <5>7, <3>1, <5>6
           <5>9. QED BY <5>6,<5>7,<5>8, <4>22, <4>08
        <4>78. JoinMsgDual'
           BY <4>10,<4>07,<4>08 DEF JoinMsgDual, Inv, TypeInvariant 
        <4>80. AtleastOneReady'
           BY <4>08 DEF AtleastOneReady, Inv            
        <4>99. QED BY <4>20,(*<4>94,*)<4>95,\*<4>96,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, 
                  <4>70, <4>71, <4>72, <4>73, <4>74, 
                  <4>75, <4>76, <4>77, <4>78, <4>80 DEF Inv       
      <3>99.QED 
        BY <1>5, <3>1 DEF ReceiveLSPrRpl, Inv
   <1>6. ASSUME NEW i \in I,
                ReceiveJoinRequest(i),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
            status[i] = "ready",
            toj[i] = i,
            LeftNeighbor(lset[i]) \in lease[i],
            RightNeighbor(lset[i]) \in lease[i],
            toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i]),
            toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i]),
            m.destination = i,
            m.mreq.type = "JoinRequest",
            m.mreq.node # i, 
            CwDist(LeftCover(lset[i]), m.mreq.node) 
               =< CwDist(LeftCover(lset[i]), RightCover(lset[i])),
            NEW newmreq, newmreq=[type |-> "JoinReply",  
                                rtable |-> m.mreq.rtable,
                                  lset |-> lset[i]],
            NEW newm, newm =  [destination |-> m.mreq.node,
                           mreq |-> newmreq ],
            receivedMsgs'= (receivedMsgs \{m}) \cup {newm},
            toj' = [toj EXCEPT ![i]= m.mreq.node],
            UNCHANGED <<status, rtable, lset, probing, failed, lease, grant>>,
                    TypeInvariant,
                 Inv
         PROVE  Inv'
            <4>0. m.mreq \in JReq
               BY SlowestZenon, <3>1 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
            <4>1. \A j \in I: /\ lset'[j] = lset[j]
                              /\ grant'[j] = grant[j]
                              /\ lease'[j] = lease[j]
               BY UNCHANGED <<status, rtable, lset, probing, failed, lease, grant>>
            <4>2. /\ NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK 
                  /\ NodesWait' = NodesWait
                  /\ NonDead'= NonDead
               BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <3>1, <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>4. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                         \/ ms.mreq.type = "LSProbeReply" 
                         \/ ms.mreq.type = "LSProbe" 
                         \/ ms.mreq.type = "BroadcastLSet"
                         \/ ms.mreq.type = "JoinRequest"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                         {newm}
                     BY <3>1 DEF TypeInvariant
                  <6>2. newm.mreq.type = "JoinReply"
                     BY <3>1
                  <6>9. QED BY <6>1,<6>2
            <4>5. receivedMsgs' = (receivedMsgs\{m}) \union {newm}
                  BY <3>1
            <4>25. HalfNeighbor' 
               <5>1. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                   \/ ms.mreq.type = "LSProbeReply"
                   \/ ms.mreq.type = "LSProbe"
                   \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in {newm}
                     BY <3>1 DEF TypeInvariant
                  <6>2. newm.mreq.type = "JoinReply"
                     BY <3>1
                  <6>9. QED BY <6>1,<6>2
               <5>2./\ ReadyOK' = ReadyOK 
                    /\ NodesWait'= NodesWait
                    /\  \A k \in I: lset'[k] = lset[k]
                  BY <3>1 DEF ReadyOK, NodesReady, NodesOK, NodesWait
               <5>3. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                  BY <5>1, <5>2, ReadyInI, <5>3 DEF HalfNeighbor
               <5>4. CASE \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                                   => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                     
                  <6>1. PICK k \in ReadyOK': ReadyOK' = {k}
                     BY <5>4, <5>2
                  <6>2.  /\ LeftNeighbor(lset'[k]) = k 
                         /\ RightNeighbor(lset'[k]) = k 
                         /\ \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                     BY <6>1, <5>4, ReadyInI, <5>2
                  
                  <6>3. /\ ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                        /\ ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1, <5>1, <5>4
                  <6>4. ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # k
                     BY <6>1, <6>2, <5>1, <5>4
                  <6>8. \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     <7>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in {newm}
                        BY <3>1 DEF TypeInvariant
                     <7>2. newm.mreq.type = "JoinReply"
                        BY <3>1
                     <7>3. k = i
                        BY <3>1, <6>1, <5>2, <5>4, ReadyInI DEF NodesReady
                     <7>4. lset[k] = EmptyLS(k)
                        <8>1. lset[k] \in LSet /\ lset[k].node = k
                           BY ReadyInI, <6>1, <5>2 DEF TypeInvariant
                        <8>2. /\ LeftNeighbor(lset[k]) = k 
                              /\ RightNeighbor(lset[k]) = k
                           BY <6>2, <6>1, ReadyInI, <5>2 
                        <8>3.QED BY <8>1, <8>2, ReadyInI, EmptyLSNoNeighbor
                     <7>5. newm.mreq.lset = lset[i]
                        BY <3>1
                     <7>8. GetLSetContent(newm.mreq.lset) = {k}
                        BY <7>4,<7>5,<7>3, EmptyLSIsEmpty
                     <7>9. QED BY <7>1, <7>2, <7>3, <7>8, <6>2  
                  <6>9. QED BY <6>1, <6>2, <5>2, <6>3, <6>4, <6>8, ReadyInI DEF HalfNeighbor
               <5>9. QED BY <5>3,<5>4, HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>26. NeighborClosest'
            \*   BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>4,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>61. GrantReady'
               BY <4>1, <4>2, ReadyInI DEF GrantReady, Inv
            <4>62. SemJRply'
               <5>1. /\\A k \in I\{i}: toj'[k] = toj[k]
                     /\ toj'[i] = m.mreq.node
                  BY <3>1 DEF TypeInvariant
               <5>6. receivedMsgs' = (receivedMsgs\{m}) \union {newm}
                  BY <3>1
               <5>7. ASSUME NEW mm \in receivedMsgs,
                        mm.mreq.type = "JoinReply"
                        PROVE SemJRply!(mm)'
                  <6>2. mm.mreq \in JRpl
                      BY SlowestZenon,<5>7 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>3. /\ mm.destination \in I
                        /\ mm.mreq.lset.node \in I
                        /\ mm.mreq.lset \in LSet
                        /\ LeftNeighbor(mm.mreq.lset) \in I
                        /\ RightNeighbor(mm.mreq.lset) \in I
                     BY <6>2, LeftNeighborType, RightNeighborType
                     DEF JRpl, LSet, DMsg, TypeInvariant
                  <6>4. /\ mm.mreq.lset.node # i
                        /\ mm.mreq.lset.node \in NodesReady
                        /\ toj[mm.mreq.lset.node] = mm.destination
                        /\ toj'[mm.mreq.lset.node] = mm.destination
                     <7>1./\ mm.mreq.lset.node \in NodesReady 
                          /\ mm.mreq.lset.node # mm.destination
                        BY <5>7, <6>3, StatusDisjoint DEF SemJRply, Inv
                     <7>2. toj[mm.mreq.lset.node] = mm.destination
                        BY <5>7 DEF SemJRply, Inv
                     <7>3. toj[i] = i
                        BY <3>1
                     <7>4. mm.mreq.lset.node # i
                        BY <7>3, <7>2, <7>1
                     <7>9. QED BY  <6>3,<5>1, <7>2, <7>4, <7>1
                  <6>5. ASSUME LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                        PROVE toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
                     <7>4. ASSUME LeftNeighbor(mm.mreq.lset) = i PROVE FALSE
                        <8>1. LeftNeighbor(lset[mm.mreq.lset.node]) = i
                           BY <7>4, <5>7 DEF SemJoinLeafSet, Inv
                        <8>2. mm.mreq.lset.node # mm.destination
                           BY <5>7, <6>3, StatusDisjoint DEF SemJRply, Inv
                        <8>3. toj[mm.mreq.lset.node] = mm.destination
                           BY <5>7 DEF SemJRply, Inv
                        <8>4. toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                           BY <3>1
                        <8>5. RightNeighbor(lset[i]) = mm.mreq.lset.node
                           \*BY <8>1, <8>2, <6>4, <5>7, <3>1, <6>3 DEF LeftNeighborToj, Inv
                           <9>1. SUFFICES ASSUME NEW rn \in I,
                                 NEW k \in I,
                                 k = mm.mreq.lset.node,
                                 rn = RightNeighbor(lset[i]),
                                 k # rn 
                                 PROVE FALSE
                              BY RightNeighborType, <6>3 DEF TypeInvariant
                           <9>2. /\ k \in NodesReady
                                 /\ i \in NodesReady 
                                 /\ i # k
                              BY <9>1, <6>4, <3>1 DEF NodesReady                                  
                           <9>3. rn \in NodesReady
                              BY <9>1, <3>1 DEF LeaseGrant, GrantReady, Inv
                           <9>6. i # rn
                              <10>1. ~\E k1 \in ReadyOK: ReadyOK = {k1}
                                  BY <9>2, ReadyInI, CardinalityTwoNode, CardinalityOneNodeExt
                              <10>2. QED BY <10>1, <9>1, <9>2, ReadyInI DEF HalfNeighbor, Inv
                           <9>7. CASE CwDist(i, rn) < CwDist(i, k)                                  
                              <10>2. CwDist(i, k) =< CwDist(rn, k)
                                  BY <8>1, <9>1, <9>2, <9>3, ReadyInI,  
                                  CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                              <10>3. CwDist(i, k) < CwDist(rn, k)
                                  BY <10>2, <9>6, CwDistType, IisNat, Unequality6, CwDistInjectivity                                
                              <10>9. QED BY <9>6, <9>1, <9>2, <9>7, <10>3, Ring3NodesPos5
                           <9>8. CASE CwDist(i, k) < CwDist(i, rn)  
                              <10>1. CwDist(i, rn) =< CwDist(i, k)
                                  BY <9>1, <9>2, ReadyInI,  
                                  CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                              <10>2. QED BY <9>8, <10>1, CwDistType, IisNat,Unequality2 
                           <9>9. QED BY <9>1, <9>7,<9>8, Antisymmetric
                        <8>6. QED BY <8>5,<8>4,<8>3, <8>2
                     <7>5. QED BY <7>4, <5>1, <6>3,  <5>7, <6>5 DEF SemJRply, Inv
                  <6>6. ASSUME RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                        PROVE toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset)
                     \* OMITTED \* Similar to <6>5 
                     <7>4. ASSUME RightNeighbor(mm.mreq.lset) = i PROVE FALSE
                        <8>1. RightNeighbor(lset[mm.mreq.lset.node]) = i
                           BY <7>4, <5>7 DEF SemJoinLeafSet, Inv
                        <8>2. mm.mreq.lset.node # mm.destination
                           BY <5>7, <6>3, StatusDisjoint DEF SemJRply, Inv
                        <8>3. toj[mm.mreq.lset.node] = mm.destination
                           BY <5>7 DEF SemJRply, Inv
                        <8>4. toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                           BY <3>1
                        <8>5. LeftNeighbor(lset[i]) = mm.mreq.lset.node
                           \*BY <8>1, <8>2, <6>4, <5>7, <3>1, <6>3 DEF LeftNeighborToj, Inv
                           <9>1. SUFFICES ASSUME NEW ln \in I,
                                 NEW k \in I,
                                 k = mm.mreq.lset.node,
                                 ln = LeftNeighbor(lset[i]),
                                 k # ln 
                                 PROVE FALSE
                              BY LeftNeighborType, <6>3 DEF TypeInvariant
                           <9>2. /\ k \in NodesReady
                                 /\ i \in NodesReady 
                                 /\ i # k
                              BY <9>1, <6>4, <3>1 DEF NodesReady                                  
                           <9>3. ln \in NodesReady
                              BY <9>1, <3>1 DEF LeaseGrant, GrantReady, Inv
                           <9>6. i # ln
                              <10>1. ~\E k1 \in ReadyOK: ReadyOK = {k1}
                                  BY <9>2, ReadyInI, CardinalityTwoNode, CardinalityOneNodeExt
                              <10>2. QED BY <10>1, <9>1, <9>2, ReadyInI DEF HalfNeighbor, Inv
                           <9>7. CASE CwDist(ln, i) < CwDist(k, i)                                  
                              <10>2. CwDist(k, i) =< CwDist(k, ln)
                                  BY <8>1, <9>1, <9>2, <9>3, ReadyInI,  
                                  CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                              <10>3. CwDist(k, i) < CwDist(k, ln)
                                  BY <10>2, <9>6, CwDistType, IisNat, Unequality6, RightDistInjectivity                                
                              <10>9. QED BY <9>6, <9>1, <9>2, <9>7, <10>3, Ring3NodesPos8
                           <9>8. CASE CwDist(k, i) < CwDist(ln, i)  
                              <10>1. CwDist(ln, i) =< CwDist(k, i)
                                  BY <9>1, <9>2, ReadyInI,  
                                  CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                              <10>2. QED BY <9>8, <10>1, CwDistType, IisNat,Unequality2 
                           <9>9. QED BY <9>1, <9>7,<9>8, Antisymmetric2
                        <8>6. QED BY <8>5,<8>4,<8>3, <8>2
                     <7>5. QED BY <7>4, <5>1, <6>3,  <5>7, <6>6 DEF SemJRply, Inv
                  <6>9. QED BY <4>1,<4>2, <5>7, <6>3, <6>4,<6>5,<6>6 DEF SemJRply, Inv
               <5>8. SemJRply!(newm)'
                  <6>1. newm.mreq \in JRpl
                      BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>2. m.mreq \in JReq
                      BY SlowestZenon,<3>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>3. /\ m.mreq.node \in I
                        /\ newm.destination \in I
                        /\ newm.mreq.lset.node \in I
                        /\ newm.mreq.lset \in LSet
                        /\ LeftNeighbor(newm.mreq.lset) \in I
                        /\ RightNeighbor(newm.mreq.lset) \in I
                     BY <6>1,<6>2, <3>1, LeftNeighborType, RightNeighborType
                     DEF JRpl, JReq, LSet, DMsg, TypeInvariant
                  <6>4. /\ newm.mreq.lset.node = i
                        /\ newm.mreq.lset = lset[i] 
                     BY <3>1, <6>3 DEF TypeInvariant
                  <6>5./\ newm.destination = m.mreq.node 
                       /\ newm.destination # i
                     BY <3>1,  <6>3, <6>4
                  <6>6. newm.mreq.lset.node \in NodesReady
                     BY <6>4, <3>1 DEF NodesReady
                  <6>7.  newm.destination \in NodesWait
                     <7>0. /\ m \in receivedMsgs
                           /\ m.mreq.type = "JoinRequest"
                        BY <3>1
                     <7>1. m.mreq.node \in NodesWait
                        BY  <7>0 DEF SemJReq, Inv
                     <7>9. QED BY  <6>5, <7>1
                  <6>8. ASSUME LeftNeighbor(newm.mreq.lset) # newm.mreq.lset.node
                        PROVE toj'[LeftNeighbor(newm.mreq.lset)] = LeftNeighbor(newm.mreq.lset)
                     <7>0. /\ lset[i].node = i
                           /\ lset[i] \in LSet
                        BY DEF TypeInvariant
                     <7>1. LeftNeighbor(lset[i]) # lset[i].node
                        BY <6>8, <7>0, <6>4
                     <7>2. LeftNeighbor(lset[i]) # i
                        BY <7>1 DEF TypeInvariant
                     <7>3. toj'[LeftNeighbor(lset[i])] = toj[LeftNeighbor(lset[i])]
                        BY <7>2, <5>1, <6>3, <6>4
                     <7>4. toj[LeftNeighbor(lset[i])] = LeftNeighbor(lset[i])
                        BY <3>1
                     <7>9. QED BY <7>4, <7>3, <6>4
                  <6>9. ASSUME RightNeighbor(newm.mreq.lset) # newm.mreq.lset.node
                        PROVE toj'[RightNeighbor(newm.mreq.lset)] = RightNeighbor(newm.mreq.lset)
                     \* OMITTED \* Similar to <6>8
                     <7>0. /\ lset[i].node = i
                           /\ lset[i] \in LSet
                        BY DEF TypeInvariant
                     <7>1. RightNeighbor(lset[i]) # lset[i].node
                        BY <6>9, <7>0, <6>4
                     <7>2. RightNeighbor(lset[i]) # i
                        BY <7>1 DEF TypeInvariant
                     <7>3. toj'[RightNeighbor(lset[i])] = toj[RightNeighbor(lset[i])]
                        BY <7>2, <5>1, <6>3, <6>4
                     <7>4. toj[RightNeighbor(lset[i])] = RightNeighbor(lset[i])
                        BY <3>1
                     <7>9. QED BY <7>4, <7>3, <6>4
                  <6>10. Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF Covers
                  <6>11. lset[newm.destination] = EmptyLS(newm.destination)
                     BY <3>1, <6>5 DEF SemJReq, Inv
                  <6>12. toj'[newm.mreq.lset.node] = newm.destination
                     BY <6>4, <6>5, <5>1 
                  <6>13. /\ RightNeighbor(lset[newm.mreq.lset.node]) \in lease[newm.mreq.lset.node]
                         /\ LeftNeighbor(lset[newm.mreq.lset.node]) \in lease[newm.mreq.lset.node]
                     BY <6>4, <3>1
                  <6>99. QED BY <4>1,<4>2,<6>3, <6>4,<6>5,<6>6,
                        <6>7,<6>8,<6>9, <6>10, <6>11,<6>13,<6>12
               <5>9. QED BY  <5>7,<5>6,<5>8, SlowestZenon DEF SemJRply, Inv
            <4>63. SemJoinNeighbor'
               <5>2. receivedMsgs' = (receivedMsgs\{m}) \union {newm}
                  BY <3>1
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>4. CASE mm = newm
                  <6>1. /\ mm.destination = m.mreq.node
                        /\ mm.mreq.lset = lset[i]
                     BY <5>4, <3>1
                  <6>2. /\ lset[i].node = i
                        /\ lset[i] \in LSet
                        /\ i # m.mreq.node
                        /\ Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF TypeInvariant, Covers
                  <6>3. mm.mreq.lset.node = i
                     BY <6>1, <6>2 
                  <6>4. m.mreq.node \in I
                     BY <4>0 DEF JReq
                  <6>5. \/ CwDist(i, m.mreq.node) < CwDist(i, RightNeighbor(lset[i]))
                        \/ CwDist(m.mreq.node, i) < CwDist(LeftNeighbor(lset[i]) , i)
                     BY NeighborCoverageCo, <6>2, <6>4 DEF TypeInvariant
                  <6>7. CASE CwDist(i, m.mreq.node) < CwDist(i, RightNeighbor(lset[i]))
                     BY <5>3, <6>1,<6>2,<6>3,<6>4, <6>7, AddToEmptyNeighborRight
                  <6>8. CASE CwDist(m.mreq.node, i) < CwDist(LeftNeighbor(lset[i]) , i)
                     BY <5>3, <6>1,<6>2,<6>3,<6>4, <6>8, AddToEmptyNeighborLeft
                  <6>9. QED BY <6>7,<6>8, <6>5
               <5>5. CASE mm # newm
                  BY <5>2, <5>5, <5>3 DEF SemJoinNeighbor, Inv
               <5>9. QED BY <5>4,<5>5
            <4>64. SemSingleReadyJoin'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. CASE mm = newm
                  <6>1. /\ mm.destination = m.mreq.node
                        /\ mm.mreq.lset = lset[i]
                     BY <5>4, <3>1
                  <6>2. /\ lset[i].node = i
                        /\ lset[i] \in LSet
                        /\ i # m.mreq.node
                        /\ Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF TypeInvariant, Covers
                  <6>3. mm.mreq.lset.node = i
                     BY <6>1, <6>2 
                  <6>4. m.mreq.node \in I
                     BY <4>0 DEF JReq
                  <6>5.  lset'[mm.mreq.lset.node] = mm.mreq.lset
                     BY <6>3, <6>1, <4>1
                  <6>6. NodesReady = {mm.mreq.lset.node}
                     <7>0. i \in NodesReady
                        BY <3>1 DEF NodesReady
                     <7>2. lset[i] = EmptyLS(i)
                        BY <5>1, <6>1, <6>3
                     <7>3. /\LeftNeighbor(lset[i]) = i
                           /\ RightNeighbor(lset[i]) = i
                        BY EmptyLSNoNeighbor, <6>2, <7>2
                     <7>4. \E k \in ReadyOK: ReadyOK = {k}
                        BY <7>0, <7>3, <6>2, ReadyInI DEF HalfNeighbor, Inv
                     <7>9. QED BY <7>0, <7>4, ReadyInI, <6>3
                  <6>9. QED BY <6>5,<6>6
               <5>5. CASE mm # newm
                  <6>1. mm \in receivedMsgs
                     BY <5>1, <5>5, <4>5
                  <6>2. mm.mreq \in JRpl
                     BY SlowestZenon, <6>1, <5>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>3. mm.mreq.lset.node \in I
                     BY <6>2 DEF JRpl, LSet
                  <6>9.QED BY <4>5, <5>5, <5>1, <6>3, <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>9. QED BY <5>4,<5>5
            <4>65. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>7. CASE mm = newm
                  <6>1. /\ mm.destination = m.mreq.node
                        /\ mm.mreq.lset = lset[i]
                     BY <5>7, <3>1
                  <6>2. /\ lset[i].node = i
                        /\ lset[i] \in LSet
                        /\ i # m.mreq.node
                        /\ Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF TypeInvariant, Covers
                  <6>3. mm.mreq.lset.node = i
                     BY <6>1, <6>2 
                  <6>4. m.mreq.node \in I
                     BY <4>0 DEF JReq
                  <6>5.  lset'[mm.mreq.lset.node] = mm.mreq.lset
                     BY <6>3, <6>1, <4>1
                  <6>7. ASSUME lset[i] = EmptyLS(i) 
                        PROVE \/ i  = LeftNeighbor(lset[i])
                              \/ i  = RightNeighbor(lset[i])
                     <7>0. i \in NodesReady
                        BY <3>1 DEF NodesReady
                     <7>3. QED BY EmptyLSNoNeighbor, <6>2, <6>7
                  <6>8. ASSUME \/ i  = LeftNeighbor(lset[i])
                               \/ i  = RightNeighbor(lset[i])
                        PROVE  lset[i] = EmptyLS(i)  
                     <7>0. i \in NodesReady
                        BY <3>1 DEF NodesReady
                     <7>4. \E k \in ReadyOK: 
                         /\ ReadyOK = {k}
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                        BY <7>0, <6>8, <6>2, ReadyInI DEF HalfNeighbor, Inv
                     <7>9. QED BY EmptyLSNoNeighbor, <7>0, ReadyInI, <6>2, <7>4
                  <6>9. QED BY <6>7, <6>8, <6>1, <6>3
               <5>8. CASE mm # newm
                  BY <4>5, <5>3, <5>8
                    DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <5>7,<5>8
            <4>66. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
               <5>7. CASE mm = newm
                  <6>1. /\ mm.destination = m.mreq.node
                        /\ mm.mreq.lset = lset[i]
                     BY <5>7, <3>1
                  <6>2. /\ lset[i].node = i
                        /\ lset[i] \in LSet
                        /\ i # m.mreq.node
                        /\ Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF TypeInvariant, Covers
                  <6>3. mm.mreq.lset.node = i
                     BY <6>1, <6>2 
                  <6>4. m.mreq.node \in I
                     BY <4>0 DEF JReq
                  <6>5.  lset'[mm.mreq.lset.node] = mm.mreq.lset
                     BY <6>3, <6>1, <4>1
                  <6>9. QED BY <6>5
               <5>8. CASE mm # newm
                  <6>1. mm \in receivedMsgs
                     BY <5>3, <5>8, <4>5
                  <6>2. mm.mreq \in JRpl
                     BY SlowestZenon, <6>1, <5>3 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>3. mm.mreq.lset.node \in I
                     BY <6>2 DEF JRpl, LSet
                  <6>9.QED BY <4>5, <5>3, <5>8, <6>3, 
                       <4>1, <4>2 DEF SemJoinLeafSet, Inv
               <5>9. QED BY <5>7,<5>8
            <4>67. GrantNeighbor'
               BY <3>1, <4>1, <4>2, ReadyInI DEF GrantNeighbor, Inv, TypeInvariant
            <4>68. SingleReadyNode'
               BY <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               <5>0. SUFFICES ASSUME NEW rr \in NodesReady', NEW kk \in I
                     PROVE CoverToJ! (rr,kk)'
                  BY DEF CoverToJ
               <5>1. /\ \A k \in I: lset'[k] = lset[k]
                     /\ \A k \in I\{i}: toj'[k] = toj[k]
                     /\ NodesReady = NodesReady' 
                     /\ i \in NodesReady
                  BY <3>1, <4>1, <4>2 DEF TypeInvariant, NodesReady
               <5>5. CASE rr = i
                  <6>1. SUFFICES ASSUME 
                        lset'[kk] # EmptyLS(kk),
                        Covers(lset'[i], kk),
                        kk # i,
                        kk # toj'[i]
                        PROVE FALSE
                     BY <5>0, <5>5 DEF CoverToJ, Inv
                  <6>2. /\ lset[kk] # EmptyLS(kk)
                        /\ Covers(lset[i], kk)
                        /\ kk # i
                     BY <6>1, <5>1
                  <6>3. toj[i] = kk
                     BY <6>2, <5>0, <5>1, <5>5 DEF CoverToJ, Inv
                  <6>4. toj[i] = i
                     BY <3>1
                  <6>9. QED BY <6>3, <6>4, <6>1
               <5>6. CASE kk = i
                  <6>1. SUFFICES ASSUME 
                        lset'[i] # EmptyLS(i),
                        Covers(lset'[rr], i),
                        i # rr 
                        PROVE FALSE
                     BY <5>0, <5>6 DEF CoverToJ, Inv
                  <6>2. /\lset[i] # EmptyLS(i)
                        /\Covers(lset[rr], i)
                        /\i # rr
                        /\ rr \in I
                        /\ lset[rr].node = rr
                        /\ lset[rr] \in LSet
                        /\ rr \in NodesReady
                        /\ i \in NodesReady
                     BY <5>1, <5>0, ReadyInI, <6>1 DEF TypeInvariant
                  <6>3. ~(Cardinality(NodesReady) =< 1)
                     BY CardinalityTwoNode, <6>1, <6>2, ReadyInI
                  <6>5. CASE CwDist(rr, i) < CwDist(rr, RightNeighbor(lset[rr]))
                     <7>1. CwDist(rr, RightNeighbor(lset[rr])) =< CwDist(rr, i)
                        BY <6>2, <6>1, <6>3, LemmaCLtoNC DEF NeighborClosest, Inv
                     <7>9. QED BY <7>1, <6>5, Unequality2, 
                           <6>2, RightNeighborType, CwDistType, IisNat
                  <6>6. CASE CwDist(i, rr) < CwDist(LeftNeighbor(lset[rr]) , rr)
                     <7>1. CwDist(LeftNeighbor(lset[rr]) , rr) =< CwDist(i, rr)
                        BY <6>2, <6>1, <6>3, LemmaCLtoNC DEF NeighborClosest, Inv
                     <7>2. QED BY <7>1, <6>6, Unequality2, 
                           <6>2, LeftNeighborType, CwDistType, IisNat
                  <6>7. QED BY <6>2,<6>5,<6>6, NeighborCoverageCo 
               <5>7. CASE rr # i /\ kk # i
                  BY <5>7,<5>1, <5>0, ReadyInI DEF CoverToJ, Inv
               <5>9. QED BY <5>5,<5>6,<5>7
            <4>70. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>71. NoCCJoin'
               <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW rr \in NodesReady', NEW kk \in I
                              PROVE  NoCCJoin!(ii,rr,kk)'
                 BY DEF NoCCJoin
               <5>1. 
                    /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ m.mreq.node \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I: lset'[k] = lset[k]
                    /\  RightNeighbor(lset'[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                    /\ NodesReady' = NodesReady
                    /\ \A k \in I\{i}: toj'[k] = toj[k]
                    /\ ii \in I
                    /\ rr \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk  
                    /\ toj'[i] = m.mreq.node       
                  <6>1. /\ lset[i] \in LSet
                        /\ m.mreq.node \in I
                        /\ \A k \in I : lset'[k] = lset[k]
                        /\ lset[i].node = i
                        /\ RightNeighbor(lset[i]) \in I
                        /\  toj'[i] = m.mreq.node    
                     BY <3>1, <4>0, <4>1, <4>2, RightNeighborType, ReadyInI DEF  TypeInvariant, JReq
                  <6>2. /\ NodesReady' = NodesReady
                        /\ \A k \in I\{i}: toj'[k] = toj[k]
                     BY <3>1 DEF TypeInvariant, NodesReady
                  <6>3. /\ ii \in I
                        /\ rr \in I
                        /\ lset[kk] \in LSet
                        /\ lset[kk].node = kk     
                     BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
                  <6>9. QED BY <6>1,<6>2,<6>3   
               <5>5. CASE ii = i
                  <6>1. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        rr = RightNeighbor(lset[i]),
                        i # rr,
                        CwDist(i, kk) < CwDist(i, rr)
                        PROVE \/ toj'[i] = kk  /\ toj[rr] = rr
                              \/ toj[rr] = kk /\ toj'[i] = i
                     BY <5>5, <5>0, <5>1 DEF NoCCJoin
                  <6>2. toj[i]' # i
                     BY <3>1, <5>1
                  <6>4. toj[rr] = rr
                     BY <6>1, <5>1, <3>1
                  <6>8. toj[i]'= kk
                     <7>1. SUFFICES ASSUME NEW on \in I,
                           toj[i]' = on,
                           on # kk
                           PROVE FALSE
                         BY <5>1
                     <7>2. m.mreq.node = on
                        BY <5>1, <7>1
                     <7>3. \/ toj[i] = kk  /\ toj[rr] = rr
                           \/ toj[rr] = kk /\ toj[i] = i
                        BY <6>1, <5>5, <5>0, <5>1  DEF NoCCJoin, Inv
                     <7>9. QED BY <6>4, <6>1, <3>1, <7>3
                  <6>9. QED BY <6>8, <6>4
               <5>6. CASE rr = i
                  <6>1. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         i = RightNeighbor(lset[ii]),
                         ii # i,
                         CwDist(ii, kk) < CwDist(ii, i)
                        PROVE \/ toj[ii] = kk  /\ toj'[i] = i
                              \/ toj'[i] = kk /\ toj[ii] = ii
                      BY <5>6, <5>0, <5>1  DEF NoCCJoin, Inv
                  <6>2. toj[i]' # i
                     BY <3>1, <5>1
                  <6>3. \/ toj[ii] = kk /\ toj[i] = i
                        \/ toj[i] = kk /\ toj[ii] = ii
                     BY <6>1, <5>0, <5>1, <5>6  DEF NoCCJoin, Inv
                  <6>4. toj[i] = i
                     BY <3>1
                  <6>5. ii = LeftNeighbor(lset[i])
                     <7>1. SUFFICES ASSUME NEW ln \in I,
                           ln = LeftNeighbor(lset[i]),
                           i = RightNeighbor(lset[ii]),
                           ii # ln,
                           ii # i
                           PROVE FALSE
                        BY <6>1, LeftNeighborType DEF TypeInvariant
                     <7>2. /\ ii \in I
                           /\ ii \in NodesReady
                           /\ i \in NodesReady
                        BY <5>1, <3>1 DEF NodesReady
                     <7>3. ln \in NodesReady
                        BY <3>1, <7>1 DEF LeaseGrant, GrantReady, Inv
                     <7>4. i # ln
                        <8>1. ~\E k1 \in ReadyOK: ReadyOK = {k1}
                           BY <6>1, <7>2, ReadyInI, CardinalityTwoNode, CardinalityOneNodeExt
                        <8>9. QED BY <8>1, ReadyInI, <7>1, <7>2 DEF HalfNeighbor, Inv
                     <7>7. CASE CwDist(ii, i) < CwDist(ln, i)
                        <8>1. CwDist(ln, i) =< CwDist(ii, i)
                           BY <7>1, <7>2, ReadyInI,  
                              CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                        <8>9. QED BY <7>2, <8>1, <7>7, CwDistType, IisNat,Unequality2 
                     <7>8. CASE CwDist(ln, i) < CwDist(ii, i)
                        <8>1. CwDist(ii, i) =< CwDist(ii, ln)
                           BY <7>1, <7>2,<7>3, ReadyInI,  
                              CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                        <8>2. CwDist(ii, i) < CwDist(ii, ln)
                           BY <7>4, <8>1,<7>2, CwDistType, IisNat, Unequality6, RightDistInjectivity
                        <8>9. QED BY Ring3NodesPos8, <8>2, <7>8, <7>4, <7>1, <7>2
                     <7>9. QED BY <7>2, <7>1, <7>7,<7>8, Antisymmetric2
                  <6>6. toj[ii] = ii
                     BY <6>5, <3>1
                  <6>9.QED BY <6>1, <6>3, <6>4, <6>6
               <5>7. CASE kk = i
                  <6>1. SUFFICES ASSUME ii # i,
                         lset'[i] # EmptyLS(i),
                         rr = RightNeighbor(lset[ii]),
                         ii # rr,
                         CwDist(ii, i) < CwDist(ii, rr)
                         PROVE FALSE
                     BY <5>7, <5>0, <5>1, SimplifyAndSolve  DEF NoCCJoin, Inv
                  <6>2. i \in NodesReady
                     BY <3>1 DEF TypeInvariant, NodesReady
                  <6>3. CwDist(ii, rr) =< CwDist(ii, i)
                        BY <5>0, <5>1, <6>1, <6>2, ReadyInI, CardinalityTwoNode, 
                         LemmaCLtoNC DEF NeighborClosest, Inv
                  <6>4. QED BY <6>3, <6>1, CwDistType, 
                        ReadyInI, <5>0, <5>1, IisNat, Unequality2
               <5>8. CASE ii # i /\ rr # i /\ kk # i
                  <6>1. SUFFICES ASSUME 
                        ii # kk,
                        lset[kk] # EmptyLS(kk),
                        rr = RightNeighbor(lset[ii]),
                        ii # rr,
                        CwDist(ii, kk) < CwDist(ii, rr)
                        PROVE 
                         \/ toj[ii] = kk /\ toj[rr] = rr
                         \/ toj[rr] = kk /\ toj[ii] = ii
                  BY <5>1, <5>0, <5>8 DEF NoCCJoin
                  <6>. QED BY <5>1,<6>1, NoCCJoin DEF NoCCJoin, Inv
               <5>. QED
                 BY <5>5, <5>6, <5>7, <5>8      
            <4>72. NoCCJoinL'
               \* OMITTED \* Similar to <4>71
               <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW ll \in NodesReady', NEW kk \in I
                              PROVE  NoCCJoinL!(ii,ll,kk)'
                 BY DEF NoCCJoinL
               <5>1. 
                    /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ m.mreq.node \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ \A k \in I: lset'[k] = lset[k]
                    /\ LeftNeighbor(lset'[i]) \in I
                    /\ LeftNeighbor(lset[i]) \in I
                    /\ NodesReady' = NodesReady
                    /\ \A k \in I\{i}: toj'[k] = toj[k]
                    /\ ii \in I
                    /\ ll \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk  
                    /\ toj'[i] = m.mreq.node       
                  <6>1. /\ lset[i] \in LSet
                        /\ m.mreq.node \in I
                        /\ \A k \in I : lset'[k] = lset[k]
                        /\ lset[i].node = i
                        /\ LeftNeighbor(lset[i]) \in I
                        /\  toj'[i] = m.mreq.node    
                     BY <3>1, <4>0, <4>1, <4>2, LeftNeighborType, ReadyInI DEF  TypeInvariant, JReq
                  <6>2. /\ NodesReady' = NodesReady
                        /\ \A k \in I\{i}: toj'[k] = toj[k]
                     BY <3>1 DEF TypeInvariant, NodesReady
                  <6>3. /\ ii \in I
                        /\ ll \in I
                        /\ lset[kk] \in LSet
                        /\ lset[kk].node = kk     
                     BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
                  <6>9. QED BY <6>1,<6>2,<6>3   
               <5>5. CASE ii = i
                  <6>1. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        ll = LeftNeighbor(lset[i]),
                        i # ll,
                        CwDist(kk, i) < CwDist(ll, i)
                        PROVE \/ toj'[i] = kk  /\ toj[ll] = ll
                              \/ toj[ll] = kk /\ toj'[i] = i
                     BY <5>5, <5>0, <5>1 DEF NoCCJoinL
                  <6>2. toj[i]' # i
                     BY <3>1, <5>1
                  <6>4. toj[ll] = ll
                     BY <6>1, <5>1, <3>1
                  <6>8. toj[i]'= kk
                     <7>1. SUFFICES ASSUME NEW on \in I,
                           toj[i]' = on,
                           on # kk
                           PROVE FALSE
                         BY <5>1
                     <7>2. m.mreq.node = on
                        BY <5>1, <7>1
                     <7>3. \/ toj[i] = kk  /\ toj[ll] = ll
                           \/ toj[ll] = kk /\ toj[i] = i
                        BY <6>1, <5>5, <5>0, <5>1  DEF NoCCJoinL, Inv
                     <7>9. QED BY <6>4, <6>1, <3>1, <7>3
                  <6>9. QED BY <6>8, <6>4
               <5>6. CASE ll = i
                  <6>1. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         i = LeftNeighbor(lset[ii]),
                         ii # i,
                         CwDist(kk, ii) < CwDist(i, ii)
                        PROVE \/ toj[ii] = kk  /\ toj'[i] = i
                              \/ toj'[i] = kk /\ toj[ii] = ii
                      BY <5>6, <5>0, <5>1  DEF NoCCJoinL, Inv
                  <6>2. toj[i]' # i
                     BY <3>1, <5>1
                  <6>3. \/ toj[ii] = kk /\ toj[i] = i
                        \/ toj[i] = kk /\ toj[ii] = ii
                     BY <6>1, <5>0, <5>1, <5>6  DEF NoCCJoinL, Inv
                  <6>4. toj[i] = i
                     BY <3>1
                  <6>5. ii = RightNeighbor(lset[i])
                     <7>1. SUFFICES ASSUME NEW rn \in I,
                           rn = RightNeighbor(lset[i]),
                           i = LeftNeighbor(lset[ii]),
                           ii # rn,
                           ii # i
                           PROVE FALSE
                        BY <6>1, LeftNeighborType, RightNeighborType DEF TypeInvariant
                     <7>2. /\ ii \in I
                           /\ ii \in NodesReady
                           /\ i \in NodesReady
                        BY <5>1, <3>1 DEF NodesReady
                     <7>3. rn \in NodesReady
                        BY <3>1, <7>1 DEF LeaseGrant, GrantReady, Inv
                     <7>4. i # rn
                        <8>1. ~\E k1 \in ReadyOK: ReadyOK = {k1}
                           BY <6>1, <7>2, ReadyInI, CardinalityTwoNode, CardinalityOneNodeExt
                        <8>9. QED BY <8>1, ReadyInI, <7>1, <7>2 DEF HalfNeighbor, Inv
                     <7>7. CASE CwDist(i, ii) < CwDist( i, rn)
                        <8>1. CwDist(i, rn) =< CwDist(i, ii)
                           BY <7>1, <7>2, ReadyInI,  
                              CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                        <8>9. QED BY <7>2, <8>1, <7>7, CwDistType, IisNat,Unequality2 
                     <7>8. CASE CwDist(i, rn) < CwDist(i, ii)
                        <8>1. CwDist(i, ii) =< CwDist(rn, ii)
                           BY <7>1, <7>2,<7>3, ReadyInI,  
                              CardinalityTwoNode, LemmaCLtoNC DEF NeighborClosest, Inv
                        <8>2. CwDist(i, ii) < CwDist(rn, ii)
                           BY <7>4, <8>1,<7>2, CwDistType, IisNat, Unequality6, CwDistInjectivity
                        <8>9. QED BY Ring3NodesPos12, <8>2, <7>8, <7>4, <7>1, <7>2
                     <7>9. QED BY <7>2, <7>1, <7>7,<7>8, Antisymmetric
                  <6>6. toj[ii] = ii
                     BY <6>5, <3>1
                  <6>9.QED BY <6>1, <6>3, <6>4, <6>6
               <5>7. CASE kk = i
                  <6>1. SUFFICES ASSUME ii # i,
                         lset'[i] # EmptyLS(i),
                         ll = LeftNeighbor(lset[ii]),
                         ii # ll,
                         CwDist(i, ii) < CwDist(ll, ii)
                         PROVE FALSE
                     BY <5>7, <5>0, <5>1, SimplifyAndSolve  DEF NoCCJoinL, Inv
                  <6>2. i \in NodesReady
                     BY <3>1 DEF TypeInvariant, NodesReady
                  <6>3. CwDist(ll, ii) =< CwDist(i, ii)
                        BY <5>0, <5>1, <6>1, <6>2, ReadyInI, CardinalityTwoNode, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                  <6>4. QED BY <6>3, <6>1, CwDistType, 
                        ReadyInI, <5>0, <5>1, IisNat, Unequality2
               <5>8. CASE ii # i /\ ll # i /\ kk # i
                  <6>1. SUFFICES ASSUME 
                        ii # kk,
                        lset[kk] # EmptyLS(kk),
                        ll = LeftNeighbor(lset[ii]),
                        ii # ll,
                        CwDist(kk, ii) < CwDist(ll, ii)
                        PROVE 
                         \/ toj[ii] = kk /\ toj[ll] = ll
                         \/ toj[ll] = kk /\ toj[ii] = ii
                  BY <5>1, <5>0, <5>8 DEF NoCCJoinL
                  <6>. QED BY <5>1,<6>1 DEF NoCCJoinL, Inv
               <5>. QED
                 BY <5>5, <5>6, <5>7, <5>8     
            <4>73. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE
                     \A k \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}:
                      lset'[k] # EmptyLS(k)  
                  BY <4>1 DEF SemJRplyLSnonEmpty, Inv
               <5>7. CASE mm = newm
                  <6>1. /\ mm.destination = m.mreq.node
                        /\ mm.mreq.lset = lset[i]
                     BY <5>7, <3>1
                  <6>2. /\ lset[i].node = i
                        /\ lset[i] \in LSet
                        /\ i # m.mreq.node
                        /\ Covers(lset[i], m.mreq.node)
                     BY <3>1 DEF TypeInvariant, Covers
                  <6>3. mm.mreq.lset.node = i
                     BY <6>1, <6>2 
                  <6>4. m.mreq.node \in I
                     BY <4>0 DEF JReq
                  <6>5.  lset'[mm.mreq.lset.node] = mm.mreq.lset
                     BY <6>3, <6>1, <4>1
                  <6>6. CASE GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} = {}
                     BY <6>6
                  <6>7. CASE GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} # {}
                     <7>1. lset[i] # EmptyLS(i)
                        BY <6>7, NonEmptyLS, <6>2, <6>1
                     <7>2. SUFFICES ASSUME NEW 
                            k \in GetLSetContent(lset[i]) \ {i}
                           PROVE lset'[k] # EmptyLS(k)
                        BY <6>3, <6>1, <4>1  
                     <7>3. /\ k \in I
                           /\ lset[k].node = k
                           /\ lset[k] \in LSet
                        BY <7>2, GetLSetContentType, <6>2 DEF TypeInvariant
                     <7>4. i \in NodesReady
                        BY <3>1 DEF NodesReady
                     <7>5. k # i
                        BY <7>2
                     <7>7. CASE k \in NodesReady
                        <8>5. ~\E kk \in ReadyOK: ReadyOK = {kk}
                           BY <7>4,<7>7,<7>5, ReadyInI, 
                              CardinalityTwoNode, CardinalityOneNodeExt
                        <8>8. LeftNeighbor(lset[k]) # k
                           BY <8>5, <7>7, ReadyInI DEF HalfNeighbor, Inv
                        <8>9. QED BY <8>8, EmptyLSNoNeighbor, <7>3, <4>1
                     <7>8. CASE ~ k \in NodesReady
                        BY <7>8, <7>3, <7>2, <4>1 DEF LSnonEmpty, Inv
                     <7>9. QED BY  <7>7,<7>8
                  <6>9. QED BY <6>6, <6>7
               <5>8. CASE mm # newm
                  <6>1. mm \in receivedMsgs
                     BY <5>3, <5>8, <4>5
                  <6>2. mm.mreq \in JRpl
                     BY SlowestZenon, <6>1, <5>3 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                  <6>3. mm.mreq.lset \in LSet
                     BY <6>2 DEF JRpl
                  <6>4. mm.mreq.lset.node \in I
                     BY <6>3 DEF LSet
                  <6>5. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                     BY <6>4,<6>3, GetLSetContentType
                  <6>9.QED BY <6>1, <5>3, <6>5, <4>1 DEF SemJRplyLSnonEmpty, Inv
               <5>9. QED BY <5>7,<5>8
            <4>74. OneJRply'
               <5>1. SUFFICES ASSUME NEW m1 \in receivedMsgs',
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type = "JoinReply",
                     m2.mreq.type = "JoinReply",
                     m1.destination = m2.destination 
                     PROVE m1 = m2
                  BY  DEF OneJRply, Inv 
               <5>5. CASE m1 = newm 
                  <6>1. SUFFICES ASSUME m2 # newm PROVE FALSE
                     BY <5>5
                  <6>2. m2 \in receivedMsgs
                     BY <6>1, <3>1
                  <6>3. m2.destination # m.mreq.node
                     BY <6>2, <5>1, <3>1 DEF JoinMsgDual, Inv
                  <6>4. m1.destination = m.mreq.node
                     BY <5>5, <3>1
                  <6>99. QED BY <6>3, <6>4, <5>1
               <5>6. CASE m2 = newm
                  <6>1. SUFFICES ASSUME m1 # newm PROVE FALSE
                     BY <5>6
                  <6>2. m1 \in receivedMsgs
                     BY <6>1, <3>1
                  <6>3. m1.destination # m.mreq.node
                     BY <6>2, <5>1, <3>1 DEF JoinMsgDual, Inv
                  <6>4. m2.destination = m.mreq.node
                     BY <5>6, <3>1
                  <6>99. QED BY <6>3, <6>4, <5>1                  
               <5>7. CASE m1 # newm /\ m2 # newm
                  <6>1. m1 \in receivedMsgs 
                     BY <5>7, <3>1
                  <6>2. m2 \in receivedMsgs
                     BY <5>7, <3>1
                  <6>9. QED BY <6>1, <6>2, <5>1 DEF OneJRply, Inv 
               <5>9. QED BY <5>5,<5>6, <5>7
            <4>75. JoinMsgDual'
               <5>1. SUFFICES ASSUME NEW m1 \in receivedMsgs',
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type  = "JoinReply",
                     m2.mreq.type = "JoinRequest"
                     PROVE m2.mreq.node # m1.destination
                  BY DEF JoinMsgDual, Inv
               <5>2. m2 \in receivedMsgs
                  BY <5>1, <4>4
               <5>7. CASE m1 = newm
                  <6>0. m # newm
                     <7>4. m.mreq.type = "JoinRequest"
                        BY <3>1
                     <7>8. newm.mreq.type = "JoinReply"
                        BY <3>1
                     <7>9. QED BY <7>4,<7>8
                  <6>1. m2 # m
                     BY <5>1, <3>1, <6>0
                  <6>2. m2.mreq.node # m.mreq.node
                     BY <6>1, <5>1, <5>2, <3>1 DEF OneJReq, Inv
                  <6>3. m1.destination = newm.destination
                     BY <5>7
                  <6>4. newm.destination = m.mreq.node
                     BY <3>1
                  <6>9. QED BY <6>4, <6>3, <6>2
               <5>8. CASE m1 # newm
                  <6>1. m1 \in receivedMsgs
                     BY <5>8, <3>1
                  <6>9. QED BY <6>1, <5>2, <5>1 DEF JoinMsgDual, Inv
               <5>9. QED BY <5>7,<5>8
            <4>76. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>2. mm \in receivedMsgs
                  BY <4>4, <5>1
               <5>3. mm.mreq \in JReq
                  BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>4. mm.mreq.node \in I
                  BY <5>3 DEF JReq
               <5>9. QED BY <5>2, <4>1, <4>2, <5>1, ReadyInI DEF SemJReq, Inv
            <4>77. SingleReadyNodeToj'
               <5>1. SUFFICES ASSUME NEW kk \in I, 
                     NEW rr \in I,
                     NodesReady' = {rr},
                     lset'[kk] # EmptyLS(kk),
                     kk # rr
                     PROVE toj'[rr] = kk
                  BY DEF SingleReadyNodeToj, Inv
               <5>2. \A k \in I\{i}: toj'[k] = toj[k]
                  BY <3>1  DEF TypeInvariant
               <5>3. /\ NodesReady = {rr}
                     /\ lset[kk] # EmptyLS(kk)
                     /\ kk # rr
                  BY <5>1, <4>1,<4>2
               <5>4. kk # rr /\ toj[i] = i
                  BY <5>3, <3>1
               <5>6. rr # i
                  BY <5>3, <5>4 DEF SingleReadyNodeToj, Inv
               <5>9. QED BY <5>6, <5>3, <5>2 DEF SingleReadyNodeToj, Inv
            <4>78. OneJReq'
               BY <4>4,<4>1,<4>2 DEF OneJReq, Inv
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv
            <4>9. QED BY <4>3,<4>25,\*<4>26,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, 
                  <4>70, <4>71, <4>72, <4>73, 
                  <4>74, <4>75, 
                  <4>76, <4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <3>1, <1>6 DEF ReceiveJoinRequest 
   <1>7. ASSUME NEW i \in I,
                ReceiveJoinReply(i),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
            status[i] = "wait",
            probing[i] = {},
            lset[i] = EmptyLS(i),
            m.destination = i,
            m.mreq.type = "JoinReply",
             m.mreq.lset.node # i,
            NEW newrtable,   
                newrtable =AddToTable(GetLSetContent(m.mreq.lset)
                             \cup GetRTableContent(m.mreq.rtable), 
                              rtable[i], i),
            NEW newlset,
                newlset = AddToLSet(GetLSetContent(m.mreq.lset), lset[i]),
            NEW toprob, toprob = GetLSetContent(newlset) \ {i},
            NEW probmsgs, probmsgs = Probe(i, newlset, {}, toprob),
            rtable'= [rtable EXCEPT ![i] = newrtable],
            lset'= [lset EXCEPT ![i] = newlset],
            probing'= [probing EXCEPT ![i] = toprob],
            failed'=[failed EXCEPT ![i] = {}],
            receivedMsgs'= (receivedMsgs \ {m}) \cup probmsgs,
            UNCHANGED <<status, lease, grant, toj>>,
            TypeInvariant,
            Inv
         PROVE  Inv'
         <4>01. m.mreq \in JRpl
            BY SlowestZenon, <3>1 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
         
         <4>03. /\ newlset \in LSet 
                /\ newlset.node = i
                /\ m.mreq.lset \in LSet
                /\ lset[i] \in LSet
                /\ GetLSetContent(m.mreq.lset) \in SUBSET I
                /\ toprob \in SUBSET I
                /\ m.mreq.lset.node \in I
                /\ lset[i].node = i
            <5>1. m.mreq.lset \in LSet
               BY <4>01 DEF JRpl
            <5>2. lset[i] \in LSet
               BY <3>1 DEF TypeInvariant
            <5>3. GetLSetContent(m.mreq.lset) \in SUBSET I
               BY <5>1, GetLSetContentType 
            <5>4. newlset \in LSet
               BY <3>1, AddToLSetType, <5>3, <5>2
            <5>5. newlset.node = i /\ lset[i].node = i
               BY <3>1, AddToLSetNoChangeNode, <5>3, <5>2 DEF TypeInvariant
            <5>6. toprob \in SUBSET I
               BY <5>4, GetLSetContentType, <3>1
            <5>7. m.mreq.lset.node \in I
               BY <5>1 DEF LSet
            <5>9.QED BY <5>1,<5>2,<5>3,<5>4, <5>5, <5>6, <5>7 DEF TypeInvariant
         <4>2./\ NodesReady' = NodesReady 
              /\ ReadyOK'= ReadyOK 
              /\ NonDead'= NonDead
              /\ NodesWait' = NodesWait
              /\ ~i \in ReadyOK
              /\ ~i \in NodesReady
               BY <3>1, ReadyInI DEF NodesReady,ReadyOK, NodesOK, NodesWait, NonDead
        <4>3. 
             /\ \A k \in ReadyOK: lset'[k] = lset[k]
             /\ \A k \in I\{i}: lset'[k] = lset[k]
           BY <4>2, ReadyInI, <3>1 DEF TypeInvariant\*, NodesReady,ReadyOK, NodesOK
        <4>4. \A k \in I:  /\grant'[k]= grant[k]
                           /\ lease'[k] = lease[k]
                           /\ toj'[k] = toj[k]
           BY <3>1 DEF TypeInvariant
        <4>5. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "JoinRequest"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
          <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                     Probe(i, newlset, {}, toprob)
             BY <3>1, <4>3 DEF TypeInvariant
          <6>2. \A ms \in Probe(i, newlset, {}, toprob): ms.mreq.type = "LSProbe"
             <7>1. ProbeMsgTypeConclusion(i, newlset, {}, toprob)
                    BY ONLY ProbeMsgType, <4>03
            <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
          <6>9. QED BY <6>1,<6>2
        <4>11. \A k \in I: ~k \in (lset[k].left) /\ ~k \in (lset[k].right) 
          BY LSetInv DEF LSetInv, Inv
        <4>12. /\ ~ lset[i].node \in (newlset.left) 
             /\ ~ lset[i].node  \in (newlset.right)
             /\ newlset.node = lset[i].node
          BY <3>1, AddNodeNotChange, NotAddSelf, <4>03 
        <4>16. lset[i]' = newlset /\ newlset.node = i
          BY <3>1, <4>03, <4>12, TypeInvariant DEF TypeInvariant
        <4>17. /\ ~ i \in (lset[i]'.left) 
             /\ ~ i \in (lset[i]'.right)
          BY <4>12, <4>16
        <4>18. \A k \in I \{i}: 
             /\ ~k \in (lset'[k].left) 
             /\ ~k \in (lset'[k].right) 
          BY <4>11, <3>1, TypeInvariant DEF TypeInvariant
        <4>20. LSetInv'
        BY <4>18,<4>17 DEF LSetInv
        <4>21.\A k \in I: lset'[k] \in LSet /\ lset[k] \in LSet
           BY <3>1, <4>03 DEF TypeInvariant 
        <4>95. HalfNeighbor' 
            <5>1. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                   \/ ms.mreq.type = "LSProbeReply" 
                   \/ ms.mreq.type = "JoinReply"
                   \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in Probe(i, newlset, {}, toprob)
                     BY <3>1 DEF TypeInvariant
                  <6>2. ProbeMsgTypeConclusion(i, newlset, {}, toprob)
                     BY ONLY ProbeMsgType, <4>03
                  <6>9. QED BY i \in I, <6>2, <6>1, GetLSetContentType DEF ProbeMsgTypeConclusion 
            <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
               BY <5>7,<4>2,<4>3 DEF HalfNeighbor
            <5>8. CASE \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                                   => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
               <6>1. PICK k \in ReadyOK: ReadyOK = {k} 
                  BY <5>8
               <6>2. /\ LeftNeighbor(lset[k]) = k 
                     /\ RightNeighbor(lset[k]) = k
                     /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                     /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs: /\ mk.mreq.type = "LSProbe" 
                                                 /\ mk.destination # k
                     /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                  BY <6>1, <5>8
               <6>3. toprob = GetLSetContent(AddToLSet({k}, EmptyLS(i))) \ {i}
                  BY <3>1, <6>2
               <6>4. i # k
                  BY <4>2, <6>1
               <6>6. toprob = {k}
                  <7>1. /\ i \in I
                        /\ k \in I
                     BY ReadyInI
                  <7>2. GetLSetContent(AddToLSet({k}, EmptyLS(i))) = {i, k}
                     BY <6>3, AddSingleNodeToEmpty, ReadyInI
                  <7>3. QED BY <6>4, <7>2, <6>3
               <6>7. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                      ms.mreq.type = "LSProbe" /\ ms.destination # k
                  <7>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in Probe(i, newlset, {}, toprob)
                     BY <3>1 DEF TypeInvariant
                  <7>2. ProbeMsgTypeConclusion(i, newlset, {}, toprob)
                     BY ONLY ProbeMsgType, <4>03
                  <7>3. \A ms \in receivedMsgs' \ receivedMsgs: 
                        ms.destination \in toprob
                     BY <7>2,<7>1 DEF ProbeMsgTypeConclusion
                  <7>4.  \A ms \in receivedMsgs' \ receivedMsgs: 
                        ms.destination = k
                     BY <7>2, <6>6, <7>1 DEF ProbeMsgTypeConclusion
                  <7>9. QED BY <6>2, <7>1, <7>4, ProbeMsgType 
               <6>8. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                  <7>1. \A w \in NodesWait'\{i}: GetLSetContent(lset'[w]) \in SUBSET {k, w}
                     BY ReadyInI, <4>2, <6>2, <4>3
                  <7>3. lset[i]' = AddToLSet({k}, EmptyLS(i))
                     BY <6>2, <3>1 DEF TypeInvariant
                  <7>4. i # k
                     BY <4>2, <7>1
                  <7>6. GetLSetContent(lset'[i]) = {i, k}
                     BY <7>3, <7>4, AddSingleNodeToEmpty, ReadyInI
                  <7>8. GetLSetContent(lset'[i]) \in SUBSET {k, i}
                     BY <7>6
                  <7>9. QED BY <7>1, <7>8
               <6>9. QED BY <5>8,<6>1, <6>2, <4>3, <4>2, <5>1, <6>7, <6>8  DEF HalfNeighbor
            <5>9. QED BY <5>7, <5>8, HalfNeighbor DEF HalfNeighbor, Inv
        \*<4>96. NeighborClosest'
        \*    BY <4>2, <4>3, NeighborClosest DEF NeighborClosest, NodesReady,ReadyOK, NodesOK, Inv
        \*<4>51. ReadyNeighbor'
        \*<4>52. IncludeNeighbor'
        \*<4>53. CrossNeighbor'
        <4>54. CompleteLeafSet'
           <5>1. /\ lset[i]' = newlset
                 /\ lset[i].node = i
                 /\ m.mreq.lset \in LSet
                 /\ m.mreq.lset.node \in I
                 /\ newlset \in LSet
                 /\ i \in I
                 /\ newlset.node = i
                 /\ m.destination = i
                 /\ m.mreq.type = "JoinReply"
                 /\ newlset = AddToLSet(GetLSetContent(m.mreq.lset), EmptyLS(m.destination))
                 /\ lset[i]'\in LSet
                 /\ lset[i]'.node = i
              BY <3>1, <4>03, <4>16
           <5>2. lset[i]' # EmptyLS(i)
              BY  <4>16, AddToLSetNoMoreEmpty, <4>03, <3>1
           <5>5. SUFFICES ASSUME
                 NEW j \in NodesReady',
                 j # i,
                 lset'[i] # EmptyLS(i),
                 NEW r \in I,
                 RightNeighbor(lset[i]')= r,
                 CwDist(i, j) < CwDist(i, r)
                 PROVE FALSE
              BY <4>16, <4>2,<4>3, <5>1, <5>2, <4>03, CompleteLeafSet DEF CompleteLeafSet, Inv
           <5>6.  m.mreq.lset.node \in NodesReady
              BY <5>5, <5>1, SemJRply DEF SemJRply, Inv
           <5>7. CASE LeftNeighbor(newlset) = m.mreq.lset.node  /\ RightNeighbor(newlset) = RightNeighbor(m.mreq.lset)
              <6>7. CASE m.mreq.lset.node = j
                 <7>2. i # r /\ j # r
                    BY <5>5, IisNat, ReadyInI, <4>2, CwDistZeroCompare
                 <7>5. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>7, <7>2, <5>5, <5>1, <6>7
                 <7>6. CwDist(LeftNeighbor(newlset), newlset.node) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>5, <5>1, <5>2,  NeighborPosition
                 <7>7. CwDist(j, i) < CwDist(j, r)
                    BY <5>7, <6>7,<4>16, <7>6, <5>5
                 <7>8. ~(CwDist(i, j) < CwDist(i, r))
                    BY  <7>2,<7>7, Ring3NodesPos,
                       <4>2, ReadyInI
                 <7>9. QED BY <7>8, <5>5
              <6>8. CASE m.mreq.lset.node # j
                 <7>1.~( Cardinality(NodesReady) =< 1)
                    BY <6>8, <5>5, <5>6,<5>1, <4>2, ReadyInI, CardinalityTwoNode
                 <7>2. m.mreq.lset.node  # RightNeighbor(lset[m.mreq.lset.node]) 
                    BY <5>6,  <7>1, ReadyInI, CardinalityOneNodeExt, HalfNeighbor DEF HalfNeighbor, Inv
                 <7>3. m.mreq.lset.node  # RightNeighbor(m.mreq.lset)
                    <8>1. SUFFICES ASSUME m.mreq.lset.node  = RightNeighbor(m.mreq.lset)
                          PROVE FALSE
                       OBVIOUS
                    <8>3. m.mreq.lset = EmptyLS(m.mreq.lset.node)
                       BY EmptyLSNoNeighbor, <8>1, <5>1,<5>6,<7>2, SemEmptyLSetSingleNode DEF SemEmptyLSetSingleNode, Inv
                    <8>4. lset[m.mreq.lset.node] = m.mreq.lset
                       BY <8>3, <5>1, SemSingleReadyJoin DEF SemSingleReadyJoin, Inv
                    <8>5. m.mreq.lset.node  # RightNeighbor(m.mreq.lset)
                       BY <7>2, <8>4
                    <8>9. QED BY <8>5, <8>1
                 <7>4. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>7, <7>3
                 <7>5. CwDist(LeftNeighbor(newlset), newlset.node) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>4, <5>1, <5>2,  NeighborPosition
                 <7>6. CwDist(m.mreq.lset.node, j) < CwDist(m.mreq.lset.node, RightNeighbor(m.mreq.lset))
                    BY <7>5, <5>5, CwDist4NodesPos,<5>7,
                       LeftNeighborType, <5>1, RightNeighborType, IisNat, ReadyInI, <4>2
                 <7>7. RightNeighbor(m.mreq.lset) = RightNeighbor(lset[m.mreq.lset.node])
                    BY <5>1,<3>1 DEF SemJoinLeafSet, Inv
                 <7>8. CwDist(m.mreq.lset.node, RightNeighbor(m.mreq.lset)) =< CwDist(m.mreq.lset.node, j)
                    BY <7>1,<4>2, <5>1, <5>5, <5>6, <6>8,<7>6,<7>7, LemmaCLtoNC, ReadyInI DEF NeighborClosest, Inv
                 <7>9. QED BY  <7>8,<7>6, Unequality2,
                    CwDistType, <5>1, ReadyInI, <4>2, IisNat, RightNeighborType 
              <6>9. QED BY <6>7, <6>8
           <5>8. CASE RightNeighbor(newlset) = m.mreq.lset.node  /\ LeftNeighbor(newlset)= LeftNeighbor(m.mreq.lset)
              \* OMITTED \* same as <5>7
              <6>6. m.mreq.lset.node = r
                 BY <5>8, <4>16, <5>5
              <6>7. CASE m.mreq.lset.node = j \* could be shorter
                 <7>2. i # r /\ j # r
                    BY <5>5, IisNat, ReadyInI, <4>2, CwDistZeroCompare
                 <7>5. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>8, <7>2, <5>5, <5>1, <6>7
                 <7>6. CwDist(LeftNeighbor(newlset), newlset.node) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>5, <5>1, <5>2,  NeighborPosition
                 <7>7. CwDist(j, i) < CwDist(j, r)
                    BY <5>8, <6>7,<4>16, <7>6, <5>5
                 <7>8. ~(CwDist(i, j) < CwDist(i, r))
                    BY  <7>2,<7>7, Ring3NodesPos,
                       <4>2, ReadyInI
                 <7>9. QED BY <7>8, <5>5
              <6>8. CASE m.mreq.lset.node # j
                 <7>1.~( Cardinality(NodesReady) =< 1)
                    BY <6>8, <5>5, <5>6,<5>1, <4>2, ReadyInI, CardinalityTwoNode
                 <7>2. m.mreq.lset.node  # LeftNeighbor(lset[m.mreq.lset.node]) 
                    BY <5>6,  <7>1, ReadyInI, CardinalityOneNodeExt, HalfNeighbor DEF HalfNeighbor, Inv
                 <7>3. m.mreq.lset.node  # LeftNeighbor(m.mreq.lset)
                    <8>1. SUFFICES ASSUME m.mreq.lset.node  = LeftNeighbor(m.mreq.lset)
                          PROVE FALSE
                       OBVIOUS
                    <8>3. m.mreq.lset = EmptyLS(m.mreq.lset.node)
                       BY EmptyLSNoNeighbor, <8>1, <5>1,<5>6,<7>2, SemEmptyLSetSingleNode DEF SemEmptyLSetSingleNode, Inv
                    <8>4. lset[m.mreq.lset.node] = m.mreq.lset
                       BY <8>3, <5>1, SemSingleReadyJoin DEF SemSingleReadyJoin, Inv
                    <8>5. m.mreq.lset.node  # LeftNeighbor(m.mreq.lset)
                       BY <7>2, <8>4
                    <8>9. QED BY <8>5, <8>1
                 <7>4. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>8, <7>3
                 <7>5. CwDist(newlset.node, RightNeighbor(newlset)) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>4, <5>1, <5>2,  NeighborPositionCo
                 <7>6. CwDist(j, m.mreq.lset.node) < CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node)
                    BY <7>5, <5>5,<6>6, CwDist4NodesPos2,<5>8,
                       LeftNeighborType, <5>1, RightNeighborType, IisNat, ReadyInI, <4>2
                 <7>7. LeftNeighbor(m.mreq.lset) = LeftNeighbor(lset[m.mreq.lset.node])
                    BY <5>1,<3>1 DEF SemJoinLeafSet, Inv
                 <7>8. CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node ) =< CwDist(j, m.mreq.lset.node)
                    BY <7>1,<4>2, <5>1, <5>5, <5>6, <6>8,<7>6,<7>7, LemmaCLtoNC, ReadyInI DEF NeighborClosest, Inv
                 <7>9. QED BY  <7>8,<7>6, Unequality2,
                    CwDistType, <5>1, ReadyInI, <4>2, IisNat, LeftNeighborType 
              <6>9. QED BY <6>7, <6>8
           <5>9. QED BY  <5>7, <5>8, <5>1,<5>5, SemJoinNeighbor DEF SemJoinNeighbor, Inv
        <4>55. CompleteLeafSetL'
           \*OMITTED \* similar to <4>54
           <5>1. /\ lset[i]' = newlset
                 /\ lset[i].node = i
                 /\ m.mreq.lset \in LSet
                 /\ m.mreq.lset.node \in I
                 /\ newlset \in LSet
                 /\ i \in I
                 /\ newlset.node = i
                 /\ m.destination = i
                 /\ m.mreq.type = "JoinReply"
                 /\ newlset = AddToLSet(GetLSetContent(m.mreq.lset), EmptyLS(m.destination))
                 /\ lset[i]'\in LSet
                 /\ lset[i]'.node = i
              BY <3>1, <4>03, <4>16
           <5>2. lset[i]' # EmptyLS(i)
              BY  <4>16, AddToLSetNoMoreEmpty, <4>03, <3>1
           <5>5. SUFFICES ASSUME
                 NEW j \in NodesReady',
                 j # i,
                 lset'[i] # EmptyLS(i),
                 NEW l \in I,
                 LeftNeighbor(lset[i]')= l,
                 CwDist(j, i) < CwDist(l, i)
                 PROVE FALSE
              BY <4>16, <4>2,<4>3, <5>1, <5>2, <4>03 DEF CompleteLeafSetL, Inv
           <5>6.  m.mreq.lset.node \in NodesReady
              BY <5>5, <5>1, SemJRply DEF SemJRply, Inv
           <5>7. CASE RightNeighbor(newlset) = m.mreq.lset.node  /\ LeftNeighbor(newlset) = LeftNeighbor(m.mreq.lset)
              <6>7. CASE m.mreq.lset.node = j
                 <7>2. i # l /\ j # l
                    BY <5>5, IisNat, ReadyInI, <4>2, CwDistZeroCompareCo
                 <7>5. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>7, <7>2, <5>5, <5>1, <6>7
                 <7>6. CwDist(newlset.node, RightNeighbor(newlset)) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>5, <5>1, <5>2,  NeighborPositionCo
                 <7>7. CwDist(i, j) < CwDist(l, j)
                    BY <5>7, <6>7,<4>16, <7>6, <5>5
                 <7>8. ~(CwDist(j, i) < CwDist(l, i))
                    BY  <7>2,<7>7, Ring3NodesPos4,
                       <4>2, ReadyInI
                 <7>9. QED BY <7>8, <5>5
              <6>8. CASE m.mreq.lset.node # j
                 <7>1.~( Cardinality(NodesReady) =< 1)
                    BY <6>8, <5>5, <5>6,<5>1, <4>2, ReadyInI, CardinalityTwoNode
                 <7>2. m.mreq.lset.node  # LeftNeighbor(lset[m.mreq.lset.node]) 
                    BY <5>6,  <7>1, ReadyInI, CardinalityOneNodeExt, HalfNeighbor DEF HalfNeighbor, Inv
                 <7>3. m.mreq.lset.node  # LeftNeighbor(m.mreq.lset)
                    <8>1. SUFFICES ASSUME m.mreq.lset.node  = LeftNeighbor(m.mreq.lset)
                          PROVE FALSE
                       OBVIOUS
                    <8>3. m.mreq.lset = EmptyLS(m.mreq.lset.node)
                       BY EmptyLSNoNeighbor, <8>1, <5>1,<5>6,<7>2, SemEmptyLSetSingleNode DEF SemEmptyLSetSingleNode, Inv
                    <8>4. lset[m.mreq.lset.node] = m.mreq.lset
                       BY <8>3, <5>1, SemSingleReadyJoin DEF SemSingleReadyJoin, Inv
                    <8>5. m.mreq.lset.node  # LeftNeighbor(m.mreq.lset)
                       BY <7>2, <8>4
                    <8>9. QED BY <8>5, <8>1
                 <7>4. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>7, <7>3
                 <7>5. CwDist(newlset.node, RightNeighbor(newlset)) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>4, <5>1, <5>2,  NeighborPositionCo
                 <7>6. CwDist(j, m.mreq.lset.node) < CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node)
                    BY <7>5, <5>5, CwDist4NodesPos3,<5>7,
                       LeftNeighborType, <5>1, RightNeighborType, IisNat, ReadyInI, <4>2
                 <7>7. LeftNeighbor(m.mreq.lset) = LeftNeighbor(lset[m.mreq.lset.node])
                    BY <5>1,<3>1 DEF SemJoinLeafSet, Inv
                 <7>8. CwDist(LeftNeighbor(m.mreq.lset), m.mreq.lset.node) =< CwDist(j, m.mreq.lset.node)
                    BY <7>1,<4>2, <5>1, <5>5, <5>6, <6>8,<7>6,<7>7, LemmaCLtoNC, ReadyInI DEF NeighborClosest, Inv
                 <7>9. QED BY  <7>8,<7>6, Unequality2,
                    CwDistType, <5>1, ReadyInI, <4>2, IisNat, LeftNeighborType 
              <6>9. QED BY <6>7, <6>8
           <5>8. CASE LeftNeighbor(newlset) = m.mreq.lset.node  /\ RightNeighbor(newlset)= RightNeighbor(m.mreq.lset)
              \* OMITTED \* same as <5>7
              <6>6. m.mreq.lset.node = l
                 BY <5>8, <4>16, <5>5
              <6>7. CASE m.mreq.lset.node = j \* could be shorter
                 <7>2. i # l /\ j # l
                    BY <5>5, IisNat, ReadyInI, <4>2, CwDistZeroCompareCo
                 <7>9. QED BY <7>2, <6>7, <6>6
              <6>8. CASE m.mreq.lset.node # j
                 <7>1.~( Cardinality(NodesReady) =< 1)
                    BY <6>8, <5>5, <5>6,<5>1, <4>2, ReadyInI, CardinalityTwoNode
                 <7>2. m.mreq.lset.node  # RightNeighbor(lset[m.mreq.lset.node]) 
                    BY <5>6,  <7>1, ReadyInI, CardinalityOneNodeExt, HalfNeighbor DEF HalfNeighbor, Inv
                 <7>3. m.mreq.lset.node  # RightNeighbor(m.mreq.lset)
                    <8>1. SUFFICES ASSUME m.mreq.lset.node  = RightNeighbor(m.mreq.lset)
                          PROVE FALSE
                       OBVIOUS
                    <8>3. m.mreq.lset = EmptyLS(m.mreq.lset.node)
                       BY EmptyLSNoNeighbor, <8>1, <5>1,<5>6,<7>2, SemEmptyLSetSingleNode DEF SemEmptyLSetSingleNode, Inv
                    <8>4. lset[m.mreq.lset.node] = m.mreq.lset
                       BY <8>3, <5>1, SemSingleReadyJoin DEF SemSingleReadyJoin, Inv
                    <8>5. m.mreq.lset.node  # RightNeighbor(m.mreq.lset)
                       BY <7>2, <8>4
                    <8>9. QED BY <8>5, <8>1
                 <7>4. LeftNeighbor(newlset) # RightNeighbor(newlset)
                    BY <5>8, <7>3
                 <7>5. CwDist(LeftNeighbor(newlset), newlset.node) < CwDist(LeftNeighbor(newlset),RightNeighbor(newlset))
                    BY <7>4, <5>1, <5>2,  NeighborPosition
                 <7>6. CwDist(m.mreq.lset.node, j) < CwDist( m.mreq.lset.node, RightNeighbor(m.mreq.lset))
                    BY <7>5, <5>5,<6>6, CwDist4NodesPos4,<5>8,
                       LeftNeighborType, <5>1, RightNeighborType, IisNat, ReadyInI, <4>2
                 <7>7. RightNeighbor(m.mreq.lset) = RightNeighbor(lset[m.mreq.lset.node])
                    BY <5>1,<3>1 DEF SemJoinLeafSet, Inv
                 <7>8. CwDist(m.mreq.lset.node, RightNeighbor(m.mreq.lset) ) =< CwDist( m.mreq.lset.node, j)
                    BY <7>1,<4>2, <5>1, <5>5, <5>6, <6>8,<7>6,<7>7, LemmaCLtoNC, ReadyInI DEF NeighborClosest, Inv
                 <7>9. QED BY  <7>8,<7>6, Unequality2,
                    CwDistType, <5>1, ReadyInI, <4>2, IisNat, RightNeighborType 
              <6>9. QED BY <6>7, <6>8
           <5>9. QED BY  <5>7, <5>8, <5>1,<5>5, SemJoinNeighbor DEF SemJoinNeighbor, Inv
        <4>56. ReadyNeighborRight'
           <5>1. /\ NodesReady' = NodesReady
                 /\ \A r \in NodesReady': lset'[r] = lset[r]
                 /\ ~i \in NodesReady
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
              BY <4>2, <4>3, ReadyInI
           <5>2. /\ m \in receivedMsgs
                 /\ m.mreq.type = "JoinReply"
                 /\ m.destination = i
                 /\ toj[m.mreq.lset.node] = m.destination
                 /\ m.mreq.lset.node \in NodesReady
              BY SemJRply, <3>1 DEF SemJRply, Inv
           <5>3. SUFFICES ASSUME NEW r \in NodesReady,
                 NEW n \in NodesReady,
                 n = m.mreq.lset.node,
                 ~RightNeighbor(lset[r]) \in  NodesReady,
                 CwDist(r, i) < CwDist(r, RightNeighbor(lset[r])),
                 r # i,
                 lset'[i] # EmptyLS(i)
                 PROVE FALSE
              BY <5>1, <5>2, ReadyInI,<4>21, IisNat, Unequality2,
                 CwDistType, RightNeighborType,  
              ReadyNeighborRight DEF ReadyNeighborRight, Inv
           <5>33. /\ lset[n] \in LSet
                  /\ lset[r] \in LSet
                  /\ lset[n].node = n
                  /\  RightNeighbor(lset[r]) \in I     
                  /\ RightNeighbor(lset[n])  \in I
                  /\ r \in I
                  /\ n \in I   
                  /\ toj[n] \in I
              BY  RightNeighborType, ReadyInI, <4>21, <5>3 DEF TypeInvariant
           <5>4. /\ n # i
                 /\ i = toj[n]
              BY <5>2,<5>3, <5>1
           <5>5. RightNeighbor(lset[n]) \in NodesReady
              BY <5>2, <5>3, <5>33,
              SemJRply, GrantReady, LeaseGrant DEF SemJRply, GrantReady, LeaseGrant, Inv
           <5>6. n # r
              BY <5>3, <5>5
           <5>7. CwDist(r, n) # CwDist(r, i)
              BY RightDistInjectivity, <5>4, <5>3, ReadyInI, IisNat
           <5>8.  /\ RightNeighbor(lset[r]) # n
                  /\ RightNeighbor(lset[r]) # r
              BY <5>3, <5>33, ReadyInI
           <5>80. CASE CwDist(r, n) < CwDist(r, i)
              <6>1. CwDist(r, n) < CwDist(r, RightNeighbor(lset[r]))
                 BY <6>1,<5>3, <5>80, CwDistType, <5>33, IisNat, TransLESS
              <6>2. ~Cardinality(NodesReady) =< 1
                 BY CardinalityTwoNode, <5>33, <5>6, <5>3, ReadyInI
              <6>3. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, n)
                 BY <6>2,<5>3, <5>6, LemmaCLtoNC DEF NeighborClosest, Inv
              <6>9. QED BY <6>3, <6>1,<5>33, CwDistType, IisNat, Unequality2
           <5>90. CASE CwDist(r, i) < CwDist(r, n)
              <6>5. CASE n = RightNeighbor(lset[r])
                 BY <5>3, <6>5
              <6>6. CASE CwDist(r, n) < CwDist(r, RightNeighbor(lset[r]))
                 <7>2. ~Cardinality(NodesReady) =< 1
                 BY CardinalityTwoNode, <5>33, <5>6, <5>3, ReadyInI
                  <7>3. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, n)
                     BY <7>2,<5>3, <5>6, LemmaCLtoNC DEF NeighborClosest, Inv
                  <7>9. QED BY <7>3, <6>6,<5>33, CwDistType, IisNat, Unequality2
              <6>7. CASE CwDist(r, RightNeighbor(lset[r])) < CwDist(r, n)
                 <7>1. RightNeighbor(lset[r]) \in GetLSetContent(lset[r])
                    BY NeigborsInLset, <5>33
                 <7>2. lset[RightNeighbor(lset[r])] # EmptyLS(RightNeighbor(lset[r]))
                    BY LSnonEmpty, <7>1, <5>8, <5>33, <5>3 DEF LSnonEmpty, Inv
                 <7>3. Covers(lset[n], i)
                    BY SemJRply, <5>2, <5>3, <5>33 DEF SemJRply, Inv
                 <7>4. CwDist(RightNeighbor(lset[r]), n) < CwDist(i, n)
                 \* BY <6>7, <5>3, <5>33, IisNat, CwDist4NodesPosCo cannot be proved
                    <8>1. /\ r \in Nat
                          /\ i \in Nat
                          /\ n \in Nat
                          /\ RightNeighbor(lset[r]) \in Nat
                       BY <5>3,<5>33, IisNat
                    <8>2. CwDist(r, i) < CwDist(r, RightNeighbor(lset[r]))
                       BY <5>3
                    <8>3. CwDist(r, RightNeighbor(lset[r])) < CwDist(r, n)
                       BY <6>7
                    <8>9. QED BY <8>1, <8>2, <8>3, CwDist4NodesPosCo
                 <7>5. Covers(lset[n], RightNeighbor(lset[r]))
                    BY CoverMonotonyNode, <5>33, <5>3, <7>4, <7>3
                 <7>6. RightNeighbor(lset[r]) = toj[n]
                    BY CoverToJ, <7>2, <7>5, <5>33, <5>8 DEF CoverToJ, Inv
                 <7>7. i = toj[n]
                    BY <5>4
                 <7>8. RightNeighbor(lset[r]) # i
                    BY <5>3, <5>33, ReadyInI
                 <7>9. QED BY <7>8, <7>6, <7>7, <5>33
              <6>9. QED BY <6>5,<6>6,<6>7, CaseAnalysis, RightDistInjectivity, 
                           <5>33, CwDistType, IisNat
           <5>99. QED BY <5>3,<5>4,<5>6,<5>7,<5>80, <5>90,
                        ReadyInI, CwDistType, IisNat, CaseAnalysis
        <4>57. ReadyNeighborLeft'
           \*OMITTED \* Similart to <4>56
           <5>1. /\ NodesReady' = NodesReady
                 /\ \A r \in NodesReady': lset'[r] = lset[r]
                 /\ ~i \in NodesReady
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
              BY <4>2, <4>3, ReadyInI
           <5>2. /\ m \in receivedMsgs
                 /\ m.mreq.type = "JoinReply"
                 /\ m.destination = i
                 /\ toj[m.mreq.lset.node] = m.destination
                 /\ m.mreq.lset.node \in NodesReady
              BY SemJRply, <3>1 DEF SemJRply, Inv
           <5>3. SUFFICES ASSUME NEW r \in NodesReady,
                 NEW n \in NodesReady,
                 n = m.mreq.lset.node,
                 ~LeftNeighbor(lset[r]) \in  NodesReady,
                 CwDist(i, r) < CwDist(LeftNeighbor(lset[r]), r),
                 r # i,
                 lset'[i] # EmptyLS(i)
                 PROVE FALSE
              BY <5>1, <5>2, ReadyInI,<4>21, IisNat, Unequality2,
                 CwDistType, LeftNeighborType 
               DEF ReadyNeighborLeft, Inv
           <5>33. /\ lset[n] \in LSet
                  /\ lset[r] \in LSet
                  /\ lset[n].node = n
                  /\  LeftNeighbor(lset[r]) \in I     
                  /\ LeftNeighbor(lset[n])  \in I
                  /\ r \in I
                  /\ n \in I   
                  /\ toj[n] \in I
              BY  LeftNeighborType, ReadyInI, <4>21, <5>3 DEF TypeInvariant
           <5>4. /\ n # i
                 /\ i = toj[n]
              BY <5>2,<5>3, <5>1
           <5>5. LeftNeighbor(lset[n]) \in NodesReady
              BY <5>2, <5>3, <5>33,
              SemJRply, GrantReady, LeaseGrant DEF SemJRply, GrantReady, LeaseGrant, Inv
           <5>6. n # r
              BY <5>3, <5>5
           <5>7. CwDist(n, r) # CwDist(i, r)
              BY CwDistInjectivity, <5>4, <5>3, ReadyInI, IisNat
           <5>8.  /\ LeftNeighbor(lset[r]) # n
                  /\ LeftNeighbor(lset[r]) # r
              BY <5>3, <5>33, ReadyInI
           <5>80. CASE CwDist(n, r) < CwDist(i, r)
              <6>1. CwDist(n, r) < CwDist(LeftNeighbor(lset[r]), r)
                 BY <6>1,<5>3, <5>80, CwDistType, <5>33, IisNat, TransLESS
              <6>2. ~Cardinality(NodesReady) =< 1
                 BY CardinalityTwoNode, <5>33, <5>6, <5>3, ReadyInI
              <6>3. CwDist(LeftNeighbor(lset[r]), r) =< CwDist(n, r)
                 BY <6>2,<5>3, <5>6, LemmaCLtoNC DEF NeighborClosest, Inv
              <6>9. QED BY <6>3, <6>1,<5>33, CwDistType, IisNat, Unequality2
           <5>90. CASE CwDist(i, r) < CwDist(n, r)
              <6>5. CASE n = LeftNeighbor(lset[r])
                 BY <5>3, <6>5
              <6>6. CASE CwDist(n, r) < CwDist(LeftNeighbor(lset[r]), r)
                 <7>2. ~Cardinality(NodesReady) =< 1
                 BY CardinalityTwoNode, <5>33, <5>6, <5>3, ReadyInI
                  <7>3. CwDist(LeftNeighbor(lset[r]), r) =< CwDist(n, r)
                     BY <7>2,<5>3, <5>6, LemmaCLtoNC DEF NeighborClosest, Inv
                  <7>9. QED BY <7>3, <6>6,<5>33, CwDistType, IisNat, Unequality2
              <6>7. CASE CwDist(LeftNeighbor(lset[r]), r) < CwDist(n, r)
                 <7>1. LeftNeighbor(lset[r]) \in GetLSetContent(lset[r])
                    BY NeigborsInLset, <5>33
                 <7>2. lset[LeftNeighbor(lset[r])] # EmptyLS(LeftNeighbor(lset[r]))
                    BY LSnonEmpty, <7>1, <5>8, <5>33, <5>3 DEF LSnonEmpty, Inv
                 <7>3. Covers(lset[n], i)
                    BY SemJRply, <5>2, <5>3, <5>33 DEF SemJRply, Inv
                 <7>4. CwDist(n, LeftNeighbor(lset[r])) < CwDist(n, i)
                 \* BY <6>7, <5>3, <5>33, IisNat, CwDist4NodesPosCo cannot be proved
                    <8>1. /\ r \in Nat
                          /\ i \in Nat
                          /\ n \in Nat
                          /\ LeftNeighbor(lset[r]) \in Nat
                       BY <5>3,<5>33, IisNat
                    <8>2. CwDist(i, r) < CwDist(LeftNeighbor(lset[r]), r)
                       BY <5>3
                    <8>3. CwDist(LeftNeighbor(lset[r]), r) < CwDist(n, r)
                       BY <6>7
                    <8>9. QED BY <8>1, <8>2, <8>3, CwDist4NodesPos5
                 <7>5. Covers(lset[n], LeftNeighbor(lset[r]))
                    BY CoverMonotonyNodeCo, <5>33, <5>3, <7>4, <7>3
                 <7>6. LeftNeighbor(lset[r]) = toj[n]
                    BY CoverToJ, <7>2, <7>5, <5>33, <5>8 DEF CoverToJ, Inv
                 <7>7. i = toj[n]
                    BY <5>4
                 <7>8. LeftNeighbor(lset[r]) # i
                    BY <5>3, <5>33, ReadyInI
                 <7>9. QED BY <7>8, <7>6, <7>7, <5>33
              <6>9. QED BY <6>5,<6>6,<6>7, CaseAnalysis, CwDistInjectivity, 
                           <5>33, CwDistType, IisNat
           <5>99. QED BY <5>3,<5>4,<5>6,<5>7,<5>80, <5>90,
                        ReadyInI, CwDistType, IisNat, CaseAnalysis
        <4>58. SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY  SimplifyAndSolve DEF SemLRply, Inv
           <5>9. QED BY <4>5,<4>4, <5>2,<4>3,<4>2 DEF SemLRply, Inv
        <4>59. LSnonEmpty'
           <5>1. SUFFICES ASSUME NEW ii \in I, 
                 NEW jj \in I,
                 ii \in GetLSetContent(lset'[jj]),
                 ~ii \in NodesReady,
                 ii # jj
                 PROVE lset'[ii] # EmptyLS(ii)
              BY <4>2 DEF LSnonEmpty, Inv
           <5>4. lset'[i] # EmptyLS(i)
              BY AddDiffNodeNoEmpty, <4>16, <4>21, <4>03, <3>1
           <5>5. CASE ii = i
              BY <5>4, <5>5
           <5>6. CASE jj = i
              <6>2. i # ii
                 BY <5>6, <5>1
              <6>3. GetLSetContent(m.mreq.lset) \in SUBSET I
                 BY <4>03
              <6>4. lset[i] = EmptyLS(i)
                 BY <3>1
              <6>5. ii \in GetLSetContent(AddToLSet(GetLSetContent(m.mreq.lset), lset[i]))
                 BY <5>1, <5>6, <3>1, <4>16
              <6>6. ii \in GetLSetContent(m.mreq.lset)
                 BY <6>2,<6>3,<6>4,<6>5, AddEffect
              <6>7. ii # m.mreq.lset.node
                 BY <5>1, <3>1 DEF SemJRply, Inv
              <6>8. lset[ii] # EmptyLS(ii)
                 BY <6>6, <6>7, <3>1 DEF SemJRplyLSnonEmpty, Inv
              <6>9. QED BY <6>8, <5>1, <5>6, <4>3 
           <5>7. CASE ii # i /\ jj # i
              BY <5>1, <5>7,<4>2, <4>3, <4>4, 
              ReadyInI DEF LSnonEmpty, Inv 
           <5>9. QED BY <5>5,<5>6,<5>7
        <4>60. LeaseGrant'
           BY <4>2, <4>3, <4>4, ReadyInI DEF LeaseGrant, Inv
        <4>61. GrantReady'
           BY <4>2, <4>3, <4>4, ReadyInI DEF GrantReady, Inv
        <4>62. SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE SemJRply!(mm)'
              BY DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <4>5, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. /\ mm.mreq.lset.node \in I
                 /\ mm.destination \in I
              BY <5>5, <5>4 DEF JRpl, LSet, DMsg, TypeInvariant
           <5>71. /\mm.mreq.lset.node \in NodesReady
                  /\mm.destination \in NodesWait
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>72.   /\ LeftNeighbor(mm.mreq.lset) \in I
                    /\ RightNeighbor(mm.mreq.lset) \in I
                    /\ mm.mreq.lset.node \in I
                    /\ \A k \in I: toj'[k] = toj[k]
                    /\ \A k \in I: lease'[k] = lease[k]
              BY <5>6, <5>7, LeftNeighborType, RightNeighborType, <4>4 DEF TypeInvariant
           <5>8. mm.mreq.lset.node # i
              BY <5>71, <3>1, StatusDisjoint DEF NodesWait 
           <5>9. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
              BY <4>3, <5>7, <5>8
           <5>50. mm.destination # i
              <6>1. SUFFICES ASSUME mm.destination = i
                    PROVE FALSE
                 OBVIOUS
              <6>4. m.mreq.lset.node = mm.mreq.lset.node
                 <7>2. Covers(lset[m.mreq.lset.node], i)
                    BY <3>1 DEF SemJRply, Inv
                 <7>3. Covers(lset[mm.mreq.lset.node], i)
                    BY <5>3,<5>4, <6>1 DEF SemJRply, Inv
                 <7>4. m.mreq.lset.node \in NodesReady
                    BY <3>1 DEF SemJRply, Inv
                 <7>5. mm.mreq.lset.node \in NodesReady
                    BY <5>3,<5>4 DEF SemJRply, Inv
                 <7>9. QED BY <7>2,<7>3, <7>4,<7>5, <4>03, ConsistentInv,
                       TypeInvariant, Inv, <5>7 DEF NodesReady
              <6>5. mm = m
                 BY <5>3, <5>4, <3>1, <6>1, <6>4 DEF OneJRply, Inv
              <6>6. \A ms \in Probe(i, newlset, {}, toprob): ms.mreq.type = "LSProbe"
                 <7>1. ProbeMsgTypeConclusion(i, newlset, {}, toprob)
                        BY ONLY ProbeMsgType, <4>03
                 <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion
              <6>7. ~ m \in Probe(i, newlset, {}, toprob)
                 BY <3>1, <6>6
              <6>8. ~ m \in receivedMsgs'
                 BY <6>7, <3>1
              <6>9. QED BY <6>5, <6>8, <5>3
           <5>51. lset'[mm.destination] = lset[mm.destination]
              BY  <4>3, <5>7,<5>50
           <5>60. QED BY <4>3, <5>7, <5>51,
                 <5>9, <5>72, <5>3, <5>4, <4>2 DEF SemJRply, Inv
        <4>63. SemJoinNeighbor'
           <5>3. SUFFICES ASSUME NEW ls \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <4>5, <5>3 DEF SemJoinNeighbor, Inv
        <4>64. SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady' = {mm.mreq.lset.node}
              BY DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <4>5, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>8. /\ lset[mm.mreq.lset.node] = mm.mreq.lset
                 /\ NodesReady = {mm.mreq.lset.node}
              BY <5>3, <5>4 DEF SemSingleReadyJoin, Inv
           <5>9. mm.mreq.lset.node # i \*Ready or not ready
              BY <3>1,<5>3,<5>4, <5>7, StatusDisjoint DEF SemJRply, Inv
           <5>10. QED BY <5>7, <5>8, <5>9, <4>3, <4>2
        <4>65. SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <4>5, <5>3
                DEF SemEmptyLSetSingleNode, Inv
        <4>66. SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY  DEF SemJoinLeafSet, Inv
           <5>4. mm \in receivedMsgs
              BY <4>5, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>71. mm.mreq.lset.node \in NodesReady
              BY <5>3, <5>4 DEF SemJRply, Inv
           <5>8. mm.mreq.lset.node # i
              BY <3>1,<5>3,<5>4, <5>7,<5>71, StatusDisjoint DEF SemJRply, Inv
           <5>9. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                 BY <4>3, <5>7, <5>8
           <5>10. QED BY <5>9, <5>3, <5>4 DEF SemJoinLeafSet, Inv
        <4>67. GrantNeighbor'
           BY <4>2, <4>3, <4>4, ReadyInI DEF GrantNeighbor, Inv
        <4>68. SingleReadyNode'
           BY <4>2, <4>3, <4>4, ReadyInI DEF SingleReadyNode, Inv
        <4>69. CoverToJ'
           <5>0. SUFFICES ASSUME NEW rr \in NodesReady', NEW kk \in I
              PROVE CoverToJ!(rr, kk)'
              BY DEF CoverToJ
           <5>1. /\ NodesReady'= NodesReady
                 /\ \A k \in I\{i}: lset'[k] = lset[k]
                 /\ \A k \in I: toj'[k] = toj[k]
                 /\ ~ i \in NodesReady
                 /\ rr \in I
                 /\ rr \in NodesReady
              BY <4>2,<4>3, <3>1, ReadyInI DEF TypeInvariant
           <5>3. CASE kk = i
              <6>1. SUFFICES ASSUME 
                    lset'[i] # EmptyLS(i), 
                    Covers(lset'[rr], i),
                    i # rr
                    PROVE i = toj'[rr]
                 BY <5>3, <5>0, <5>1 DEF CoverToJ, Inv
              <6>2. Covers(lset[rr], i)
                 BY <5>1, <6>1 
              <6>3.  m.mreq.lset.node \in NodesReady
                 BY <3>1 DEF SemJRply, Inv     
              <6>4. /\ TypeInvariant
                    /\ Inv
                    /\ Covers(lset[m.mreq.lset.node], i) 
                 BY <3>1 DEF SemJRply, Inv     
              <6>7. rr = m.mreq.lset.node
                 BY ConsistentInv, <6>2, <5>1, <5>0,<6>3, <6>4, ReadyInI DEF NodesReady
              <6>8. toj[rr] = i 
                 BY <3>1, <6>7, <5>0, <5>1 DEF SemJRply, Inv
              <6>9. QED BY <6>8, <5>1
           <5>4. CASE kk # i
              <6>1. rr # i
                 BY <5>1, <5>0
              <6>2. QED BY <6>1, <5>4, <5>0, <5>1, ReadyInI DEF CoverToJ, Inv
           <5>9. QED BY <5>3,<5>4
        <4>70. SemPRplyNonEmpty'
           <5>1. lset'[i] # EmptyLS(i)
              <6>1. lset'[i] # EmptyLS(lset[i].node)
                 BY <3>1, AddToLSetNoMoreEmpty, <4>03, <4>16, <6>1
              <6>9. QED BY <4>03,<6>1
           <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 \/ mm.mreq.type = "LSProbeReply"
                 \/ mm.mreq.type = "LSProbe"
                 PROVE lset'[mm.mreq.node] # EmptyLS(mm.mreq.node) 
              BY DEF SemPRplyNonEmpty, Inv
           <5>5. CASE mm \in receivedMsgs
              <6>1. mm.mreq \in Prb \/ mm.mreq \in PRpl
                 BY SlowestZenon,<5>5, <5>2 DEF TypeInvariant, DMsg, 
                 MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
              <6>2. mm.mreq.node \in I
                BY <6>1 DEF PRpl,Prb
              <6>4. CASE mm.mreq.node = i
                 BY <5>1, <6>4 
              <6>5. CASE mm.mreq.node # i
                 BY <5>5, <5>2, <6>5, <4>3, <6>2 DEF SemPRplyNonEmpty, Inv
              <6>9. QED BY <6>4,<6>5
           <5>6. CASE ~ mm \in receivedMsgs
              <6>1. mm \in Probe(i, newlset, {}, toprob)
                 BY <5>6, <3>1, <5>2
              <6>3. \A ms \in Probe(i, newlset, {}, toprob): ms.mreq.node = i
                 <7>1. ProbeMsgTypeConclusion(i, newlset, {}, toprob)
                    BY ONLY ProbeMsgType, <4>03
                 <7>2. QED BY <7>1 DEF ProbeMsgTypeConclusion             
              <6>9. QED BY <5>1, <6>3, <6>1
           <5>9. QED BY <5>5,<5>6
        <4>71. NoCCJoin'
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW rr \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoin!(ii,rr,kk)'
                 BY DEF NoCCJoin
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ m.mreq.lset.node \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I\{i}: lset'[k] = lset[k]
                /\ RightNeighbor(lset'[i]) \in I
                /\ RightNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ rr \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk     
              <6>1. /\ lset[i] \in LSet
                    /\ lset'[i] \in LSet 
                    /\ m.mreq.lset.node \in I
                    /\ \A k \in I\{i} : lset'[k] = lset[k]
                    /\ lset[i].node = i
                    /\ lset'[i].node = i
                    /\ RightNeighbor(lset[i]) \in I
                    /\ RightNeighbor(lset'[i]) \in I
                 BY <3>1, <4>03, <4>3, <4>16, RightNeighborType
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>2, <3>1 DEF TypeInvariant
              <6>3. /\ ii \in I
                    /\ rr \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3 
           <5>2. ~i \in NodesReady
              BY <4>2, <4>3, ReadyInI
           <5>3. /\ i # ii
                 /\ i # rr
              BY <5>2, <5>1, <5>0
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     rr = RightNeighbor(lset[ii]),
                     ii # rr,
                     CwDist(ii, i) < CwDist(ii, rr)
                     PROVE 
                      \/ toj[ii] = kk /\ toj[rr] = rr
                      \/ toj[rr] = kk /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1, <4>03  DEF NoCCJoin, Inv
              <6>2. /\ m.mreq.lset.node \in I
                    /\ m.mreq.lset.node # i
                    /\ lset[m.mreq.lset.node].node = m.mreq.lset.node
                    /\ lset[m.mreq.lset.node] \in LSet
                    /\ RightNeighbor(lset[m.mreq.lset.node]) \in I
                    /\ LeftNeighbor(lset[m.mreq.lset.node]) \in I
                 BY <3>1,<5>1, RightNeighborType, LeftNeighborType DEF TypeInvariant
              <6>3. /\ Covers(lset[m.mreq.lset.node], i)
                    /\ toj[m.mreq.lset.node] = i
                    /\ m.mreq.lset.node \in NodesReady
                 BY <3>1 DEF SemJRply, Inv
              <6>4. \/ CwDist(m.mreq.lset.node, i) 
                      < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                    \/ CwDist(i, m.mreq.lset.node) 
                      < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) 
                 BY <6>3, <3>1, <6>1, <6>2, NeighborCoverageCo
              <6>7. CASE CwDist(m.mreq.lset.node, i) 
                      < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                 <7>3. ii = m.mreq.lset.node
                     <8>1. SUFFICES ASSUME ii # m.mreq.lset.node 
                           PROVE FALSE
                        OBVIOUS
                     <8>2. /\ i # m.mreq.lset.node
                           /\ ii # i
                           /\ i \in I
                           /\ ii \in I
                           /\ m.mreq.lset.node \in I
                           /\ ii \in NodesReady
                           /\ rr \in NodesReady
                           /\ m.mreq.lset.node \in NodesReady
                        BY <6>2, <5>1, <5>3, <5>0, <6>3
                     <8>3. \/ CwDist(m.mreq.lset.node , ii) < CwDist(m.mreq.lset.node , i)
                           \/ CwDist(ii, m.mreq.lset.node ) < CwDist(ii, i) 
                        BY Ring3NodesPos, <8>2, <8>1
                     <8>4. CwDist(ii, rr)
                              =<  CwDist(ii, m.mreq.lset.node)
                        BY <8>1, <8>2, <6>1, CardinalityTwoNode, ReadyInI,
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>6. CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                              =<  CwDist(m.mreq.lset.node, ii)
                        BY <8>1, <8>2, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>7. CASE CwDist(m.mreq.lset.node , ii) < CwDist(m.mreq.lset.node , i)
                        <9>0. /\ ii \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ RightNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ CwDist(m.mreq.lset.node , ii) \in Nat
                              /\ CwDist(m.mreq.lset.node , i) \in Nat
                              /\ CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node])) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(m.mreq.lset.node, ii) 
                             < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                           BY <8>7, <6>7, TransLESS, <9>0
                        <9>9. QED BY <8>6, <9>1, CwDistType, <8>2, <6>2, IisNat, Unequality2
                     <8>8. CASE CwDist(ii, m.mreq.lset.node ) < CwDist(ii, i) 
                        <9>2. CwDist(ii, m.mreq.lset.node) < CwDist(ii, rr)
                           BY <8>8,<6>1,  TransLESS, <8>2, <6>2, ReadyInI, IisNat, CwDistType
                        <9>9. QED BY <9>2, <8>4, CwDistType, ReadyInI, <8>2, <6>2, IisNat, Unequality2
                     <8>9. QED BY <8>3, <8>7, <8>8
                 <7>4. toj[ii] = kk
                     BY <7>3, <6>3, <5>7
                 <7>5. toj[rr] = rr
                        BY <3>1, <7>3, <6>1 DEF SemJRply, SemJoinLeafSet, Inv
                 <7>9. QED BY <7>4, <7>5
              <6>8. CASE CwDist(i, m.mreq.lset.node) 
                          < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) 
                 <7>1. ii # m.mreq.lset.node
                    <8>1. SUFFICES ASSUME ii = m.mreq.lset.node PROVE FALSE
                       OBVIOUS
                    <8>2. /\ rr \in Nat
                          /\ m.mreq.lset.node \in Nat
                          /\ LeftNeighbor(lset[m.mreq.lset.node]) \in Nat
                          /\ i \in Nat
                          /\ ii \in Nat
                          /\ CwDist(rr , m.mreq.lset.node) \in Nat
                          /\ CwDist(i , m.mreq.lset.node) \in Nat
                          /\ CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) \in Nat
                       BY <6>2, <5>1, <5>3, <5>0, <6>3, IisNat, CwDistType
                    <8>3. CwDist(ii, i) < CwDist(ii, RightNeighbor(lset[ii]))
                       BY <6>1
                    <8>4. CwDist(i, ii) 
                          < CwDist(LeftNeighbor(lset[ii]), ii) 
                       BY <6>8, <8>1
                    <8>9. QED BY <8>3,<8>4, <5>3, LeftRightDisjoint, <6>2, <8>1
                 <7>2. rr = m.mreq.lset.node 
                    <8>1. SUFFICES ASSUME rr # m.mreq.lset.node PROVE FALSE
                         OBVIOUS
                    <8>2.  /\ i # m.mreq.lset.node
                           /\ rr # i
                           /\ i # ii
                           /\ i \in I
                           /\ rr \in I
                           /\ ii \in I
                           /\ m.mreq.lset.node \in I
                           /\ ii \in NodesReady
                           /\ rr \in NodesReady
                           /\ m.mreq.lset.node \in NodesReady
                        BY <6>2, <5>1, <5>3, <5>0, <6>3
                     <8>3. \/ CwDist(rr , m.mreq.lset.node) < CwDist(i , m.mreq.lset.node)
                           \/ CwDist(m.mreq.lset.node, rr ) < CwDist(i, rr) 
                        BY Ring3NodesPos4, <8>2, <8>1
                     <8>4. CwDist(ii, rr)
                              =<  CwDist(ii, m.mreq.lset.node)
                        BY <8>1, <8>2, <7>1, <6>1, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>6. CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node)
                              =< CwDist(rr, m.mreq.lset.node) 
                        BY <8>1, <8>2, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>7. CASE CwDist(rr , m.mreq.lset.node) < CwDist(i , m.mreq.lset.node)
                        <9>0. /\ rr \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ LeftNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ ii \in Nat
                              /\ CwDist(rr , m.mreq.lset.node) \in Nat
                              /\ CwDist(i , m.mreq.lset.node) \in Nat
                              /\ CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(rr, m.mreq.lset.node) 
                                < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node)
                           BY <8>7, <6>8, TransLESS, <9>0, SlowestZenon
                        <9>9. QED BY <8>6, <9>1, CwDistType, <8>2, <6>2, IisNat, Unequality2
                     <8>8. CASE  CwDist(m.mreq.lset.node, rr ) < CwDist(i, rr) 
                        <9>0. /\ rr \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ LeftNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ ii \in Nat
                              /\ CwDist(rr , m.mreq.lset.node) \in Nat
                              /\ CwDist(i , m.mreq.lset.node) \in Nat
                              /\ CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(i, rr) # CwDist(ii, rr)
                           BY <8>2, <9>0, CwDistInjectivity
                        <9>2. CwDist(i, rr) =< CwDist(ii, rr)
                           BY <6>1, <9>0, CwDistProp
                        <9>3. CwDist(i, rr) < CwDist(ii, rr)
                           BY <9>0, <9>1,<9>2,CwDistType, Unequality6
                        <9>4. CwDist(m.mreq.lset.node, rr ) < CwDist(ii, rr)
                           BY <8>8,<9>3,  TransLESS, <9>0, CwDistType
                        <9>5. CwDist(ii, m.mreq.lset.node) =< CwDist(ii, rr)
                           BY <9>4, CwDistPropCo, <9>0
                        <9>6. CwDist(ii, m.mreq.lset.node) # CwDist(ii, rr)
                           BY <8>1, <9>0,  RightDistInjectivity
                        <9>7. CwDist(ii, m.mreq.lset.node) < CwDist(ii, rr)
                           BY <9>0,<9>5,<9>6, CwDistType, Unequality6
                        <9>9. QED BY <9>7, <8>4, CwDistType, ReadyInI, <8>2, <6>2, IisNat, Unequality2
                     <8>9. QED BY <8>3, <8>7, <8>8
                 <7>3. ii = LeftNeighbor(lset[rr])
                    <8>1. SUFFICES ASSUME ii # LeftNeighbor(lset[rr]) PROVE FALSE
                        OBVIOUS
                    <8>2. /\ rr \in I
                          /\ m.mreq.lset.node \in I
                          /\ LeftNeighbor(lset[rr]) \in I
                          /\ ii \in I
                          /\ CwDist(ii, LeftNeighbor(lset[rr])) \in Nat
                          /\ CwDist(ii, rr) \in Nat
                          /\  CwDist(LeftNeighbor(lset[rr]), rr) \in Nat
                          /\ ii # rr
                          /\ ii \in NodesReady
                          /\ rr \in NodesReady
                       <9>1.  /\ ii # rr
                              /\ ii \in NodesReady
                              /\ rr \in NodesReady
                              /\ m.mreq.lset.node \in I
                              /\ LeftNeighbor(lset[rr]) \in I
                              /\ ii \in I
                              /\ rr \in I
                          BY  <6>1, <5>1, <5>0, ReadyInI, <6>2, <7>2
                       <9>2.  /\ CwDist(ii, LeftNeighbor(lset[rr])) \in Nat
                              /\ CwDist(ii, rr) \in Nat
                              /\  CwDist(LeftNeighbor(lset[rr]), rr) \in Nat
                          BY <9>1, CwDistType, IisNat
                       <9>3. QED BY <9>1,<9>2
                    <8>3. LeftNeighbor(lset[rr]) \in NodesReady
                       <9>1.  LeftNeighbor(lset[m.mreq.lset.node]) \in NodesReady 
                          BY <3>1,  <6>2 DEF SemJRply, LeaseGrant, GrantReady, Inv
                       <9>9. QED BY <9>1, <7>2
                    <8>4. \/ CwDist(ii, LeftNeighbor(lset[rr])) < CwDist(ii, rr)
                          \/ CwDist(ii, rr) < CwDist(LeftNeighbor(lset[rr]), rr)
                       <9>1. rr # LeftNeighbor(lset[rr])
                          BY <8>2,  ReadyInI, 
                             CardinalityOneNodeExt, CardinalityTwoNode DEF HalfNeighbor, Inv
                       <9>9. QED BY <6>2, Ring3NodesPos5, <8>2, <8>1, <9>1
                    <8>5. CwDist(ii, rr) =< CwDist(ii, LeftNeighbor(lset[rr]))
                       BY <6>1, <8>2, CardinalityTwoNode, ReadyInI, <8>3, <8>1, 
                       LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>6. CwDist(LeftNeighbor(lset[rr]), rr) =< CwDist(ii, rr)
                       BY <8>2, CardinalityTwoNode, ReadyInI, 
                       LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>7. CASE  CwDist(ii, LeftNeighbor(lset[rr])) < CwDist(ii, rr)
                       BY <8>5, <8>7, Unequality2, <8>2, IisNat
                    <8>8. CASE CwDist(ii, rr) < CwDist(LeftNeighbor(lset[rr]), rr)
                       BY <8>6, <8>8, Unequality2, <8>2, IisNat
                    <8>9. QED BY <8>4, <8>7, <8>8
                 <7>4. toj[rr] = kk
                     BY <7>2, <6>3, <5>7
                 <7>5. ii = LeftNeighbor(m.mreq.lset)
                    BY <7>2,<7>3, <3>1 DEF SemJoinLeafSet, Inv 
                 <7>6. LeftNeighbor(m.mreq.lset) # m.mreq.lset.node 
                    BY <7>5, <7>1
                 <7>8. toj[ii] = ii 
                        BY <3>1, <7>5, <7>6 DEF SemJRply, Inv
                 <7>9. QED BY <7>4, <7>8
              <6>9. QED BY <6>4, <6>7,<6>8
           <5>8. CASE kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    rr = RightNeighbor(lset[ii]),
                    ii # rr,
                    CwDist(ii, kk) < CwDist(ii, rr)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[rr] = rr
                     \/ toj[rr] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>3, <5>8 DEF NoCCJoin
              <6>. QED BY <5>1,<6>1, NoCCJoin DEF NoCCJoin, Inv
           <5>. QED
             BY <5>3, <5>7, <5>8      
        <4>72. NoCCJoinL'
              \* OMITTED \* similar to <4>71
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW ll \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoinL!(ii,ll,kk)'
                 BY DEF NoCCJoinL
           <5>1. 
                /\ lset'[i] \in LSet 
                /\ lset[i] \in LSet
                /\ m.mreq.lset.node \in I
                /\ lset'[i].node = i
                /\ lset[i].node = i
                /\ \A k \in I\{i}: lset'[k] = lset[k]
                /\ LeftNeighbor(lset'[i]) \in I
                /\ LeftNeighbor(lset[i]) \in I
                /\ NodesReady' = NodesReady
                /\ \A k \in I: toj'[k] = toj[k]
                /\ ii \in I
                /\ ll \in I
                /\ lset[kk] \in LSet
                /\ lset[kk].node = kk     
              <6>1. /\ lset[i] \in LSet
                    /\ lset'[i] \in LSet 
                    /\ m.mreq.lset.node \in I
                    /\ \A k \in I\{i} : lset'[k] = lset[k]
                    /\ lset[i].node = i
                    /\ lset'[i].node = i
                    /\ LeftNeighbor(lset[i]) \in I
                    /\ LeftNeighbor(lset'[i]) \in I
                 BY <3>1, <4>03, <4>3, <4>16, LeftNeighborType
              <6>2. /\ NodesReady' = NodesReady
                    /\ \A k \in I: toj'[k] = toj[k]
                 BY <4>2, <3>1 DEF TypeInvariant
              <6>3. /\ ii \in I
                    /\ ll \in I
                    /\ lset[kk] \in LSet
                    /\ lset[kk].node = kk     
                 BY <5>0, ReadyInI, <6>2 DEF TypeInvariant 
              <6>9. QED BY <6>1,<6>2,<6>3 
           <5>2. ~i \in NodesReady
              BY <4>2, <4>3, ReadyInI
           <5>3. /\ i # ii
                 /\ i # ll
              BY <5>2, <5>1, <5>0
           <5>7. CASE kk = i
              <6>1. SUFFICES ASSUME ii # i,
                     lset'[i] # EmptyLS(i),
                     ll = LeftNeighbor(lset[ii]),
                     ii # ll,
                     CwDist(i, ii) < CwDist(ll, ii)
                     PROVE 
                      \/ toj[ii] = kk /\ toj[ll] = ll
                      \/ toj[ll] = kk /\ toj[ii] = ii
                 BY <5>7, <5>0, <5>1, <4>03  DEF NoCCJoinL, Inv
              <6>2. /\ m.mreq.lset.node \in I
                    /\ m.mreq.lset.node # i
                    /\ lset[m.mreq.lset.node].node = m.mreq.lset.node
                    /\ lset[m.mreq.lset.node] \in LSet
                    /\ RightNeighbor(lset[m.mreq.lset.node]) \in I
                    /\ LeftNeighbor(lset[m.mreq.lset.node]) \in I
                 BY <3>1,<5>1, RightNeighborType, LeftNeighborType DEF TypeInvariant
              <6>3. /\ Covers(lset[m.mreq.lset.node], i)
                    /\ toj[m.mreq.lset.node] = i
                    /\ m.mreq.lset.node \in NodesReady
                 BY <3>1 DEF SemJRply, Inv
              <6>4. \/ CwDist(m.mreq.lset.node, i) 
                      < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                    \/ CwDist(i, m.mreq.lset.node) 
                      < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) 
                 BY <6>3, <3>1, <6>1, <6>2, NeighborCoverageCo
              <6>7. CASE CwDist(i, m.mreq.lset.node) 
                      < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) 
                 <7>3. ii = m.mreq.lset.node
                     <8>1. SUFFICES ASSUME ii # m.mreq.lset.node 
                           PROVE FALSE
                        OBVIOUS
                     <8>2. /\ i # m.mreq.lset.node
                           /\ ii # i
                           /\ i \in I
                           /\ ii \in I
                           /\ m.mreq.lset.node \in I
                           /\ ii \in NodesReady
                           /\ ll \in NodesReady
                           /\ m.mreq.lset.node \in NodesReady
                        BY <6>2, <5>1, <5>3, <5>0, <6>3
                     <8>3. \/ CwDist(ii, m.mreq.lset.node) < CwDist(i, m.mreq.lset.node)
                           \/ CwDist(m.mreq.lset.node, ii ) < CwDist(i, ii) 
                        BY Ring3NodesPos4, <8>2, <8>1
                     <8>4. CwDist(ll, ii)
                              =<  CwDist(m.mreq.lset.node, ii)
                        BY <8>1, <8>2, <6>1, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>6. CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node)
                              =<  CwDist(ii, m.mreq.lset.node)
                        BY <8>1, <8>2, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>7. CASE CwDist(ii, m.mreq.lset.node) < CwDist(i, m.mreq.lset.node)
                        <9>0. /\ ii \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ LeftNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ CwDist(ii, m.mreq.lset.node) \in Nat
                              /\ CwDist(i, m.mreq.lset.node) \in Nat
                              /\ CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(ii, m.mreq.lset.node) 
                             < CwDist(LeftNeighbor(lset[m.mreq.lset.node]), m.mreq.lset.node)
                           BY <8>7, <6>7, TransLESS, <9>0, SlowestZenon
                        <9>9. QED BY <8>6, <9>1, CwDistType, <8>2, <6>2, IisNat, Unequality2
                     <8>8. CASE CwDist(m.mreq.lset.node, ii ) < CwDist(i, ii) 
                        <9>2. CwDist(m.mreq.lset.node, ii ) < CwDist(ll, ii) 
                           BY <8>8,<6>1,  TransLESS, <8>2, <6>2, ReadyInI, IisNat, CwDistType
                        <9>9. QED BY <9>2, <8>4, CwDistType, ReadyInI, <8>2, <6>2, IisNat, Unequality2
                     <8>9. QED BY <8>3, <8>7, <8>8
                 <7>4. toj[ii] = kk
                     BY <7>3, <6>3, <5>7
                 <7>5. toj[ll] = ll
                        BY <3>1, <7>3, <6>1 DEF SemJRply, SemJoinLeafSet, Inv
                 <7>9. QED BY <7>4, <7>5
              <6>8. CASE CwDist(m.mreq.lset.node, i) 
                      < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                 <7>1. ii # m.mreq.lset.node
                    <8>1. SUFFICES ASSUME ii = m.mreq.lset.node PROVE FALSE
                       OBVIOUS
                    <8>2. /\ ll \in Nat
                          /\ m.mreq.lset.node \in Nat
                          /\ RightNeighbor(lset[m.mreq.lset.node]) \in Nat
                          /\ i \in Nat
                          /\ ii \in Nat
                          /\ CwDist( m.mreq.lset.node, ll) \in Nat
                          /\ CwDist(m.mreq.lset.node, i) \in Nat
                          /\ CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node])) \in Nat
                       BY <6>2, <5>1, <5>3, <5>0, <6>3, IisNat, CwDistType
                    <8>3. CwDist(i, ii) < CwDist(LeftNeighbor(lset[ii]), ii)
                       BY <6>1
                    <8>4. CwDist(ii, i) 
                          < CwDist(ii, RightNeighbor(lset[ii])) 
                       BY <6>8, <8>1
                    <8>9. QED BY <8>3,<8>4, <5>3, LeftRightDisjoint, <6>2, <8>1
                 <7>2. ll = m.mreq.lset.node 
                    <8>1. SUFFICES ASSUME ll # m.mreq.lset.node PROVE FALSE
                         OBVIOUS
                    <8>2.  /\ i # m.mreq.lset.node
                           /\ ll # i
                           /\ i # ii
                           /\ i \in I
                           /\ ll \in I
                           /\ ii \in I
                           /\ m.mreq.lset.node \in I
                           /\ ii \in NodesReady
                           /\ ll \in NodesReady
                           /\ m.mreq.lset.node \in NodesReady
                        BY <6>2, <5>1, <5>3, <5>0, <6>3
                     <8>3. \/ CwDist( m.mreq.lset.node, ll) < CwDist(m.mreq.lset.node, i)
                           \/ CwDist(ll, m.mreq.lset.node) < CwDist(ll, i) 
                        BY Ring3NodesPos, <8>2, <8>1
                     <8>4. CwDist(ll, ii)
                              =<  CwDist(m.mreq.lset.node, ii)
                        BY <8>1, <8>2, <7>1, <6>1, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>6. CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                              =< CwDist(m.mreq.lset.node, ll) 
                        BY <8>1, <8>2, CardinalityTwoNode, ReadyInI, 
                        LemmaCLtoNC DEF NeighborClosest, Inv
                     <8>7. CASE CwDist(m.mreq.lset.node, ll) < CwDist(m.mreq.lset.node, i)
                        <9>0. /\ ll \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ RightNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ ii \in Nat
                              /\ CwDist(m.mreq.lset.node, ll) \in Nat
                              /\ CwDist(m.mreq.lset.node, i) \in Nat
                              /\ CwDist( m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node])) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(m.mreq.lset.node, ll) 
                                < CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node]))
                           BY <8>7, <6>8, TransLESS, <9>0, SimplifyAndSolve, IsaT(60)
                        <9>9. QED BY <8>6, <9>1, CwDistType, <8>2, <6>2, IisNat, Unequality2
                     <8>8. CASE  CwDist(ll, m.mreq.lset.node) < CwDist(ll, i) 
                        <9>0. /\ ll \in Nat
                              /\ m.mreq.lset.node \in Nat
                              /\ RightNeighbor(lset[m.mreq.lset.node]) \in Nat
                              /\ i \in Nat
                              /\ ii \in Nat
                              /\ CwDist(m.mreq.lset.node, ll) \in Nat
                              /\ CwDist(m.mreq.lset.node, i) \in Nat
                              /\ CwDist(m.mreq.lset.node, RightNeighbor(lset[m.mreq.lset.node])) \in Nat
                           BY <8>2, <6>2, IisNat, CwDistType
                        <9>1. CwDist(ll, i) # CwDist(ll, ii)
                           BY <8>2, <9>0, RightDistInjectivity
                        <9>2. CwDist(ll, i) =< CwDist(ll, ii)
                           BY <6>1, <9>0, CwDistPropCo
                        <9>3. CwDist(ll, i) < CwDist(ll, ii)
                           BY <9>0, <9>1,<9>2,CwDistType, Unequality6
                        <9>4. CwDist(ll, m.mreq.lset.node) < CwDist(ll, ii)
                           BY <8>8,<9>3,  TransLESS, <9>0, CwDistType
                        <9>5. CwDist(m.mreq.lset.node, ii) =< CwDist(ll, ii)
                           BY <9>4, CwDistProp, <9>0
                        <9>6. CwDist(m.mreq.lset.node, ii) # CwDist(ll, ii)
                           BY <8>1, <9>0,  CwDistInjectivity
                        <9>7. CwDist(m.mreq.lset.node, ii) < CwDist(ll, ii)
                           BY <9>0,<9>5,<9>6, CwDistType, Unequality6
                        <9>9. QED BY <9>7, <8>4, CwDistType, ReadyInI, <8>2, <6>2, IisNat, Unequality2
                     <8>9. QED BY <8>3, <8>7, <8>8
                 <7>3. ii = RightNeighbor(lset[ll])
                    <8>1. SUFFICES ASSUME ii # RightNeighbor(lset[ll]) PROVE FALSE
                        OBVIOUS
                    <8>2. /\ ll \in I
                          /\ m.mreq.lset.node \in I
                          /\ RightNeighbor(lset[ll]) \in I
                          /\ ii \in I
                          /\ CwDist(RightNeighbor(lset[ll]), ii) \in Nat
                          /\ CwDist(ll, ii) \in Nat
                          /\  CwDist(ll, RightNeighbor(lset[ll])) \in Nat
                          /\ ii # ll
                          /\ ii \in NodesReady
                          /\ ll \in NodesReady
                       <9>1.  /\ ii # ll
                              /\ ii \in NodesReady
                              /\ ll \in NodesReady
                              /\ m.mreq.lset.node \in I
                              /\ RightNeighbor(lset[ll]) \in I
                              /\ ii \in I
                              /\ ll \in I
                          BY  <6>1, <5>1, <5>0, ReadyInI, <6>2, <7>2
                       <9>2.  /\ CwDist( RightNeighbor(lset[ll]), ii) \in Nat
                              /\ CwDist(ll, ii) \in Nat
                              /\  CwDist(ll, RightNeighbor(lset[ll])) \in Nat
                          BY <9>1, CwDistType, IisNat
                       <9>3. QED BY <9>1,<9>2
                    <8>3. RightNeighbor(lset[ll]) \in NodesReady
                       <9>1.  RightNeighbor(lset[m.mreq.lset.node]) \in NodesReady 
                          BY <3>1,  <6>2 DEF SemJRply, LeaseGrant, GrantReady, Inv
                       <9>9. QED BY <9>1, <7>2
                    <8>4. \/ CwDist(RightNeighbor(lset[ll]), ii) < CwDist(ll, ii)
                          \/ CwDist(ll, ii) < CwDist(ll, RightNeighbor(lset[ll]))
                       <9>1. ll # RightNeighbor(lset[ll])
                          BY <8>2,  ReadyInI, 
                             CardinalityOneNodeExt, CardinalityTwoNode DEF HalfNeighbor, Inv
                       <9>9. QED BY <6>2, Ring3NodesPos10, <8>2, <8>1, <9>1
                    <8>5. CwDist(ll, ii) =< CwDist(RightNeighbor(lset[ll]), ii)
                       BY <6>1, <8>2, CardinalityTwoNode, ReadyInI, <8>3, <8>1, 
                       LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>6. CwDist(ll, RightNeighbor(lset[ll])) =< CwDist(ll, ii)
                       BY <8>2, CardinalityTwoNode, ReadyInI, 
                       LemmaCLtoNC DEF NeighborClosest, Inv
                    <8>7. CASE  CwDist(RightNeighbor(lset[ll]), ii) < CwDist(ll, ii)
                       BY <8>5, <8>7, Unequality2, <8>2, IisNat
                    <8>8. CASE CwDist(ll, ii) < CwDist(ll, RightNeighbor(lset[ll]))
                       BY <8>6, <8>8, Unequality2, <8>2, IisNat
                    <8>9. QED BY <8>4, <8>7, <8>8
                 <7>4. toj[ll] = kk
                     BY <7>2, <6>3, <5>7
                 <7>5. ii = RightNeighbor(m.mreq.lset)
                    BY <7>2,<7>3, <3>1 DEF SemJoinLeafSet, Inv 
                 <7>6. RightNeighbor(m.mreq.lset) # m.mreq.lset.node 
                    BY <7>5, <7>1
                 <7>8. toj[ii] = ii 
                        BY <3>1, <7>5, <7>6 DEF SemJRply, Inv
                 <7>9. QED BY <7>4, <7>8
              <6>9. QED BY <6>4, <6>7,<6>8
           <5>8. CASE kk # i
              <6>1. SUFFICES ASSUME 
                    ii # kk,
                    lset[kk] # EmptyLS(kk),
                    ll = LeftNeighbor(lset[ii]),
                    ii # ll,
                    CwDist(kk, ii) < CwDist(ll, ii)
                    PROVE 
                     \/ toj[ii] = kk /\ toj[ll] = ll
                     \/ toj[ll] = kk /\ toj[ii] = ii
              BY <5>1, <5>0, <5>3, <5>8 DEF NoCCJoinL
              <6>. QED BY <5>1,<6>1 DEF NoCCJoinL, Inv
           <5>. QED
             BY <5>3, <5>7, <5>8      
        <4>73. SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY  DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <4>5, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. ASSUME i \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}
                 PROVE lset'[i] # EmptyLS(i)
              <6>1. /\ m.mreq.lset \in LSet
                    /\ lset[i] \in LSet
                    /\ m.mreq.lset.node # lset[i].node 
                 BY <5>6, <4>03, <4>16, <3>1
              <6>2. AddToLSet(GetLSetContent(m.mreq.lset), lset[i]) # EmptyLS(lset[i].node)
                 BY <6>1, AddToLSetNoMoreEmpty
              <6>9. QED BY <6>2, <3>1, <4>03, <4>16
           <5>99. QED BY <5>3, <5>4, <5>8, <5>9, <4>3
                DEF SemJRplyLSnonEmpty, Inv
        <4>74. SingleReadyNodeToj'
           <5>1. SUFFICES ASSUME NEW kk \in I, 
                 NEW rr \in I,
                 kk # rr,
                 NodesReady = {rr},
                 lset'[kk] # EmptyLS(kk)
                 PROVE toj[rr] = kk
              BY <4>2,<4>4 DEF SingleReadyNodeToj, Inv
           <5>2. /\ lset[i] \in LSet
                 /\ lset[i].node = i
              BY <3>1, <4>2, <4>3,<4>03,<4>16 
           <5>4. lset'[i] # EmptyLS(i)
              BY <3>1, AddToLSetNoMoreEmpty, <4>03, <4>16, <5>2
           <5>5. CASE kk = i
              <6>5. m.mreq.lset.node \in NodesReady
                 BY <3>1 DEF SemJRply, Inv
              <6>6. m.mreq.lset.node = rr
                 BY <6>5, <5>1
              <6>9. QED BY <6>6, <5>5, <3>1 DEF SemJRply, Inv
           <5>6. CASE rr = i
              <6>2. lset[kk] # EmptyLS(kk)
                 BY <5>1, <4>3, <5>6
              <6>3. QED BY <6>2, <5>1, SingleReadyNodeToj DEF SingleReadyNodeToj, Inv
           <5>7. CASE kk # i /\ rr # i
              BY <5>7, <4>3, <5>1 DEF SingleReadyNodeToj, Inv
           <5>9. QED BY <5>5,<5>6,<5>7
        <4>75. OneJReq'
           BY <4>5,<4>4, <4>3,<4>2 DEF OneJReq, Inv
        <4>76. OneJRply'
           BY <4>5,<4>4, <4>3,<4>2 DEF OneJRply, Inv
        <4>77. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>2. mm \in receivedMsgs
                  BY <4>5, <5>1
               <5>3. mm.mreq \in JReq
                  BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>4. mm.mreq.node \in I
                  BY <5>3 DEF JReq
               <5>5. mm.mreq.node # i
                  BY <5>1, <5>2, <3>1 DEF JoinMsgDual, Inv
               <5>9. QED BY <5>5,<5>2, <4>3, <4>2, <5>1, ReadyInI 
                  DEF SemJReq, Inv
        <4>78. JoinMsgDual'
               BY <4>5,<4>4, <4>3,<4>2 DEF JoinMsgDual, Inv
        <4>80. AtleastOneReady'
           BY <4>2 DEF AtleastOneReady, Inv
        <4>99. QED BY <4>20,<4>95,\*<4>96,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69, <4>70, 
                  <4>71, <4>72, <4>73,<4>74,
                  <4>75,<4>76,
                  <4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <3>1, <1>7\*, SimplifyAndSolve
           DEF ReceiveJoinReply\*, Inv
   <1>8. ASSUME NEW i \in I,
                RequestLease(i),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME NEW ln, ln = LeftNeighbor(lset[i]),
            NEW rn, rn =  RightNeighbor(lset[i]),
            NEW llr, llr =   IF ~ (ln \in lease[i])
                             THEN {[destination |-> ln, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]]}
                             ELSE {},
            NEW rlr, rlr =   IF ~ (rn \in lease[i])
                             THEN {[destination |-> rn, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]]}
                             ELSE {},
            receivedMsgs' = (receivedMsgs \cup llr) \cup rlr,
            lset[i] # EmptyLS(i),
            status[i] \in {"ok", "ready"},
            ~ (ln \in lease[i] /\ rn \in lease[i]),
            UNCHANGED <<status, lset, rtable, probing, failed, lease, grant, toj>>,
                    TypeInvariant,
                 Inv
         PROVE  Inv'
            <4>1. \A j \in I: /\lset'[j] = lset[j]
                              /\ toj'[j] = toj[j]
                              /\ grant'[j]= grant[j]
                              /\ lease'[j] = lease[j]
               BY <3>1, UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>
            <4>2./\ NodesReady' = NodesReady 
                 /\ ReadyOK' = ReadyOK 
                 /\ NonDead'= NonDead
                 /\ NodesWait' = NodesWait
               BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <3>1, <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>4. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
                     \/ ms.mreq.type = "JoinRequest"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: 
                         \/ ms \in 
                             {[destination |-> ln, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]]}
                         \/ ms \in {[destination |-> rn, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]]}
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> ln, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]].mreq.type = "LeaseRequest"
                     OBVIOUS
                  <6>3. [destination |-> rn, 
                                  mreq |-> [type |-> "LeaseRequest",
                                            node |-> i]].mreq.type = "LeaseRequest"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2,<6>3
            <4>5. HalfNeighbor' 
               <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                  BY <5>7,<4>2,<4>1, ReadyInI DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: /\ mk.mreq.type = "LSProbe" 
                                                     /\ mk.destination # k
                         /\ ~\E mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               /\ GetLSetContent(mj.mreq.lset) # {k}
                         /\~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                  <6>1. PICK k \in ReadyOK: 
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                     BY <5>8
                  <6>2. k = i
                     BY <3>1, <6>1, ReadyInI DEF NodesOK, ReadyOK, NodesReady
                  <6>3. lset[k].node = k /\ lset[k] \in LSet /\ k \in I
                     BY <6>1, ReadyInI DEF TypeInvariant
                  <6>4. lset[k] = EmptyLS(k)
                     BY <6>1, <6>3, EmptyLSNoNeighbor
                  <6>9. QED BY <6>2,<6>4,<3>1
               <5>9. QED BY <5>7, <5>8, HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>6. NeighborClosest'
              \* BY <4>1, <4>2, ReadyInI DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <4>1, <4>2, ReadyInI DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <4>1, <4>2, ReadyInI DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>4,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>61. GrantReady'
               BY <4>1, <4>2, ReadyInI DEF GrantReady, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>4, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>4, <5>3, <5>6, <5>8, <4>1, <4>2, 
                    SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>4, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>4, <5>3, <5>7, <5>6, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady'
                         /\ mm.destination \in NodesWait'
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>4,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <4>1, <4>2, ReadyInI DEF GrantNeighbor, Inv
            <4>68. SingleReadyNode'
               BY <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               BY <4>1, <4>2, ReadyInI, SlowestZenon DEF CoverToJ, Inv
            <4>70. NoCCJoin'
               BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve DEF NoCCJoin, Inv
            <4>71. NoCCJoinL'
               BY <3>1, <4>1, <4>2, ReadyInI, SimplifyAndSolve DEF NoCCJoinL, Inv
            <4>72. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>73. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>4, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>74. OneJRply'
               BY <4>4,<4>1,<4>2 DEF OneJRply, Inv
            <4>75. OneJReq'
               BY <4>4,<4>1,<4>2 DEF OneJReq, Inv
            <4>76. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>2. mm \in receivedMsgs
                  BY <4>4, <5>1
               <5>3. mm.mreq \in JReq
                  BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>4. mm.mreq.node \in I
                  BY <5>3 DEF JReq
               <5>9. QED BY <5>2, <4>1, <4>2, <5>1, ReadyInI DEF SemJReq, Inv
            <4>77. SingleReadyNodeToj'
               BY <3>1, <4>1, <4>2, ReadyInI, SingleReadyNodeToj DEF SingleReadyNodeToj, Inv
            <4>78. JoinMsgDual'
               BY <4>4,<4>1,<4>2 DEF JoinMsgDual, Inv
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv 
            <4>9. QED BY <4>3,<4>5,\*<4>6,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76, <4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <1>8, <3>1 DEF RequestLease, Inv
   <1>9. ASSUME NEW i \in I,
                ReceiveLReq(i),
                TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
            m.destination = i,
            lset[i] # EmptyLS(i),
            m.mreq.node # i,
            status[i] = "ready",
            m.mreq.type = "LeaseRequest",
            receivedMsgs' = (receivedMsgs \ {m})
                  \cup {[destination |-> m.mreq.node, 
                              mreq |-> [type |-> "BroadcastLSet",
                                        lset |-> lset[i],
                                        grant|-> m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                                                 ]]},
            grant' = [grant EXCEPT ![i] = 
                  IF m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                  THEN @ \cup {m.mreq.node} 
                  ELSE @],
            
            UNCHANGED <<status, rtable, lset, probing, failed, lease, toj>>,
                    TypeInvariant,
                 Inv
         PROVE  Inv'
         <4>0. m.mreq.node \in I 
            <5>1. m \in DMsg
               BY <3>1 DEF TypeInvariant
            <5>2. m.mreq \in LReq
               BY SlowestZenon, <3>1 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
            <5>3. QED BY <5>2 DEF LReq
         <4>. DEFINE Status == {0,1,2}
        <4>1. \A j \in I: /\ lset'[j] = lset[j]
                          /\ toj'[j] = toj[j]
                          /\ lease'[j] = lease[j]
           BY <3>1, UNCHANGED <<status, rtable, lset, probing, failed, lease, toj>>
        <4>2. /\ NodesReady' = NodesReady 
              /\ ReadyOK' = ReadyOK 
              /\ NonDead'= NonDead
              /\ NodesWait' = NodesWait
           BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
        <4>21. \A k \in I\{i}: grant'[k] = grant[k]
           BY <3>1 DEF TypeInvariant
        <4>22. \A k \in I: k \in grant[i]=>  k \in grant'[i]
           BY <3>1 DEF TypeInvariant
        <4>3. LSetInv'
           BY <3>1, <4>1, <4>2, LSetInv DEF LSetInv, Inv
        <4>4. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "JoinRequest"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: 
                         ms \in 
                             {[destination |-> m.mreq.node, 
                              mreq |-> [type |-> "BroadcastLSet",
                                        lset |-> lset[i],
                                        grant|-> m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                                        ]]}
                                                 \*/\ toj[i] \in {i, m.mreq.node}
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> m.mreq.node, 
                              mreq |-> [type |-> "BroadcastLSet",
                                        lset |-> lset[i],
                                        grant|-> m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                                        ]].mreq.type = "BroadcastLSet"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2
        <4>5. HalfNeighbor' 
               <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                  BY <5>7,<4>2,<4>1, ReadyInI DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: /\ mk.mreq.type = "LSProbe" 
                                                     /\ mk.destination # k
                         /\ ~\E mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               /\ GetLSetContent(mj.mreq.lset) # {k}
                         /\~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                  <6>1. PICK k \in ReadyOK: 
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                     BY <5>8
                  <6>2. k = i
                     BY <3>1, <6>1, ReadyInI DEF NodesOK, ReadyOK, NodesReady
                  <6>3. lset[k].node = k /\ lset[k] \in LSet /\ k \in I
                     BY <6>1, ReadyInI DEF TypeInvariant
                  <6>4. lset[k] = EmptyLS(k)
                     BY <6>1, <6>3, EmptyLSNoNeighbor
                  <6>9. QED BY <6>2,<6>4,<3>1
               <5>9. QED BY <5>7, <5>8, HalfNeighbor DEF HalfNeighbor, Inv
        \*<4>6. NeighborClosest'
          \* BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
        \*<4>51. ReadyNeighbor'
        \*<4>52. IncludeNeighbor'
        \*<4>53. CrossNeighbor'
        <4>54. CompleteLeafSet'
           BY <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
        <4>55. CompleteLeafSetL'
           BY <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
        <4>56. ReadyNeighborRight'
           BY <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
        <4>57. ReadyNeighborLeft'
           BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
        <4>58. SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                 NEW nn \in I, 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "BroadcastLSet",
                 mm.mreq.lset.node = nn,
                 mm.destination = ii
                PROVE
                    /\ nn \in NodesReady
                    /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
              BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv

           <5>4. CASE mm = [destination |-> m.mreq.node, 
                              mreq |-> [type |-> "BroadcastLSet",
                                        lset |-> lset[i],
                                        grant|-> m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                                        ]]
              <6>1. mm.mreq.lset.node = i
                 BY <5>4 DEF TypeInvariant
              <6>2. nn = i /\ nn \in NodesReady
                 BY <5>2, <6>1, <3>1 DEF NodesReady
              <6>4. CASE m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}                    
                 <7>2. m.mreq.node \in grant'[i]
                    BY <6>4, <3>1 DEF TypeInvariant
                 <7>3. mm.destination = m.mreq.node
                    BY <5>4
                 <7>9. QED BY <7>2, <6>2, <5>2, <7>3 DEF SemLRply, Inv
              <6>5. CASE ~m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}                   
                 BY <5>4, <6>2, <6>5 DEF   SemLRply, Inv
              <6>9. QED BY <6>5,<6>4
           <5>5. CASE mm # [destination |-> m.mreq.node, 
                              mreq |-> [type |-> "BroadcastLSet",
                                        lset |-> lset[i],
                                        grant|-> m.mreq.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])}
                                        ]]
              <6>1. mm \in receivedMsgs
                 BY <5>2, <5>5, <3>1
              <6>4. (mm.mreq.grant =TRUE => ii \in grant[nn])
                 BY <6>1, <5>2 DEF SemLRply, Inv
              <6>5. nn \in NodesReady
                 BY <6>1, <5>2 DEF  SemLRply, Inv
              <6>6. CASE nn = i
                 BY <6>4, <4>22, <6>5, <6>6
              <6>7. CASE nn # i
                 BY <4>21, <6>4, <6>5, <6>7
              <6>9. QED BY <6>6, <6>7
           <5>9. QED BY <5>4,<5>5
        <4>59. LSnonEmpty'
           BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
        <4>60. LeaseGrant'
           BY <3>1, <4>0, <4>2 DEF LeaseGrant, Inv, TypeInvariant, NodesReady
        <4>61. GrantReady'
           BY <3>1, <4>0, <4>2 DEF GrantReady, Inv, TypeInvariant, NodesReady
        <4>62. SemJoinNeighbor'
           <5>3. SUFFICES ASSUME NEW ls \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <4>4, <5>3 DEF SemJoinNeighbor, Inv
        <4>63. SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady = {mm.mreq.lset.node}
              BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <4>4, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
            <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>8. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>9. QED BY <4>4, <5>3, <5>6, <5>8, <4>1, <4>2, 
                SlowestZenon DEF SemSingleReadyJoin, Inv
        <4>64. SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY <4>1 DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <4>4, <5>3, <4>1, <4>2 
                DEF SemEmptyLSetSingleNode, Inv
        <4>65. SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY <4>1 DEF SemJoinLeafSet, Inv
            <5>4. mm \in receivedMsgs
              BY <4>4, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
              BY <4>1, <5>6
           <5>9. QED BY <4>4, <5>3, <5>6, <5>7, <4>1, SimplifyAndSolve
                DEF SemJoinLeafSet, Inv
        <4>66. SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE
                     /\ lset'[mm.destination] = EmptyLS(mm.destination)
                     /\ toj'[mm.mreq.lset.node] = mm.destination
                     /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                         => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                     /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                        => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                     /\ mm.mreq.lset.node \in NodesReady
                     /\ mm.destination \in NodesWait
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
              BY <4>1, <4>2 DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <4>4, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.destination \in I
              BY <5>4 DEF TypeInvariant, DMsg
           <5>8. /\ mm.mreq.lset.node \in I
                 /\ LeftNeighbor(mm.mreq.lset) \in I
                 /\ RightNeighbor(mm.mreq.lset) \in I          
              BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
           <5>9. QED BY <4>4,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
        
        <4>67. GrantNeighbor'
           \* idea: when kk = i, it grants ii when it is the left or right neighbor
           <5>1. \A k \in I\{i}: grant'[k] = grant[k]
              BY <3>1 DEF TypeInvariant
           <5>2. SUFFICES ASSUME 
                    NEW kk \in NodesReady, 
                    NEW ii \in I,
                    ~ii \in NodesReady,
                    ii \in grant'[kk],
                    ii # LeftNeighbor(lset[kk]),
                    ii # RightNeighbor(lset[kk])
                 PROVE FALSE
              BY <4>1, <4>2, ReadyInI DEF GrantNeighbor, Inv
           <5>5. CASE kk = i
              <6>3. CASE ii \in grant[i]
                 BY <6>3, <5>5, <5>1, <5>2, ReadyInI DEF GrantNeighbor, Inv
              <6>4. CASE ~ii \in grant[i]
                 BY <6>4, <5>5,<5>2, <3>1 DEF TypeInvariant
              <6>9. QED BY <6>3,<6>4
           <5>6. CASE kk # i 
              BY <5>1, <5>2, <5>6, ReadyInI DEF GrantNeighbor, Inv  
           <5>9. QED BY <5>5, <5>6
        <4>68. SingleReadyNode'
           BY <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
        <4>69. CoverToJ'
           BY <4>1, <4>2, ReadyInI DEF CoverToJ, Inv
        <4>70. NoCCJoin'
           BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve DEF NoCCJoin, Inv
        <4>71. NoCCJoinL' 
           BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoinL, SimplifyAndSolve DEF NoCCJoinL, Inv
        <4>72. SemPRplyNonEmpty'
           <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 \/ mm.mreq.type = "LSProbeReply"
                 \/ mm.mreq.type = "LSProbe"
                 PROVE 
                    lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
              BY <4>1 DEF SemPRplyNonEmpty, Inv
           <5>4. mm \in receivedMsgs
              BY <4>4, <5>2
           <5>5. CASE mm.mreq.type = "LSProbeReply"
               <6>5. mm.mreq \in PRpl
                  BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <6>6. mm.mreq.node \in I
                  BY <6>5 DEF PRpl
               <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                  BY <4>1, <6>6
               <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                    DEF SemPRplyNonEmpty, Inv
           <5>6. CASE mm.mreq.type = "LSProbe"
              <6>5. mm.mreq \in Prb
                  BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <6>6. mm.mreq.node \in I
                  BY <6>5 DEF Prb, LSet
               <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                  BY <4>1, <6>6
               <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                    DEF SemPRplyNonEmpty, Inv
           <5>9. QED BY <5>2,<5>5,<5>6
        <4>73. SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY <4>1 DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <4>4, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. QED BY <4>4, <5>3, <5>8, <4>1
                DEF SemJRplyLSnonEmpty, Inv
        <4>74. OneJRply'
           BY <4>4 DEF OneJRply, Inv
        <4>75. OneJReq'
           BY <4>4 DEF OneJReq, Inv
        <4>76. SemJReq'
           <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinRequest"
                 PROVE  /\ mm.mreq.node \in NodesWait'
                        /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY DEF SemJReq, Inv
           <5>2. mm \in receivedMsgs
              BY <4>4, <5>1
           <5>3. mm.mreq \in JReq
              BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>4. mm.mreq.node \in I
              BY <5>3 DEF JReq
           <5>9. QED BY <5>2, <4>1, <4>2, <5>1, ReadyInI DEF SemJReq, Inv
        <4>77. SingleReadyNodeToj'
           BY <3>1, <4>1, <4>2, ReadyInI DEF SingleReadyNodeToj, Inv
        <4>78. JoinMsgDual'
           BY <4>4 DEF JoinMsgDual, Inv
        <4>80. AtleastOneReady'
           BY <4>2 DEF AtleastOneReady, Inv
        <4>9. QED BY <4>3,<4>5,\*<4>6,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76,<4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <1>9, <3>1 DEF ReceiveLReq
   <1>10. ASSUME NEW i \in I,
                ReceiveBLS(i),
                TypeInvariant,
                Inv
         PROVE  Inv'
      <3>1. ASSUME NEW m \in receivedMsgs,
            m.destination = i,
            m.mreq.type = "BroadcastLSet",
             status[i] \in {"ok", "ready"},
             m.mreq.lset.node # i,
             lset[i]#EmptyLS(i),
             IsComplete(lset[i]),
             m.mreq.lset.node \in {LeftNeighbor(lset[i]), RightNeighbor(lset[i])},
            NEW ls, 
            ls = lset[i],
            NEW ln, ln = LeftNeighbor(ls),
            NEW rn, rn = RightNeighbor(ls),
            NEW newlease, 
            newlease = [lease EXCEPT ![i] = 
                       IF m.mreq.grant = TRUE
                       THEN @ \cup {m.mreq.lset.node}
                       ELSE @],
            NEW okToReady, 
            okToReady =       /\ ln \in newlease[i] 
                              /\ rn \in newlease[i]
                              /\ status[i]="ok",
            
            lease' = newlease,
            status' = [status EXCEPT ![i] =   IF okToReady
                                              THEN "ready"
                                              ELSE @ ],
            receivedMsgs' = (receivedMsgs \ {m}),
            toj'= [toj EXCEPT ![i]= IF toj[i] = m.mreq.lset.node 
                               THEN i
                               ELSE @],
            UNCHANGED <<rtable, probing, failed, grant, lset>>,
                    TypeInvariant,
            Inv         
         PROVE  Inv'
         <4>1. m.mreq \in BLS
            BY SlowestZenon, <3>1 DEF TypeInvariant, DMsg, MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
         <4>2. m.mreq.lset \in LSet
            BY <4>1 DEF BLS
         <4>3. GetLSetContent(m.mreq.lset)\ failed[i] \in SUBSET I
            BY <4>1, <4>2, GetLSetContentType DEF TypeInvariant
         <4>4. \A k \in I: lset[k] \in LSet
            BY TypeInvariant DEF TypeInvariant
         <4>5. ls.node = i /\ lset[i].node = i
            BY ls = lset[i] DEF TypeInvariant
         <4>6. \A k \in I: lset'[k].node = k
            BY <4>5, UNCHANGED <<rtable, probing, failed, grant, lset>> DEF TypeInvariant
         <4>7. ls \in LSet
            BY <3>1, <4>1, <4>2, AddToLSetType, GetLSetContentType DEF TypeInvariant
        <4>8.  /\ ls \in LSet
               /\ lset[i] \in LSet
           BY <4>4, <4>7 DEF TypeInvariant
        <4>9. lset[i]'= ls
           BY <3>1 DEF TypeInvariant
        <4>10. ln # i
           BY AddToLSetNotEmpty, 
              CompleteHasNeighbor,
              IsComplete(lset[i]),
              ln = LeftNeighbor(ls), 
              ls = lset[i],
              GetLSetContent(m.mreq.lset) \ failed[i] \in SUBSET I,
              <4>5, <4>8
        <4>101.rn # i
            BY AddToLSetNotEmpty, 
              CompleteHasNeighbor,
              IsComplete(lset[i]),
              rn = RightNeighbor(ls), 
              ls = lset[i],
              GetLSetContent(m.mreq.lset) \ failed[i] \in SUBSET I,
              <4>5, <4>8
        <4>11. \A k \in I: ~k \in (lset[k].left) /\ ~k \in (lset[k].right)
          BY LSetInv DEF LSetInv, Inv
        <4>12. ln \in I
           BY  ln = LeftNeighbor(ls), 
              ls = lset[i],
              m.mreq.lset \in LSet,
              GetLSetContentType, 
              TypeInvariant,
              AddToLSetType,
              LeftNeighborType
           DEF TypeInvariant
        <4>13. rn \in I
           BY  rn = RightNeighbor(ls), 
              ls = lset[i],
              m.mreq.lset \in LSet,
              GetLSetContentType, 
              TypeInvariant,
              AddToLSetType,
              RightNeighborType
           DEF TypeInvariant
        <4>14. /\ ~ i \in (ls.left) 
             /\ ~ i  \in (ls.right)
          BY <4>3, <4>4,<4>5, <4>11, NotAddSelf, <3>1
        <4>16. lset[i]' = ls /\ ls.node = i
          BY <3>1, <4>5, TypeInvariant DEF TypeInvariant
        <4>17. /\ ~ i \in (lset[i]'.left) 
             /\ ~ i \in (lset[i]'.right)
          BY <4>14, <4>16
        <4>18. \A k \in I \{i}: 
             /\ ~k \in (lset'[k].left) 
             /\ ~k \in (lset'[k].right) 
          BY <4>11, <3>1, TypeInvariant DEF TypeInvariant
        <4>19. \A k \in I:
               /\lset[k]' = lset[k]
               /\grant'[k] = grant[k]
           BY UNCHANGED <<rtable, probing, failed, grant, lset>> DEF TypeInvariant
        <4>20. LSetInv'
        BY <4>18,<4>17 DEF LSetInv
        \* \A i \in I: ~i \in (lset[i].left) /\ ~i \in (lset[i].right)
        <4>21.\A k \in I: lset'[k] \in LSet /\ lset[k] \in LSet
           BY <3>1, <4>8 DEF TypeInvariant 
        <4>22. i \in ReadyOK 
           BY <3>1 DEF NodesReady, NodesOK, ReadyOK
        <4>23./\ ReadyOK' = ReadyOK 
              /\ NodesReady' \in SUBSET ReadyOK'
              /\ NonDead'= NonDead
              /\ NodesReady \in SUBSET NodesReady'
              /\ NodesReady'\{i} = NodesReady\{i}
              /\ NodesWait'= NodesWait
              /\ \A k \in NodesReady: k \in NodesReady'
           BY <3>1, ReadyInI DEF TypeInvariant, NodesReady, NodesOK, ReadyOK, NodesWait, NonDead

        <4>25. newlease[i] = 
                          IF
                          /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                          THEN lease[i]\cup {m.mreq.lset.node}
                          ELSE lease[i]
           BY <3>1 DEF TypeInvariant
        <4>26. \A k \in lease[i]: 
               /\ k \in newlease[i]
               /\ k \in lease'[i]
           BY <4>25, <3>1 DEF TypeInvariant 
        <4>27. \A a, b \in I: a \in lease[b] => a \in lease'[b]
           BY <4>26, <3>1 DEF TypeInvariant
        <4>28. \A k \in I\{i}: toj'[k] = toj[k]
           BY <3>1 DEF TypeInvariant
        <4>44. \A ms \in receivedMsgs': ms \in receivedMsgs
           BY <3>1

  
        <4>95. HalfNeighbor' 
              <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                 <7>1. \A k \in ReadyOK' \ {i}: RightNeighbor(lset'[k]) # k /\ LeftNeighbor(lset'[k]) # k
                    BY <4>23, ReadyInI, <3>1, <5>7 DEF TypeInvariant
                <7>2. RightNeighbor(lset[i]) # lset[i].node /\ LeftNeighbor(lset[i]) #  lset[i].node
                   BY <5>7, <3>1, <4>5 DEF NodesReady, NodesOK, ReadyOK 
                <7>3. RightNeighbor(lset[i]') # lset[i].node /\ LeftNeighbor(lset[i]') #lset[i].node 
                   BY AddToLSetNotEmpty,  <4>3, <4>8, <7>2, <4>22, <4>16, <3>1
                <7>4. RightNeighbor(lset[i]') # i /\ LeftNeighbor(lset[i]') #i 
                   BY <7>3, <4>5
                <7>9. QED BY <7>1, <7>4, HalfNeighbor DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: /\ mk.mreq.type = "LSProbe" 
                                                     /\ mk.destination # k
                         /\ ~\E mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               /\ GetLSetContent(mj.mreq.lset) # {k}
                         /\~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                  <6>1. PICK k \in ReadyOK: 
                         /\ ReadyOK = {k} 
                         /\ LeftNeighbor(lset[k]) = k 
                         /\ RightNeighbor(lset[k]) = k
                     BY <5>8
                  <6>2. k = i
                     BY <3>1, <6>1, ReadyInI DEF NodesOK, ReadyOK, NodesReady
                  <6>3. lset[k].node = k /\ lset[k] \in LSet /\ k \in I
                     BY <6>1, ReadyInI DEF TypeInvariant
                  <6>4. lset[k] = EmptyLS(k)
                     BY <6>1, <6>3, EmptyLSNoNeighbor
                  <6>9. QED BY <6>2,<6>4,<3>1 \*contradiction
               <5>9. QED BY <5>7, <5>8, HalfNeighbor DEF HalfNeighbor, Inv
 
        \*<4>52. IncludeNeighbor'
        \*<4>53. CrossNeighbor'
        <4>54. CompleteLeafSet'
           <5>0. CASE okToReady
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>0, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1, <6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12 
                              DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1 
                          DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>40. \A k \in NodesReady:
                    ~(/\ k # i
                      /\ lset'[i] # EmptyLS(i)
                      /\ CwDist(i, k) < CwDist(i, RightNeighbor(lset[i]'))
                     )
                 <7>0. lset[i] # EmptyLS(i)
                    BY <3>1
                 <7>1. SUFFICES ASSUME NEW k \in NodesReady,
                       k # i
                       PROVE ~ CwDist(i, k) < CwDist(i, RightNeighbor(lset'[i]))
                    OBVIOUS
                 <7>2. ~ CwDist(i, k)< CwDist(i, RightNeighbor(lset[i]))
                    BY <7>0, <7>1, RightNeighborType, <4>8 DEF CompleteLeafSet, Inv
                 <7>3. CwDist(i, RightNeighbor(lset[i]))=< CwDist(i, k)
                    BY <7>2, RightNeighborType, <4>8, <7>1, ReadyInI, IisNat, CwDistType, Unequality2
                 \*<7>4. CwDist(i, RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i])))=< CwDist(i, RightNeighbor(lset[i]))
                 \*   BY <3>1, AddToLSetInvCo, <4>5, <4>8, <4>3
                 \*<7>5. CwDist(i, RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i]))) =< CwDist(i, k)
                 \*   BY <7>4, <7>3, IisNat, CwDistTransLEQ,  
                 \*   RightNeighborType, <4>21, ReadyInI, <7>1, <3>1, <4>1, <4>2, AddToLSetType, GetLSetContentType DEF TypeInvariant
                 <7>6. CwDist(i, RightNeighbor(lset[i]')) =< CwDist(i, k)
                    BY <7>3, <4>19
                 <7>7. QED BY <7>6, RightNeighborType, <4>21, IisNat, CwDistType, Unequality2, <7>1, ReadyInI
              <6>4. SUFFICES ASSUME NEW k \in I,
                    k # i,
                    NEW r \in I,
                    RightNeighbor(lset[k]) = r,
                    lset[k] # EmptyLS(k),
                    CwDist(k, i) < CwDist(k, r)
                    PROVE FALSE
                 BY <4>4,<4>19, <6>0, <6>40, RightNeighborType, 
                    CompleteLeafSet DEF CompleteLeafSet, Inv
              <6>5. k # ln
                 \*BY <6>1, <6>4
                 <7>1. SUFFICES ASSUME k = ln
                       PROVE FALSE
                    OBVIOUS
                 <7>2. i = RightNeighbor(lset[k])
                    BY <7>1, <6>1, <6>31
                 <7>3. CwDist(k, i) < CwDist(k, i)
                    BY <7>2, <6>4
                 <7>4. QED BY <7>3
              <6>6. ln # i
                 BY <6>1, <6>3
              <6>7. CwDist(ln, i) =< CwDist(ln, k)
                 BY <6>4, <6>3, <6>1,<6>31, <6>5
                   DEF ReadyNeighborRight, Inv
              <6>8. CwDist(k, ln) < CwDist(k, i)
                 BY <6>5, <6>7,<6>4,<6>6,Ring3NodesCo, 
                    <6>3, ReadyInI, IisNat
              <6>9. CwDist(k, ln) < CwDist(k, RightNeighbor(lset[k]))
                 BY <6>8, <6>4,<6>5,<6>6, CwDistTransLess,
                    RightNeighborType, <6>3, ReadyInI, <4>4,IisNat 
              <6>10. QED BY <6>9, <6>4,<6>5, <6>6, <6>3,
                     RightNeighborType, <4>4,
                     CompleteLeafSet DEF CompleteLeafSet, Inv
           <5>1. CASE ~okToReady
              <6>0. /\ RightNeighbor(lset[i]) \in I
                    /\ lset[i] \in LSet
                    /\ rn \in I
                    /\ rn = RightNeighbor(ls)
                    \*/\ ls = AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i])
                    /\ ls = lset[i]
                    /\ GetLSetContent(m.mreq.lset)\ failed[i] \in SUBSET I
                    /\ lset[i].node = i
                    /\ ls \in LSet
                    /\ ls.node = i
                 BY <4>8,<3>1,<4>5, <4>3,RightNeighborType
              <6>1. status' = status
                 BY <3>1, <5>1 DEF TypeInvariant
              <6>2. NodesReady' = NodesReady
                 BY <6>1 DEF NodesReady
              <6>3. lset'[i] # EmptyLS(i)
                 BY AddToLSetComplete, <3>1, <4>3,<4>8,<4>9, CompleteNotEmpty
              <6>4. RightNeighbor(lset'[i]) = rn
                 BY <3>1, <4>9
              \*<6>5. CwDist(i, rn) =< CwDist(i, RightNeighbor(lset[i])) 
              \*   BY AddToLSetInvCo, <6>0
              <6>5. CwDist(i, rn) = CwDist(i, RightNeighbor(lset[i]))
                 BY <6>0
              <6>6. SUFFICES ASSUME NEW j \in NodesReady,
                    i # j,
                    CwDist(i, j) < CwDist(i, rn)
                    PROVE FALSE
                 BY <6>2, <6>3, <6>4, <4>19, CompleteLeafSet DEF CompleteLeafSet, Inv
              <6>7. CwDist(i, j) < CwDist(i, RightNeighbor(lset[i]))
                 BY <6>0, ReadyInI, <6>6, <6>5, CwDistType, IisNat, TransLESS
              <6>9. QED BY <3>1, <6>1,<6>6, <6>7, <6>0, 
                    CompleteLeafSet DEF CompleteLeafSet, Inv
           <5>9. QED BY <5>0, <5>1
        <4>55. CompleteLeafSetL'
           \*OMITTED \* similar to <4>54
           <5>0. CASE okToReady
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>0, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1, <6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12 
                              DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1 
                          DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>40. \A k \in NodesReady:
                    ~(/\ k # i
                      /\ lset'[i] # EmptyLS(i)
                      /\ CwDist(k, i) < CwDist(LeftNeighbor(lset[i]'), i)
                     )
                 <7>0. lset[i] # EmptyLS(i)
                    BY <3>1
                 <7>1. SUFFICES ASSUME NEW k \in NodesReady,
                       k # i
                       PROVE ~ CwDist(k, i) < CwDist(LeftNeighbor(lset'[i]), i)
                    OBVIOUS
                 <7>2. ~ CwDist(k, i)< CwDist(LeftNeighbor(lset[i]), i)
                    BY <7>0, <7>1, LeftNeighborType, <4>8 DEF CompleteLeafSetL, Inv
                 <7>3. CwDist(LeftNeighbor(lset[i]), i)=< CwDist(k, i)
                    BY <7>2, LeftNeighborType, <4>8, <7>1, ReadyInI, IisNat, CwDistType, Unequality2
                 \*<7>4. CwDist(i, RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i])))=< CwDist(i, RightNeighbor(lset[i]))
                 \*   BY <3>1, AddToLSetInvCo, <4>5, <4>8, <4>3
                 \*<7>5. CwDist(i, RightNeighbor(AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i]))) =< CwDist(i, k)
                 \*   BY <7>4, <7>3, IisNat, CwDistTransLEQ,  
                 \*   RightNeighborType, <4>21, ReadyInI, <7>1, <3>1, <4>1, <4>2, AddToLSetType, GetLSetContentType DEF TypeInvariant
                 <7>6. CwDist(LeftNeighbor(lset[i]'), i) =< CwDist(k, i)
                    BY <7>3, <4>19
                 <7>7. QED BY <7>6, LeftNeighborType, <4>21, IisNat, CwDistType, Unequality2, <7>1, ReadyInI
              <6>4. SUFFICES ASSUME NEW k \in I,
                    k # i,
                    NEW l \in I,
                    LeftNeighbor(lset[k]) = l,
                    lset[k] # EmptyLS(k),
                    CwDist(i, k) < CwDist(l, k)
                    PROVE FALSE
                 BY <4>4,<4>19, <6>0, <6>40, LeftNeighborType
                     DEF CompleteLeafSetL, Inv
              <6>5. k # rn
                 \*BY <6>1, <6>4
                 <7>1. SUFFICES ASSUME k = rn
                       PROVE FALSE
                    OBVIOUS
                 <7>2. i = LeftNeighbor(lset[k])
                    BY <7>1, <6>1, <6>31
                 <7>3. CwDist(i, k) < CwDist(i, k)
                    BY <7>2, <6>4
                 <7>4. QED BY <7>3
              <6>6. rn # i
                 BY <6>1, <6>3
              <6>7. CwDist(i, rn) =< CwDist(k, rn)
                 BY <6>4, <6>3, <6>1,<6>31, <6>5
                   DEF ReadyNeighborLeft, Inv
              <6>8. CwDist(rn, k) < CwDist(i, k)
                 BY <6>5, <6>7,<6>4,<6>6,Ring3NodesPos11, 
                    <6>3, ReadyInI, IisNat
              <6>9. CwDist(rn, k) < CwDist(LeftNeighbor(lset[k]), k)
                 BY <6>8, <6>4, TransLESS, CwDistType,
                    LeftNeighborType, <6>3, ReadyInI, <4>4,IisNat 
              <6>10. QED BY <6>9, <6>4,<6>5, <6>6, <6>3,
                     LeftNeighborType, <4>4
                      DEF CompleteLeafSetL, Inv
           <5>1. CASE ~okToReady
              <6>0. /\ LeftNeighbor(lset[i]) \in I
                    /\ lset[i] \in LSet
                    /\ ln \in I
                    /\ ln = LeftNeighbor(ls)
                    \*/\ ls = AddToLSet(GetLSetContent(m.mreq.lset)\ failed[i] , lset[i])
                    /\ ls = lset[i]
                    /\ GetLSetContent(m.mreq.lset)\ failed[i] \in SUBSET I
                    /\ lset[i].node = i
                    /\ ls \in LSet
                    /\ ls.node = i
                 BY <4>8,<3>1,<4>5, <4>3,LeftNeighborType
              <6>1. status' = status
                 BY <3>1, <5>1 DEF TypeInvariant
              <6>2. NodesReady' = NodesReady
                 BY <6>1 DEF NodesReady
              <6>3. lset'[i] # EmptyLS(i)
                 BY AddToLSetComplete, <3>1, <4>3,<4>8,<4>9, CompleteNotEmpty
              <6>4. LeftNeighbor(lset'[i]) = ln
                 BY <3>1, <4>9
              \*<6>5. CwDist(i, rn) =< CwDist(i, RightNeighbor(lset[i])) 
              \*   BY AddToLSetInvCo, <6>0
              <6>5. CwDist(ln, i) = CwDist(LeftNeighbor(lset[i]), i)
                 BY <6>0
              <6>6. SUFFICES ASSUME NEW j \in NodesReady,
                    i # j,
                    CwDist(j, i) < CwDist(ln, i)
                    PROVE FALSE
                 BY <6>2, <6>3, <6>4, <4>19 DEF CompleteLeafSetL, Inv
              <6>7. CwDist(j, i) < CwDist(LeftNeighbor(lset[i]), i)
                 BY <6>0, ReadyInI, <6>6, <6>5, CwDistType, IisNat, TransLESS
              <6>9. QED BY <3>1, <6>1,<6>6, <6>7, <6>0, 
                    CompleteLeafSetL DEF CompleteLeafSetL, Inv
           <5>9. QED BY <5>0, <5>1           
        <4>56. ReadyNeighborRight'
           <5>0. /\ RightNeighbor(lset'[i]) = rn
                 /\ \A k \in I\{i}: lset'[k] = lset[k] 
                 /\ \A k \in I: lset[k] \in LSet
                 /\ \A k \in I: lset'[k] \in LSet
              BY <3>1, <4>19, <4>21
           <5>1. CASE okToReady
              <6>0. lset[i] # EmptyLS(i)
                 BY <3>1
              <6>1.\* same as completeleafset-<5>0 
                  /\ NodesReady= NodesReady' \ {i} 
                  /\ rn # i
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  /\  \A k \in NodesReady: lset'[k] = lset[k] 
                 <7>1.  /\ ~i \in NodesReady
                        /\ i \in NodesReady'
                        /\ NodesReady = NodesReady'\{i} 
                 BY <5>1, <3>1 DEF NodesReady, NodesOK, TypeInvariant
                 <7>2.    /\ ln \in newlease[i] 
                          /\ rn \in newlease[i]
                          /\ ln = LeftNeighbor(lset[i])
                          /\ rn = RightNeighbor(lset[i])
                          
                  BY <5>1, <3>1 DEF NodesReady, NodesOK
                 <7>3. \A k \in NodesReady: lset'[k] = lset[k] 
                    BY ReadyInI, <7>1, <4>19
                 <7>4. QED BY <4>101, <7>1,<7>2,<7>3
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1 \* same as completeleafset-<5>0 
              <6>3. rn \in NodesReady
              \* same as completeleafset-<5>0 
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>5. \A r \in NodesReady, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, k)
                 BY <5>0, <6>1, ReadyInI DEF ReadyNeighborRight, Inv
              <6>6. \A r \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, i)
                 <7>1. SUFFICES ASSUME NEW r \in NodesReady,
                       lset'[i] # EmptyLS(i),
                       r # i,
                       ~RightNeighbor(lset[r]) \in NodesReady
                       PROVE CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                    BY <6>1
                 <7>8. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                    BY <6>0,<7>1 DEF ReadyNeighborRight, Inv
                 <7>9. QED BY <7>8
              <6>7. \A k \in I: 
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~RightNeighbor(lset'[i]) \in NodesReady' 
                  => CwDist(i, RightNeighbor(lset'[i])) =< CwDist(i, k)
                 BY <6>3, <6>1, <5>0
              <6>9. QED BY <6>5, <6>6,<6>7,<6>1 DEF ReadyNeighborRight
           <5>2. CASE ~okToReady
              <6>1. /\ \A k \in I: lset'[k] =  lset[k] 
                    /\ lset[i] # EmptyLS(i)
                    /\ lset'[i] \in LSet
                 BY <4>19, <3>1, <4>21
              <6>2. NodesReady' = NodesReady
                 BY <3>1, <5>2 DEF NodesReady, TypeInvariant
              <6>5. CASE i \in NodesReady
                 BY <6>1,<6>2, <6>5, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
              <6>6. CASE ~i \in NodesReady
                 <7>1. \A k \in NodesReady: lset'[k] = lset[k]
                    BY <6>1, ReadyInI, <6>6  
                 <7>2.\A r \in NodesReady, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~RightNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(r, RightNeighbor(lset'[r])) =< CwDist(r, k)
                 BY <6>6, <6>2,<7>2, <6>1, <7>1, ReadyInI, ReadyNeighborRight 
                    DEF ReadyNeighborRight, Inv
                 <7>7. SUFFICES ASSUME NEW r \in NodesReady,
                       lset'[i] # EmptyLS(i),
                       r # i,
                       ~RightNeighbor(lset[r]) \in NodesReady
                       PROVE CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                    BY <6>6, <6>2,<7>2, <6>1,<7>1, ReadyNeighborRight 
                       DEF ReadyNeighborRight, Inv
                 <7>8. CwDist(r, RightNeighbor(lset[r])) =< CwDist(r, i)
                    BY <6>1,<7>7,ReadyNeighborRight DEF ReadyNeighborRight, Inv
                 <7>9. QED BY <7>8  
              <6>9. QED BY <6>5,<6>6
           <5>9. QED BY <5>1,<5>2
        <4>57. ReadyNeighborLeft'
           \* OMITTED \*similar to <4>56 
           <5>0. /\ LeftNeighbor(lset'[i]) = ln
                 /\ \A k \in I\{i}: lset'[k] = lset[k] 
                 /\ \A k \in I: lset[k] \in LSet
                 /\ \A k \in I: lset'[k] \in LSet
              BY <3>1, <4>19, <4>21
           <5>1. CASE okToReady
              <6>0. lset[i] # EmptyLS(i)
                 BY <3>1
              <6>1.\* same as completeleafset-<5>0 
                  /\ NodesReady= NodesReady' \ {i} 
                  /\ ln # i
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  /\  \A k \in NodesReady: lset'[k] = lset[k] 
                 <7>1.  /\ ~i \in NodesReady
                        /\ i \in NodesReady'
                        /\ NodesReady = NodesReady'\{i} 
                 BY <5>1, <3>1 DEF NodesReady, NodesOK, TypeInvariant
                 <7>2.    /\ ln \in newlease[i] 
                          /\ rn \in newlease[i]
                          /\ ln = LeftNeighbor(lset[i])
                          /\ rn = RightNeighbor(lset[i])
                          
                  BY <5>1, <3>1 DEF NodesReady, NodesOK
                 <7>3. \A k \in NodesReady: lset'[k] = lset[k] 
                    BY ReadyInI, <7>1, <4>19
                 <7>4. QED BY <4>10, <7>1,<7>2,<7>3
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1 \* same as completeleafset-<5>0 
              <6>3. ln \in NodesReady
              \* same as completeleafset-<5>0 
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>5. \A r \in NodesReady, k \in I\{i}:
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~LeftNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(k, r)
                 BY <5>0, <6>1, ReadyInI DEF ReadyNeighborLeft, Inv
              <6>6. \A r \in NodesReady:
                       /\ lset'[i] # EmptyLS(i)
                       /\ i # r
                       /\~LeftNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(i, r)
                 <7>1. SUFFICES ASSUME NEW r \in NodesReady,
                       lset'[i] # EmptyLS(i),
                       r # i,
                       ~LeftNeighbor(lset[r]) \in NodesReady
                       PROVE CwDist(LeftNeighbor(lset[r]), r) =< CwDist(i, r)
                    BY <6>1
                 <7>8. CwDist(LeftNeighbor(lset[r]), r) =< CwDist(i, r)
                    BY <6>0,<7>1 DEF ReadyNeighborLeft, Inv
                 <7>9. QED BY <7>8
              <6>7. \A k \in I: 
                       /\ lset'[k] # EmptyLS(k)
                       /\ k # i
                       /\~LeftNeighbor(lset'[i]) \in NodesReady' 
                  => CwDist(LeftNeighbor(lset'[i]), i) =< CwDist(k, i)
                 BY <6>3, <6>1, <5>0
              <6>9. QED BY <6>5, <6>6,<6>7,<6>1 DEF ReadyNeighborLeft
           <5>2. CASE ~okToReady
              <6>1. /\ \A k \in I: lset'[k] =  lset[k] 
                    /\ lset[i] # EmptyLS(i)
                    /\ lset'[i] \in LSet
                 BY <4>19, <3>1, <4>21
              <6>2. NodesReady' = NodesReady
                 BY <3>1, <5>2 DEF NodesReady, TypeInvariant
              <6>5. CASE i \in NodesReady
                 BY <6>1,<6>2, <6>5, ReadyInI DEF ReadyNeighborLeft, Inv
              <6>6. CASE ~i \in NodesReady
                 <7>1. \A k \in NodesReady: lset'[k] = lset[k]
                    BY <6>1, ReadyInI, <6>6  
                 <7>2.\A r \in NodesReady, k \in I\{i}:
                       /\lset'[k] # EmptyLS(k)
                       /\ k # r
                       /\~LeftNeighbor(lset'[r]) \in NodesReady 
                  => CwDist(LeftNeighbor(lset'[r]), r) =< CwDist(k, r)
                 BY <6>6, <6>2,<7>2, <6>1, <7>1, ReadyInI, ReadyNeighborLeft
                    DEF ReadyNeighborLeft, Inv
                 <7>7. SUFFICES ASSUME NEW r \in NodesReady,
                       lset'[i] # EmptyLS(i),
                       r # i,
                       ~LeftNeighbor(lset[r]) \in NodesReady
                       PROVE CwDist(LeftNeighbor(lset[r]), r) =< CwDist(i, r)
                    BY <6>6, <6>2,<7>2, <6>1,<7>1, ReadyNeighborLeft
                       DEF ReadyNeighborLeft, Inv
                 <7>8. CwDist(LeftNeighbor(lset[r]), r) =< CwDist(i, r)
                    BY <6>1,<7>7 DEF ReadyNeighborLeft, Inv
                 <7>9. QED BY <7>8  
              <6>9. QED BY <6>5,<6>6
           <5>9. QED BY <5>1,<5>2           
        <4>58. SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                 NEW nn \in I, 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "BroadcastLSet",
                 mm.mreq.lset.node = nn,
                 mm.destination = ii
                PROVE
                    /\ nn \in NodesReady'
                    /\ (mm.mreq.grant =TRUE => ii \in grant[nn])
              BY <4>19 DEF SemLRply, Inv
           <5>9. QED BY <4>44, <4>19, <4>23, <5>2 DEF SemLRply, Inv
        <4>59. LSnonEmpty'
           <5>1. SUFFICES ASSUME NEW ii \in I, 
                   NEW jj \in I,
                   ii \in GetLSetContent(lset[jj]),
                   ii # jj,
                   ~ii \in NodesReady'
                 PROVE lset[ii] # EmptyLS(ii)
              BY <4>19 DEF LSnonEmpty, Inv
           <5>5. CASE ii = i
              BY <5>1,<5>5,<3>1
           <5>7. CASE ii # i 
              BY <5>1, <5>7, <4>23 DEF LSnonEmpty, Inv
           <5>9. QED BY <5>5,<5>7
        <4>60. LeaseGrant'
           <5>1. \A k \in I\{i}: lease'[k] = lease[k]
              BY <3>1 DEF TypeInvariant
           <5>2. lease'[i] = IF
                          /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                          THEN lease[i]\cup {m.mreq.lset.node}
                          ELSE lease[i]
              BY <3>1 DEF TypeInvariant
           <5>3. SUFFICES ASSUME NEW ii \in I, NEW nn \in I,
                 nn \in lease'[ii]
                 PROVE ii \in grant[nn]
              BY <4>19 DEF LeaseGrant, Inv
           <5>5. CASE /\ m.mreq.lset.node \in {ln, rn}
                      /\ m.mreq.grant = TRUE
              <6>1. m.mreq.lset.node \in I 
                 BY <4>2 DEF LSet
              <6>2. i \in grant[m.mreq.lset.node]
                 BY <3>1, <5>5, <6>1 DEF SemLRply, Inv
              <6>3. m.mreq.lset.node # i
                 BY <3>1
              <6>4. \A k \in lease[i]: i \in grant[k]
                 BY DEF LeaseGrant, Inv, TypeInvariant
              <6>5. CASE ii = i /\ nn # i
                 <7>1. nn \in lease[i]\cup {m.mreq.lset.node}
                    BY <6>5, <5>2, <5>5, <5>3
                 <7>2. CASE nn \in lease[i]
                    BY <6>4, <6>5, <7>2
                 <7>3. CASE nn = m.mreq.lset.node
                    BY <6>2, <6>5, <7>3
                 <7>9. QED BY <7>1,<7>2,<7>3
              <6>6. CASE ii # i 
                 BY <5>1, <5>3, <6>6 DEF LeaseGrant, Inv
              <6>8. CASE ii = i /\ nn = i
                 <7>7. i \in lease[i]
                    BY <5>2, <5>5, <5>3, <6>3, <6>8
                 <7>8. i \in grant[i]
                    BY <7>7 DEF LeaseGrant, Inv
                 <7>9. QED BY <7>8, <6>8
              <6>9. QED BY <6>5,<6>6,<6>8
           <5>6. CASE ~(/\ m.mreq.lset.node \in {ln, rn}
                        /\ m.mreq.grant = TRUE)
              <6>1. lease'[i] = lease[i]
                 BY <5>6, <5>2
              <6>9. QED BY <6>1, <5>1, <5>3,<4>19
                 DEF LeaseGrant, Inv
           <5>9. QED BY <5>5,<5>6
        <4>61. GrantReady'
           <5>1. \/ NodesReady' = NodesReady
                 \/ NodesReady' = NodesReady \cup {i}
              BY <3>1 DEF NodesReady, NodesOK, TypeInvariant
           <5>9. QED BY <5>1, <4>19 DEF GrantReady, Inv
        <4>62. SemJoinNeighbor'
           <5>3. SUFFICES ASSUME NEW ls1 \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls1 = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls1) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls1) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls1) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls1) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <4>44, <5>3 DEF SemJoinNeighbor, Inv
        <4>63. SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady' = {mm.mreq.lset.node}
              BY DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <4>44, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
            <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>8. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>9. lset'[mm.mreq.lset.node] = mm.mreq.lset
               BY <4>44, <5>3, <5>6, <5>8, <4>19, <4>23
                DEF SemSingleReadyJoin, Inv
           <5>10. ASSUME NodesReady' # NodesReady PROVE FALSE
              <6>1. /\ i \in NodesReady' 
                    /\ ~ i \in NodesReady
                    /\ mm.mreq.lset \in LSet
                 BY <3>1, <5>10, <5>6 DEF NodesReady, NodesOK, TypeInvariant
              <6>2. /\m.mreq.lset.node \in I 
                    /\m.mreq.lset.node \in NodesReady
                 BY <6>1, <3>1, <4>2 DEF SemLRply, Inv, LSet
              <6>3. /\ m.mreq.lset.node = mm.mreq.lset.node
                    /\ lset[m.mreq.lset.node] = mm.mreq.lset
                    /\ i # mm.mreq.lset.node
                 BY <6>2, <5>3, <5>4, <6>1 DEF SemSingleReadyJoin, Inv
              <6>4. CASE m.mreq.lset.node \in lease[i] /\ okToReady
                 <7>3. i \in grant[m.mreq.lset.node]
                    BY <6>4, <6>2 DEF LeaseGrant, Inv
                 <7>4. \/ i = LeftNeighbor(lset[m.mreq.lset.node]) 
                       \/ i = RightNeighbor(lset[m.mreq.lset.node])
                    BY <7>3, <6>3, <6>1, <6>2 DEF GrantNeighbor, Inv
                 <7>5.  mm.mreq.lset # EmptyLS(mm.mreq.lset.node)
                    BY <7>4, EmptyLSNoNeighbor, <6>3, <6>1
                 <7>9. QED BY  <7>5, <5>3, <6>3, <6>1
              <6>5. CASE ~m.mreq.lset.node \in lease[i] /\ okToReady
                 <7>1. m.mreq.lset.node \in newlease[i]
                    BY <6>5, <3>1
                 <7>2. m.mreq.grant = TRUE
                    BY <6>5, <7>1, <4>25
                 <7>3. i \in grant[m.mreq.lset.node]
                    BY <6>3, <3>1, <7>2, <5>8 DEF SemLRply, Inv
                 <7>4. \/ i = LeftNeighbor(lset[m.mreq.lset.node]) 
                       \/ i = RightNeighbor(lset[m.mreq.lset.node])
                    BY <7>3, <6>3, <6>1, <6>2 DEF GrantNeighbor, Inv
                 <7>5.  mm.mreq.lset # EmptyLS(mm.mreq.lset.node)
                    BY <7>4, EmptyLSNoNeighbor, <6>3, <6>1
                 <7>9. QED BY  <7>5, <5>3, <6>3, <6>1
              <6>6. CASE ~okToReady
                 BY <6>6, <5>10, <3>1 DEF TypeInvariant, NodesReady
              <6>9. QED BY <6>4,<6>5,<6>6
           <5>99. QED BY <5>9,<5>10, <4>44, <5>3, <5>6, <5>8, <4>19, <4>23
                DEF SemSingleReadyJoin, Inv
                
        <4>64. SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY <4>1 DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <4>44, <5>3, <4>19, <4>23 
                DEF SemEmptyLSetSingleNode, Inv
        <4>65. SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY <4>19 DEF SemJoinLeafSet, Inv
            <5>4. mm \in receivedMsgs
              BY <4>44, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
              BY <4>19, <5>6
           <5>9. QED BY <4>44, <5>3, <5>7, <4>19, SimplifyAndSolve
                DEF SemJoinLeafSet, Inv
        <4>66. SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE
                     /\ lset'[mm.destination] = EmptyLS(mm.destination)
                     /\ toj'[mm.mreq.lset.node] = mm.destination
                     /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                         => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                     /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                        => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                     /\ mm.mreq.lset.node \in NodesReady'
                     /\ mm.destination \in NodesWait'
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
              BY DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <4>44, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.destination \in I
              BY <5>4 DEF TypeInvariant, DMsg
           <5>8. /\ mm.mreq.lset.node \in I
                 /\ LeftNeighbor(mm.mreq.lset) \in I
                 /\ RightNeighbor(mm.mreq.lset) \in I          
                 /\ lset'[mm.mreq.lset.node] \in LSet
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in I
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in I
              BY <5>6, LeftNeighborType, RightNeighborType, <4>19 DEF LSet, TypeInvariant
           <5>9. /\ lset[mm.destination] = EmptyLS(mm.destination)
                 /\ RightNeighbor(lset[mm.mreq.lset.node]) \in lease[mm.mreq.lset.node]
                 /\ LeftNeighbor(lset[mm.mreq.lset.node]) \in lease[mm.mreq.lset.node]
                 /\ Covers(lset[mm.mreq.lset.node], mm.destination)
                 /\ toj[mm.mreq.lset.node] = mm.destination
                 /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                     => toj[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                 /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                    => toj[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                 /\ mm.mreq.lset.node \in NodesReady
                 /\ mm.destination \in NodesWait
              BY <5>3,<5>4 DEF SemJRply, Inv
           <5>10./\ lset'[mm.destination] = EmptyLS(mm.destination)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                 /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
              BY <5>7, <5>8, <5>9, <4>19, <4>27
           <5>11.  toj'[mm.mreq.lset.node] = mm.destination
              <6>3. CASE mm.mreq.lset.node = i
                 <7>1. toj[i] = mm.destination
                    BY <5>9, <6>3
                 <7>2. m.mreq.lset.node \in NodesReady
                    BY <3>1, <4>2 DEF SemLRply, Inv, LSet
                 <7>5. toj[i] # m.mreq.lset.node
                    BY <7>2, <5>7, <7>1, StatusDisjoint, <5>9
                 <7>6. toj'[i] = toj[i]
                    BY <3>1, <7>5 DEF TypeInvariant
                 <7>9. QED BY <7>6, <7>1, <6>3
              <6>4. CASE mm.mreq.lset.node # i
                 BY <4>28, <5>9, <6>4, <5>8
              <6>9. QED BY <6>3,<6>4
           <5>12. LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                         => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset)
              <6>3. SUFFICES ASSUME 
                    LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                    PROVE toj'[LeftNeighbor(mm.mreq.lset)] 
                           = LeftNeighbor(mm.mreq.lset)
                 OBVIOUS
              <6>4. CASE LeftNeighbor(mm.mreq.lset) = i
                 <7>1. toj[i] = i
                    BY <6>4, <5>9, <6>3
                 <7>2. toj[i]' = i
                    BY <3>1, <7>1 DEF TypeInvariant
                 <7>9. QED BY <7>2, <6>4
              <6>5. CASE LeftNeighbor(mm.mreq.lset) # i
                 BY  <4>28, <5>9, <6>5, <5>8, <6>3
              <6>9. QED BY <6>5,<6>4 
           <5>13. ASSUME RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                  PROVE toj'[RightNeighbor(mm.mreq.lset)] 
                          = RightNeighbor(mm.mreq.lset)
              <6>4. CASE RightNeighbor(mm.mreq.lset) = i
                 <7>1. toj[i] = i
                    BY <6>4, <5>9, <5>13
                 <7>2. toj[i]' = i
                    BY <3>1, <7>1 DEF TypeInvariant
                 <7>9. QED BY <7>2, <6>4
              <6>5. CASE RightNeighbor(mm.mreq.lset) # i
                 BY  <4>28, <5>9, <6>5, <5>8, <5>13
              <6>9. QED BY <6>5,<6>4 
           <5>14. mm.mreq.lset.node \in NodesReady'
              BY <4>23, <5>9, <5>8
           <5>15. mm.destination \in NodesWait'
              <6>1. mm.destination # i
                 BY <5>8, <5>9, <3>1
              <6>9. QED BY <6>1, <3>1, <5>8,<5>9, <4>23 DEF TypeInvariant, NodesReady 
           <5>99. QED BY <5>10, <5>11,<5>12,<5>13,<5>14,<5>15
        <4>67. GrantNeighbor'
           <5>2. SUFFICES ASSUME 
                    NEW kk \in NodesReady', 
                    NEW ii \in I,
                    ~ii \in NodesReady',
                    ii \in grant[kk],
                    ii # LeftNeighbor(lset[kk]),
                    ii # RightNeighbor(lset[kk])
                 PROVE FALSE
              BY <4>19, ReadyInI, <4>23 DEF GrantNeighbor, Inv
           <5>4. CASE kk = i /\ kk \in NodesReady
              BY <5>2, <5>4, <4>23, ReadyInI DEF GrantNeighbor, Inv
           <5>5. CASE kk = i /\ ~kk \in NodesReady
              BY <5>5, <5>2, ReadyInI DEF GrantReady, Inv
           <5>6. CASE ii = i
              <6>1. kk # i
                 BY <5>2, <5>6
              <6>2. ~i \in NodesReady
                 BY <5>2, <5>6, <4>23 
              <6>3. kk \in NodesReady
                 BY <6>1, <4>23
              <6>9. QED  BY <6>1,<6>2,<6>3, <5>2, <5>6,ReadyInI DEF GrantNeighbor, Inv
           <5>7. CASE kk # i /\ ii # i
              BY <5>2, <5>4, <4>23, <5>7, ReadyInI DEF GrantNeighbor, Inv
           <5>9. QED BY <5>4, <5>5,<5>6,<5>7
        <4>68. SingleReadyNode'
           <5>1. /\ m.mreq.lset.node \in I
                 /\ m.mreq.lset.node \in NodesReady
                 /\ m.mreq.lset.node # i
              BY <4>2, <3>1 DEF LSet, SemLRply, Inv
           <5>2. SUFFICES ASSUME NEW ii \in I,
                 NodesReady'= {ii}
                 PROVE LeftNeighbor(lset[ii]) = RightNeighbor(lset[ii])
              BY <4>19 DEF SingleReadyNode, Inv
           <5>4. \/ NodesReady' = NodesReady
                 \/ NodesReady' = NodesReady \cup {i}
              BY <3>1 DEF NodesReady, NodesOK, TypeInvariant
           <5>5. CASE okToReady
              <6>1. i \in NodesReady'
                 BY <5>5, <3>1 DEF NodesReady, TypeInvariant
              <6>2. m.mreq.lset.node \in NodesReady'
                 BY <5>1, <5>4
              <6>3. ~(Cardinality(NodesReady') =< 1)
                    BY <6>1, <6>2, <5>1, <5>4, ReadyInI, 
                    <5>2, CardinalityTwoNode
              <6>9. QED BY <6>3, CardinalityOneNodeExt, <5>2
           <5>6. CASE ~okToReady
              <6>1. NodesReady' = NodesReady
                 BY <5>6,  <3>1 DEF NodesReady, TypeInvariant
              <6>2. QED BY <6>1, <5>2 DEF SingleReadyNode, Inv
           <5>9. QED BY <5>5,<5>6
        <4>69. CoverToJ'     
           <5>0. SUFFICES ASSUME NEW rr \in NodesReady', NEW kk \in I
                          PROVE CoverToJ!(rr, kk)'
                 BY DEF CoverToJ
           <5>1. /\ \A k \in I: lset'[k] = lset[k]
                 /\ \A k \in I\{i}: toj'[k] = toj[k]
                 /\ \A k \in NodesReady': k # i => k \in NodesReady
                 /\ rr \in I
                 /\ \A k \in I: /\ lset'[k] \in LSet 
                                /\ lset[k] \in LSet
                                /\ lset[k].node = k
                 /\ lset[i].node = i
                 /\ lset[i] \in LSet
               <6>1. \A k \in NodesReady': k # i => k \in NodesReady
                  BY <3>1, ReadyInI DEF NodesReady, TypeInvariant
               <6>2. rr \in I
                  BY <6>1, <5>0, ReadyInI
               <6>3. /\ \A k \in I: lset'[k] = lset[k]
                     /\ \A k \in I: /\ lset'[k] \in LSet 
                                /\ lset[k] \in LSet
                                /\ lset[k].node = k
                     /\ lset[i].node = i
                     /\ lset[i] \in LSet
                  BY <3>1, <4>19, <4>21 DEF TypeInvariant
               <6>4. \A k \in I\{i}: toj'[k] = toj[k]
                  BY <3>1 DEF TypeInvariant
               <6>9. QED BY <6>1,<6>2,<6>3,<6>4
           <5>3. CASE rr = i /\ ~ i \in NodesReady /\ ~okToReady
              BY <5>0, <5>3, <3>1 DEF NodesReady, NodesOK, TypeInvariant
           <5>4. CASE rr = i /\ ~ i \in NodesReady /\ okToReady
              <6>0. SUFFICES ASSUME 
                    lset[kk] # EmptyLS(kk), 
                    Covers(lset[i], kk),
                    kk # i
                    PROVE FALSE
                 BY <5>1, ReadyInI, <5>4 DEF CoverToJ, Inv
              <6>1.\* same as completeleafset-<5>0 
                  /\ NodesReady= NodesReady' \ {i} 
                  /\ rn # i
                  /\ ln # i
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  /\  \A k \in NodesReady: lset'[k] = lset[k] 
                  /\ okToReady
                 <7>1.  /\ ~i \in NodesReady
                        /\ i \in NodesReady'
                        /\ NodesReady = NodesReady'\{i} 
                        /\ okToReady
                 BY <5>4, <3>1 DEF NodesReady, NodesOK, TypeInvariant
                 <7>2.    /\ ln \in newlease[i] 
                          /\ rn \in newlease[i]
                          /\ ln = LeftNeighbor(lset[i])
                          /\ rn = RightNeighbor(lset[i])
                          
                  BY <5>1, <3>1, <5>4 DEF NodesReady, NodesOK
                 <7>3. \A k \in NodesReady: lset'[k] = lset[k] 
                    BY ReadyInI, <7>1, <4>19
                 <7>4. QED BY <4>101,<4>10, <7>1,<7>2,<7>3
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1, <6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12,CompleteLeafSet DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12,CompleteLeafSetL DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>1,<6>3, <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>1,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>7. CASE CwDist(i, kk) < CwDist(i, rn)
                 <7>2. rn # kk
                    BY <6>7, <6>3, ReadyInI, RightDistInjectivity, <6>1
                 <7>3. CwDist(kk, rn) < CwDist(i, rn)
                    BY Ring3NodesPos6, <6>7, <6>0, <6>1, <7>2, <6>3, ReadyInI
                 <7>5. lset[kk] # EmptyLS(kk)
                    BY <6>0
                 <7>8. CwDist(i, rn) =< CwDist(kk, rn)
                    BY <6>1, <7>5, <7>2, <6>3, <6>31 DEF ReadyNeighborLeft, Inv
                 <7>9. QED BY <7>8, <7>3, Unequality2, <6>3, <5>1, CwDistType, IisNat, ReadyInI 
              <6>8. CASE CwDist(kk, i) < CwDist(ln , i)  
                 <7>2. ln # kk
                    BY <6>8, <6>3, ReadyInI, CwDistInjectivity, <6>1
                 <7>3. CwDist(ln, kk) < CwDist(ln, i)
                    BY Ring3NodesPos6, <6>8, <6>0, <6>1, <7>2, <6>3, ReadyInI
                 <7>5. lset[kk] # EmptyLS(kk)
                    BY <6>0
                 <7>8. CwDist(ln, i) =< CwDist(ln, kk)
                    BY <6>1, <7>5, <7>2, <6>3, <6>31 DEF ReadyNeighborRight, Inv
                 <7>9. QED BY <7>8, <7>3, Unequality2, <6>3, <5>1, CwDistType, IisNat, ReadyInI 
              <6>9. QED BY <6>7, <6>8, <6>0, <5>1, NeighborCoverageCo, <6>1
           <5>5. CASE rr = i /\ i \in NodesReady
              <6>1. SUFFICES ASSUME 
                    lset[kk] # EmptyLS(kk), 
                    Covers(lset[i], kk),
                    kk # i
                    PROVE kk = toj'[i]
                 BY <5>1, ReadyInI, <5>5 DEF CoverToJ, Inv
              <6>3. kk = toj[i]
                 BY <6>1, <5>5 DEF CoverToJ, Inv
              <6>8. kk # m.mreq.lset.node
                 <7>1. SUFFICES ASSUME kk = m.mreq.lset.node PROVE FALSE
                    OBVIOUS
                 <7>6. /\ kk # RightNeighbor(lset[i])
                       /\ kk # LeftNeighbor(lset[i])
                    BY NoCoverNeighbor, <6>1, <5>1
                 <7>8. kk \in {RightNeighbor(lset[i]), LeftNeighbor(lset[i])}
                    BY <3>1, <7>1
                 <7>9. QED BY <7>6, <7>8
              <6>9. QED BY <6>3, <6>8, <3>1 DEF TypeInvariant
           <5>6. CASE kk = i
              <6>1. SUFFICES ASSUME 
                    lset[i] # EmptyLS(i), 
                    Covers(lset[rr], i),
                    rr # i
                    PROVE i = toj'[rr]
                 BY <5>1, ReadyInI, <5>6 DEF CoverToJ, Inv
              <6>9. QED BY <5>6,<6>1, <5>1 DEF CoverToJ, Inv
           <5>7. CASE rr # i /\ kk # i
              BY <5>1, <5>7, ReadyInI DEF CoverToJ, Inv
           <5>9. QED BY <5>4,<5>3,<5>5, <5>6, <5>7
        <4>70. NoCCJoin'
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW rr \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoin!(ii,rr,kk)'
             BY DEF NoCCJoin
           <5>1.    /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ lset[kk] \in LSet
                    /\ m.mreq.lset.node \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ lset[kk].node = kk  
                    /\ \A k \in I: lset'[k] = lset[k]
                    /\ RightNeighbor(lset'[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                    /\ \A k \in NodesReady': k # i => k \in NodesReady
                    /\ \A k \in I\{i}: toj'[k] = toj[k]
                    /\ ii \in I
                    /\ rr \in I  
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ lset[kk] \in LSet
                    /\ \A k \in I: lset'[k] = lset[k]
                 BY <4>21, <4>19
              <6>2. /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ lset[kk].node = kk  
                 BY <4>5,<4>6, <6>1
              <6>3. m.mreq.lset.node \in I
                 BY <4>2 DEF LSet
              <6>4. \A k \in NodesReady': k # i => k \in NodesReady
                 BY <4>23, <3>1 DEF NodesReady, TypeInvariant
              <6>5. \A k \in I\{i}: toj'[k] = toj[k]
                 BY <3>1 DEF TypeInvariant
              <6>6. /\ ii \in I
                    /\ rr \in I
                 BY <5>0, ReadyInI, <4>23
              <6>7. /\ RightNeighbor(lset'[i]) \in I
                    /\ RightNeighbor(lset[i]) \in I
                 BY <6>1, RightNeighborType
              <6>9. QED BY <6>1,<6>2,<6>3,<6>4,<6>5,<6>6,<6>7 
           <5>3. CASE ~ i \in NodesReady /\ ~ okToReady
              <6>0. /\ ~ i \in NodesReady'
                    /\ NodesReady'= NodesReady
                 BY <5>3, <3>1 DEF TypeInvariant, NodesReady
              <6>1. /\ ii \in NodesReady
                    /\ rr \in NodesReady
                 BY <6>0, <3>1, <5>1 
              <6>2. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         rr = RightNeighbor(lset[ii]),
                         ii # rr,
                         CwDist(ii, kk) < CwDist(ii, rr)
                    PROVE \/ toj'[ii] = kk  /\ toj'[rr] = rr
                          \/ toj'[rr] = kk /\ toj'[ii] = ii
                  BY <5>3, <5>0, <5>1, <6>1  DEF NoCCJoin, Inv
              <6>5. \/ toj[ii] = kk  /\ toj[rr] = rr
                    \/ toj[rr] = kk /\ toj[ii] = ii
                 BY <3>1, <6>2, <6>1 DEF NoCCJoin, Inv
              <6>9. QED BY <6>1, <5>1, <6>5, <5>3, ReadyInI
           <5>4. CASE ii = i /\ ~ i \in NodesReady /\ okToReady
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>4, <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>4, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1,<6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12
                               DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>4. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        rr = rn,
                        i # rr,
                        CwDist(i, kk) < CwDist(i, rr)
                        PROVE FALSE
                     BY <5>4, <5>0, <5>1, <6>1 DEF NoCCJoin, Inv
              <6>5. kk # rr
                 BY <6>4, CwDistZeroCompare, <5>1, IisNat
              <6>6. CwDist(kk, rr) < CwDist(i, rr)
                 BY Ring3NodesPos6, <5>1, <6>4, <6>5, IisNat
              <6>7. lset[kk] # EmptyLS(kk)
                 BY <6>4, <5>1
              <6>8. CwDist(i, rr) =< CwDist(kk, rr)
                 BY <6>3,<6>31, <6>4,<6>5, <6>7,<5>4, <5>1 
                    DEF ReadyNeighborLeft, Inv
              <6>9. QED BY <6>6,<6>8, Unequality2, CwDistType, <5>1, IisNat
           <5>5. CASE rr = i /\ ~ i \in NodesReady /\ okToReady
           \* similar to <5>4
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>5, <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>5, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1, <6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12 
                              DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>4. SUFFICES ASSUME ii # kk,
                        lset[kk] # EmptyLS(kk),
                        i = RightNeighbor(lset[ii]),
                        ii # i,
                        CwDist(ii, kk) < CwDist(ii, i)
                        PROVE FALSE
                     BY <5>5, <5>0, <5>1, <6>1 DEF NoCCJoin, Inv
              <6>5. kk # i
                 BY <6>4, CwDistZeroCompare, <5>1, IisNat
              <6>7. lset[kk] # EmptyLS(kk)
                 BY <6>4
              <6>8. CwDist(ii, i) =< CwDist(ii, kk)
                 BY <6>3,<6>31, <6>4,<6>5, <6>7,<5>5, <5>1 
                    DEF ReadyNeighborRight, Inv
              <6>9. QED BY <6>4,<6>8, Unequality2, CwDistType, <5>1, IisNat
           <5>6. CASE ii = i /\ i \in NodesReady
              <6>1. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        rr = RightNeighbor(lset[i]),
                        i # rr,
                        CwDist(i, kk) < CwDist(i, rr)
                        PROVE \/ toj'[i] = kk  /\ toj[rr] = rr
                              \/ toj[rr] = kk /\ toj'[i] = i
                     BY <5>6, <5>0, <5>1 DEF NoCCJoin, Inv
              <6>3. \/ toj[i] = kk  /\ toj[rr] = rr
                    \/ toj[rr] = kk /\ toj[i] = i
                 BY <6>1, <5>6, <5>1,  NoCCJoin DEF NoCCJoin, Inv
              <6>4. kk # m.mreq.lset.node 
                 <7>1. CASE  m.mreq.lset.node = LeftNeighbor(lset[i])
                    <8>1. SUFFICES ASSUME kk = m.mreq.lset.node PROVE FALSE
                          OBVIOUS
                    <8>3. kk # rr
                          BY <6>1, <5>1, CwDistZeroCompare, IisNat
                    <8>5. CwDist(kk, i) < CwDist(kk, rr)
                          BY <5>1, <6>1, <7>1, <8>1, NeighborPosition
                    <8>9. QED BY <6>1, <8>3, <5>1,<8>5, Ring3NodesPos
                 <7>2. CASE  m.mreq.lset.node = RightNeighbor(lset[i])
                    BY <6>1, CwDistInjectivity, <7>2, <5>1
                 <7>9. QED BY <7>1,<7>2, <3>1
              <6>5. CASE toj[i] = kk  /\ toj[rr] = rr
                 BY <6>4, <6>5, <3>1 DEF TypeInvariant
              <6>6. CASE toj[rr] = kk /\ toj[i] = i
                 BY <6>6, <3>1 DEF TypeInvariant
              <6>9. QED BY <6>6,<6>5,<6>3
           <5>7. CASE rr = i /\  i \in NodesReady
              <6>1. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         i = RightNeighbor(lset[ii]),
                         ii # i,
                         CwDist(ii, kk) < CwDist(ii, i)
                        PROVE \/ toj[ii] = kk  /\ toj'[i] = i
                              \/ toj'[i] = kk /\ toj[ii] = ii
                  BY <5>7, <5>0, <5>1  DEF NoCCJoin, Inv
              <6>3. \/ toj[ii] = kk  /\ toj[i] = i
                    \/ toj[i] = kk /\ toj[ii] = ii
                 BY <6>1, <5>7, <5>1 DEF NoCCJoin, Inv
              <6>4. kk # m.mreq.lset.node  
                 <7>1. SUFFICES ASSUME kk = m.mreq.lset.node PROVE FALSE
                       OBVIOUS
                 <7>2. kk \in NodesReady
                          BY <7>1, <3>1 DEF SemLRply, Inv
                 <7>5. CwDist(ii, i) =< CwDist(ii, kk)
                          BY <5>0, <5>1, <7>2, CardinalityTwoNode, 
                            ReadyInI, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>5, <6>1, IisNat, <5>1, CwDistType, Unequality2
              <6>7. CASE  toj[ii] = kk  /\ toj[i] = i
                 BY <6>7, <3>1 DEF TypeInvariant
              <6>8. CASE toj[i] = kk /\ toj[ii] = ii
                 BY <6>4, <6>8, <3>1 DEF TypeInvariant
              <6>9. QED BY <6>3,<6>7,<6>8
           <5>8. CASE ii # i /\ rr # i
              <6>1. /\ ii \in NodesReady
                    /\ rr \in NodesReady
                 BY <5>8, <3>1, <5>1 
              <6>2. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         rr = RightNeighbor(lset[ii]),
                         ii # rr,
                         CwDist(ii, kk) < CwDist(ii, rr)
                    PROVE \/ toj'[ii] = kk  /\ toj'[rr] = rr
                          \/ toj'[rr] = kk /\ toj'[ii] = ii
                  BY <5>8, <5>0, <5>1, <6>1  DEF NoCCJoin, Inv
              <6>5. \/ toj[ii] = kk  /\ toj[rr] = rr
                    \/ toj[rr] = kk /\ toj[ii] = ii
                 BY <3>1, <6>2, <6>1 DEF NoCCJoin, Inv
              <6>9. QED BY <6>1, <5>1, <6>5, <5>8, ReadyInI
           <5>9. QED BY <5>3,<5>4,<5>5,<5>6,<5>7,<5>8
        <4>71. NoCCJoinL'
           \*OMITTED \* similar to <4>70
           <5>0. SUFFICES ASSUME NEW ii \in NodesReady', NEW ll \in NodesReady', NEW kk \in I
                          PROVE  NoCCJoinL!(ii,ll,kk)'
             BY DEF NoCCJoinL
           <5>1.    /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ lset[kk] \in LSet
                    /\ m.mreq.lset.node \in I
                    /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ lset[kk].node = kk  
                    /\ \A k \in I: lset'[k] = lset[k]
                    /\ LeftNeighbor(lset'[i]) \in I
                    /\ LeftNeighbor(lset[i]) \in I
                    /\ \A k \in NodesReady': k # i => k \in NodesReady
                    /\ \A k \in I\{i}: toj'[k] = toj[k]
                    /\ ii \in I
                    /\ ll \in I  
              <6>1. /\ lset'[i] \in LSet 
                    /\ lset[i] \in LSet
                    /\ lset[kk] \in LSet
                    /\ \A k \in I: lset'[k] = lset[k]
                 BY <4>21, <4>19
              <6>2. /\ lset'[i].node = i
                    /\ lset[i].node = i
                    /\ lset[kk].node = kk  
                 BY <4>5,<4>6, <6>1
              <6>3. m.mreq.lset.node \in I
                 BY <4>2 DEF LSet
              <6>4. \A k \in NodesReady': k # i => k \in NodesReady
                 BY <4>23, <3>1 DEF NodesReady, TypeInvariant
              <6>5. \A k \in I\{i}: toj'[k] = toj[k]
                 BY <3>1 DEF TypeInvariant
              <6>6. /\ ii \in I
                    /\ ll \in I
                 BY <5>0, ReadyInI, <4>23
              <6>7. /\ LeftNeighbor(lset'[i]) \in I
                    /\ LeftNeighbor(lset[i]) \in I
                 BY <6>1, LeftNeighborType
              <6>9. QED BY <6>1,<6>2,<6>3,<6>4,<6>5,<6>6,<6>7 
           <5>3. CASE ~ i \in NodesReady /\ ~ okToReady
              <6>0. /\ ~ i \in NodesReady'
                    /\ NodesReady'= NodesReady
                 BY <5>3, <3>1 DEF TypeInvariant, NodesReady
              <6>1. /\ ii \in NodesReady
                    /\ ll \in NodesReady
                 BY <6>0, <3>1, <5>1 
              <6>2. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         ll = LeftNeighbor(lset[ii]),
                         ii # ll,
                         CwDist(kk,ii) < CwDist(ll, ii)
                    PROVE \/ toj'[ii] = kk  /\ toj'[ll] = ll
                          \/ toj'[ll] = kk /\ toj'[ii] = ii
                  BY <5>3, <5>0, <5>1, <6>1  DEF NoCCJoinL, Inv
              <6>5. \/ toj[ii] = kk  /\ toj[ll] = ll
                    \/ toj[ll] = kk /\ toj[ii] = ii
                 BY <3>1, <6>2, <6>1 DEF NoCCJoinL, Inv
              <6>9. QED BY <6>1, <5>1, <6>5, <5>3, ReadyInI
           <5>4. CASE ii = i /\ ~ i \in NodesReady /\ okToReady
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>4, <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>4, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1,<6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12
                               DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>4. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        ll = ln,
                        i # ll,
                        CwDist(kk, i) < CwDist(ll, i)
                        PROVE FALSE
                     BY <5>4, <5>0, <5>1, <6>1 DEF NoCCJoinL, Inv
              <6>5. kk # ll
                 BY <6>4, CwDistZeroCompareCo, <5>1, IisNat
              <6>6. CwDist(ll, kk) < CwDist(ll, i)
                 BY Ring3NodesPos6, <5>1, <6>4, <6>5, IisNat
              <6>7. lset[kk] # EmptyLS(kk)
                 BY <6>4
              <6>8. CwDist(ll, i) =< CwDist(ll, kk)
                 BY <6>3,<6>31, <6>4,<6>5, <6>7,<5>4, <5>1 
                    DEF ReadyNeighborRight, Inv
              <6>9. QED BY <6>6,<6>8, Unequality2, CwDistType, <5>1, IisNat
           <5>5. CASE ll = i /\ ~ i \in NodesReady /\ okToReady
           \* similar to <5>4
              <6>0. /\ ~i \in NodesReady
                    /\ i \in NodesReady'
                    /\ NodesReady = NodesReady'\{i} 
                 BY <5>5, <5>0, <3>1 DEF NodesReady, NodesOK, TypeInvariant
              <6>1. 
                  /\ ln \in newlease[i] 
                  /\ rn \in newlease[i]
                  /\ ln = LeftNeighbor(lset[i])
                  /\ rn = RightNeighbor(lset[i])
                  /\ ~i \in NodesReady
                  BY <5>5, <3>1 DEF NodesReady, NodesOK
                 
              <6>2. \/ /\ ln \in lease[i]
                       /\ rn = m.mreq.lset.node
                    \/ /\ rn \in lease[i]
                       /\ ln = m.mreq.lset.node
                    \/ /\ ln \in lease[i]
                       /\ rn \in lease[i]
                    \/ /\ ~ln \in lease[i]
                       /\ ~rn \in lease[i]
                 BY <4>25, <6>1
              <6>21. /\ i \in  grant[ln]
                     /\ i \in grant[rn]
                 <7>1./\ m \in receivedMsgs
                      /\ m.mreq.type = "BroadcastLSet"
                      /\ m.destination = i
                      /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                 <7>7. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>7
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>7
                    <8>9. QED BY <8>5, <7>7, <8>1, <8>2,<7>1, SemLRply DEF SemLRply, Inv
                 <7>8. CASE \/ /\ ln \in lease[i]
                               /\ rn = m.mreq.lset.node
                            \/ /\ rn \in lease[i]
                               /\ ln = m.mreq.lset.node
                            \/ /\ ln \in lease[i]
                               /\ rn \in lease[i]
                    BY SemLRply, LeaseGrant,<4>12, <4>13, <7>8, <3>1, <6>1, <4>25,<7>1 DEF SemLRply, LeaseGrant, Inv
                 <7>9. QED BY <7>7,<7>8,<6>2
              <6>3. /\ ln \in NodesReady
                    /\ rn \in NodesReady
                 <7>0.    /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet       
                 <7>1. CASE /\ ~ln \in lease[i]
                            /\ ~rn \in lease[i]
                    <8>1. ln = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>2. rn = m.mreq.lset.node
                       BY <6>1, <4>25, <7>1
                    <8>5. /\ m.mreq.lset.node \in {ln, rn}
                          /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>1
                    <8>6. /\ m \in receivedMsgs
                          /\ m.mreq.type = "BroadcastLSet"
                          /\ m.destination = i
                          /\ m.mreq.lset.node \in I
                       BY <3>1, <4>2 DEF LSet
                    <8>7. m.mreq.lset.node \in NodesReady
                       BY <8>5, <8>6, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                    <8>9. QED BY <8>1, <8>2,<8>7
                 <7>2. CASE /\ ln \in lease[i]
                            /\ rn = m.mreq.lset.node
                    <8>1. ln  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>2, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE rn \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>2, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ rn \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>2, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>2, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>2
                    <8>9. QED BY <8>2,<8>3
                 <7>3. CASE /\ rn \in lease[i]
                            /\ ln = m.mreq.lset.node
                    <8>1. rn  \in NodesReady
                       BY GrantReady, LeaseGrant,<4>12, <4>13, <7>3, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                    <8>2. CASE ln \in lease[i]
                       BY <8>1, <8>2,<4>12, <4>13, <7>3, <6>1,
                         GrantReady, LeaseGrant  DEF GrantReady, LeaseGrant, Inv
                    <8>3. CASE ~ ln \in lease[i]
                       <9>1. /\ m.mreq.lset.node \in {ln, rn}
                             /\ m.mreq.grant = TRUE
                       BY <6>1, <4>25, <7>3, <8>3
                       <9>2. m.mreq.lset.node \in NodesReady
                       BY <9>1, <7>0,<7>3, SemLRply,GrantReady
                        DEF SemLRply,GrantReady, Inv
                       <9>9. QED BY <8>1, <9>2, <7>3
                    <8>9. QED BY <8>2,<8>3        
                 <7>4. CASE /\ ln \in lease[i]
                            /\ rn \in lease[i]
                     BY GrantReady, LeaseGrant,<4>12, <4>13, <7>4, <6>1
                       DEF GrantReady, LeaseGrant, Inv
                 <7>9. QED BY <7>1,<7>2,<6>2, <7>3, <7>4
              <6>31. /\ i = RightNeighbor(lset[ln])
                     /\ i = LeftNeighbor(lset[rn])
                 <7>1. /\ lset[rn].node = rn
                       /\ lset[ln].node = ln
                       /\ lset[ln] \in LSet
                       /\ lset[rn] \in LSet
                    BY <4>12, <4>13 DEF TypeInvariant
                 <7>4. CASE ln = rn
                    <8>3. NodesReady = {rn}
                       <9>1.  SUFFICES ASSUME NEW k \in NodesReady,
                              k # rn
                              PROVE FALSE
                          BY <6>3
                       <9>2. /\ lset[i] # EmptyLS(i)
                             /\ i # k
                             /\ RightNeighbor(lset[i]) = rn
                             /\ LeftNeighbor(lset[i]) = ln
                             /\ ln \in I
                          BY <3>1, <6>1, <9>1,<4>13, <7>4 DEF Inv
                       <9>7. CASE CwDist(i, k) < CwDist(i, rn)
                          BY <9>2,<9>1,<6>1, <9>7, <7>4, <4>12 DEF CompleteLeafSet, Inv
                       <9>8. CASE CwDist(k, i) < CwDist(k, rn)
                          <10>1. CwDist(k, i) < CwDist(k, ln)
                              BY <9>8, <7>4
                          <10>3. CwDist(k, i) < CwDist(ln, i)
                              BY <10>1, Ring3NodesPos3, <9>1, ReadyInI, <9>2, <4>10, <7>4
                          <10>9. QED BY <10>3, <7>4,<9>2,<9>1,<6>1,<4>12 
                              DEF CompleteLeafSetL, Inv
                       <9>9. QED BY <9>7,<9>8,
                            <6>1, <9>1, <4>101, ReadyInI, <6>3, Ring3NodesPos
                    <8>4. RightNeighbor(lset[rn]) = LeftNeighbor(lset[rn])
                       BY <8>3, ReadyInI,<3>1, SingleReadyNode DEF SingleReadyNode, Inv
                    <8>5. QED BY <8>4,<7>4, <6>21, <6>0,<6>3, <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>5. CASE ln # rn
                    <8>1. CwDist(ln, i) < CwDist(ln, rn)
                       BY <7>5, NeighborPosition, <4>10, <4>101, <6>1, <4>8, <4>5
                    <8>2. CwDist(rn, ln) < CwDist(rn, i)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos1
                    <8>22.CwDist(rn, ln) < CwDist(i, ln)
                       BY <4>12, <4>13,<7>5, <4>10, <4>101, <8>1, Ring3NodesPos2
                    <8>3. ~(Cardinality(NodesReady) =< 1)
                          BY CardinalityTwoNode, ReadyInI, <6>3, <7>5
                    <8>4. i # RightNeighbor(lset[rn])
                       <9>1. SUFFICES ASSUME i = RightNeighbor(lset[rn])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(rn, i) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>2, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>5. i # LeftNeighbor(lset[ln])
                       <9>1. SUFFICES ASSUME i = LeftNeighbor(lset[ln])
                             PROVE FALSE
                          OBVIOUS 
                       <9>2. CwDist(i, ln) =< CwDist(rn, ln)
                          BY <9>1,<8>3, <6>3,<7>1, <7>5, LemmaCLtoNC DEF NeighborClosest, Inv
                       <9>3. QED BY <9>2, <8>22, Unequality2, CwDistType, ReadyInI, IisNat,<6>3
                    <8>9. QED BY <8>4,<8>5, <6>21, <6>0,<6>3, 
                           <3>1, GrantNeighbor DEF GrantNeighbor, Inv
                 <7>9. QED BY <7>4,<7>5
              <6>4. SUFFICES ASSUME ii # kk,
                        lset[kk] # EmptyLS(kk),
                        i = LeftNeighbor(lset[ii]),
                        ii # i,
                        CwDist(kk, ii) < CwDist(i, ii)
                        PROVE FALSE
                     BY <5>5, <5>0, <5>1, <6>1 DEF NoCCJoinL, Inv
              <6>5. kk # i
                 BY <6>4, CwDistZeroCompareCo, <5>1, IisNat
              <6>7. lset[kk] # EmptyLS(kk)
                 BY <6>4
              <6>8. CwDist(i, ii) =< CwDist(kk, ii)
                 BY <6>3,<6>31, <6>4,<6>5, <6>7,<5>5, <5>1 
                    DEF ReadyNeighborLeft, Inv
              <6>9. QED BY <6>4,<6>8, Unequality2, CwDistType, <5>1, IisNat
           <5>6. CASE ii = i /\ i \in NodesReady
              <6>1. SUFFICES ASSUME i # kk,
                        lset[kk] # EmptyLS(kk),
                        ll = LeftNeighbor(lset[i]),
                        i # ll,
                        CwDist(kk, i) < CwDist(ll, i)
                        PROVE \/ toj'[i] = kk  /\ toj[ll] = ll
                              \/ toj[ll] = kk /\ toj'[i] = i
                     BY <5>6, <5>0, <5>1 DEF NoCCJoinL, Inv
              <6>3. \/ toj[i] = kk  /\ toj[ll] = ll
                    \/ toj[ll] = kk /\ toj[i] = i
                 BY <6>1, <5>6, <5>1,  NoCCJoin DEF NoCCJoinL, Inv
              <6>4. kk # m.mreq.lset.node 
                 <7>1. CASE  m.mreq.lset.node = RightNeighbor(lset[i])
                    <8>1. SUFFICES ASSUME kk = m.mreq.lset.node PROVE FALSE
                          OBVIOUS
                    <8>3. kk # ll
                          BY <6>1, <5>1, CwDistZeroCompareCo, IisNat
                    <8>5. CwDist(i, kk) < CwDist(ll, kk)
                          BY <5>1, <6>1, <7>1, <8>1, NeighborPositionCo
                    <8>9. QED BY <6>1, <8>3, <5>1,<8>5, Ring3NodesPos9
                 <7>2. CASE  m.mreq.lset.node = LeftNeighbor(lset[i])
                    BY <6>1, RightDistInjectivity, <7>2, <5>1
                 <7>9. QED BY <7>1,<7>2, <3>1
              <6>5. CASE toj[i] = kk  /\ toj[ll] = ll
                 BY <6>4, <6>5, <3>1 DEF TypeInvariant
              <6>6. CASE toj[ll] = kk /\ toj[i] = i
                 BY <6>6, <3>1 DEF TypeInvariant
              <6>9. QED BY <6>6,<6>5,<6>3
           <5>7. CASE ll = i /\  i \in NodesReady
              <6>1. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         i = LeftNeighbor(lset[ii]),
                         ii # i,
                         CwDist(kk, ii) < CwDist(i, ii)
                        PROVE \/ toj[ii] = kk  /\ toj'[i] = i
                              \/ toj'[i] = kk /\ toj[ii] = ii
                  BY <5>7, <5>0, <5>1  DEF NoCCJoinL, Inv
              <6>3. \/ toj[ii] = kk  /\ toj[i] = i
                    \/ toj[i] = kk /\ toj[ii] = ii
                 BY <6>1, <5>7, <5>1,  NoCCJoinL DEF NoCCJoinL, Inv
              <6>4. kk # m.mreq.lset.node  
                 <7>1. SUFFICES ASSUME kk = m.mreq.lset.node PROVE FALSE
                       OBVIOUS
                 <7>2. kk \in NodesReady
                          BY <7>1, <3>1 DEF SemLRply, Inv
                 <7>5. CwDist(i, ii) =< CwDist(kk, ii)
                          BY <5>0, <5>1, <7>2, CardinalityTwoNode, 
                            ReadyInI, <6>1, LemmaCLtoNC DEF NeighborClosest, Inv
                 <7>9. QED BY <7>5, <6>1, IisNat, <5>1, CwDistType, Unequality2
              <6>7. CASE  toj[ii] = kk  /\ toj[i] = i
                 BY <6>7, <3>1 DEF TypeInvariant
              <6>8. CASE toj[i] = kk /\ toj[ii] = ii
                 BY <6>4, <6>8, <3>1 DEF TypeInvariant
              <6>9. QED BY <6>3,<6>7,<6>8
           <5>8. CASE ii # i /\ ll # i
              <6>1. /\ ii \in NodesReady
                    /\ ll \in NodesReady
                 BY <5>8, <3>1, <5>1 
              <6>2. SUFFICES ASSUME ii # kk,
                         lset[kk] # EmptyLS(kk),
                         ll = LeftNeighbor(lset[ii]),
                         ii # ll,
                         CwDist(kk, ii) < CwDist(ll, ii)
                    PROVE \/ toj'[ii] = kk  /\ toj'[ll] = ll
                          \/ toj'[ll] = kk /\ toj'[ii] = ii
                  BY <5>8, <5>0, <5>1, <6>1  DEF NoCCJoinL, Inv
              <6>5. \/ toj[ii] = kk  /\ toj[ll] = ll
                    \/ toj[ll] = kk /\ toj[ii] = ii
                 BY <3>1, <6>2, <6>1 DEF NoCCJoinL, Inv
              <6>9. QED BY <6>1, <5>1, <6>5, <5>8, ReadyInI
           <5>9. QED BY <5>3,<5>4,<5>5,<5>6,<5>7,<5>8
        <4>72. SemPRplyNonEmpty'
           <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 \/ mm.mreq.type = "LSProbeReply"
                 \/ mm.mreq.type = "LSProbe"
                 PROVE 
                    lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
              BY <4>1 DEF SemPRplyNonEmpty, Inv
           <5>4. mm \in receivedMsgs
              BY <4>44, <5>2
           <5>5. CASE mm.mreq.type = "LSProbeReply"
               <6>5. mm.mreq \in PRpl
                  BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <6>6. mm.mreq.node \in I
                  BY <6>5 DEF PRpl
               <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                  BY <4>19, <6>6
               <6>9. QED BY <4>44, <5>2, <6>7, <6>6
                    DEF SemPRplyNonEmpty, Inv
           <5>6. CASE mm.mreq.type = "LSProbe"
              <6>5. mm.mreq \in Prb
                  BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <6>6. mm.mreq.node \in I
                  BY <6>5 DEF Prb, LSet
               <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                  BY <4>19, <6>6
               <6>9. QED BY <4>44, <5>2, <6>7, <6>6
                    DEF SemPRplyNonEmpty, Inv
           <5>9. QED BY <5>2,<5>5,<5>6
        <4>73. SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY <4>19 DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <4>44, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. QED BY <4>44, <5>3, <5>8, <4>19
                DEF SemJRplyLSnonEmpty, Inv
        <4>74. OneJRply'
            BY <4>44 DEF OneJRply, Inv
        <4>75. OneJReq'
            BY <4>44 DEF OneJReq, Inv
        <4>76. SemJReq'
           <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinRequest"
                 PROVE  /\ mm.mreq.node \in NodesWait'
                        /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY DEF SemJReq, Inv
           <5>2. mm \in receivedMsgs
              BY <4>44, <5>1
           <5>3. mm.mreq \in JReq
              BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>4. mm.mreq.node \in I
              BY <5>3 DEF JReq
           <5>9. QED BY <5>2, <4>19, <4>23, <5>1,
                 <5>4, ReadyInI DEF SemJReq, Inv
        <4>77. SingleReadyNodeToj'
           <5>1. SUFFICES ASSUME NEW kk \in I,
                 NEW rr \in I,
                 NodesReady' = {rr},
                 lset'[kk] # EmptyLS(kk),
                 kk # rr
                 PROVE toj'[rr] = kk
              BY DEF SingleReadyNodeToj, Inv
           <5>2. /\ lset[kk] # EmptyLS(kk)
                 /\ kk # rr
              BY <5>1, <4>19 DEF SingleReadyNodeToj, Inv
           <5>4. NodesReady = {rr}
              BY <4>23, <5>1 DEF AtleastOneReady, Inv
           <5>5. toj[rr] = kk
              BY <5>1, <5>2, <5>4 DEF SingleReadyNodeToj, Inv
           <5>6. m.mreq.lset.node \in NodesReady
              BY <3>1, <4>2 DEF LSet, SemLRply, Inv
           <5>7. kk # m.mreq.lset.node
              BY <5>6, <5>2, <5>4
           <5>9. QED BY <5>5, <5>7, <3>1 DEF TypeInvariant
        <4>78. JoinMsgDual'
           BY <4>44 DEF JoinMsgDual, Inv
        <4>80. AtleastOneReady'
           BY <4>23 DEF AtleastOneReady, Inv
        <4>1001. TypeInvariant'
           BY NextTypeCorrect
        \*<4>96. NeighborClosest'
          \* BY LemmaCLtoNCp, <4>1001, <4>95, <4>54, <4>55
        <4>99. QED BY <4>20,<4>95,\*<4>96,
                  \*<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76,<4>77, <4>78, <4>80 DEF Inv  
      <3>9. QED BY <1>10, <3>1,  SimplifyAndSolve
            DEF ReceiveBLS

   <1>22. ASSUME NEW i \in I,
                 NEW j \in I,
                 Lookup(i, j), 
                 TypeInvariant,
                 Inv
          PROVE Inv'
      <3>1. ASSUME
                receivedMsgs'= receivedMsgs \cup 
                    {[destination |-> i, 
                      mreq |-> [
                                type |-> "Lookup", 
                                node |-> j]
                     ]},
                 UNCHANGED <<status, rtable, lset, probing, failed, lease, grant, toj>>,
                 TypeInvariant,
                 Inv
         PROVE  Inv'
            <4>1. \A k \in I: /\ lset'[k] = lset[k]
                              /\ grant'[k] = grant[k]
                              /\ lease'[k] = lease[k]
                              /\ toj'[k] = toj[k]
               BY UNCHANGED  <<status, rtable, lset, probing, failed, lease, grant, toj>>
            <4>2. /\ NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK 
                  /\ NonDead'= NonDead
                  /\ NodesWait' = NodesWait
               BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <3>1, <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>4. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "JoinRequest"
                     \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: 
                         ms \in 
                             {[destination |-> i, 
                              mreq |-> [
                                        type |-> "Lookup", 
                                        node |-> j]
                             ]}
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> i, 
                              mreq |-> [
                                        type |-> "Lookup", 
                                        node |-> j]
                             ].mreq.type = "Lookup"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2
            <4>5. HalfNeighbor' 
               <5>1. \A ms \in receivedMsgs' \ receivedMsgs:
                     ms \in {[destination |-> i, 
                      mreq |-> [
                                type |-> "Lookup", 
                                node |-> j]
                     ]}
                  BY <3>1 DEF TypeInvariant
               <5>2. [destination |-> i, 
                      mreq |-> [
                                type |-> "Lookup", 
                                node |-> j]
                     ].mreq.type = "Lookup"
                  OBVIOUS
               <5>3.  ~\E ms \in receivedMsgs' \ receivedMsgs: 
                      \/ ms.mreq.type = "LSProbeReply"
                      \/ ms.mreq.type = "JoinReply"
                      \/ ms.mreq.type = "LSProbe"
                      \/ ms.mreq.type = "BroadcastLSet"
                  BY <5>2, <5>1
               <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                  BY <4>1, <4>2, <5>7, ReadyInI DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                     /\ ReadyOK = {k} 
                     /\ LeftNeighbor(lset[k]) = k 
                     /\ RightNeighbor(lset[k]) = k
                     /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                     /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                     /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                  <6>1. PICK k \in ReadyOK: ReadyOK' = {k}
                     BY <5>8, <4>2
                  <6>2. 
                     /\ LeftNeighbor(lset'[k]) = k 
                     /\ RightNeighbor(lset'[k]) = k
                     /\ \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                     BY <6>1,<5>8, <4>1,<4>2, ReadyInI, SlowestZenon
                  <6>3. 
                     /\ ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # k
                     /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1,<5>8, <4>1,<4>2, ReadyInI, <5>3
                  <6>9. QED BY <6>1,<6>2,<6>3 DEF HalfNeighbor, Inv
               <5>9. QED BY <5>7,<5>8, HalfNeighbor DEF HalfNeighbor, Inv
           \* <4>6. NeighborClosest'
             \*  BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>4,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>61. GrantReady'
               BY <4>1, <4>2, ReadyInI DEF GrantReady, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>4, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>4, <5>3, <5>6, <5>8, <4>1, <4>2, 
                    SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>4, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>4, <5>3, <5>7, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>4,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <3>1, <4>1, <4>2, ReadyInI, GrantNeighbor DEF GrantNeighbor, Inv, TypeInvariant
            <4>68. SingleReadyNode'
               BY <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               BY <3>1, <4>1, <4>2, ReadyInI, CoverToJ DEF CoverToJ, Inv, TypeInvariant
            <4>70. NoCCJoin'
               BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve DEF NoCCJoin, Inv, TypeInvariant
            <4>71. NoCCJoinL'
               BY <3>1, <4>1, <4>2, ReadyInI, SimplifyAndSolve DEF NoCCJoinL, Inv, TypeInvariant
            <4>72. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>73. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>4, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>74. OneJRply'
                BY <4>4 DEF OneJRply, Inv
            <4>75. OneJReq'
                BY <4>4 DEF OneJReq, Inv
            <4>76. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>2. mm \in receivedMsgs
                  BY <4>4, <5>1
               <5>3. mm.mreq \in JReq
                  BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>4. mm.mreq.node \in I
                  BY <5>3 DEF JReq
               <5>9. QED BY <5>2, <4>1, <4>2, <5>1, ReadyInI DEF SemJReq, Inv
            <4>77. SingleReadyNodeToj'
                BY <3>1, <4>1, <4>2, ReadyInI DEF SingleReadyNodeToj, Inv
            <4>78. JoinMsgDual'
               BY <4>4 DEF JoinMsgDual, Inv
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv
            <4>9. QED BY <4>3,<4>5,\*<4>6,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76,<4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <1>22, <3>1 DEF Lookup, Inv   
   <1>26. ASSUME NEW i \in I, NEW seed \in I,
                 Join(i, seed),
                 TypeInvariant,
                 Inv
         PROVE  Inv'
      <3>1. ASSUME   
                status[i] = "dead" /\ status[seed] ="ready",
                lset[i]= EmptyLS(i),   
                  receivedMsgs'= receivedMsgs \cup 
                    {[destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]},
                    status' = [status EXCEPT ![i]="wait"],
                    UNCHANGED <<rtable, lset, probing, failed, lease,grant, toj>>,
                    TypeInvariant,
                 Inv
         PROVE  Inv'
            <4>0. status[i] = "dead" /\ status[i]'= "wait"
                 /\ \A k \in I\{i}: status'[k] = status[k]
               BY <3>1 DEF TypeInvariant
            <4>1. \A k \in I: /\ lset'[k] = lset[k]
                              /\ grant'[k] = grant[k]
                              /\ lease'[k] = lease[k]
                              /\ toj'[k] = toj[k]
               BY UNCHANGED <<rtable, lset, probing, failed, lease, grant, toj>> DEF TypeInvariant
            <4>2. /\ NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK
                  /\ \A k \in NodesWait: k \in NodesWait'
               BY <4>0, ReadyInI DEF NodesReady, ReadyOK, NodesOK, NodesWait
            <4>3. LSetInv'
               BY <3>1, <4>1, <4>2, LSetInv DEF LSetInv, Inv
            <4>4. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: 
                         ms \in 
                             {[destination |-> seed, 
                              mreq |-> [
                                        type |-> "JoinRequest", 
                                       rtable |-> InitRTable, 
                                       node |-> i]
                             ]}
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> seed, 
                              mreq |-> [
                                        type |-> "JoinRequest", 
                                       rtable |-> InitRTable, 
                                       node |-> i]
                             ].mreq.type = "JoinRequest"
                     OBVIOUS
                  <6>9. QED BY <6>1,<6>2
            <4>5. HalfNeighbor' 
               <5>1./\  ~i \in ReadyOK' 
                    /\  ~i \in ReadyOK 
                    /\ ReadyOK' = ReadyOK
                    /\ NodesWait'\{i} = NodesWait
                  BY <3>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, TypeInvariant
               <5>2. \A k \in I \ {i}:  lset'[k] = lset[k]
                  BY <3>1 DEF TypeInvariant
               <5>3. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                   \/ ms.mreq.type = "LSProbeReply"
                   \/ ms.mreq.type = "LSProbe"
                   \/ ms.mreq.type = "JoinReply"
                   \/ ms.mreq.type = "BroadcastLSet"
                  <6>1. \A ms \in receivedMsgs' \ receivedMsgs: ms \in 
                     {[destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]}
                     BY <3>1 DEF TypeInvariant
                  <6>2. [destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ].mreq.type = "JoinRequest"
                     OBVIOUS
                  
                  <6>9. QED BY <6>1, <6>2,<3>1
               <5>7. CASE \A k \in ReadyOK : RightNeighbor(lset[k]) # k /\ LeftNeighbor(lset[k]) # k
                    BY  <5>7, <4>2, <4>1, ReadyInI, HalfNeighbor DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                     /\ ReadyOK = {k} 
                     /\ LeftNeighbor(lset[k]) = k 
                     /\ RightNeighbor(lset[k]) = k
                     /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                     /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                     /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w}
                  <6>1. PICK k \in ReadyOK': ReadyOK = {k} 
                    BY <5>8, <5>1, <4>2
                  <6>2.  /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                                   => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                     BY <5>8, <6>1
                  <6>3.  /\ ReadyOK' = {k} 
                         /\ ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # k
                         /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                                                   => GetLSetContent(mj.mreq.lset) = {k}
                         /\ ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>2, <5>3, <6>1, <5>1
                  <6>4. lset[i] = EmptyLS(i)
                     BY <3>1
                  <6>5. \A w \in NodesWait'\{i}: GetLSetContent(lset'[w]) \in SUBSET {k, w}
                     BY <5>1, <5>8, <6>1, ReadyInI, <5>2
                  <6>6.  /\ LeftNeighbor(lset'[k]) = k 
                         /\ RightNeighbor(lset'[k]) = k
                     BY <4>1, <4>2, ReadyInI, <6>1, <5>8
                  <6>7. GetLSetContent(lset[i]) \in SUBSET {k, i}
                     BY <6>4,  EmptyLSIsEmpty DEF TypeInvariant
                  <6>8. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                     BY <6>5, <6>7, <4>1, ReadyInI
                  <6>9. QED BY <6>1,<6>3, <6>6, <6>8 DEF HalfNeighbor, Inv
               <5>9. QED BY <5>7,<5>8,HalfNeighbor DEF HalfNeighbor, Inv
            \*<4>6. NeighborClosest'
              \* BY <4>1, <4>2, ReadyInI, NeighborClosest DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSet DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSetL DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborRight DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>1, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>4,<5>2,<4>1,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2, ReadyInI DEF LeaseGrant, Inv
            <4>61. GrantReady'
               BY <4>1, <4>2, ReadyInI DEF GrantReady, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>4, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>4, <5>3, <5>6, <5>8, <4>1, <4>2, 
                    SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>4, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>4, <5>3, <5>7, <4>1
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply\*, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>4,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <3>1, <4>1, <4>2, ReadyInI, GrantNeighbor DEF GrantNeighbor, Inv, TypeInvariant
            <4>68. SingleReadyNode'
               BY <4>1, <4>2, ReadyInI DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               BY <3>1, <4>1, <4>2, ReadyInI, CoverToJ DEF CoverToJ, Inv, TypeInvariant
            <4>70. NoCCJoin'
               BY <3>1, <4>1, <4>2, ReadyInI, NoCCJoin, SimplifyAndSolve DEF NoCCJoin, Inv, TypeInvariant
            <4>71. NoCCJoinL'
               BY <3>1, <4>1, <4>2, ReadyInI, SimplifyAndSolve DEF NoCCJoinL, Inv, TypeInvariant
            <4>72. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>4, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>4, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>73. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>4, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>4, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>74. OneJRply'
               BY <4>4 DEF OneJRply, Inv
            <4>75. OneJReq'
               <5>1. SUFFICES ASSUME 
                     NEW m1 \in receivedMsgs', 
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type = "JoinRequest",
                     m2.mreq.type = "JoinRequest",
                     m1.mreq.node = m2.mreq.node
                  PROVE m1 = m2
                  BY DEF OneJReq, Inv
               <5>6. CASE m1 = [destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]
                  <6>1. SUFFICES ASSUME m2 # m1 PROVE FALSE
                     OBVIOUS
                  <6>2. m2 \in receivedMsgs
                     BY <3>1, <6>1, <5>6
                  <6>3. m1.mreq.node = i
                     BY <5>6
                  <6>4. m2.mreq.node = i
                     BY <6>3, <5>1
                  <6>5. m2.mreq.node \in NodesWait
                     BY <6>2, <5>1 DEF SemJReq, Inv
                  <6>6. status[i] = "dead"
                     BY <3>1
                  <6>9. QED BY <6>5, <6>6,<6>4, StatusDisjoint
               <5>7. CASE m2 = [destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]
                  <6>1. SUFFICES ASSUME m2 # m1 PROVE FALSE
                     OBVIOUS
                  <6>2. m1 \in receivedMsgs
                     BY <3>1, <6>1, <5>6
                  <6>3. m2.mreq.node = i
                     BY <5>7
                  <6>4. m1.mreq.node = i
                     BY <6>3, <5>1
                  <6>5. m1.mreq.node \in NodesWait
                     BY <6>2, <5>1 DEF SemJReq, Inv
                  <6>6. status[i] = "dead"
                     BY <3>1
                  <6>9. QED BY <6>5, <6>6,<6>4, StatusDisjoint
               <5>8. CASE /\ m1 # [destination |-> seed, 
                              mreq |-> [
                                        type |-> "JoinRequest", 
                                       rtable |-> InitRTable, 
                                       node |-> i]
                             ]
                          /\ m2 # [destination |-> seed, 
                              mreq |-> [
                                        type |-> "JoinRequest", 
                                       rtable |-> InitRTable, 
                                       node |-> i]
                             ]
                  BY <3>1, <5>8, <5>1 DEF OneJReq, Inv
               <5>9. QED BY <5>6,<5>7,<5>8
            <4>76. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>2. CASE mm \in receivedMsgs
                   <6>3. mm.mreq \in JReq
                      BY SlowestZenon,<5>2, <5>1 DEF TypeInvariant, DMsg, 
                          MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>4. mm.mreq.node \in I
                      BY <6>3 DEF JReq
                   <6>9. QED BY <5>2, <4>1, <4>2, <5>1, ReadyInI DEF SemJReq, Inv
               <5>3. CASE ~mm \in receivedMsgs
                  <6>1. mm = [destination |-> seed, 
                              mreq |-> [
                                        type |-> "JoinRequest", 
                                       rtable |-> InitRTable, 
                                       node |-> i]
                             ]
                     BY <3>1, <5>3, <5>1
                  <6>2. mm.mreq.node = i
                     BY <6>1
                  <6>3. i \in NodesWait'
                     BY <3>1 DEF NodesWait, TypeInvariant
                  <6>4. lset[i] = EmptyLS(i)
                     BY <3>1
                  <6>9. QED BY <6>4, <6>3, <4>1, <6>2
               <5>9. QED BY <5>3,<5>2
            <4>77. SingleReadyNodeToj'
                BY <3>1, <4>1, <4>2, ReadyInI DEF SingleReadyNodeToj, Inv
            <4>78. JoinMsgDual'
               <5>1. SUFFICES ASSUME NEW m1 \in receivedMsgs',
                     NEW m2 \in receivedMsgs',
                     m1.mreq.type  = "JoinReply",
                     m2.mreq.type = "JoinRequest"
                     PROVE m2.mreq.node # m1.destination
                  BY DEF JoinMsgDual, Inv
               <5>2. m1 \in receivedMsgs
                  BY <5>1, <4>4
               <5>7. CASE m2 = [destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]
                  <6>1. m2.mreq.node = i
                     BY <5>7
                  <6>2. m1.destination \in NodesWait
                     BY <5>1, <5>2 DEF SemJRply, Inv
                  <6>3. QED BY <6>1, <6>2, <3>1, StatusDisjoint DEF NodesWait 
               <5>8. CASE m2 # [destination |-> seed, 
                      mreq |-> [
                                type |-> "JoinRequest", 
                               rtable |-> InitRTable, 
                               node |-> i]
                     ]
                  <6>1. m2 \in receivedMsgs
                     BY <5>8, <3>1
                  <6>9. QED BY <6>1, <5>2, <5>1 DEF JoinMsgDual, Inv
               <5>9. QED BY <5>7,<5>8
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv
            <4>9. QED BY <4>3,<4>5,\*<4>6,
                  \*<4>51,<4>52,<4>53,
                  <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76,<4>77, <4>78, <4>80 DEF Inv  
      <3>9.QED BY <1>26, <3>1, SimplifyAndSolve DEF Join
   <1>27. ASSUME MsgLost, 
              TypeInvariant,
              Inv
          PROVE Inv'
      <2>1. SUFFICES ASSUME NEW m \in receivedMsgs,
            receivedMsgs' = receivedMsgs \ {m},
            UNCHANGED  <<status, rtable, lset, probing, failed, lease, grant, toj>>,
            TypeInvariant,
            Inv
            PROVE Inv'
         BY <1>27 DEF MsgLost
      <2>2. ~\E ms \in receivedMsgs' \ receivedMsgs: 
                     \/ ms.mreq.type = "LSProbe" 
                     \/ ms.mreq.type = "LSProbeReply"
                     \/ ms.mreq.type = "JoinRequest"
                     \/ ms.mreq.type = "JoinReply"
                     \/ ms.mreq.type = "BroadcastLSet"
         BY <2>1
      <2>3.   /\ NodesReady' = NodesReady 
              /\ ReadyOK' = ReadyOK 
              /\ NonDead'= NonDead
              /\ NodesWait' = NodesWait
           BY <2>1, ReadyInI DEF NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
      <2>4. \A j \in I: /\ lset'[j] = lset[j]
                        /\ lset'[j].node = j
                        /\ lset[j].node = j
                        /\ grant'[j] = grant[j]
                        /\ lease'[j] = lease[j]
                        /\ toj'[j] = toj[j]
         BY <2>1 DEF TypeInvariant
      \*<2>60. NeighborClosest'
        \* BY <2>1, <2>3, <2>4 DEF NeighborClosest, Inv
      <2>61. CompleteLeafSet'
         BY <2>1, <2>3, <2>4 DEF CompleteLeafSet, Inv
      <2>62. CompleteLeafSetL'
         BY <2>1, <2>3, <2>4 DEF CompleteLeafSetL, Inv
      <2>63.ReadyNeighborRight'
         BY <2>1, <2>3, <2>4 DEF ReadyNeighborRight, Inv
      <2>64.ReadyNeighborLeft'
         BY <2>1, <2>3, <2>4 DEF ReadyNeighborLeft, Inv
      <2>65.NoCCJoinL'
         BY <2>1, <2>3, <2>4 DEF NoCCJoinL, Inv
      <2>66.NoCCJoin'
         BY <2>1, <2>3, <2>4 DEF NoCCJoin, Inv
      <2>67.LSetInv'
         BY <2>1, <2>3, <2>4 DEF LSetInv, Inv
      <2>68.HalfNeighbor'
           <5>2./\ ReadyOK' = ReadyOK 
                /\ NodesWait'= NodesWait
                /\ \A jj \in I: lset'[jj] = lset[jj]
              BY <2>3, <2>4
           <5>4. CASE \A jj \in ReadyOK : RightNeighbor(lset[jj]) # jj 
                      /\ LeftNeighbor(lset[jj]) # jj
              BY <5>2, <5>4, ReadyInI DEF HalfNeighbor
           <5>5. CASE  \E jj \in ReadyOK:
                         /\ ReadyOK = {jj} 
                         /\ LeftNeighbor(lset[jj]) = jj
                         /\ RightNeighbor(lset[jj]) = jj
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # jj
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                               => GetLSetContent(mj.mreq.lset) = {jj}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {jj, w}
              <6>1. PICK jj \in ReadyOK:
                         /\ ReadyOK = {jj} 
                         /\ LeftNeighbor(lset[jj]) = jj
                         /\ RightNeighbor(lset[jj]) = jj
                         /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                         /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination #jj
                         /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                               => GetLSetContent(mj.mreq.lset) = {jj}
                         /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                         /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {jj, w}
                 BY <5>5
              <6>2. /\ReadyOK'= {jj}
                    /\ LeftNeighbor(lset'[jj]) = jj
                    /\ RightNeighbor(lset'[jj]) = jj
                 BY <6>1, <5>2, ReadyInI
              <6>3. ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                 BY <6>1, <2>2
              <6>4. ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # jj
                 BY <6>1, <2>2, <5>5
              <6>5. /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                           => GetLSetContent(mj.mreq.lset) = {jj}
                 BY <6>1, <2>2, <5>5
              <6>6. ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                 BY <6>1, <2>2, <5>5
              <6>7. \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {jj, w}
                 BY <5>2, ReadyInI, <6>1
              <6>9. QED BY <6>1, <6>2, <6>3, <6>4, <6>5, <6>6, <6>7 DEF HalfNeighbor
           <5>9. QED BY <5>4, <5>5, HalfNeighbor DEF HalfNeighbor, Inv
      <2>69.GrantNeighbor'
         BY <2>1, <2>3, <2>4 DEF GrantNeighbor, Inv
      <2>70.GrantReady'
         BY <2>1, <2>3, <2>4 DEF GrantReady, Inv
      <2>71.LeaseGrant'
         BY <2>1, <2>3, <2>4 DEF LeaseGrant, Inv
      <2>72.CoverToJ'
         BY <2>1, <2>3, <2>4 DEF CoverToJ, Inv
      <2>73.SingleReadyNodeToj'
         BY <2>1, <2>3, <2>4 DEF SingleReadyNodeToj, Inv
      <2>74.SingleReadyNode'
         BY <2>1, <2>3, <2>4 DEF SingleReadyNode, Inv
      <2>75.LSnonEmpty'
         BY <2>1, <2>3, <2>4 DEF LSnonEmpty, Inv
      <2>76.SemPRplyNonEmpty'
         BY <2>1, <2>2, <2>3, <2>4 DEF SemPRplyNonEmpty, Inv
      <2>77.SemJRplyLSnonEmpty'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                              lset'[kk] # EmptyLS(kk)  
               BY <2>1 DEF SemJRplyLSnonEmpty, Inv
            <5>4. mm \in receivedMsgs
              BY <2>2, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.mreq.lset.node \in I
              BY <5>6 DEF LSet
           <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
              BY <5>7, <5>6, GetLSetContentType
           <5>9. QED BY <2>2, <5>3, <5>8, <2>1
                DEF SemJRplyLSnonEmpty, Inv
      <2>78.SemLRply'
           <5>2. SUFFICES ASSUME NEW ii \in I, 
                 NEW nn \in I, 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "BroadcastLSet",
                 mm.mreq.lset.node = nn,
                 mm.destination = ii
                PROVE
                    /\ nn \in NodesReady'
                    /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
              BY <2>3, <2>4, SimplifyAndSolve DEF SemLRply, Inv
           <5>9. QED BY <2>2,<5>2,<2>3,<2>4 DEF SemLRply, Inv
      <2>79.SemJRply'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                PROVE
                     /\ lset'[mm.destination] = EmptyLS(mm.destination)
                     /\ toj'[mm.mreq.lset.node] = mm.destination
                     /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                         => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                     /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                        => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                     /\ mm.mreq.lset.node \in NodesReady
                     /\ mm.destination \in NodesWait
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                     /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
              BY <2>1, <2>3 DEF SemJRply, Inv
           <5>4. mm \in receivedMsgs
              BY <2>2, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>7. mm.destination \in I
              BY <5>4 DEF TypeInvariant, DMsg
           <5>8. /\ mm.mreq.lset.node \in I
                 /\ LeftNeighbor(mm.mreq.lset) \in I
                 /\ RightNeighbor(mm.mreq.lset) \in I          
              BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
           <5>9. QED BY <2>2,<5>3,<5>7, <5>8,<2>1,<2>3, SlowestZenon DEF SemJRply, Inv

      <2>80.SemJoinNeighbor'
            <5>3. SUFFICES ASSUME NEW ls \in LSet,
                 NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                 PROVE 
                 \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                    /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                 \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                    /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
              BY DEF SemJoinNeighbor, Inv
           <5>9. QED BY <2>2, <5>3 DEF SemJoinNeighbor, Inv
      <2>81.SemJoinLeafSet'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE 
                 /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                 /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
              BY <2>4 DEF SemJoinLeafSet, Inv
            <5>4. mm \in receivedMsgs
              BY <2>2, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
              BY <2>4, <5>6
           <5>9. QED BY <2>2, <5>3, <5>7, <2>4, SimplifyAndSolve
                DEF SemJoinLeafSet, Inv
      <2>82.SemSingleReadyJoin'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply",
                 mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                 PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                       /\ NodesReady = {mm.mreq.lset.node}
              BY <2>4, <2>3 DEF SemSingleReadyJoin, Inv
           <5>4. mm \in receivedMsgs
              BY <2>2, <5>3
           <5>5. mm.mreq \in JRpl
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
            <5>6. mm.mreq.lset \in LSet
              BY <5>5 DEF JRpl
           <5>8. mm.mreq.lset.node \in I
              BY <5>5 DEF JRpl, LSet
           <5>9. QED BY <2>1, <2>2, <5>3, <5>6, <5>8, 
              <2>4, <2>3, SlowestZenon DEF SemSingleReadyJoin, Inv
      <2>83.SemEmptyLSetSingleNode'
           <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                 mm.mreq.type = "JoinReply"
                 PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                         \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                   <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
              BY <2>4 DEF SemEmptyLSetSingleNode, Inv
           <5>9. QED BY <2>2, <5>3, <2>4, <2>3 
                DEF SemEmptyLSetSingleNode, Inv
      <2>84.SemJReq'
           <5>3. SUFFICES ASSUME 
                 NEW mm \in receivedMsgs',
                 mm.mreq.type  = "JoinRequest"
                 PROVE
                   /\ mm.mreq.node \in NodesWait'
                   /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY <2>4, <2>3 DEF SemJReq, Inv
            <5>4. mm \in receivedMsgs
              BY <2>2, <5>3
           <5>5. mm.mreq \in JReq
              BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
              MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
           <5>6. mm.mreq.node \in I
              BY <5>5 DEF JReq
           <5>7. /\ mm.mreq.node \in NodesWait
                 /\ lset[mm.mreq.node] = EmptyLS(mm.mreq.node)
              BY <5>3, <5>4 DEF SemJReq, Inv
           <5>9. QED BY <2>2,<5>3,<5>6,<2>4,<2>3, <5>7
      <2>85.JoinMsgDual'
         <3>1. SUFFICES ASSUME 
               NEW m1 \in receivedMsgs',
               NEW m2 \in receivedMsgs',
               m1.mreq.type  = "JoinReply",
               m2.mreq.type = "JoinRequest"
               PROVE m2.mreq.node # m1.destination
            BY DEF JoinMsgDual
         <3>2. /\ m1 \in receivedMsgs
               /\ m2 \in receivedMsgs
            BY <3>1, <2>2
         <3>9. QED BY <3>2, <3>1 DEF JoinMsgDual, Inv
      <2>86.OneJRply'
         <3>1. SUFFICES ASSUME 
               NEW m1 \in receivedMsgs',
               NEW m2 \in receivedMsgs',
               m1.mreq.type = "JoinReply",
               m2.mreq.type = "JoinReply",
               m1.destination = m2.destination 
               PROVE m1 = m2
            BY DEF OneJRply
         <3>2. /\ m1 \in receivedMsgs
               /\ m2 \in receivedMsgs
            BY <3>1, <2>2
         <3>9. QED BY <3>2, <3>1 DEF OneJRply, Inv
      <2>87.OneJReq'
           <3>1. SUFFICES ASSUME 
                 NEW m1 \in receivedMsgs', 
                 NEW m2 \in receivedMsgs',
                 m1.mreq.type = "JoinRequest",
                 m2.mreq.type = "JoinRequest",
                 m1.mreq.node = m2.mreq.node
              PROVE m1 = m2
              BY DEF OneJReq, Inv
           <3>2. /\ m1 \in receivedMsgs
                 /\ m2 \in receivedMsgs
            BY <3>1, <2>2
         <3>9. QED BY <3>2, <3>1 DEF OneJReq, Inv
      <2>88.AtleastOneReady'
         BY <2>3 DEF AtleastOneReady, Inv
      <2>99. QED BY \* <2>60, 
             <2>61, <2>62, <2>63, <2>64, <2>65, 
             <2>66, <2>67, <2>68, <2>69, 
             <2>70, <2>71, <2>72, <2>73, <2>74, 
             <2>75, <2>76, <2>77, <2>78, <2>79, 
             <2>80, <2>81, <2>82, <2>83, <2>84, 
             <2>85, <2>86, <2>87, <2>88 DEF Inv 
   <1>90. CASE UNCHANGED vars
            <4>0. \A j \in I: toj'[j] = toj[j]
                BY <1>90, UNCHANGED  toj DEF vars
            <4>1. \A j \in I: lset'[j] = lset[j]
               BY <1>90, UNCHANGED  lset DEF vars
            <4>2. /\ NodesReady' = NodesReady 
                  /\ ReadyOK' = ReadyOK 
                  /\ NonDead'= NonDead
                  /\ NodesWait' = NodesWait
                  /\ receivedMsgs' = receivedMsgs
               BY <1>90, ReadyInI DEF vars, NodesReady, NodesOK, ReadyOK, NodesWait, NonDead
            <4>3. LSetInv'
               BY <1>90, <4>1, <4>2, LSetInv 
               DEF LSetInv, Inv
            <4>4. \A j \in I: /\ grant'[j] = grant[j]
                              /\ lease'[j] = lease[j]
               BY <1>90, UNCHANGED  <<grant, lease>> DEF vars
            <4>5. HalfNeighbor' 
               <5>3.  ~\E ms \in receivedMsgs' \ receivedMsgs: 
                      \/ ms.mreq.type = "LSProbeReply"
                      \/ ms.mreq.type = "JoinReply"
                      \/ ms.mreq.type = "LSProbe"
                      \/ ms.mreq.type = "BroadcastLSet"
                  BY <1>90 DEF vars
               <5>7. CASE  \A k \in ReadyOK : 
                      /\ RightNeighbor(lset[k]) # k 
                      /\ LeftNeighbor(lset[k]) # k
                  BY <4>1, <4>2, <5>7, ReadyInI DEF HalfNeighbor, Inv
               <5>8. CASE \E k \in ReadyOK:
                     /\ ReadyOK = {k} 
                     /\ LeftNeighbor(lset[k]) = k 
                     /\ RightNeighbor(lset[k]) = k
                     /\ \A w \in NodesWait: GetLSetContent(lset[w]) \in SUBSET {k, w} 
                     /\ ~\E ms \in receivedMsgs: ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs: mk.mreq.type = "LSProbe" /\ mk.destination # k
                     /\ \A mj \in receivedMsgs: mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs: mb.mreq.type = "BroadcastLSet"
                  <6>1. PICK k \in ReadyOK: ReadyOK' = {k}
                     BY <5>8, <4>2
                  <6>2. 
                     /\ LeftNeighbor(lset'[k]) = k 
                     /\ RightNeighbor(lset'[k]) = k
                     /\ \A w \in NodesWait': GetLSetContent(lset'[w]) \in SUBSET {k, w}
                     BY <6>1,<5>8, <4>1,<4>2, ReadyInI
                  <6>3. 
                     /\ ~\E ms \in receivedMsgs': ms.mreq.type = "LSProbeReply" 
                     /\ ~\E mk \in receivedMsgs': mk.mreq.type = "LSProbe" /\ mk.destination # k
                     /\ \A mj \in receivedMsgs': mj.mreq.type = "JoinReply"
                                               => GetLSetContent(mj.mreq.lset) = {k}
                     /\ ~\E mb \in receivedMsgs': mb.mreq.type = "BroadcastLSet"
                     BY <6>1,<5>8, <4>1,<4>2, ReadyInI, <5>3
                  <6>9. QED BY <6>1,<6>2,<6>3 DEF HalfNeighbor, Inv
               <5>9. QED BY <5>7,<5>8, HalfNeighbor DEF Inv, HalfNeighbor
            \*<4>6. NeighborClosest'
              \* BY <4>1, <4>2, ReadyInI, NeighborClosest 
               \*DEF NeighborClosest, Inv
            \*<4>51. ReadyNeighbor'
            \*<4>52. IncludeNeighbor'
            \*<4>53. CrossNeighbor'
            <4>54. CompleteLeafSet'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSet 
               DEF CompleteLeafSet, Inv
            <4>55. CompleteLeafSetL'
               BY <4>1, <4>2, ReadyInI, CompleteLeafSetL 
               DEF CompleteLeafSetL, Inv
            <4>56. ReadyNeighborRight'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborRight 
               DEF ReadyNeighborRight, Inv
            <4>57. ReadyNeighborLeft'
               BY <4>1, <4>2, ReadyInI, ReadyNeighborLeft 
               DEF ReadyNeighborLeft, Inv
            <4>58. SemLRply'
               <5>2. SUFFICES ASSUME NEW ii \in I, 
                     NEW nn \in I, 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "BroadcastLSet",
                     mm.mreq.lset.node = nn,
                     mm.destination = ii
                    PROVE
                        /\ nn \in NodesReady'
                        /\ (mm.mreq.grant =TRUE => ii \in grant'[nn])
                  BY <4>4, <4>2, SimplifyAndSolve DEF SemLRply, Inv
               <5>9. QED BY <4>4,<5>2,<4>2 DEF SemLRply, Inv
            <4>59. LSnonEmpty'
               BY <4>1, <4>2, ReadyInI 
               DEF LSnonEmpty, Inv
            <4>60. LeaseGrant'
               BY <4>1, <4>2,<4>4, ReadyInI 
               DEF LeaseGrant, Inv
            <4>61. GrantReady'
               BY <4>1, <4>2,<4>4, ReadyInI 
               DEF GrantReady, Inv
            <4>62. SemJoinNeighbor'
               <5>3. SUFFICES ASSUME NEW ls \in LSet,
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     ls = AddToLSet(GetLSetContent(mm.mreq.lset), EmptyLS(mm.destination))
                     PROVE 
                     \/ /\ RightNeighbor(ls) = mm.mreq.lset.node
                        /\ LeftNeighbor(ls) = LeftNeighbor(mm.mreq.lset)
                     \/ /\ LeftNeighbor(ls) = mm.mreq.lset.node 
                        /\ RightNeighbor(ls) = RightNeighbor(mm.mreq.lset)
                  BY DEF SemJoinNeighbor, Inv
               <5>9. QED BY <4>2, <5>3 DEF SemJoinNeighbor, Inv
            <4>63. SemSingleReadyJoin'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply",
                     mm.mreq.lset = EmptyLS(mm.mreq.lset.node)
                     PROVE /\ lset'[mm.mreq.lset.node] = mm.mreq.lset
                           /\ NodesReady = {mm.mreq.lset.node}
                  BY <4>1, <4>2 DEF SemSingleReadyJoin, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>2, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>8. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>9. QED BY <4>4, <5>3, <5>6, <5>8, <4>1, <4>2, 
                    SlowestZenon DEF SemSingleReadyJoin, Inv
            <4>64. SemEmptyLSetSingleNode'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE ( \/ mm.mreq.lset.node  = LeftNeighbor(mm.mreq.lset)
                             \/ mm.mreq.lset.node  = RightNeighbor(mm.mreq.lset)
                       <=> mm.mreq.lset = EmptyLS(mm.mreq.lset.node))
                  BY <4>1 DEF SemEmptyLSetSingleNode, Inv
               <5>9. QED BY <4>4, <5>3, <4>1, <4>2 
                    DEF SemEmptyLSetSingleNode, Inv
            <4>65. SemJoinLeafSet'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     /\ LeftNeighbor(lset'[mm.mreq.lset.node]) = LeftNeighbor(mm.mreq.lset)
                     /\ RightNeighbor(lset'[mm.mreq.lset.node]) = RightNeighbor(mm.mreq.lset)
                  BY <4>1 DEF SemJoinLeafSet, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>2, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset.node \in I
                  BY <5>5 DEF JRpl, LSet
               <5>7. lset'[mm.mreq.lset.node] = lset[mm.mreq.lset.node]
                  BY <4>1, <5>6
               <5>9. QED BY <4>2, <5>3, <5>7, <4>1, SimplifyAndSolve
                    DEF SemJoinLeafSet, Inv
            <4>66. SemJRply'
               <5>3. SUFFICES ASSUME 
                     NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                    PROVE
                         /\ lset'[mm.destination] = EmptyLS(mm.destination)
                         /\ toj'[mm.mreq.lset.node] = mm.destination
                         /\ (LeftNeighbor(mm.mreq.lset) # mm.mreq.lset.node 
                             => toj'[LeftNeighbor(mm.mreq.lset)] = LeftNeighbor(mm.mreq.lset))
                         /\ (RightNeighbor(mm.mreq.lset) # mm.mreq.lset.node
                            => toj'[RightNeighbor(mm.mreq.lset)] = RightNeighbor(mm.mreq.lset))
                         /\ mm.mreq.lset.node \in NodesReady
                         /\ mm.destination \in NodesWait
                         /\ RightNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ LeftNeighbor(lset'[mm.mreq.lset.node]) \in lease'[mm.mreq.lset.node]
                         /\ Covers(lset'[mm.mreq.lset.node], mm.destination)
                  BY <4>1, <4>2 DEF SemJRply, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>2, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.destination \in I
                  BY <5>4 DEF TypeInvariant, DMsg
               <5>8. /\ mm.mreq.lset.node \in I
                     /\ LeftNeighbor(mm.mreq.lset) \in I
                     /\ RightNeighbor(mm.mreq.lset) \in I          
                  BY <5>6, LeftNeighborType, RightNeighborType DEF LSet
               <5>9. QED BY <4>0, <4>4,<5>3,<5>7,<5>8,<4>1,<4>2, SlowestZenon DEF SemJRply, Inv
            <4>67. GrantNeighbor'
               BY <1>90, <4>2 DEF GrantNeighbor, Inv, vars
            <4>68. SingleReadyNode'
               BY <4>1, <4>2, ReadyInI 
               DEF SingleReadyNode, Inv
            <4>69. CoverToJ'
               BY <1>90, <4>2 DEF CoverToJ, Inv, vars
            <4>70. NoCCJoin'
              BY <1>90, <4>2, SimplifyAndSolve DEF NoCCJoin, Inv, vars
            <4>71. NoCCJoinL'
               BY <1>90, <4>2, SimplifyAndSolve DEF NoCCJoinL, Inv, vars
            <4>72. SemPRplyNonEmpty'
               <5>2. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     \/ mm.mreq.type = "LSProbeReply"
                     \/ mm.mreq.type = "LSProbe"
                     PROVE 
                        lset'[mm.mreq.node] # EmptyLS(mm.mreq.node)
                  BY <4>1 DEF SemPRplyNonEmpty, Inv
               <5>4. mm \in receivedMsgs
                  BY <4>2, <5>2
               <5>5. CASE mm.mreq.type = "LSProbeReply"
                   <6>5. mm.mreq \in PRpl
                      BY SlowestZenon,<5>2, <5>4, <5>5 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF PRpl
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>2, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>6. CASE mm.mreq.type = "LSProbe"
                  <6>5. mm.mreq \in Prb
                      BY SlowestZenon,<5>2, <5>4, <5>6 DEF TypeInvariant, DMsg, 
                      MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
                   <6>6. mm.mreq.node \in I
                      BY <6>5 DEF Prb, LSet
                   <6>7. lset'[mm.mreq.node] = lset[mm.mreq.node]
                      BY <4>1, <6>6
                   <6>9. QED BY <4>2, <5>2, <6>7, <6>6
                        DEF SemPRplyNonEmpty, Inv
               <5>9. QED BY <5>2,<5>5,<5>6
            <4>73. SemJRplyLSnonEmpty'
               <5>3. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinReply"
                     PROVE 
                     \A kk \in GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node}: 
                                  lset'[kk] # EmptyLS(kk)  
                   BY <4>1 DEF SemJRplyLSnonEmpty, Inv
                <5>4. mm \in receivedMsgs
                  BY <4>2, <5>3
               <5>5. mm.mreq \in JRpl
                  BY SlowestZenon,<5>3, <5>4 DEF TypeInvariant, DMsg, 
                  MReq, JReq, JRpl, Prb, PRpl, Look, LReq, BLS, NoLR
               <5>6. mm.mreq.lset \in LSet
                  BY <5>5 DEF JRpl
               <5>7. mm.mreq.lset.node \in I
                  BY <5>6 DEF LSet
               <5>8. GetLSetContent(mm.mreq.lset) \ {mm.mreq.lset.node} \in SUBSET I
                  BY <5>7, <5>6, GetLSetContentType
               <5>9. QED BY <4>2, <5>3, <5>8, <4>1
                    DEF SemJRplyLSnonEmpty, Inv
            <4>74. OneJRply'
               BY <1>90, <4>2 DEF OneJRply, Inv
            <4>75. OneJReq'
               BY <1>90, <4>2 DEF OneJReq, Inv
               
            <4>76. SemJReq'
               <5>1. SUFFICES ASSUME NEW mm \in receivedMsgs',
                     mm.mreq.type = "JoinRequest"
                     PROVE  /\ mm.mreq.node \in NodesWait'
                            /\ lset'[mm.mreq.node] = EmptyLS(mm.mreq.node)
                  BY DEF SemJReq, Inv
               <5>9. QED BY <4>2, <4>1, <5>1, ReadyInI
                    DEF SemJReq, Inv
            <4>77. SingleReadyNodeToj'
                BY <1>90, <4>0, <4>1, <4>2,<4>4, ReadyInI DEF SingleReadyNodeToj, Inv, vars
            <4>78. JoinMsgDual'
               BY <1>90, <4>2 DEF JoinMsgDual, Inv
            <4>80. AtleastOneReady'
               BY <4>2 DEF AtleastOneReady, Inv
            <4>9. QED BY <4>3,<4>5,\*<4>6,
                 \*<4>51,<4>52,<4>53,
                 <4>54,<4>55,<4>56,<4>57,<4>58,
                  <4>59,<4>60,<4>61,<4>62,<4>63,<4>64,
                  <4>65, <4>66,<4>67,<4>68,<4>69,
                  <4>70,<4>71, <4>72, <4>73, <4>74,<4>75,
                  <4>76,<4>77, <4>78, <4>80 DEF Inv  

<1>99. QED
     BY <1>1, <1>2, <1>3, <1>4, <1>5, <1>6,<1>7,<1>8,<1>9,<1>10,
               \*<1>11,<1>12, <1>122, <1>13,<1>14,<1>15,
               \*<1>17,<1>18,<1>19,<1>21,
               <1>22, <1>26,<1>27, <1>90, SlowestZenon DEF Next
THEOREM Invariance == Spec => []Inv
 

=============================================================================
\* Modification History
\* Last modified Wed Apr 03 17:36:56 CEST 2013 by Administrator
\* Last modified Fri May 25 11:42:03 CEST 2012 by tianlu
\* Last modified Thu Mar 08 15:01:54 CET 2012 by tianlu
\* Last modified Thu Mar 08 10:10:41 CET 2012 by merz
\* Last modified Wed Mar 07 18:31:17 CET 2012 by merz
\* Last modified Wed Mar 07 13:54:55 CET 2012 by tianlu
\* Last modified Tue Mar 06 15:18:40 CET 2012 by merz
\* Last modified Sat Feb 18 09:17:41 CET 2012 by tianlu
\* Last modified Tue Feb 14 14:51:58 CET 2012 by merz
\* Last modified Wed Feb 08 14:37:18 CET 2012 by tianlu
\* Last modified Wed Feb 08 09:18:04 CET 2012 by merz
\* Last modified Tue Jan 31 21:49:51 CET 2012 by tianlu
\* Last modified Tue Jan 31 12:19:57 CET 2012 by merz
\* Last modified Thu Jan 19 14:41:51 CET 2012 by tianlu
\* Last modified Thu Jan 19 11:08:01 CET 2012 by merz
\* Last modified Wed Jan 11 18:33:05 CET 2012 by tianlu
\* Last modified Wed Jan 11 15:26:38 CET 2012 by merz
\* Last modified Fri Jan 06 19:56:15 CET 2012 by tianlu
\* Last modified Fri Jan 06 16:22:45 CET 2012 by merz
\* Last modified Mon Apr 04 15:01:14 CEST 2011 by tianlu
\* Last modified Mon Feb 14 17:32:49 CET 2011 by merz
\* Last modified Mon Jan 10 13:35:48 CET 2011 by tianlu
\* Last modified Fri Jan 07 15:41:32 CET 2011 by merz
\* Created Wed Dec 15 21:43:48 CET 2010 by tianlu
